! See copyright notice in the COPYRIGHT file.
! ****************************************************************************** !
!> This module contains boundary type definitions, different boundary treatment
!! types, abtract interface and routine to load boundary table from config file
!!
!! The boundary conditions are stored in boundary type definitions for each 
!! boundary which is defined in the main lua file. 
!!```lua
!! boundary_condition = {  }
!!```
!!
!! The definitions there must equal the boundary definition within the mesh on 
!! disk, in terms of the amount of boundary conditions and the labels.
!! A detailed description on the implementation details are given in 
!! [[tem_bc_module]] "boundary condition implementation"
!!
!! This module will be used as a header to use boundary definitions in other
!! modules.
!!
?? include 'header/lbm_macros.inc'
module mus_bc_header_module

  ! include treelm modules
  use env_module,               only: rk, LabelLen
  use tem_aux_module,           only: tem_abort
  use tem_property_module,      only: prp_solid
  use tem_bc_module,            only: tem_bc_state_type, tem_load_bc_state
  use tem_bc_header_module,     only: tem_load_bc_header, tem_bc_header_type
  use tem_bc_prop_module,       only: tem_bc_prop_type
  use tem_tools_module,         only: tem_horizontalSpacer
  use tem_time_module,          only: tem_time_type
  use treelmesh_module,         only: treelmesh_type
  use tem_varSys_module,        only: tem_varSys_type
  use tem_logging_module,       only: logUnit
  use tem_debug_module,         only: dbgUnit
  use tem_construction_module,  only: tem_levelDesc_type
  use tem_geometry_module,      only: tem_BaryOfId,tem_ElemSizeLevel
  use tem_stencil_module,       only: tem_stencilHeader_type
  use tem_stringKeyValuePair_module, only: init, truncate,                     &
    &                                 grw_stringKeyValuePairArray_type

  ! include aotus modules
  use aotus_module,     only: flu_State, aoterr_Fatal, aoterr_NonExistent,     &
    &                         aoterr_WrongType, aot_top_get_val, aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close

  ! include musubi modules
  use mus_field_prop_module,     only: mus_field_prop_type
  use mus_scheme_layout_module,  only: mus_scheme_layout_type
  use mus_derVarPos_type_module, only: mus_derVarPos_type
  ! use mus_param_module,          only: mus_param_type
  use mus_physics_module,        only: mus_physics_type
  use mus_mixture_module,        only: mus_mixture_type

  implicit none

  private

  public :: boundary_type
  public :: glob_boundary_type
  public :: mus_load_bc
  public :: mus_allocate_bc_elems
  public :: check_solid_in_bc
  public :: rearrange_bc_elems
  public :: debug_glob_boundary_type
  public :: mus_set_posInNghElems
  public :: mus_set_bcLinks
  public :: mus_set_bouzidi
  public :: mus_alloc_bouzidi
  public :: mus_alloc_fieldBC
  public :: mus_set_inletUbb
  public :: mus_set_inletBfl
  public :: mus_set_outletExpol
  public :: unload_boundary_type
  !public :: mus_set_localBC

  !> information needed for moments based boundary condition
  type bc_moments_type
    !> Number of unknown state links to update
    integer :: nUnKnownPdfs
    !> known moments position in moments vector
    integer, allocatable :: knownMom_pos(:)
    !> inverse matrix of unknown pdf matrix
    real(kind=rk), allocatable :: unKnownPdfs_MatInv(:,:)

  end type bc_moments_type

  !> Level wise boundary elements information
  type bc_elems_type

    !> Positions in levelDesc total list
    !! Its purpose is to get the treeID of BC elements
    !! size: globBC%nElems
    !! to use: levelDesc( iLevel )%total( globBC%elems( iLevel )%elem( iElem ) )
    integer, allocatable :: elem(:)

    !> Position of this boundary in the bc_elemBuffer
    !! size: globBC%nElems
    !! It is allcoated in update_BCLists of mus_construction_module
    !! Only non-wall BC needs elemBuffer
    integer, allocatable :: elemBuffer(:)

    !> Normal vector for each boundary element pointing into the domain.
    !! size: 3, nElems
    integer, allocatable :: normal(:,:)

    !> which is the index in the stencil corresponding to the normal vector?
    integer, allocatable :: normalInd(:)

    !> bit mask for each node holding directions which have to be updated.
    !! The bitmask points into the incoming direction into the flow domain, 
    !!  which actually we want to update
    !! * For PUSH, we write to the incoming position, 
    !!   as the kernel reads it from there without propagation.
    !! * For PULL, we need to write to the inverse direction, as the kernel 
    !!   performs a bounce back before reading it. 
    !!   However, this bounced back direction actually comes from the 
    !!   non-existent boundary element and would point into the incoming 
    !!   direction, so the value has to be treated and set as if it points 
    !!   really into the incoming direction.
    logical, allocatable :: bitmask(:,:)

    !> The q-Values for the exact wall positions
    !! Its size: QQN, nElemsBC
    !! It is allocated in routine: allocateBCList
    !! assigned in routine: assignBCList
    real(kind=rk), allocatable :: qVal(:,:)
    real(kind=rk), allocatable ::  cIn(:,:)
    real(kind=rk), allocatable :: cOut(:,:)
    real(kind=rk), allocatable :: cNgh(:,:)
    real(kind=rk), allocatable :: cVel(:,:)

  end type bc_elems_type

  !> contains information needed to treat corner nodes or nodes intersected
  !! by more than one boundary type
  type mus_bc_corner_type
    !> number of local boundary elements per level
    integer, allocatable :: nElems(:)
    !> boundary neighbor in different refinement level
    type( bc_elems_type ), allocatable :: elems(:)
    !> boundary id in each direction for each corner node
    !! size layout%stencil%QQ, nElems
    integer, allocatable :: bcid(:,:)
  end type mus_bc_corner_type

  !> Wall bouzidi data for one level
  type bc_wall_bouzidi_type

    !> size: links(iLevel)%nVals
    real(kind=rk), allocatable :: qVal(:)
    real(kind=rk), allocatable ::  cIn(:)
    real(kind=rk), allocatable :: cOut(:)
    real(kind=rk), allocatable :: cNgh(:)

    !> size: links(iLevel)%nVals
    !! set in routine: mus_set_bouzidi
    integer, allocatable ::  inPos(:) ! position in bcBuffer
    integer, allocatable :: outPos(:) ! position in bcBuffer
    integer, allocatable :: nghPos(:) ! position in computeNeighBuf

  end type bc_wall_bouzidi_type

  !> Provides bouzidi coefficients and q-values needed for link-wise
  !! implementation of certain inlet boundary conditions.
  type bc_inlet_bouzidi_type

    !> size: links(iLevel)%nVals
    real(kind=rk), allocatable :: qVal(:) ! q-Values
    real(kind=rk), allocatable ::  cIn(:) ! coefficient for incoming directions
    real(kind=rk), allocatable :: cOut(:) ! coefficient for outgoing directions
    real(kind=rk), allocatable :: cNgh(:) ! coefficient for neighbours

    !> size: links(iLevel)%nVals
    !! set in routine: mus_set_bouzidi
    integer, allocatable ::  inPos(:) ! position in bcBuffer
    integer, allocatable :: outPos(:) ! position in bcBuffer
    integer, allocatable :: nghPos(:) ! position in computeNeighBuf

    !> size: links(iLevel)%nVals
    real(kind=rk), allocatable ::    cVel(:) ! coefficient for eqPlus
    real(kind=rk), allocatable ::   coord(:,:) ! coordination of boundary node

    !> size: links(iLevel)%nVals
    !! set in routine: mus_set_bouzidi
    integer, allocatable :: iDir(:) ! incoming direction from boundary

  end type bc_inlet_bouzidi_type

  !> Provides needed positions for link-wise implementation of outlet boundary
  !> condition outlet_expol
  type bc_outlet_type

    !> size: links(iLevel)%nVals
    integer, allocatable ::  statePos(:) ! position of state
    integer, allocatable ::  iElem(:) ! position of state

  end type bc_outlet_type

!  type bc_local_type ! this is for the local boundary condition

!    !> size: links(iLevel)%nVals
!    real, allocatable :: K(:,:) ! K-matrix for all pairs of velocity indices
!    real, allocatable :: L(:,:) ! L matrix at least a diagonal matrix
!                                ! will have size of nLinks * nLinks
!
!  end type bc_local_type

  !> variable definition for non-reflective type of boundary conditions\n
  !! These boundary condition is taken from the paper:
  !!```lua
  !!   S. Izquierdo and N. Fueyo,
  !!   "Characteristic nonreflecting boundary
  !!   conditions for open boundaries in lattice Boltzmann methods,"
  !!   Physical Review E, vol. 78, no. 46707, 2008.
  !!```
  type bc_outlet_nrbc_type

    !> specific heat ratio
    !! appears in eq 3: cs = sqrt( kappa * R * T )
    real(kind=rk) :: kappa

    !> from eq 17: k = sigma * ( 1 - Ma^2 ) * cs / L
    real(kind=rk) :: K_mod

    !> used in eq 17
    real(kind=rk) :: cs_mod

    !> used in eq 17
    real(kind=rk) :: sigma

    !> length between inlet and outlet, represented by L in the paper
    real(kind=rk) :: lodi_length

  end type bc_outlet_nrbc_type

  !> Contains variables for black-box membrane boundary condition
  type bc_blackbox_membrane_type
    !> type of membrance (AEM or CEM)
    character(len=labelLen) :: label
    !> transference number defining transfer property
    !! of ionic species on the membrance
    !! KM @todo check transNr constrain
    !!          sumOfIonicSpecies(transNr) = 1
    real(kind=rk) :: transNr
    !> osmotic solvent transport coefficient
    real(kind=rk) :: solvTrans_coeff
  end type bc_blackbox_membrane_type

  !> contains all possible bc state variables needed for boundary
  type bc_states_type

    !> Velocity at boundary
    type(tem_bc_state_type) :: velocity

    !> mass flow rate at boundary node for inlet_mfr (inlet mass flow rate)
    !! boundary condition
    type(tem_bc_state_type) :: massFlowRate

    !> Pressure at boundary
    type(tem_bc_state_type) :: pressure

    !> mole fraction for species boundary
    type(tem_bc_state_type) :: moleFrac

    !> molar flux for species (n_i*u_i)
    type(tem_bc_state_type) :: moleFlux

    !> molar diffusion flux for species
    type(tem_bc_state_type) :: moleDiff_flux

    !> probability density function for boundary
    type(tem_bc_state_type) :: pdf

  end type bc_states_type

  !> Simple array type which include array and its size
  type array_type

    !> array
    integer, allocatable :: val(:)

    !> size
    integer :: nVals

  end type array_type


  !> Information about boundary elements
  !!
  !! This derived type encapsulates the definition of a boundary
  !! with all necessary parameters and data fields
  type glob_boundary_type

    !> name of this boundary
    character(len=LabelLen) :: label

    !> elements list of this boundary in different refinement level
    !! size: minLevel -> maxLevel
    type( bc_elems_type ), allocatable :: elems(:)

    !> list corner boundary elements
    type( mus_bc_corner_type ) :: cornerBC

    !> number of local (this process)    boundary elements per level
    integer, allocatable :: nElems(:)

    !> number of total (total processes) boundary elements per level
    integer, allocatable :: nElems_totalLevel(:)

    !> number of local boundary elements
    integer :: nElems_local = 0

    !> number of total boundary elements
    integer :: nElems_total = 0

    !> whether this boundary requires qVal
    logical :: hasQVal = .false.

    !> has qVal initialized with default qVal = 0.5
    logical :: qValInitialized = .false.

    !> if this is a wall BC (implicit treated during streaming)
    logical :: isWall = .false.

  end type glob_boundary_type

  !> Level wise information about neighbor of boundary elements
  !! for a single field
  type bc_neigh_type

    !> Neighbor position in level-wise state array
    !! size: nNeighs, nElems
    !! allocated in routine: setFieldBCNeigh, mus_construction_module
    integer, allocatable :: posInState(:,:)

    !> Pre-collision state values of neighbors on normal direction
    !! size: nNeighs, nElems*stencil%QQ
    !! It is allocated in routine update_BClists
    !! It use AOS data layout (iElem-1)%QQ+iDir
    real(kind=rk), allocatable :: neighBufferPre(:,:)

    !> Post-coliision state values of neighbors on normal direction
    !! size: nNeighs, nElems*stencil%QQ
    !! It is allocated in routine update_BClists
    !! It is filled in fill_neighBuffer
    real(kind=rk), allocatable :: neighBufferPost(:,:)

    !> Post-collision state values of neighbors on all directions
    !! size: nElems * computeStencil%QQ
    !! It is allocated in routine update_BClists (mus_construction_module),
    !! filled in routine: fill_computeNeighBuf (mus_bc_general_module).
    !! it use AOS layout always!
    real(kind=rk), allocatable :: computeNeighBuf(:)

  end type bc_neigh_type

  !> Contains field specific element information for each level
  type bc_field_elems_type
    !> Position of stencil of each element in array of stencils 
    !! in layout.
    !! Unique stencil label for boundary stencils are created with boundary label
    !! and stencil%cxDir therefore each stencil is limited to one boundary type
    !! Dimension: globBC%nElems(iLevel)
    !! allocated in mus_build_BCStencils (mus_construction_module)
    integer, allocatable :: stencilPos(:)

    !> LODI array for the NRBC
    !! Dimension: 4, 3, globBC%nElems(iLevel)
    !! allocated in init_nrbc routine
    real(kind=rk), allocatable :: lodi(:,:,:)

    !> defined moments position in moments matrix and
    !! inverse of unknown pdf matrix
    !! Dimension: bc%nElems(iLevel)
    type( bc_moments_type ), allocatable :: moments(:)

    !> Position in LevelDesc%neigh%NghElems
    !! size: globBC%nElems(iLevel)
    !! allocated in mus_build_BCStencils routine, mus_construction_module
    integer, allocatable :: posInNghElems(:)

  end type bc_field_elems_type

  !> Boundary treatment data type.
  !! It include boundary treatment for only ONE field.
  !! For the same boundary elements, each field may have a different treatment.
  !! It includes function pointers and neighbor elements required by this
  !! boundary treatment
  type boundary_type

    !> Boundary ID
    integer :: bc_id

    !> kind of this boundary
    character(len=LabelLen) :: BC_kind

    !> name of this boundary
    character(len=LabelLen) :: label  

    !> number of neighbors fluid element needed for extrapolation to
    !! get data at boundary element
    integer :: nNeighs =  0

    !> Level wise neighbor information of boundary elements
    !! Each field may have a different handlement on the same boundary
    !! Thus requires a different neighborhood
    !! allocated in mus_alloc_fieldBC
    type( bc_neigh_type ), allocatable :: neigh(:)

    !> Level wise element information of boundary elements
    !! like stencil position in array of stencils,
    !! moments type for momentBC and LODI array for NRBC
    !! allocated in mus_alloc_fieldBC
    type( bc_field_elems_type ), allocatable :: elems(:)

    !> function pointer to boundary routine to call
    procedure( boundaryRoutine ),pointer :: fnct => null()

    !> musubi bc state variables
    type( bc_states_type ) :: BC_states

    !> Dictionary of boundary variables with 
    !! varDict%val()%key is the name of boundary variable and
    !! varDict%val()%value is the name of spacetime function variable
    type(grw_stringKeyValuePairArray_type) :: varDict

    !> order of boundary which determines number of neighbor need for this boundary
    integer :: order

    !> standard q-value for higher order boundaries without any definition
    !! in seeder
    real(kind=rk) :: qVal

    !> Non-reflective boundary type
    type( bc_outlet_nrbc_type ) :: nrbc

    !> slip wall factor
    real(kind=rk) :: slip_fac

    !> black box membrane boundary
    type( bc_blackbox_membrane_type ) :: blackbox_mem

    !> whether this boundary needs neighbors on compute stencil
    logical :: useComputeNeigh = .false.

    !> Is true if this boundary requires boundary variables
    !! on boundary surface linkwise.
    !! Required to compute real coordinates points on boundary surface 
    logical :: evalBcVar_link = .false.

    !> Is true only if boundary requires neighbuf Pre-collision values
    logical :: requireNeighBufPre = .false.

    !> Is true only if boundary requires neighbuf Post-collision values
    logical :: requireNeighBufPost = .false. 

    !> Level-wise links that are to be updated
    !! size: minLevel:maxLevel
    !! allocated in mus_alloc_fieldBC
    type( array_type ), allocatable :: links(:)

    !> Level wise wall bouzidi data
    !! size: minLevel:maxLevel
    type( bc_wall_bouzidi_type ), allocatable :: bouzidi(:)

    !> Level wise inletUbbQVal data
    !! size: minLevel:maxLevel
    type( bc_inlet_bouzidi_type ), allocatable :: inletUbbQVal(:)

    !> Level wise inletBfl data
    !! size: minLevel:maxLevel
    type( bc_inlet_bouzidi_type ), allocatable :: inletBFL(:)

    !> Level wise outletExpol data
    !! size: minLevel:maxLevel
    type( bc_outlet_type ), allocatable :: outletExpol(:)

  end type boundary_type

  !> Interface definition for the boundary condition subroutines
  abstract interface
    ! This interface definition must match all the boundary routine interfaces.
    ! these are all defined via the macro bc_routineHeader( bc_name )
    subroutine boundaryRoutine( me, state, bcBuffer, globBC, levelDesc, tree, nSize, &
      & iLevel, sim_time, neigh, layout, fieldProp, varPos, nScalars,&
      & varSys, derVarPos, physics, iField, mixture )
      import :: boundary_type, rk, mus_field_prop_type, mus_scheme_layout_type,&
        & tem_time_type, tem_leveldesc_type, treelmesh_type, mus_derVarPos_type,  &
        & tem_varSys_type, glob_boundary_type, mus_physics_type, mus_mixture_type
      !> field boundary type
      class( boundary_type ) :: me
      !> State array
      real(kind=rk), intent(inout) :: state(:)
      !> size of state array ( in terms of elements )
      integer, intent(in) :: nSize
      !> state values of boundary elements of all fields of iLevel
      real(kind=rk),intent(in) :: bcBuffer(:)
      !> gloabl type contains iLevel descriptor
      type( tem_levelDesc_type ), intent(in) :: levelDesc
      !> global treelm mesh
      type( treelmesh_type ), intent(in) ::tree
      !> level which invokes boundary
      integer,intent(in) :: iLevel
      !> global time information
      type( tem_time_type ), intent(in)  :: sim_time
      !> global parameters
      ! type(mus_param_type),intent(in) :: params
      integer,intent(in)  :: neigh(:)  !< connectivity array
      !> scheme layout
      type( mus_scheme_layout_type ),intent(in) :: layout
      !> Fluid property
      type( mus_field_prop_type ), intent(in)  :: fieldProp
      !> pointer to field variable in the state vector
      integer, intent(in) :: varPos(:)
      !> number of Scalars in the scheme var system
      integer, intent(in) :: nScalars
      !> scheme variable system
      type( tem_varSys_type ), intent(in) :: varSys
      !> position of derived quantities in varsys
      type( mus_derVarPos_type ), intent(in) :: derVarPos
      !> scheme global boundary type
      type( glob_boundary_type ), intent(in) :: globBC
      !> scheme global boundary type
      type( mus_physics_type ), intent(in) :: physics
      !> current field
      integer, intent(in) :: iField
      !> mixture info
      type(mus_mixture_type), intent(in) :: mixture
    end subroutine boundaryRoutine
  end interface

contains

! ****************************************************************************** !
  !> Read in the boundary conditions from the LUA parameter file.\n
  !!
  subroutine mus_load_bc( me, bc_prop, conf, parent, varSys, stencil )
    !---------------------------------------------------------------------------
    !> field boundary type
    type( boundary_type ), allocatable   :: me(:)
    !> boundary data from mesh
    type( tem_bc_prop_type ), intent(in) :: bc_prop
    !> lua flu state
    type( flu_state )               :: conf
    !> bc parent handle
    integer, intent(in), optional   :: parent
    !> Global variable system required to append annoymous boundary variables
    type(tem_varSys_type), intent(inout) :: varSys
    !> stencil information
    type(tem_stencilHeader_type) :: stencil
    !---------------------------------------------------------------------------
    integer :: bc_handle
    integer :: sub_handle
    integer :: iError
    integer :: errFatal(3)
    ! loop variable
    integer :: iBC, myBCID
    ! boundary header type
    type( tem_bc_header_type ) :: bc_header
    ! ---------------------------------------------------------------------------
    errFatal = aoterr_Fatal

    ! if fluid informations in scheme table parentHandle /= 0
    ! load bounary labels  in lua config, and match them with those of
    ! loaded from mesh (information in tem_bc_prop_type)
    ! If a BC in bc_prop can not be match by lua, then code abort
    ! If a BC in lua can not be match by bc_prop, it gives a warning.
    call tem_load_bc_header( me           = bc_header, &
      &                      conf         = conf,      &
      &                      parentHandle = parent,    &
      &                      BC_prop      = bc_prop   )
    call aot_table_open( L       = conf,                &
      &                  parent  = parent,              &
      &                  thandle = bc_handle,           &
      &                  key     = 'boundary_condition' )

    call tem_horizontalSpacer(fUnit = logUnit(1))
    write(logUnit(1),'(A,I0)') ' Loading boundary definition for nBCs: ', &
      &                        bc_header%nBCs

    ! load tables inside the boundary table for bc kind and their state
    ! variables boundaries defined in lua are match with bc label read from
    ! mesh using bc_header
    allocate( me( bc_header%nBCs ))

    do iBC = 1, bc_header%nBCs
      myBCID = bc_header%BC_ID(iBC)
      call aot_table_open( L       = conf,                                   &
        &                  parent  = bc_handle,                              &
        &                  thandle = sub_handle,                             &
        &                  pos     = iBC )
      ! Skip BC definitions that are not present in the mesh.
      if ( (sub_handle /= 0) .and. (myBCID > 0) ) then
        ! matching the boundary label from mesh to boundary label read for lua
        ! file bc_header%BC_ID(iBC) returns the position of boundary label iBC
        ! in the mesh.
        me( myBCID )%bc_id = myBCID
        me( myBCID )%label   = bc_header%label(iBC)
        me( myBCID )%BC_kind = bc_header%BC_kind(iBC)

        ! Initialize varDict for current boundary
        call init( me = me(myBCID)%varDict )

        write(logUnit(1),"(A,I0)") '   Boundary: ',iBC
        select case( trim( bc_header%BC_kind(iBC) ))
          case( 'wall' )
            write(logUnit(1),*) '    Simple bounce back treatment for '//    &
              &  'no-slip wall '// trim(me( myBCID )%label)
            me( myBCID )%nNeighs = 0

          case( 'spc_bb_wall' )
            write(logUnit(1),*) '    Species bounce back treatment for '//    &
              &  'no-slip wall '// trim(me( myBCID )%label)
            me( myBCID )%nNeighs = 1

          case( 'moments_wall', 'spc_moments_wall' )
            write(logUnit(1),*) '    Moments wall treatment for '//    &
              &  'no-slip wall '// trim(me( myBCID )%label)
            me( myBCID )%nNeighs = 0

          case( 'spc_moments_outflow' )
            write(logUnit(1),*) '    Spc moments outflow treatment for '//    &
              &  trim(me( myBCID )%label)
            me( myBCID )%nNeighs = 1

          case( 'wall_multiReflection','wall_mr')
            write(logUnit(1),*) '    Wall multi reflection BC does NOT work!'
            call tem_abort()

            bc_header%BC_kind(iBC) = 'wall_multiReflection'
            me( myBCID )%BC_kind = bc_header%BC_kind(iBC)
            write(logUnit(1),*) '    Wall multi reflection for no-slip wall '&
              &             //trim(me( myBCID )%label)
            call aot_get_val( L       = conf,                                &
              &               thandle = sub_handle,                          &
              &               val     = me( myBCID )%qVal,                   &
              &               ErrCode = iError,                              &
              &               key     = 'qVal',                              &
              &               default = 0.5_rk )
            me( myBCID )%nNeighs = 2

          case( 'wall_linearInterpolation',     &
            &   'wall_libb', 'wall_libb_link',  &
            &   'wall_bouzidi' )
            bc_header%BC_kind(iBC) = 'wall_libb'
            me( myBCID )%BC_kind = bc_header%BC_kind(iBC)
            write(logUnit(1),*) '    Wall LIBB treatment: '// &
              &                 trim(me( myBCID )%label)
            call aot_get_val( L       = conf,                                &
              &               thandle = sub_handle,                          &
              &               val     = me( myBCID )%qVal,                   &
              &               ErrCode = iError,                              &
              &               key     = 'qVal',                              &
              &               default = 0.5_rk )
            me( myBCID )%nNeighs = 0
            me( myBCID )%useComputeNeigh = .true.

          case( 'slip_wall', 'spc_slip_wall' )
            write(logUnit(1),*) '   '//trim( bc_header%BC_kind(iBC) ) //     &
              &    ' treatment for '//trim(me( myBCID )%label)
            call aot_get_val( L       = conf,                                &
              &               thandle = sub_handle,                          &
              &               val     = me( myBCID )%slip_fac,               &
              &               ErrCode = iError,                              &
              &               key     = 'fac',                               &
              &               default = 1.0_rk )
            me( myBCID )%nNeighs = 0
          
          case( 'bc_pdf' )
            write(logUnit(1),*) '   '//trim( bc_header%BC_kind(iBC) ) //     &
              &    ' treatment for '//trim(me( myBCID )%label)

            call tem_load_bc_state( bc         = me(myBCID)%BC_states%pdf, &
              &                     state_name = 'pdf',                    &
              &                     nComp      = stencil%QQ,               &
              &                     conf       = conf,                     &
              &                     bc_handle  = sub_handle,               &
              &                     varDict    = me(myBCID)%varDict,       &
              &                     varSys     = varSys                    )

          case( 'inlet_ubb', 'inlet_eq', 'spc_inlet_eq', 'spc_outlet_vel',   &
            &   'spc_vel_bb', 'moments_vel', 'moments_inflow', 'spc_inlet',  &
            &   'spc_moments_vel', 'spc_bb_vel_test', 'inlet_bfl', 'vel_neq' )
            write(logUnit(1),*) '   '//trim( bc_header%BC_kind(iBC) ) //     &
              &    ' treatment for '//trim(me( myBCID )%label)

            call tem_load_bc_state( bc         = me(myBCID)%BC_states &
              &                                            %velocity, &
              &                     state_name = 'velocity',          &
              &                     nComp      = 3,                   &
              &                     conf       = conf,                &
              &                     bc_handle  = sub_handle,          &
              &                     varDict    = me(myBCID)%varDict,  &
              &                     varSys     = varSys               )

            call aot_get_val( L       = conf,                                &
              &               thandle = sub_handle,                          &
              &               val     = me( myBCID )%order,                  &
              &               ErrCode = iError,                              &
              &               key     = 'order',                             &
              &               default = 1 )

            me(myBCID)%nNeighs = me(myBCID)%order

            ! load molefraction in addition for spc_inlet_eq
            select case( trim( bc_header%BC_kind(iBC) ))
              case( 'spc_inlet_eq', 'spc_inlet' )
                call tem_load_bc_state( bc         = me(myBCID)%BC_states &
                  &                                            %moleFrac, &
                  &                     state_name = 'mole_fraction',     &
                  &                     nComp      = 1,                   &
                  &                     conf       = conf,                &
                  &                     bc_handle  = sub_handle,          &
                  &                     varDict    = me(myBCID)%varDict,  &
                  &                     varSys     = varSys               )

              case( 'moments_inflow' )
                call tem_load_bc_state( bc         = me(myBCID)%BC_states &
                  &                                            %pressure, &
                  &                     state_name = 'pressure',          &
                  &                     nComp      = 1,                   &
                  &                     conf       = conf,                &
                  &                     bc_handle  = sub_handle,          &
                  &                     varDict    = me(myBCID)%varDict,  &
                  &                     varSys     = varSys               )

              !case ('inlet_bouzidi', 'inlet_ubb_qval', 'inlet_bfl')
              case ('inlet_ubb', 'inlet_bfl')
                write(logUnit(1),*) '    Inlet higher order with linear'   //  &
                  &             ' interpolation treatment for velocity bc '//  &
                  &            trim(me( myBCID )%label)
                call aot_get_val( L       = conf,                              &
                  &               thandle = sub_handle,                        &
                  &               val     = me( myBCID )%qVal,                 &
                  &               ErrCode = iError,                            &
                  &               key     = 'qVal',                            &
                  &               default = 0.5_rk )

                me( myBCID )%nNeighs = 0
                me( myBCID )%useComputeNeigh = .true.
            end select

          ! inlet mass flow rate boundary condition
          case( 'inlet_mfr', 'inlet_mfr_eq' )
            write(logUnit(1),*) '   '//trim( bc_header%BC_kind(iBC) ) //     &
              &                 ' treatment for '//trim(me( myBCID )%label)

            call tem_load_bc_state( bc         = me(myBCID)%BC_states     &
              &                                            %massFlowRate, &
              &                     state_name = 'mass_flowrate',         &
              &                     nComp      = 1,                       &
              &                     conf       = conf,                    &
              &                     bc_handle  = sub_handle,              &
              &                     varDict    = me(myBCID)%varDict,      &
              &                     varSys     = varSys                   )

            call aot_get_val( L       = conf,                                &
              &               thandle = sub_handle,                          &
              &               val     = me( myBCID )%order,                  &
              &               ErrCode = iError,                              &
              &               key     = 'order',                             &
              &               default = 1 )

            me(myBCID)%nNeighs = me(myBCID)%order

          case( 'outlet_eq', 'outlet_pab', 'outlet_bouzidi', 'press_neq', &
            &   'outlet_expol', 'moments_press', 'outlet_expol_slow' )
            select case( trim( bc_header%BC_kind(iBC) ))
            case( 'outlet_eq' )
              write(logUnit(1),*) '    Outlet pressure equilibrium type for '&
                &             //trim(me( myBCID )%label)
              me( myBCID )%requireNeighBufPre = .true.
              me( myBCID )%nNeighs = 2

            case( 'outlet_pab' )
              write(logUnit(1),*) '    Outlet pressure anti bounce back for '&
                &            //trim(me( myBCID )%label)
              me( myBCID )%requireNeighBufPre = .true.
              me( myBCID )%nNeighs = 1

            case( 'outlet_bouzidi' )
              write(logUnit(1),*) '    Outlet bouzidi boundary condition for'&
                &             //' '//trim(me( myBCID )%label)
              me( myBCID )%nNeighs = 1

            case( 'press_neq' )
              write(logUnit(1),*) '    Pressure non-equilibrium boundary ' &
                &            //'condition for '//trim(me( myBCID )%label)
              me( myBCID )%nNeighs = 1

            case( 'outlet_expol', 'outlet_expol_slow' )
              write(logUnit(1),*) '  Outlet extrapolation type for '//  &
                &                      trim(me( myBCID )%label)
              me( myBCID )%requireNeighBufPre = .true.
              me( myBCID )%nNeighs = 2

            case( 'moments_press' )
              write(logUnit(1),*) '    Moments based pressure boundary '//   &
                &                 'type for '//                              &
                &                 trim(me( myBCID )%label)
              me( myBCID )%nNeighs = 0
            end select

            call tem_load_bc_state( bc         = me(myBCID)%BC_states &
              &                                            %pressure, &
              &                     state_name = 'pressure',          &
              &                     conf       = conf,                &
              &                     bc_handle  = sub_handle,          &
              &                     varDict    = me(myBCID)%varDict,  &
              &                     varSys     = varSys               )
          case( 'spc_outlet_eq', 'spc_outlet_expol' )
            select case( trim( bc_header%BC_kind(iBC) ))
            case( 'spc_outlet_eq' )
              write(logUnit(1),*) '    Species outlet pressure '//           &
                &                 'equilibrium type for '//                  &
                &                  trim(me( myBCID )%label)
              me( myBCID )%nNeighs = 2

            case( 'spc_outlet_expol' )
              write(logUnit(1),*) '    Species outlet extrapolation type for ' &
                &                 // trim(me( myBCID )%label)
              me( myBCID )%requireNeighBufPre = .true.
              me( myBCID )%nNeighs = 2
            end select

          case( 'outlet_nrbc', 'outlet_nrbc_eq' )
            me( myBCID )%requireNeighBufPost = .true.
            select case( trim( bc_header%BC_kind(iBC) ))
            case( 'outlet_nrbc' )
              write(logUnit(1),*) '    Characteristic based non-reflective ' &
                &                 //'boundary condition '//   &
                &                 trim(me( myBCID )%label)
              me( myBCID )%nNeighs = 2

            case( 'outlet_nrbc_eq' )
              write(logUnit(1),*) '    Eq-type Characteristic based '   //   &
                &             'non-reflective boundary condition '      //   &
                &             trim(me( myBCID )%label)
              me( myBCID )%nNeighs = 2
            end select

            call aot_get_val( L       = conf,                                &
              &               thandle = sub_handle,                          &
              &               val     = me( myBCID )%nrbc%kappa,             &
              &               ErrCode = iError,                              &
              &               key     = 'kappa',                             &
              &               default = 1.0_rk )
            call aot_get_val( L       = conf,                                &
              &               thandle = sub_handle,                          &
              &               val     = me( myBCID )%nrbc%sigma,             &
              &               ErrCode = iError,                              &
              &               key     = 'sigma',                             &
              &               default = 0.0001_rk )
            call aot_get_val(                                                &
              &         L       = conf,                                      &
              &         thandle = sub_handle,                                &
              &         val     = me( myBCID )%nrbc%lodi_length,             &
              &         ErrCode = iError,                                    &
              &         key     = 'L',                                       &
              &         default = 100._rk )
            write(logUnit(1),"(A,F9.2)")' kappa: ', me( myBCID )%nrbc%kappa
            write(logUnit(1),"(A,F9.5)")' sigma: ', me( myBCID )%nrbc%sigma
            write(logUnit(1),"(A,F9.2)")' lodi L:', me( myBCID )%nrbc%lodi_length

            call tem_load_bc_state( bc         = me(myBCID)%BC_states &
              &                                            %pressure, &
              &                     state_name = 'pressure',          &
              &                     conf       = conf,                &
              &                     bc_handle  = sub_handle,          &
              &                     varDict    = me(myBCID)%varDict,  &
              &                     varSys     = varSys               )

          case( 'outlet_zero_prsgrd', 'spc_outlet_zero_prsgrd' )
            write(logUnit(1),*) '    Outlet zero pressure gradient for '  // &
              &                      trim(me( myBCID )%label)
            me( myBCID )%nNeighs = 1

          case( 'outlet_dnt' )
            write(logUnit(1),*) '    Outlet do-nothing type for '// &
              &                      trim(me( myBCID )%label)
            me( myBCID )%requireNeighBufPre = .true.
            me( myBCID )%nNeighs = 2

          case( 'moments_outflow' )
            write(logUnit(1),*) '    Outlet moments type for '// &
              &                      trim(me( myBCID )%label)
            me( myBCID )%requireNeighBufPre = .true.
            me( myBCID )%nNeighs = 2

          case( 'spc_molefrac', 'spc_molefrac_wtdf', 'spc_moments_molefrac', &
            &   'spc_molefrac_eq' )
            write(logUnit(1),*) '    Species '//trim(bc_header%BC_kind(iBC)) &
              &                 //'  treatment for '//trim(me( myBCID )%label)

            call tem_load_bc_state( bc         = me(myBCID)%BC_states &
              &                                            %moleFrac, &
              &                     state_name = 'mole_fraction',     &
              &                     conf       = conf,                &
              &                     bc_handle  = sub_handle,          &
              &                     varDict    = me(myBCID)%varDict,  &
              &                     varSys     = varSys               )

            me( myBCID )%nNeighs = 0

          case( 'spc_moleflux', 'spc_moleflux_eq', 'spc_moments_moleflux' )
            write(logUnit(1),*) '    Species molar flux treatment for '//    &
              &                 trim(me( myBCID )%label)

            call tem_load_bc_state( bc         = me(myBCID)%BC_states &
              &                                            %moleFlux, &
              &                     state_name = 'mole_flux',         &
              &                     nComp      = 3,                   &
              &                     conf       = conf,                &
              &                     bc_handle  = sub_handle,          &
              &                     varDict    = me(myBCID)%varDict,  &
              &                     varSys     = varSys               )

            me( myBCID )%nNeighs = 0

          case( 'spc_blackbox_mem_ion', 'spc_blackbox_mem_solvent' )
            write(logUnit(1),*) '    Species black box membrance '//         &
              &                 'treatment for '//trim(me( myBCID )%label)

            call aot_get_val( L       = conf,                                &
              &               thandle = sub_handle,                          &
              &               val     = me( myBCID )%blackbox_mem%transNr,   &
              &               ErrCode = iError,                              &
              &               key     = 'transference_number' )
            if( btest( iError, aoterr_Fatal ))then
              write(logUnit(1),*)'FATAL Error occured in definition of '//   &
                &            'black box membrance boundary condition'
              write(logUnit(1),*)'while retrieving transference number'
              write(logUnit(1),*)'STOPPING'
              call tem_abort()
            end if

            select case( trim( bc_header%BC_kind(iBC) ))
              case( 'spc_blackbox_mem_solvent' )
                call aot_get_val( L       = conf,                            &
                  &               thandle = sub_handle,                      &
                  &               val     = me( myBCID )       &
                  &                         %blackbox_mem%solvTrans_coeff,   &
                  &               ErrCode = iError,                          &
                  &               key     = 'transference_number' )
                if( btest( iError, aoterr_Fatal ))then
                  write(logUnit(1),*)'FATAL Error occured in definition of ' &
                    &            //'black box '//                            &
                    &            'membrance solvent boundary condition'
                  write(logUnit(1),*)'while retrieving solv_trans_coeff'
                  write(logUnit(1),*)'STOPPING'
                  call tem_abort()
                end if
            end select
          case( 'spc_molediff_flux' )
            write(logUnit(1),*) '    Species molar diffusion flux '//        &
              &                 'treatment for '//trim(me( myBCID )%label)

            call tem_load_bc_state( bc         = me(myBCID)%BC_states &
              &                                            %moleFrac, &
              &                     state_name = 'mole_fraction',     &
              &                     conf       = conf,                &
              &                     bc_handle  = sub_handle,          &
              &                     varDict    = me(myBCID)%varDict,  &
              &                     varSys     = varSys               )

            call tem_load_bc_state( bc         = me(myBCID)%BC_states      &
              &                                            %moleDiff_flux, &
              &                     state_name = 'mole_diff_flux',         &
              &                     nComp      = 3,                        &
              &                     conf       = conf,                     &
              &                     bc_handle  = sub_handle,               &
              &                     varDict    = me(myBCID)%varDict,       &
              &                     varSys     = varSys                    )

            me( myBCID )%nNeighs = 0

          case( 'flekkoy_outlet' )
            write(logUnit(1),*) '   Outlet BCs for the flekkoy '//           &
              &                 'diffusion model'
            me( myBCID )%nNeighs = 2

          case( 'flekkoy_inlet' )
            write(logUnit(1),*) '   Inlet BCs for the flekkoy '//            &
              &                 'diffusion model'
            me( myBCID )%nNeighs = 0

          case( 'test' )
            write(logUnit(1),*) '   WARNING!! ' 
            write(logUnit(1),*) '   Test boundary conditions for testing ' //&
              &             'neighbor info exchange' 
            me( myBCID )%nNeighs = 20

          case default
            write(logUnit(1),*) 'Unknown Boundary condition kind!'
            write(logUnit(1),*) 'Stopping ...' 
            call tem_abort()
        end select
        write(logUnit(1),"(A,I0)")'   required neighbors: ', me( myBCID )%nNeighs

        call truncate( me = me(myBCID)%varDict )
      end if  ! sub_handle/=0
      call aot_table_close( L=conf, thandle=sub_handle )
    end do ! iBC
    call tem_horizontalSpacer(fUnit = logUnit(1))

  end subroutine mus_load_bc
! ****************************************************************************** !


! ****************************************************************************** !
  subroutine mus_allocate_bc_elems( me, nElems, QQ, hasQVal )
    ! ---------------------------------------------------------------------------
    type( bc_elems_type ) :: me
    integer, intent(in) :: nElems, QQ
    logical, intent(in) :: hasQval
    ! ---------------------------------------------------------------------------

    if ( allocated( me%elem ) ) then
      deallocate( me%elem       )
      deallocate( me%elemBuffer )
      deallocate( me%normal     )
      deallocate( me%normalInd  )
      deallocate( me%bitMask    )
    end if

    allocate( me%elem(nElems)        )
    allocate( me%elemBuffer(nElems)  )
    allocate( me%bitMask(QQ, nElems) )
    allocate( me%normal(3, nElems)   )
    allocate( me%normalInd(nElems)   )

    if ( hasQVal ) then
      if ( allocated( me%cIn ) ) then
        deallocate( me%qVal )
        deallocate( me%cIn  )
        deallocate( me%cOut )
        deallocate( me%cNgh )
        deallocate( me%cVel )
      end if

      allocate( me%qVal(QQ-1,nElems) )
      allocate( me%cIn (QQ-1,nElems) )
      allocate( me%cOut(QQ-1,nElems) )
      allocate( me%cNgh(QQ-1,nElems) )
      allocate( me%cVel(QQ-1,nElems) )
      ! Initialize with zero
      me%qVal = 0.0_rk
      me%cIn  = 0.0_rk
      me%cOut = 0.0_rk
      me%cNgh = 0.0_rk
      me%cVel = 0.0_rk
    end if

  end subroutine mus_allocate_bc_elems
! ****************************************************************************** !

! ****************************************************************************** !
  !> It removes non-valid elements while still maintaining the origianl order.
  !! The given bc elements (elems) contains both valid and non-valid elements.
  !! Position of valid elements are given by posInBCElem.
  !! Valid elements are moved towards the start of the elems so that they become
  !! continuous in the elems.
  !!
  subroutine rearrange_bc_elems( minLevel, maxLevel, nValid, posInBCElem, &
    &                            nElems, elems )
    ! ---------------------------------------------------------------------------
    integer, intent(in) :: minLevel, maxLevel
    !> number of valid BC elements
    integer, intent(in) :: nValid(minLevel:maxLevel)
    !> BC elements information
    integer, intent(in) :: nElems(minLevel:maxLevel)
    !> their position in bc_elems_type
    integer, intent(in) :: posInBCElem(maxval(nElems),minLevel:maxLevel)
    !> BC elements information
    type( bc_elems_type ), intent(inout) :: elems(minLevel:maxLevel)
    ! ---------------------------------------------------------------------------
    integer :: iElem, iLevel, oldPos
    ! ---------------------------------------------------------------------------

    write(dbgUnit(1), *) 'Inside remove solid in bcElems routine ...'

    do iLevel = minLevel, maxLevel

      ! Re arrange bc elements only when any solid are found
      if ( nValid(iLevel) < nElems(iLevel) ) then

        do iElem = 1, nValid( iLevel )
          oldPos = posInBCElem(iElem,iLevel)
          elems(iLevel)%elem( iElem )      = elems(iLevel)%elem( oldPos )
          elems(iLevel)%bitMask( :,iElem ) = elems(iLevel)%bitmask( :,oldPos )
          elems(iLevel)%normal( :,iElem )  = elems(iLevel)%normal( :,oldPos )
          elems(iLevel)%normalInd( iElem ) = elems(iLevel)%normalInd( oldPos )
          write(dbgUnit(6), *) 'elem: ', elems( iLevel )%elem( iElem )
        end do

      end if ! nValid(iLevel) == nElems(iLevel)

    end do ! iLevel

    write(dbgUnit(1), *) 'Leave  remove solid in bcElems routine ...'
    write(dbgUnit(1), *) ''

  end subroutine rearrange_bc_elems
! ****************************************************************************** !

! ****************************************************************************** !
  !> It count valid (non-solid) elements in BC elements list.
  !! Input:
  !!   minLevel, maxLevel
  !!   LevelPointer
  !!   LevelDesc
  !!   nElems - number of BC elements
  !!   elems - positions of BC elements in tree or levelPointer
  !! Output:
  !!   nValid - number of valid BC elements
  !!   posInBCElem - positions of valid elements in BC elements list
  !!
  subroutine check_solid_in_bc( minLevel, maxLevel, levelPointer, levelDesc, &
    &                           nElems, elems, nValid, posInBCElem )
    ! ---------------------------------------------------------------------------
    integer, intent(in) :: minLevel, maxLevel
    !> Level Pointer ( position of a element in level desc )
    integer, intent(in) :: levelPointer(:)
    !> Level Descriptor
    type( tem_levelDesc_type ), intent(in) :: levelDesc(minLevel:maxLevel)
    !> number of BC elements
    integer, intent(in) :: nElems(minLevel:maxLevel)
    !> BC elements list that contains their position in levelPointer
    type( bc_elems_type ), intent(in) :: elems(minLevel:maxLevel)
    !> number of valid (non-solid) elements
    integer, intent(out) :: nValid(minLevel:maxLevel)
    !> positions of valid elements in globBC elements list
    integer, intent(out) :: posInBCElem(maxval(nElems),minLevel:maxLevel)
    ! ---------------------------------------------------------------------------
    integer :: iElem, iLevel, posInTotal
    ! ---------------------------------------------------------------------------

    write(dbgUnit(1), *) 'Inside check solid in bc routine ...'
    write(dbgUnit(1), *) 'If any solid are found, they are listed as below.'

    posInBCElem(:,:) = 0
    nValid(:) = 0

    do iLevel = minLevel, maxLevel

      do iElem = 1, nElems( iLevel )
        posInTotal = levelPointer( elems( iLevel )%elem( iElem ) )
        if ( btest( levelDesc( iLevel )%property( posInTotal ), prp_solid ) ) then
          write(dbgUnit(3), "(A,I2,2(A,I0))") 'found a Solid! level: ', iLevel, &
            &           ', posInTotal: ', posInTotal, &
            &           ', treeID: ', levelDesc( iLevel )%total( posInTotal )
        ! else if ( count( levelDesc( iLevel )%neigh(1)%nghElems(:,posInTotal) > 0 ) < 2 ) then
        !   write(logUnit(9), "(A)")    &
        !     &    ' found an element having less than two neighbors!'
        !   write(logUnit(9), "(A,I0)") &
        !     &    ' treeID: ',levelDesc( iLevel )%total( posInTotal )
        else
          nValid(iLevel) = nValid(iLevel) + 1
          posInBCElem( nValid(iLevel), iLevel ) = iElem
        end if
        write(dbgUnit(6), *) 'elem: ', elems( iLevel)%elem( iElem )

      end do ! iElem

      if ( nElems(iLevel) /= nValid( iLevel) ) then
        write(logUnit(5),"(2(A,I0))") ' Removed ', nElems(iLevel) - nValid(iLevel), &
          &                           ' elements on level ', iLevel
      end if
      write(dbgUnit(3), *) 'nElems: ', nElems( iLevel )
      write(dbgUnit(3), *) 'nValid: ', nValid( iLevel )

    end do ! iLevel

    write(dbgUnit(1), *) 'Leave  check solid in bc routine ...'
    write(dbgUnit(1), *) ''

  end subroutine check_solid_in_bc
! ****************************************************************************** !

! ****************************************************************************** !
  subroutine debug_glob_boundary_type( nBCs, minLevel, maxLevel, me )
    ! ---------------------------------------------------------------------------
    integer,                    intent(in) :: nBCs, minLevel, maxLevel
    type( glob_boundary_type ), intent(in) :: me( nBCs )
    ! ---------------------------------------------------------------------------
    integer :: iBC, iLevel, iElem
    ! ---------------------------------------------------------------------------

    write(dbgUnit(1),"(A)")'-- DEBUG glob_boundary_type  ----------------------'
    do iBC = 1, nBCs
      write(dbgUnit(5), "(I0,A)") iBC, ', BC label: '//trim(me(iBC)%label)
      write(dbgUnit(5), "(A,L)")  'hasQVal: ', me(iBC)%hasQVal
      write(dbgUnit(5), "(A,L)")  'qValInitialized: ', me(iBC)%qValInitialized
      write(dbgUnit(5), "(A,L)")  'isWall: ', me(iBC)%isWall
      write(dbgUnit(5), "(A,I0)") 'nElems_local: ', me(iBC)%nElems_local
      write(dbgUnit(5), "(A,I0)") 'nElems_total: ', me(iBC)%nElems_total
      if ( .not. me( iBC )%isWall ) then
        do iLevel = minLevel, maxLevel
          write(dbgUnit(5), "(A,I0)") 'iLevel: ', iLevel
          write(dbgUnit(5), "(A,I0)") 'nElems: ', me(iBC)%nElems(iLevel)
          write(dbgUnit(5), "(A,I0)") 'nElems_totalLevel: ', me(iBC)%nElems_totalLevel(iLevel)
          write(dbgUnit(5), "(4A10)") 'elem', 'elemBuf', 'normalInd', 'normal'
          do iElem = 1, me(iBC)%nElems(iLevel)
            write(dbgUnit(5), "(6I10)") me(iBC)%elems(iLevel)%elem(iElem), &
              &                         me(iBC)%elems(iLevel)%elemBuffer(iElem), &
              &                         me(iBC)%elems(iLevel)%normalInd(iElem), &
              &                         me(iBC)%elems(iLevel)%normal(1:3,iElem)
          end do ! iElem
          write(dbgUnit(1), "(A)") ''
        end do ! iLevel
      else ! isWall == true
        write(dbgUnit(5),"(A)") 'This is wall. No need to check bc elems'
      end if ! isWall
      write(dbgUnit(1), "(A)") ''
    end do ! iBC

  end subroutine debug_glob_boundary_type
! ****************************************************************************** !

! ****************************************************************************** !
!> Set BC elements positions in LevelDesc%neigh%nghElems
!!
  subroutine mus_set_posInNghElems( minLevel, maxLevel, nStencils, &
      &                             globBC, BC )
    ! ---------------------------------------------------------------------------
    integer, intent(in) :: minLevel, maxLevel, nStencils
    type( glob_boundary_type ), intent(in) :: globBC
    type( boundary_type ),   intent(inout) :: BC
    ! ---------------------------------------------------------------------------
    integer :: iLevel, iElem
    integer :: counter(minLevel:maxLevel,nStencils), stencilPos
    ! ---------------------------------------------------------------------------

    if ( BC%nNeighs > 0 ) then
      counter = 0
      do iLevel = minLevel, maxLevel
        do iElem = 1, globBC%nElems(iLevel)

          stencilPos = BC%elems(iLevel)%stencilPos(iElem)
          counter(iLevel, stencilPos) = counter(iLevel, stencilPos) + 1
          BC%elems(iLevel)%posInNghElems(iElem) = counter( iLevel, stencilPos )

        end do ! iElem
      end do ! iLevel
    end if

  end subroutine mus_set_posInNghElems
! ****************************************************************************** !

! ****************************************************************************** !
  subroutine mus_set_bcLinks( iField, QQ, nScalars, nElems, elems, nSize, &
    &                         neigh, links )
    ! ---------------------------------------------------------------------------
    integer,                intent(in) :: iField, QQ, nScalars
    !> number of BC elements
    integer,                intent(in) :: nElems, nSize
    type( bc_elems_type ),  intent(in) :: elems
    integer,                intent(in) :: neigh(:)
    type( array_type ),    intent(out) :: links
    ! ---------------------------------------------------------------------------
    integer :: iElem, iDir, iLink, posInTotal
    integer, allocatable :: dirs(:)
    ! ---------------------------------------------------------------------------
    allocate( dirs( nElems * QQ ) )
    iLink = 0

write(dbgUnit(7), "(A,I0)") 'nSize: ', nSize

    do iElem = 1, nElems
      posInTotal = elems%elem( iElem )
write(dbgUnit(7), "(2(A,I0))") 'iElem: ', iElem, ', posInTotal: ', posInTotal
      do iDir = 1, QQ
        if( elems%bitmask( iDir, iElem )) then
          iLink = iLink + 1
          dirs(iLink) = ?FETCH?( iDir, iField, posInTotal, QQ, nScalars, nSize, neigh )
        end if
      end do ! iDir
    end do ! iElem

    allocate( links%val( iLink ) )
    links%val(1:iLink) = dirs(1:iLink)
    links%nVals        = iLink

    deallocate( dirs )

  end subroutine mus_set_bcLinks
! ****************************************************************************** !

! ****************************************************************************** !
  !> Set necessary data for Wall Bouzidi BC
  !! bouzidi should be allocated beforehand
  subroutine mus_set_bouzidi( iLevel, QQ, nScalars, globBC, cxDirInv,          &
    &                         varPos, bouzidi )
    ! ---------------------------------------------------------------------------
    integer, intent(in) :: iLevel, QQ, nScalars
    type( glob_boundary_type ), intent(in) :: globBC
    integer, intent(in)  :: cxDirInv(QQ)
    integer, intent(in)  :: varPos(:)
    !> cIn, cOut, cNgh should have size of nLinks
    type( bc_wall_bouzidi_type ), intent(inout) :: bouzidi
    ! ---------------------------------------------------------------------------
    integer :: iLink, iDir, invDir, iElem, posInBCBuf
    ! ---------------------------------------------------------------------------

    iLink = 0

    do iElem = 1, globBC%nElems(iLevel)

      posInBCBuf = globBC%elems( iLevel )%elemBuffer( iElem )

      do iDir = 1, QQ
        if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
          invDir = cxDirInv(iDir)
          iLink = iLink + 1
          bouzidi%  qVal( iLink ) = globBC%elems( iLevel )%qVal( invDir, iElem )
          bouzidi%   cIn( iLink ) = globBC%elems( iLevel )%cIn(  invDir, iElem )
          bouzidi%  cOut( iLink ) = globBC%elems( iLevel )%cOut( invDir, iElem )
          bouzidi%  cNgh( iLink ) = globBC%elems( iLevel )%cNgh( invDir, iElem )
          ! inPos and outPos are used in bcBuffer
          bouzidi% inPos( iLink ) = varPos(iDir)   + ( posInBCBuf-1 ) * nScalars
          bouzidi%outPos( iLink ) = varPos(invDir) + ( posInBCBuf-1 ) * nScalars
          ! nghPos is used in computeNeighBuf
          bouzidi%nghPos( iLink ) = invDir + (iElem-1)*QQ
        end if
      end do ! iDir

    end do ! iElem

  end subroutine mus_set_bouzidi
! ****************************************************************************** !

! ****************************************************************************** !
  subroutine mus_alloc_bouzidi( me, nVals, minLevel, maxLevel )
    ! ---------------------------------------------------------------------------
    type( bc_wall_bouzidi_type ), allocatable :: me(:)
    integer, intent(in) :: minLevel, maxLevel
    integer, intent(in) :: nVals(minLevel:maxLevel)
    ! ---------------------------------------------------------------------------
    integer :: iLevel
    ! ---------------------------------------------------------------------------

    allocate( me( minLevel:maxLevel ) )
    write(logUnit(7), "(2(A,I0))") 'Allocated wall libb type level: ', &
      &                             minLevel, ' - ', maxLevel

    do iLevel = minLevel, maxLevel
      if ( nVals(iLevel) > 0 ) then
        write(logUnit(9), "(2(A,I0))") 'level: ', iLevel, ', nVal: ', nVals(iLevel)
      end if
      if ( nVals(iLevel) >= 0 ) then
        allocate( me(iLevel)%  qVal( nVals(iLevel) ) )
        allocate( me(iLevel)%   cIn( nVals(iLevel) ) )
        allocate( me(iLevel)%  cOut( nVals(iLevel) ) )
        allocate( me(iLevel)%  cNgh( nVals(iLevel) ) )
        allocate( me(iLevel)% inPos( nVals(iLevel) ) )
        allocate( me(iLevel)%outPos( nVals(iLevel) ) )
        allocate( me(iLevel)%nghPos( nVals(iLevel) ) )
        ! Intialize with zero
        me(iLevel)%  qVal=0.0_rk
        me(iLevel)%   cIn=0.0_rk
        me(iLevel)%  cOut=0.0_rk
        me(iLevel)%  cNgh=0.0_rk
        me(iLevel)% inPos=0
        me(iLevel)%outPos=0
        me(iLevel)%nghPos=0
      end if
    end do ! iLevel

  end subroutine mus_alloc_bouzidi
! ****************************************************************************** !

! ****************************************************************************** !
  subroutine mus_alloc_fieldBC( me, minLevel, maxLevel )
    ! ---------------------------------------------------------------------------
    type( boundary_type ) :: me
    integer, intent(in) :: minLevel, maxLevel
    ! ---------------------------------------------------------------------------

    if ( .not. allocated( me%elems ) ) allocate( me%elems(minLevel:maxLevel) )
    if ( .not. allocated( me%neigh ) ) allocate( me%neigh(minLevel:maxLevel) )
    if ( .not. allocated( me%links ) ) allocate( me%links(minLevel:maxLevel) )

  end subroutine mus_alloc_fieldBC
! ****************************************************************************** !

! ****************************************************************************** !
  !> Set necessary data for BC inlet_ubb_qval
  !! for bc with computed values at the boundary nodes instead of barycenters
  !! cIn, cOut, cNgh should have size of nLinks
  subroutine mus_set_inletUbb( me, tree, stencil, nScalars,      &
    &                          globBC, levelDesc, varPos, nLinks,         &
    &                          minLevel, maxLevel                          )
    ! ---------------------------------------------------------------------------
    !> setting type for bc
    type(bc_inlet_bouzidi_type), allocatable, intent(out) :: me(:)
    !> using mesh information
    type(treelmesh_type), intent(in)         :: tree
    !> for directions
    type(tem_stencilHeader_type), intent(in) :: stencil
    !> number of scalars
    integer, intent(in)                      :: nScalars
    !> for number of elements in boundary and position in buffer
    type(glob_boundary_type), intent(in)  :: globBC
    !> minimum and maximum level
    integer, intent(in)                   :: minLevel, maxLevel
    !> level descriptor
    type(tem_levelDesc_type), intent(in)  :: levelDesc(minLevel:maxLevel)
    !> for position of outgoing elements
    integer, intent(in)                   :: varPos(:)
    !> for linkwise treatment
    integer, intent(in)                   :: nLinks(minLevel:maxLevel)
    ! ---------------------------------------------------------------------------
    integer :: iLink, iDir, invDir, iElem, posInBCBuf
    integer :: iLevel
    real(kind=rk) :: dx, qVal
    real(kind=rk), dimension(3) :: bary
    ! ---------------------------------------------------------------------------
    write(logUnit(10), "(A)") 'Setting inletUbb type ...'

    allocate( me(minLevel:maxLevel) )

    ! loop over levels
    do iLevel = minLevel, maxLevel
      if (nLinks(iLevel) > 0) then
        write(logUnit(10), "(2(A,I0))") 'level: ', iLevel, ', nVal: ', nLinks(iLevel)
      end if
      if (nLinks(iLevel) >= 0) then
        ! allocate arrays
        allocate( me(iLevel)%  qVal(nLinks(iLevel)) )
        allocate( me(iLevel)%outPos(nLinks(iLevel)) )
        allocate( me(iLevel)%  iDir(nLinks(iLevel)) )
        allocate( me(iLevel)% coord(nLinks(iLevel), 3) )
        ! Intialize arrays with zero
        do iLink = 1, nLinks(iLevel)
          me(iLevel)%  qVal(iLink) = 0.0_rk
          me(iLevel)%outPos(iLink) = 0
          me(iLevel)%  iDir(iLink) = 0
          me(iLevel)%coord (iLink,1) = 0.0_rk
          me(iLevel)%coord (iLink,2) = 0.0_rk
          me(iLevel)%coord (iLink,3) = 0.0_rk
        end do
      end if

      ! fill arrays
      iLink = 0
      dx    = tem_ElemSizeLevel(tree, iLevel)

      ! loop over elements
      do iElem = 1, globBC%nElems(iLevel)
        ! position in buffer and position of barycenter
        posInBCBuf = globBC%elems(iLevel)%elemBuffer(iElem)
        bary(:)    = tem_BaryOfId(tree, levelDesc(iLevel)        &
          &                        %total(globBC%elems(iLevel)   &
          &                        %elem(iElem) ) )

        ! loop over directions
        do iDir = 1, stencil%QQN
          !> Bitmask is true for incoming direction
          if ( globBC%elems(iLevel)%bitmask(iDir, iElem) ) then
            invDir  = stencil%cxDirInv(iDir)
            iLink   = iLink + 1
            ! qValues
            qVal    = globBC%elems(iLevel)%qVal(invDir, iElem)
            !me(iLevel)%   qVal(iLink) = qVal needed here?
            ! position for outgoing elememts
            me(iLevel)% outPos(iLink) = varPos(invDir)            &
              &                                   + (posInBCBuf-1) * nScalars
            ! store directions to access stencil weight and cxDirRK
            me(iLevel)%   iDir(iLink) = iDir
            ! position of boundary surface 
            me(iLevel)%coord(iLink, :) = bary(:) - dx * qVal    &
              &                                  * stencil%cxDirRK(:, iDir)

          end if
        end do ! iDir

      end do ! iElem

    end do ! iLevel

  end subroutine mus_set_inletUbb
! ****************************************************************************** !

! ****************************************************************************** !
  subroutine mus_set_inletBfl( me, tree, stencil, nScalars, globBC,      &
    &                          levelDesc, varPos, nLinks, minLevel, maxLevel  )
    ! ---------------------------------------------------------------------------
    !> setting type for bc
    type(bc_inlet_bouzidi_type), allocatable, intent(out) :: me(:)
    !> using mesh information
    type(treelmesh_type), intent(in)         :: tree
    !> for directions
    type(tem_stencilHeader_type), intent(in) :: stencil
    !> number of scalars
    integer, intent(in)                      :: nScalars
    !> for number of elements in boundary and position in buffer
    type(glob_boundary_type), intent(in)     :: globBC
    !> minimum and maximum level
    integer, intent(in)                      :: minLevel, maxLevel
    !> level descriptor
    type(tem_levelDesc_type), intent(in)     :: levelDesc(minLevel:maxLevel)
    !> for position of elements
    integer, intent(in)                      :: varPos(:)
    !> for linkwise treatment
    integer, intent(in)          :: nLinks(minLevel:maxLevel)
    ! ---------------------------------------------------------------------------
    integer :: iLink, iDir, invDir, iElem, posInBCBuf
    integer :: iLevel
    real(kind=rk) :: dx, qVal
    real(kind=rk), dimension(3) :: bary
    ! ---------------------------------------------------------------------------

    write(logUnit(10), "(A)") 'Setting inletBFL ...'

    allocate( me(minLevel:maxLevel) )

    ! loop over levels
    do iLevel = minLevel, maxLevel
      if (nLinks(iLevel) > 0) then
        write(logUnit(10),"(2(A,I0))") 'level: ', iLevel, ', nVal: ', nLinks(iLevel)
      end if
      if (nLinks(iLevel) >= 0) then
        ! allocate arrays
        ! allocate( me(iLevel)%  qVal(nLinks(iLevel)) )
        allocate( me(iLevel)%   cIn(nLinks(iLevel)) )
        allocate( me(iLevel)%  cOut(nLinks(iLevel)) )
        allocate( me(iLevel)%  cNgh(nLinks(iLevel)) )
        allocate( me(iLevel)%  cVel(nLinks(iLevel)) )
        allocate( me(iLevel)% inPos(nLinks(iLevel)) )
        allocate( me(iLevel)%outPos(nLinks(iLevel)) )
        allocate( me(iLevel)%nghPos(nLinks(iLevel)) )
        allocate( me(iLevel)%  iDir(nLinks(iLevel)) )
        allocate( me(iLevel)% coord(nLinks(iLevel), 3) )
        ! Intialize arrays with zero
        ! me(iLevel)%  qVal = 0.0_rk
        do iLink = 1, nLinks(iLevel)
          me(iLevel)%   cIn(iLink) = 0.0_rk
          me(iLevel)%  cOut(iLink) = 0.0_rk
          me(iLevel)%  cNgh(iLink) = 0.0_rk
          me(iLevel)% inPos(iLink) = 0
          me(iLevel)%outPos(iLink) = 0
          me(iLevel)%nghPos(iLink) = 0
          me(iLevel)%  iDir(iLink) = 0
          me(iLevel)% coord(iLink,1) = 0.0_rk
          me(iLevel)% coord(iLink,2) = 0.0_rk
          me(iLevel)% coord(iLink,3) = 0.0_rk
        end do
      end if

      ! fill arrays
      iLink = 0
      dx    = tem_ElemSizeLevel( tree, iLevel )

      ! loop over elements
      do iElem = 1, globBC%nElems(iLevel)

        ! position in buffer and position of barycenter
        posInBCBuf = globBC%elems( iLevel )%elemBuffer( iElem )
        bary(:)    = tem_BaryOfId( tree, levelDesc( iLevel )      &
          &                        %total(globBC%elems ( iLevel ) &
          &                        %elem( iElem ) ) )

        ! loop over directions
        do iDir = 1, stencil%QQ
          !> Bitmask is true for incoming direction
          if( globBC%elems(iLevel)%bitmask( iDir, iElem )) then
            invDir  = stencil%cxDirInv(iDir)
            iLink   = iLink + 1
            ! qValues
            qVal = globBC%elems( iLevel )%qVal( invDir, iElem )
            ! coefficients for state values
            me(iLevel)%    cIn(iLink) = globBC%elems(iLevel) &
              &                               % cIn(invDir, iElem)
            me(iLevel)%   cOut(iLink) = globBC%elems(iLevel) &
              &                               %cOut(invDir, iElem)
            me(iLevel)%   cNgh(iLink) = globBC%elems(iLevel) &
              &                               %cNgh(invDir, iElem)
            me(iLevel)%   cVel(iLink) = globBC%elems(iLevel) &
              &                               %cVel(invDir, iElem)
            ! positions of incoming, outgoing and neighbour elemets
            ! inPos and outPos are used in bcBuffer
            me(iLevel)%  inPos(iLink) = varPos(iDir)                &
              &                               + (posInBCBuf-1) * nScalars
            me(iLevel)% outPos(iLink) = varPos(invDir)              &
              &                               + (posInBCBuf-1) * nScalars
            ! nghPos is used in computeNeighBuf
            me(iLevel)% nghPos(iLink) = invDir + (iElem-1)*stencil%QQ
            ! store directions to access stencil weight and cxDirRK
            me(iLevel)%   iDir(iLink) = iDir
            ! coord of boundary node
            me(iLevel)%coord(iLink, :) = bary(:) - dx * qVal    &
              &                              * stencil%cxDirRK(:, iDir)

          end if
        end do ! iDir

      end do ! iElem

    end do ! iLevel


  end subroutine mus_set_inletBfl
! ****************************************************************************** !


! ****************************************************************************** !
  subroutine mus_set_outletExpol( me, stencil, globBC,      &
    &                             nLinks, minLevel, maxLevel )
    ! ---------------------------------------------------------------------------
    !> setting type for bc
    type(bc_outlet_type), allocatable, intent(out) :: me(:)
    !> for directions
    type(tem_stencilHeader_type), intent(in) :: stencil
    !> number of scalars
    ! integer, intent(in)                      :: nScalars
    !> for number of elements in boundary and position in buffer
    type(glob_boundary_type), intent(in)     :: globBC
    !> minimum and maximum level
    integer, intent(in) :: minLevel, maxLevel
    !> for position of elements
    ! integer, intent(in)                      :: varPos(:)
    !> Number of links on each level
    integer, intent(in) :: nLinks(minLevel:maxLevel)
    ! ---------------------------------------------------------------------------
    integer :: iLevel, iLink, iElem, iDir
    ! ---------------------------------------------------------------------------

    write(logUnit(10), "(A)") 'Setting outlet expol ...'

    allocate( me(minLevel:maxLevel) )

    ! loop over levels
    do iLevel = minLevel, maxLevel
      if ( nLinks(iLevel) > 0) then
        write(logUnit(10), '(2(A,I0))') 'level: ', iLevel, &
          &                             ', nLinks: ', nLinks(iLevel)
      end if
      if ( nLinks(iLevel) >= 0) then
        ! allocate arrays
        allocate( me(iLevel)%statePos(nLinks(iLevel)) )
        allocate( me(iLevel)%iElem(nLinks(iLevel)) )
        ! Intialize arrays with zero
        me(iLevel)%statePos = 0
      end if

      iLink = 0

      ! loop over elements
      do iElem = 1, globBC%nElems(iLevel)
        ! position in buffer is needed afterwards
        ! posInBCBuf = globBC%elems(iLevel)%elemBuffer(iElem)

        ! loop over directions
        do iDir = 1, stencil%QQ
          ! check bitmask
          if ( globBC%elems(iLevel)%bitmask(iDir, iElem) ) then
            ! invDir  = stencil%cxDirInv(iDir)
            iLink   = iLink + 1
            ! store position
            me(iLevel)%statePos(iLink) = iDir + (iElem-1)*stencil%QQ
            me(iLevel)%iElem(iLink) = iElem
          end if ! check bitmask

        end do ! iDir

      end do ! iElem

    end do ! iLevel

  end subroutine mus_set_outletExpol
! ****************************************************************************** !

! ****************************************************************************** !
  subroutine unload_boundary_type( me, nBCs, minLevel, maxLevel )
    ! ---------------------------------------------------------------------------
    integer, intent(in) :: nBCs, minLevel, maxLevel
    type( boundary_type ), intent(inout) :: me(nBCs)
    ! ---------------------------------------------------------------------------
    integer :: iBC, iLevel
    ! ---------------------------------------------------------------------------

    write(dbgUnit(1),*) "Enter unload_boundary_type"

    do iBC = 1, nBCs
! write(logUnit(10),"(A,I0)") 'iBC: ', iBC
      if( me(iBC)%nNeighs > 0 ) then
! write(logUnit(10),"(A)") 'deallocate posInState, neighBUfferPre, neighBufferPost'
! write(logUnit(10),"(A)") 'deallocate stencilPos, posInNghElems'
        do iLevel = minLevel, maxLevel
          deallocate( me(iBC)%neigh( iLevel )%posInState )
          if( me(iBC)%requireNeighBufPre  )deallocate( me(iBC)%neigh( iLevel )%neighBufferPre )
          if( me(iBC)%requireNeighBufPost ) deallocate( me(iBC)%neigh( iLevel )%neighBufferPost )
          deallocate( me(iBC)%elems( iLevel )%stencilPos )
          deallocate( me(iBC)%elems( iLevel )%posInNghElems )
        end do
! write(logUnit(10),"(A)") 'done'
      end if

      if( me(iBC)%useComputeNeigh ) then
! write(logUnit(10),"(A)") 'deallocate computeNeighBuf'
        do iLevel = minLevel, maxLevel
          deallocate( me(iBC)%neigh( iLevel )%computeNeighBuf )
        end do
! write(logUnit(10),"(A)") 'done'
      end if

      do iLevel = minLevel, maxLevel

! write(logUnit(10),"(A,I0)") 'iLevel: ', iLevel
        ! @todo: for pure wall, following array is not allocated
        !        we should have better way to detect this.
        if ( allocated( me(iBC)%links( iLevel )%val ) ) then
! write(logUnit(10),"(A)") 'deallocate links val'
          deallocate( me(iBC)%links( iLevel )%val )
! write(logUnit(10),"(A)") 'done'
        end if

        if ( allocated( me(iBC)%elems( iLevel )%lodi ) ) then
! write(logUnit(10),"(A)") 'deallocate elems lodi'
          deallocate( me(iBC)%elems( iLevel )%lodi )
! write(logUnit(10),"(A)") 'done'
        end if

        if ( allocated( me(iBC)%elems( iLevel )%moments ) ) then
! write(logUnit(10),"(A)") 'deallocate elem moments'
          deallocate( me(iBC)%elems( iLevel )%moments )
! write(logUnit(10),"(A)") 'done'
        end if

        if ( allocated( me(iBC)%bouzidi ) ) then
! write(logUnit(10),"(A)") 'deallocate bouzidi qVal ...'
          deallocate( me(iBC)%bouzidi( iLevel )%qVal )
          deallocate( me(iBC)%bouzidi( iLevel )%cIn  )
          deallocate( me(iBC)%bouzidi( iLevel )%cOut )
          deallocate( me(iBC)%bouzidi( iLevel )%cNgh )
          deallocate( me(iBC)%bouzidi( iLevel )%inPos )
          deallocate( me(iBC)%bouzidi( iLevel )%outPos )
          deallocate( me(iBC)%bouzidi( iLevel )%nghPos )
! write(logUnit(10),"(A)") 'done'
        end if

        if ( allocated( me(iBC)%inletUbbQVal ) ) then
! write(logUnit(10),"(A)") 'deallocate inletUBBqVal qVal ...'
          deallocate( me(iBC)%inletUbbQVal( iLevel )%qVal )
          ! deallocate( me(iBC)%inletUbbQVal( iLevel )%cIn )
          ! deallocate( me(iBC)%inletUbbQVal( iLevel )%cOut )
          ! deallocate( me(iBC)%inletUbbQVal( iLevel )%cNgh )
          ! deallocate( me(iBC)%inletUbbQVal( iLevel )%inPos )
          deallocate( me(iBC)%inletUbbQVal( iLevel )%outPos )
          ! deallocate( me(iBC)%inletUbbQVal( iLevel )%nghPos )
          ! deallocate( me(iBC)%inletUbbQVal( iLevel )%cVel )
          deallocate( me(iBC)%inletUbbQVal( iLevel )%coord )
          deallocate( me(iBC)%inletUbbQVal( iLevel )%iDir )
! write(logUnit(10),"(A)") 'done'
        end if

        if ( allocated( me(iBC)%inletBFL ) ) then
! write(logUnit(10),"(A)") 'deallocate inletBFL qVal ...'
          ! deallocate( me(iBC)%inletBFL( iLevel )%qVal )
          deallocate( me(iBC)%inletBFL( iLevel )%cIn )
          deallocate( me(iBC)%inletBFL( iLevel )%cOut )
          deallocate( me(iBC)%inletBFL( iLevel )%cNgh )
          deallocate( me(iBC)%inletBFL( iLevel )%inPos )
          deallocate( me(iBC)%inletBFL( iLevel )%outPos )
          deallocate( me(iBC)%inletBFL( iLevel )%nghPos )
          deallocate( me(iBC)%inletBFL( iLevel )%cVel )
          deallocate( me(iBC)%inletBFL( iLevel )%coord )
          deallocate( me(iBC)%inletBFL( iLevel )%iDir )
! write(logUnit(10),"(A)") 'done'
        end if

        if ( allocated( me(iBC)%outletExpol ) ) then
! write(logUnit(10),"(A)") 'deallocate outletExpol'
          deallocate( me(iBC)%outletExpol( iLevel )%statePos )
          deallocate( me(iBC)%outletExpol( iLevel )%iElem )
! write(logUnit(10),"(A)") 'done'
        end if

      end do ! iLevel = minLevel, maxLevel

      if ( allocated( me(iBC)%links ) ) then
! write(logUnit(10),"(A)") 'deallocate neigh, elems, links'
        deallocate( me(iBC)%neigh )
        deallocate( me(iBC)%elems )
        deallocate( me(iBC)%links )
! write(logUnit(10),"(A)") 'done'
      end if

      if ( allocated( me(iBC)%bouzidi ) ) then
! write(logUnit(10),"(A)") 'deallocate bouzidi'
        deallocate( me(iBC)%bouzidi )
! write(logUnit(10),"(A)") 'done'
      end if
      if ( allocated( me(iBC)%inletUBBqVal ) ) then
! write(logUnit(10),"(A)") 'deallocate inletUBBqVal'
        deallocate( me(iBC)%inletUBBqVal )
! write(logUnit(10),"(A)") 'done'
      end if
      if ( allocated( me(iBC)%inletBFL ) ) then
! write(logUnit(10),"(A)") 'deallocate inletBFL'
        deallocate( me(iBC)%inletBFL )
! write(logUnit(10),"(A)") 'done'
      end if
      if ( allocated( me(iBC)%outletExpol ) ) then
! write(logUnit(10),"(A)") 'deallocate outletExpol'
        deallocate( me(iBC)%outletExpol )
! write(logUnit(10),"(A)") 'done'
      end if
    end do

    write(dbgUnit(1),*) "Done!"

  end subroutine unload_boundary_type
! ****************************************************************************** !


!! new subroutine for local boundary condition
!! compute here all preprocessing like K(i,k) and L(i,k)
!! ****************************************************************************** !
!  subroutine mus_set_localBC(                                                  &
!    &                                                                          &
!    &                                                                          )
!    ! ---------------------------------------------------------------------------
!    !> setting type for bc
!    type(bc_inlet_..._type), allocatable, intent(out) :: inlet...(:)
!    !> for number of elements in boundary and position in buffer
!    type(glob_boundary_type), intent(in)     :: globBC
!    !> minimum and maximum level
!    integer, intent(in)                      :: minLevel, maxLevel
!    !> for linkwise treatment
!    type(array_type), intent(in)          :: bcLinks(minLevel:maxLevel)
!    !> for directions
!    type(tem_stencilHeader_type), intent(in) :: stencil
!    ! ---------------------------------------------------------------------------
!    integer :: iLink, iDir, invDir, iElem, posInBCBuf
!    integer :: kLink, kDir, kinvDir
!    integer :: iLevel
!    integer :: alpha ! depends on the stencil dimension (1 ... 3)
!    real    :: theta ! [0,1] controls implicitness
!  
!!to compute size of matrix, use 'count' function of bc_general_module and count
!!links with bitmask==true
!!compute L matrix only for checked bitmask links, so loop not over all directions
!!but over all necessary links (iLink, kLink)
!!to compute inverse of matrix, use 'invert_matrix(L)' from tem_math_module
!!alpha, theta, dimension, kroneckerdelta
!
!  
!    write(logUnit(10), "(A)") 'Setting inlet... type .. .'
!
!    allocate( inlet...(minLevel:maxLevel) )
!    
!    ! loop over levels
!    do iLevel = minLevel, maxLevel
!      if (bcLinks(iLevel)%nVals >= 0) then
!        write(logUnit(10), *) 'level: ', iLevel, ', nVal: ', bcLinks(iLevel)%nVals
!        ! allocate arrays
!        ! Intialize arrays with zero
!
!      ! fill arrays
!      iLink = 0
!      kLink = 0
!      alpha = stencil%nDims
!
!      ! Factor for implicitness theta
!      theta = 0.7 ! theta = [0,1]
!
!      ! Evaluate K(i,k) for all pairs of velocity indices i,k:
!         do iElem = 1, globBC%nElems( iLevel )
!         
!            do iDir = 1, stencil%QQN
!              do kDir = 1, stencil%QQN    
!      
!            ! make formula easier to view:
!                c(:,iDir)     = stencil%cxDirRK(:,iDir)  
!                c(:,kDir)     = stencil%cxDirRK(:,kDir)
!                c(alpha,iDir) = stencil%cxDir(alpha,iDir)
!      
!                K(iDir,kDir) = -3.0_rk / 2.0_rk * ( 3.0_rk - 6.0_rk             &
!                  &          * globBC%elems(iLevel)%qVal(iDir, iElem) )         &
!                  &          * layout%weight(iDir)                              &
!                  &          * ( DOT_PRODUCT( c(:,iDir) , c(:,kDir) )** 2.0_rk  &
!                  &          - ABS( c(:,iDir) ) ** 2.0_rk / 3.0_rk              &
!                  &          - c(alpha,iDir) ** 2.0_rk                          &
!                  &          * ( ABS( c(:,kDir) ) ** 2.0_rk - stencil%nDims / 3.0_rk ) )
!      
!                ! Using Kronecker delta           
!                if ( iDir /= kDir ) then 
!                  delta = 0.0_rk
!                elseif ( iDir == kDir ) then
!                  delta = 1.0_rk
!                end if
!      
!      
!                ! Matrix L(i,k) only for incoming links. 
!                !Check for bitmask. 
!                if( globBC%elems(iLevel)%bitmask(kDir, iElem )) .and. &
!                & ( globBC%elems(iLevel)%bitmask(iDir, iElem )) then
!                  
!                  L(iDir,kDir) = delta + theta * K(iDir,kDir) ! L-Matrix
!                  kinvDir  = stencil%cxDirInv(kDir)
!                  kLink   = kLink + 1
!                
!                end if
!      
!      
!                ! Compute the inverse of L(i,k)
!                L_inv = invert_matrix(L) ! from tem_math_module
!      
!                ! store directions to access stencil weight and cxDirRK
!                inlet...(iLevel)%   kDir(kLink) = kDir
!                !Evaluate sigma_k(n,j) for all directions k
!                

!              end do ! kDir
!
!              invDir  = stencil%cxDirInv(iDir)
!              iLink   = iLink + 1
!              ! store directions to access stencil weight and cxDirRK
!              inlet...(iLevel)%   iDir(iLink) = iDir
!            end do ! iDir
!
!
!        ! element position in state array and treeID list
!        elemPos = globBC%elems(iLevel)%elem(iElem)
!        ! normal of this current node
!        normal = globBC%elems(iLevel)%normal(:,iElem)
!        ! Count the number of links which are incoming
!        nLinks = count(globBC%elems(iLevel)%bitmask(:, iElem))
!
! end subroutine mus_set_localBC
!! ****************************************************************************** !
end module mus_bc_header_module
! ****************************************************************************** !
