! See copyright notice in the COPYRIGHT file.
! ****************************************************************************** !
!> author: Jiaxing Qi
!! Routines and parameter definitions for the standard D3Q27 model
?? include 'header/lbm_macros.inc'
?? include 'header/lbm_interfaceMacros.inc'
?? include 'header/lbm_d3q27Macros.inc'
module mus_d3q27_module

  ! include treelm modules
  use env_module,            only: rk
  use tem_varSys_module,     only: tem_varSys_type, tem_varSys_op_type
  use tem_param_module,      only: div1_2, div1_3, div1_36, div1_8, div3_4h, div2_3
  use tem_logging_module,    only: logUnit

  ! include musubi modules
  use mus_field_prop_module,    only: mus_field_prop_type
  use mus_scheme_layout_module, only: mus_scheme_layout_type
  use mus_scheme_type_module,   only: mus_scheme_type
  use mus_param_module,         only: mus_param_type

  implicit none

  private

  public :: feq_d3q27
  public :: bgk_d3q27
  public :: central_moment
  public :: central_moment_split
  public :: cm_to_pdf
  public :: cumulant_d3q27
  public :: cascaded_d3q27
  public :: trt_d3q27

  integer,parameter :: QQ   = 27  !< number of pdf directions
?? copy :: q27

contains

! ****************************************************************************** !
  !> D3Q27 equilibirium definition in product form.
  !! Given in paper: The cumulant lattice Boltzmann equation in three dimensions:
  !! Theory and validation, Martin Geier .et al,
  !! Computers and Mathematics with Applicaitons, 2015.
  !! @todo: this can be optimized
  pure function feq_d3q27( i, j, k, rho, u, v, w ) result ( eq )
    !> i, j, k are direction index, taking the value of -1, 0, 1
    integer,       intent(in) :: i, j, k
    !> density and velocity
    real(kind=rk), intent(in) :: rho, u, v, w
    !> equilibirium density
    real(kind=rk)             :: eq
    real(kind=rk)             :: rho_part, u_part, v_part, w_part, cs

    cs = div1_3

    rho_part = -rho * (-1)**(abs(i)+abs(j)+abs(k)) &
      &        / ( (abs(i)+1) * (abs(j)+1) * (abs(k)+1) )

    u_part = abs(i) - 1 + cs + i*u + u*u
    v_part = abs(j) - 1 + cs + j*v + v*v
    w_part = abs(k) - 1 + cs + k*w + w*w

    eq = rho_part * u_part * v_part * w_part

  end function feq_d3q27
! ****************************************************************************** !


! ****************************************************************************** !
?? copy :: compute_routineHeader( bgk_d3q27 )
?? copy :: compute_routineParams
    ! ---------------------------------------------------------------------------
    integer :: iElem
    integer :: nScalars
?? copy :: var27( f )
    real(kind=rk) :: u, v, w ! , u2, v2, w2
    real(kind=rk) :: rho, rho_omg
    real(kind=rk) :: inv_rho
    real(kind=rk) :: omega, cmpl_o, fac
    real(kind=rk) :: X0, X1, XN
    real(kind=rk) :: Y0, Y1, YN
    real(kind=rk) :: Z0, Z1, ZN
    ! real(kind=rk) :: m200, m020, m002
    ! real(kind=rk) :: Gx, Gy, Gz
    real(kind=rk) :: sum1, sum2, sum3, sum4, sum5, sum6, sum7, sum8, sum9, sum0
    real(kind=rk) :: x0y0, x0y1, x0yn, x1y0, xny0, x1y1, x1yn, xny1, xnyn
    ! ---------------------------------------------------------------------------

    nScalars = varSys%nScalars

    omega = fieldProp(1)%fluid%omLvl( level )
    cmpl_o = 1._rk - omega
    fac = 4.5d0 - 2.25d0 * omega

!$omp do schedule(static)
    !CDIR NODEP
?? IF (SOA) THEN
?? copy :: dir_vector
?? ELSE
?? copy :: dir_novec
?? END IF
    nodeloop: do iElem = 1, nSolve

      fN00 = inState(?FETCH?( qN00, 1, iElem, QQ, nScalars, nElems,neigh ))
      f0N0 = inState(?FETCH?( q0N0, 1, iElem, QQ, nScalars, nElems,neigh ))
      f00N = inState(?FETCH?( q00N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f100 = inState(?FETCH?( q100, 1, iElem, QQ, nScalars, nElems,neigh ))
      f010 = inState(?FETCH?( q010, 1, iElem, QQ, nScalars, nElems,neigh ))
      f001 = inState(?FETCH?( q001, 1, iElem, QQ, nScalars, nElems,neigh ))

      f0NN = inState(?FETCH?( q0NN, 1, iElem, QQ, nScalars, nElems,neigh ))
      f0N1 = inState(?FETCH?( q0N1, 1, iElem, QQ, nScalars, nElems,neigh ))
      f01N = inState(?FETCH?( q01N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f011 = inState(?FETCH?( q011, 1, iElem, QQ, nScalars, nElems,neigh ))
      fN0N = inState(?FETCH?( qN0N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f10N = inState(?FETCH?( q10N, 1, iElem, QQ, nScalars, nElems,neigh ))
      fN01 = inState(?FETCH?( qN01, 1, iElem, QQ, nScalars, nElems,neigh ))
      f101 = inState(?FETCH?( q101, 1, iElem, QQ, nScalars, nElems,neigh ))
      fNN0 = inState(?FETCH?( qNN0, 1, iElem, QQ, nScalars, nElems,neigh ))
      fN10 = inState(?FETCH?( qN10, 1, iElem, QQ, nScalars, nElems,neigh ))
      f1N0 = inState(?FETCH?( q1N0, 1, iElem, QQ, nScalars, nElems,neigh ))
      f110 = inState(?FETCH?( q110, 1, iElem, QQ, nScalars, nElems,neigh ))

      fNNN = inState(?FETCH?( qNNN, 1, iElem, QQ, nScalars, nElems,neigh ))
      fNN1 = inState(?FETCH?( qNN1, 1, iElem, QQ, nScalars, nElems,neigh ))
      fN1N = inState(?FETCH?( qN1N, 1, iElem, QQ, nScalars, nElems,neigh ))
      fN11 = inState(?FETCH?( qN11, 1, iElem, QQ, nScalars, nElems,neigh ))
      f1NN = inState(?FETCH?( q1NN, 1, iElem, QQ, nScalars, nElems,neigh ))
      f1N1 = inState(?FETCH?( q1N1, 1, iElem, QQ, nScalars, nElems,neigh ))
      f11N = inState(?FETCH?( q11N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f111 = inState(?FETCH?( q111, 1, iElem, QQ, nScalars, nElems,neigh ))

      f000 = inState(?FETCH?( q000, 1, iElem, QQ, nScalars, nElems,neigh ))

      rho =   fNNN + fNN1 + fN1N + fN11 + f1NN + f1N1 + f11N + f111 &
        &   + f0NN + f0N1 + f01N + f011 &
        &   + fN0N + f10N + fN01 + f101 &
        &   + fNN0 + fN10 + f1N0 + f110 &
        &   + fN00 + f100 + f0N0 + f010 + f00N + f001 + f000

      ! inverse local density
      inv_rho = 1._rk / rho
      rho_omg = rho * omega

      sum1 = f10N - fN01
      sum2 = f101 - fN0N
      sum3 = f1N0 - fN10
      sum4 = f110 - fNN0
      sum5 = f1NN - fN11
      sum6 = f1N1 - fN1N
      sum7 = f11N - fNN1
      sum8 = f111 - fNNN
      sum9 = f01N - f0N1
      sum0 = f011 - f0NN

      u = (f100-fN00) + sum1 + sum2 + sum3 + sum4 + sum5 + sum6 + sum7 + sum8
      v = (f010-f0N0) + sum9 + sum0 - sum3 + sum4 - sum6 - sum5 + sum7 + sum8
      w = (f001-f00N) - sum9 + sum0 - sum1 + sum2 - sum7 - sum5 + sum6 + sum8

      ! transfer moments to velocities
      u = u * inv_rho
      v = v * inv_rho
      w = w * inv_rho

      ! u v square
      ! u2 = u*u
      ! v2 = v*v
      ! w2 = w*w

      !----------------------------------------------------------------
      ! ! calculate Galilean correction term, by equation A.13 and A.14
      ! m200 = ( c200 + c202 + c220 + m222 ) * inv_rho
      ! m020 = ( c020 + c220 + c022 + m222 ) * inv_rho
      ! m002 = ( c002 + c022 + c202 + m222 ) * inv_rho

      ! Gx = fac * u2 * ( m200 - div1_3 - u2 )
      ! Gy = fac * v2 * ( m020 - div1_3 - v2 )
      ! Gz = fac * w2 * ( m002 - div1_3 - w2 )
      !----------------------------------------------------------------

      ! X Y Z components of eq
      ! by equation A.19 - A.21
      X0 = -div2_3 + u*u ! + Gx
      X1 = - ( X0 + 1.0d0 + u ) * 0.5d0
      XN = X1 + u

      Y0 = -div2_3 + v*v ! + Gy
      Y1 = - ( Y0 + 1.0d0 + v ) * 0.5d0
      YN = Y1 + v

      Z0 = -div2_3 + w*w ! + Gz
      Z1 = - ( Z0 + 1.0d0 + w ) * 0.5d0
      ZN = Z1 + w

      X0 = -rho_omg * X0
      X1 = -rho_omg * X1
      XN = -rho_omg * XN

! fEq000 = X0 * Y0 * Z0
! fEq00N = X0 * Y0 * ZN
! fEq001 = X0 * Y0 * Z1
x0y0 = x0 * y0
outState( ?SAVE?( q000,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f000 + x0y0*z0
outState( ?SAVE?( q00N,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f00N + x0y0*zn
outState( ?SAVE?( q001,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f001 + x0y0*z1

! fEq010 = X0 * Y1 * Z0
! fEq011 = X0 * Y1 * Z1
! fEq01N = X0 * Y1 * ZN
x0y1 = x0 * y1
outState( ?SAVE?( q010,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f010 + X0Y1*Z0
outState( ?SAVE?( q011,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f011 + X0Y1*Z1
outState( ?SAVE?( q01N,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f01N + X0Y1*ZN

! fEq0N0 = X0 * YN * Z0
! fEq0N1 = X0 * YN * Z1
! fEq0NN = X0 * YN * ZN
x0yn = x0 * yn
outState( ?SAVE?( q0N0,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f0N0 + X0YN*Z0
outState( ?SAVE?( q0N1,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f0N1 + X0YN*Z1
outState( ?SAVE?( q0NN,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f0NN + X0YN*ZN

! fEq100 = X1 * Y0 * Z0
! fEq10N = X1 * Y0 * ZN
! fEq101 = X1 * Y0 * Z1
x1y0 = x1 * y0
outState( ?SAVE?( q100,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f100 + X1Y0*Z0
outState( ?SAVE?( q101,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f101 + X1Y0*Z1
outState( ?SAVE?( q10N,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f10N + X1Y0*ZN

! fEqN00 = XN * Y0 * Z0
! fEqN01 = XN * Y0 * Z1
! fEqN0N = XN * Y0 * ZN
xny0 = xn * y0
outState( ?SAVE?( qN00,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*fN00 + XNY0*Z0
outState( ?SAVE?( qN01,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*fN01 + XNY0*Z1
outState( ?SAVE?( qN0N,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*fN0N + XNY0*ZN

! fEq110 = X1 * Y1 * Z0
! fEq111 = X1 * Y1 * Z1
! fEq11N = X1 * Y1 * ZN
x1y1 = x1 * y1
outState( ?SAVE?( q110,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f110 + X1Y1*Z0
outState( ?SAVE?( q111,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f111 + X1Y1*Z1
outState( ?SAVE?( q11N,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f11N + X1Y1*ZN

! fEq1N0 = X1 * YN * Z0
! fEq1N1 = X1 * YN * Z1
! fEq1NN = X1 * YN * ZN
x1yn = x1 * yn
outState( ?SAVE?( q1N0,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f1N0 + X1YN*Z0
outState( ?SAVE?( q1N1,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f1N1 + X1YN*Z1
outState( ?SAVE?( q1NN,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*f1NN + X1YN*ZN

! fEqN10 = XN * Y1 * Z0
! fEqN11 = XN * Y1 * Z1
! fEqN1N = XN * Y1 * ZN
xny1 = xn * y1
outState( ?SAVE?( qN10,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*fN10 + XNY1*Z0
outState( ?SAVE?( qN11,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*fN11 + XNY1*Z1
outState( ?SAVE?( qN1N,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*fN1N + XNY1*ZN

! fEqNN0 = XN * YN * Z0
! fEqNN1 = XN * YN * Z1
! fEqNNN = XN * YN * ZN
xnyn = xn * yn
outState( ?SAVE?( qNN0,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*fNN0 + XNYN*Z0
outState( ?SAVE?( qNN1,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*fNN1 + XNYN*Z1
outState( ?SAVE?( qNNN,1,iElem,QQ,QQ,nElems,neigh )) = cmpl_o*fNNN + XNYN*ZN

    end do nodeloop
!$omp end do nowait

  end subroutine bgk_d3q27
! ****************************************************************************** !

! ****************************************************************************** !
  !> Calculating central moment by spliting among directions.
  !! This follows equations 43, 44, 45 in cumulent paper (Geier .et al 2015)
  !! We first do x direction for better performance.
  pure function central_moment_split( f, a, b, g, ux, uy, uz ) result ( kappa )
    !> PDF
    real(kind=rk), intent(in) :: f(-1:1, -1:1, -1:1)
    !> order of central moments
    integer, intent(in) :: a, b, g
    real(kind=rk), intent(in) :: ux, uy, uz
    ! ---------------------------------------------------------------------------
    real(kind=rk) :: ka(-1:1, -1:1), kb(-1:1)
    real(kind=rk) :: kappa
    integer :: ii, jj, kk
    real(kind=rk), parameter :: ii_rk(-1:1) = [ -1.d0, 0.d0, 1.d0 ]
    ! ---------------------------------------------------------------------------

    ka = 0.0d0
    kb = 0.0d0
    kappa = 0.0d0

    do kk = -1, 1
      do jj = -1, 1
        do ii = -1, 1
          ka(jj,kk) = ka(jj,kk) + f( ii, jj, kk ) * ( ( ii_rk(ii) - ux ) ** a )
        end do
        kb(kk) = kb(kk) + ka(jj,kk) * ( ( ii_rk(jj) - uy ) ** b )
      end do
      kappa = kappa + kb(kk) * ( ( ii_rk(kk) - uz ) ** g )
    end do

  end function
! ****************************************************************************** !

! ****************************************************************************** !
  !> Calculating central moment.
  !! This follows equations 21 in cumulent paper ( Geier .et al 2015 )
  pure function central_moment( f, a, b, g, ux, uy, uz ) result ( kappa )
    !> PDF
    real(kind=rk), intent(in) :: f(-1:1, -1:1, -1:1)
    !> order of central moments
    integer, intent(in) :: a, b, g
    real(kind=rk), intent(in) :: ux, uy, uz
    ! ---------------------------------------------------------------------------
    real(kind=rk) :: kappa
    integer :: ii, jj, kk
    real(kind=rk), parameter :: ii_rk(-1:1) = [ -1.d0, 0.d0, 1.d0 ]
    ! ---------------------------------------------------------------------------

    kappa = 0.0d0

    do kk = -1, 1
      do jj = -1, 1
        do ii = -1, 1
          kappa = kappa + f( ii, jj, kk ) * ( ( ii_rk(ii) - ux ) ** a )  &
            &                             * ( ( ii_rk(jj) - uy ) ** b )  &
            &                             * ( ( ii_rk(kk) - uz ) ** g )
        end do
      end do
    end do

  end function
! ****************************************************************************** !

! ****************************************************************************** !
?? copy :: compute_routineHeader( cumulant_d3q27 )
?? copy :: compute_routineParams
    ! ---------------------------------------------------------------------------
    integer :: iElem, ii, jj, kk, nScalars
    real(kind=rk) :: ux, uy, uz, rho, inv_rho, omega
    real(kind=rk) :: dxu, dyv, dzw, AA, BB, CC
    ! PDF
    real(kind=rk) :: f(-1:1,-1:1,-1:1)
    ! k = central moment, c = cumulant
    real(kind=rk) :: k(0:2,0:2,0:2), c(0:2,0:2,0:2)
    ! ---------------------------------------------------------------------------

    nScalars = varSys%nScalars
    ! this omega refers to the w1 in the paper
    ! other relaxation parameters are assumed to be 1, thus omitted
    omega = fieldProp(1)%fluid%omLvl( level )

    nodeloop: do iElem = 1, nSolve

      ! perform streaming step
      f(-1, 0, 0) = inState(?FETCH?( qN00, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0,-1, 0) = inState(?FETCH?( q0N0, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 0,-1) = inState(?FETCH?( q00N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 0, 0) = inState(?FETCH?( q100, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 1, 0) = inState(?FETCH?( q010, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 0, 1) = inState(?FETCH?( q001, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0,-1,-1) = inState(?FETCH?( q0NN, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0,-1, 1) = inState(?FETCH?( q0N1, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 1,-1) = inState(?FETCH?( q01N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 1, 1) = inState(?FETCH?( q011, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 0,-1) = inState(?FETCH?( qN0N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 0,-1) = inState(?FETCH?( q10N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 0, 1) = inState(?FETCH?( qN01, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 0, 1) = inState(?FETCH?( q101, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1,-1, 0) = inState(?FETCH?( qNN0, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 1, 0) = inState(?FETCH?( qN10, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1,-1, 0) = inState(?FETCH?( q1N0, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 1, 0) = inState(?FETCH?( q110, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1,-1,-1) = inState(?FETCH?( qNNN, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1,-1, 1) = inState(?FETCH?( qNN1, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 1,-1) = inState(?FETCH?( qN1N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 1, 1) = inState(?FETCH?( qN11, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1,-1,-1) = inState(?FETCH?( q1NN, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1,-1, 1) = inState(?FETCH?( q1N1, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 1,-1) = inState(?FETCH?( q11N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 1, 1) = inState(?FETCH?( q111, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 0, 0) = inState(?FETCH?( q000, 1, iElem, QQ, nScalars, nElems,neigh ))

      ! calculate rho and velocity ----------------------------------------------
      rho = sum( f )
      inv_rho = 1.0_rk / rho
      ux  = ( sum(f(1,:,:)) - sum(f(-1, :, :)) ) * inv_rho
      uy  = ( sum(f(:,1,:)) - sum(f( :,-1, :)) ) * inv_rho
      uz  = ( sum(f(:,:,1)) - sum(f( :, :,-1)) ) * inv_rho
      ! calculate rho and velocity ----------------------------------------------

      ! Central moments ---------------------------------------------------------
      ! eq 43 - 45
      do kk = 0,2
        do jj = 0,2
          do ii = 0,2
            k( ii, jj, kk ) = central_moment_split( f, ii, jj, kk, ux, uy, uz )
          end do
        end do
      end do
      ! Central moments ---------------------------------------------------------

      ! Cumulants and Collision -------------------------------------------------
      ! Not sure how to calcuate the post collision value for 0th and 1st order
      c(0,0,0) = k(0,0,0)
      c(1,0,0) = k(1,0,0)
      c(0,1,0) = k(0,1,0)
      c(0,0,1) = k(0,0,1)
      ! Relaxation for 2nd order terms
      c(1,1,0) = ( 1.0_rk - omega ) * k(1,1,0)
      c(1,0,1) = ( 1.0_rk - omega ) * k(1,0,1)
      c(0,1,1) = ( 1.0_rk - omega ) * k(0,1,1)

      ! eq 58 - 60
      Dxu = -omega * 0.5_rk * inv_rho * ( 2.0_rk * k(2,0,0) - k(0,2,0) - k(0,0,2))&
        &   - 0.5_rk * inv_rho * ( k(2,0,0) + k(0,2,0) + k(0,0,2) - k(0,0,0) )
      Dyv = Dxu + 3.0_rk * omega * 0.5_rk * inv_rho * ( k(2,0,0) - k(0,2,0) )
      Dzw = Dxu + 3.0_rk * omega * 0.5_rk * inv_rho * ( k(2,0,0) - k(0,0,2) )

      ! eq 61 - 63
      ! first calculate the rights parts
      AA = ( 1.0_rk - omega ) * ( k(2,0,0) - k(0,2,0) ) &
        & - 3.0_rk * rho * ( 1.0_rk - omega*0.5_rk ) * ( ux*ux*Dxu - uy*uy*Dyv )
      BB = ( 1.0_rk - omega ) * ( k(2,0,0) - k(0,0,2) ) &
        & - 3.0_rk * rho * ( 1.0_rk - omega*0.5_rk ) * ( ux*ux*Dxu - uz*uz*Dzw )
      CC = k(0,0,0) - 1.5_rk * rho * ( ux*ux*Dxu + uy*uy*Dyv + uz*uz*Dzw )

      ! then solve the three moments
      c(2,0,0) = (AA+BB+CC) * div1_3
      c(0,2,0) = (BB+CC) * div1_3 - AA * div2_3
      c(0,0,2) = (AA+CC) * div1_3 - BB * div2_3

      ! for 3rd order or higher, as relaxation == 1, they become simply 0
      c(2,1,0) = 0.0_rk ! 3rd order
      c(2,0,1) = 0.0_rk
      c(1,2,0) = 0.0_rk
      c(0,2,1) = 0.0_rk
      c(1,0,2) = 0.0_rk
      c(0,1,2) = 0.0_rk
      c(1,1,1) = 0.0_rk

      c(2,1,1) = 0.0_rk ! 4th order
      c(1,2,1) = 0.0_rk
      c(1,1,2) = 0.0_rk
      c(2,2,0) = 0.0_rk
      c(2,0,2) = 0.0_rk
      c(0,2,2) = 0.0_rk

      c(1,2,2) = 0.0_rk ! 5th order
      c(2,1,2) = 0.0_rk
      c(2,2,1) = 0.0_rk
      c(2,2,2) = 0.0_rk ! 6th order
      ! Collision --------------------------------------------------------------

      ! Back to central moment -------------------------------------------------
      k = c
      ! k(1,0,0) = -k(1,0,0)
      ! k(0,1,0) = -k(0,1,0)
      ! k(0,0,1) = -k(0,0,1)
      ! get 4th and 6th order from 2nd order
      k(2,1,1) = ( k(2,0,0)*k(0,1,1) + 2.0_rk*k(1,1,0)*k(1,0,1) ) * inv_rho
      k(1,2,1) = ( k(0,2,0)*k(1,0,1) + 2.0_rk*k(1,1,0)*k(0,1,1) ) * inv_rho
      k(1,1,2) = ( k(0,0,2)*k(1,1,0) + 2.0_rk*k(1,0,1)*k(0,1,1) ) * inv_rho
      k(2,2,0) = ( k(2,0,0)*k(0,2,0) + 2.0_rk*k(1,1,0)*k(1,1,0) ) * inv_rho
      k(2,0,2) = ( k(2,0,0)*k(0,0,2) + 2.0_rk*k(1,0,1)*k(1,0,1) ) * inv_rho
      k(0,2,2) = ( k(0,2,0)*k(0,0,2) + 2.0_rk*k(0,1,1)*k(0,1,1) ) * inv_rho
      k(2,2,2) = - (   16.0_rk*k(1,1,0)*k(1,0,1)*k(0,1,1) &
        &          + 4.0_rk*(k(1,0,1)**2*k(0,2,0) + k(0,1,1)**2*k(2,0,0) + k(1,1,0)**2*k(0,0,2) ) &
        &          + 2.0_rk*k(2,0,0)*k(0,2,0)*k(0,0,2) ) * inv_rho * inv_rho
      ! Back to central moment -------------------------------------------------

      ! Back to PDF ------------------------------------------------------------
      f = cm_to_pdf( k, ux, uy, uz )
      ! Back to PDF ------------------------------------------------------------

      ! write to state array
      outState( ?SAVE?( q000,1,iElem,QQ,QQ,nElems,neigh )) = f( 0, 0, 0)
      outState( ?SAVE?( qN00,1,iElem,QQ,QQ,nElems,neigh )) = f(-1, 0, 0)
      outState( ?SAVE?( q100,1,iElem,QQ,QQ,nElems,neigh )) = f( 1, 0, 0)
      outState( ?SAVE?( q0N0,1,iElem,QQ,QQ,nElems,neigh )) = f( 0,-1, 0)
      outState( ?SAVE?( q010,1,iElem,QQ,QQ,nElems,neigh )) = f( 0, 1, 0)
      outState( ?SAVE?( q00N,1,iElem,QQ,QQ,nElems,neigh )) = f( 0, 0,-1)
      outState( ?SAVE?( q001,1,iElem,QQ,QQ,nElems,neigh )) = f( 0, 0, 1)
      outState( ?SAVE?( q0NN,1,iElem,QQ,QQ,nElems,neigh )) = f( 0,-1,-1)
      outState( ?SAVE?( q01N,1,iElem,QQ,QQ,nElems,neigh )) = f( 0, 1,-1)
      outState( ?SAVE?( q0N1,1,iElem,QQ,QQ,nElems,neigh )) = f( 0,-1, 1)
      outState( ?SAVE?( q011,1,iElem,QQ,QQ,nElems,neigh )) = f( 0, 1, 1)
      outState( ?SAVE?( qN0N,1,iElem,QQ,QQ,nElems,neigh )) = f(-1, 0,-1)
      outState( ?SAVE?( q10N,1,iElem,QQ,QQ,nElems,neigh )) = f( 1, 0,-1)
      outState( ?SAVE?( qN01,1,iElem,QQ,QQ,nElems,neigh )) = f(-1, 0, 1)
      outState( ?SAVE?( q101,1,iElem,QQ,QQ,nElems,neigh )) = f( 1, 0, 1)
      outState( ?SAVE?( qNN0,1,iElem,QQ,QQ,nElems,neigh )) = f(-1,-1, 0)
      outState( ?SAVE?( q1N0,1,iElem,QQ,QQ,nElems,neigh )) = f( 1,-1, 0)
      outState( ?SAVE?( qN10,1,iElem,QQ,QQ,nElems,neigh )) = f(-1, 1, 0)
      outState( ?SAVE?( q110,1,iElem,QQ,QQ,nElems,neigh )) = f( 1, 1, 0)
      outState( ?SAVE?( q1NN,1,iElem,QQ,QQ,nElems,neigh )) = f( 1,-1,-1)
      outState( ?SAVE?( q11N,1,iElem,QQ,QQ,nElems,neigh )) = f( 1, 1,-1)
      outState( ?SAVE?( q1N1,1,iElem,QQ,QQ,nElems,neigh )) = f( 1,-1, 1)
      outState( ?SAVE?( q111,1,iElem,QQ,QQ,nElems,neigh )) = f( 1, 1, 1)
      outState( ?SAVE?( qNNN,1,iElem,QQ,QQ,nElems,neigh )) = f(-1,-1,-1)
      outState( ?SAVE?( qN1N,1,iElem,QQ,QQ,nElems,neigh )) = f(-1, 1,-1)
      outState( ?SAVE?( qNN1,1,iElem,QQ,QQ,nElems,neigh )) = f(-1,-1, 1)
      outState( ?SAVE?( qN11,1,iElem,QQ,QQ,nElems,neigh )) = f(-1, 1, 1)

    end do nodeloop

  end subroutine cumulant_d3q27
! ****************************************************************************** !

! ****************************************************************************** !
  pure function cm_to_pdf( k, ux, uy, uz ) result ( f )
    real(kind=rk), intent(in) :: k(0:2,0:2,0:2)
    real(kind=rk), intent(in) :: ux, uy, uz

    real(kind=rk) :: f(-1:1,-1:1,-1:1)
    real(kind=rk) :: g(-1:1, 0:2, 0:2)
    real(kind=rk) :: h(-1:1,-1:1, 0:2)
    integer :: ii, jj
    real(kind=rk) :: uu, mm, nn, pp, qq, rr, u2

    uu = ux*ux
    mm = 1.0_rk - uu
    nn = uu + ux
    pp = uu - ux
    u2 = 2.0_rk * ux
    qq = u2 - 1.0_rk
    rr = u2 + 1.0_rk
    do jj = 0,2
      do ii = 0,2
        g( 0,ii,jj) =   k(0,ii,jj)*mm - k(1,ii,jj)*u2 - k(2,ii,jj)
        g(-1,ii,jj) = ( k(0,ii,jj)*pp + k(1,ii,jj)*qq + k(2,ii,jj) ) * div1_2
        g( 1,ii,jj) = ( k(0,ii,jj)*nn + k(1,ii,jj)*rr + k(2,ii,jj) ) * div1_2
      end do
    end do

    uu = uy*uy
    mm = 1.0_rk - uu
    nn = uu + uy
    pp = uu - uy
    u2 = 2.0_rk * uy
    qq = u2 - 1.0_rk
    rr = u2 + 1.0_rk
    do jj = 0,2
      do ii = -1,1
        h(ii, 0,jj) =   g(ii,0,jj)*mm - g(ii,1,jj)*u2 - g(ii,2,jj)
        h(ii,-1,jj) = ( g(ii,0,jj)*pp + g(ii,1,jj)*qq + g(ii,2,jj) ) * div1_2
        h(ii, 1,jj) = ( g(ii,0,jj)*nn + g(ii,1,jj)*rr + g(ii,2,jj) ) * div1_2
      end do
    end do

    uu = uz*uz
    mm = 1.0_rk - uu
    nn = uu + uz
    pp = uu - uz
    u2 = 2.0_rk * uz
    qq = u2 - 1.0_rk
    rr = u2 + 1.0_rk
    do jj = -1,1
      do ii = -1,1
        f(ii, jj, 0) =   h(ii,jj,0)*mm - h(ii,jj,1)*u2 - h(ii,jj,2)
        f(ii, jj,-1) = ( h(ii,jj,0)*pp + h(ii,jj,1)*qq + h(ii,jj,2) ) * div1_2
        f(ii, jj, 1) = ( h(ii,jj,0)*nn + h(ii,jj,1)*rr + h(ii,jj,2) ) * div1_2
      end do
    end do
  end function
! ****************************************************************************** !

! ****************************************************************************** !
?? copy :: compute_routineHeader( cascaded_d3q27 )
?? copy :: compute_routineParams
    ! ---------------------------------------------------------------------------
    integer :: iElem, ii, jj, kk, nScalars
    real(kind=rk) :: ux, uy, uz, rho, inv_rho, omega
    real(kind=rk) :: dxu, dyv, dzw, AA, BB, CC
    real(kind=rk) ::  f(-1:1,-1:1,-1:1)
    real(kind=rk) :: ff(-1:1,-1:1,-1:1)
    ! k = central moment, c = cumulant
    real(kind=rk) :: k(0:2,0:2,0:2), c(0:2,0:2,0:2)
    ! ---------------------------------------------------------------------------

    nScalars = varSys%nScalars
    omega = fieldProp(1)%fluid%omLvl( level )

    nodeloop: do iElem = 1, nSolve

      f(-1, 0, 0) = inState(?FETCH?( qN00, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0,-1, 0) = inState(?FETCH?( q0N0, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 0,-1) = inState(?FETCH?( q00N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 0, 0) = inState(?FETCH?( q100, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 1, 0) = inState(?FETCH?( q010, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 0, 1) = inState(?FETCH?( q001, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0,-1,-1) = inState(?FETCH?( q0NN, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0,-1, 1) = inState(?FETCH?( q0N1, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 1,-1) = inState(?FETCH?( q01N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 1, 1) = inState(?FETCH?( q011, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 0,-1) = inState(?FETCH?( qN0N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 0,-1) = inState(?FETCH?( q10N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 0, 1) = inState(?FETCH?( qN01, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 0, 1) = inState(?FETCH?( q101, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1,-1, 0) = inState(?FETCH?( qNN0, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 1, 0) = inState(?FETCH?( qN10, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1,-1, 0) = inState(?FETCH?( q1N0, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 1, 0) = inState(?FETCH?( q110, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1,-1,-1) = inState(?FETCH?( qNNN, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1,-1, 1) = inState(?FETCH?( qNN1, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 1,-1) = inState(?FETCH?( qN1N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 1, 1) = inState(?FETCH?( qN11, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1,-1,-1) = inState(?FETCH?( q1NN, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1,-1, 1) = inState(?FETCH?( q1N1, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 1,-1) = inState(?FETCH?( q11N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 1, 1) = inState(?FETCH?( q111, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 0, 0) = inState(?FETCH?( q000, 1, iElem, QQ, nScalars, nElems,neigh ))

      ! calculate rho and velocity ----------------------------------------------
      rho = sum( f )
      inv_rho = 1.0_rk / rho
      ux  = ( sum(f(1,:,:)) - sum(f(-1, :, :)) ) * inv_rho
      uy  = ( sum(f(:,1,:)) - sum(f( :,-1, :)) ) * inv_rho
      uz  = ( sum(f(:,:,1)) - sum(f( :, :,-1)) ) * inv_rho
      ! calculate rho and velocity ----------------------------------------------

      ! get all central moments
      ! eq 43 - 45
      do kk = 0,2
        do jj = 0,2
          do ii = 0,2
            k( ii, jj, kk ) = central_moment_split( f, ii, jj, kk, ux, uy, uz )
          end do
        end do
      end do

      ! Collision --------------------------------------------------------------
      c(0,0,0) = k(0,0,0)
      c(1,0,0) = k(1,0,0)
      c(0,1,0) = k(0,1,0)
      c(0,0,1) = k(0,0,1)
      c(1,1,0) = ( 1.0_rk - omega ) * k(1,1,0)
      c(1,0,1) = ( 1.0_rk - omega ) * k(1,0,1)
      c(0,1,1) = ( 1.0_rk - omega ) * k(0,1,1)

      ! eq D.1 - D.3
      Dxu = -omega * 0.5_rk * inv_rho * ( 2.0_rk * k(2,0,0) - k(0,2,0) - k(0,0,2))&
        &   - 0.5_rk * inv_rho * ( k(2,0,0) + k(0,2,0) + k(0,0,2) - rho )
      Dyv = Dxu + 3.0_rk * omega * 0.5_rk * inv_rho * ( k(2,0,0) - k(0,2,0) )
      Dzw = Dxu + 3.0_rk * omega * 0.5_rk * inv_rho * ( k(2,0,0) - k(0,0,2) )

      ! eq D.4 - D.6
      ! first calculate the rights parts
      AA = ( 1.0_rk - omega ) * ( k(2,0,0) - k(0,2,0) ) &
        & - 3.0_rk * rho * ( 1.0_rk - omega*0.5_rk ) * ( ux*ux*Dxu - uy*uy*Dyv )
      BB = ( 1.0_rk - omega ) * ( k(2,0,0) - k(0,0,2) ) &
        & - 3.0_rk * rho * ( 1.0_rk - omega*0.5_rk ) * ( ux*ux*Dxu - uz*uz*Dzw )
      CC = rho - 1.5_rk * rho * ( ux*ux*Dxu + uy*uy*Dyv + uz*uz*Dzw )

      ! then solve the three moments
      c(2,0,0) = (AA+BB+CC) * div1_3
      c(0,2,0) = (BB+CC) * div1_3 - AA * div2_3
      c(0,0,2) = (AA+CC) * div1_3 - BB * div2_3

      ! for 3rd order or higher, set to 0
      c(2,1,0) = 0.0_rk ! 3rd order
      c(2,0,1) = 0.0_rk
      c(1,2,0) = 0.0_rk
      c(0,2,1) = 0.0_rk
      c(1,0,2) = 0.0_rk
      c(0,1,2) = 0.0_rk
      c(1,1,1) = 0.0_rk

      c(2,1,1) = 0.0_rk ! 4th order
      c(1,2,1) = 0.0_rk
      c(1,1,2) = 0.0_rk
      c(2,2,0) = rho * div1_3 * div1_3
      c(2,0,2) = rho * div1_3 * div1_3
      c(0,2,2) = rho * div1_3 * div1_3

      c(1,2,2) = 0.0_rk ! 5th order
      c(2,1,2) = 0.0_rk
      c(2,2,1) = 0.0_rk
      c(2,2,2) = rho * div1_3 * div1_3 * div1_3 ! 6th order
      ! Collision --------------------------------------------------------------

      ! Back to PDF ------------------------------------------------------------
      ff = cm_to_pdf( c, ux, uy, uz )
      ! Back to PDF ------------------------------------------------------------

      ! write to state array
outState( ?SAVE?( q000,1,iElem,QQ,QQ,nElems,neigh )) = ff( 0,  0,  0)
outState( ?SAVE?( qN00,1,iElem,QQ,QQ,nElems,neigh )) = ff(-1,  0,  0)
outState( ?SAVE?( q100,1,iElem,QQ,QQ,nElems,neigh )) = ff( 1,  0,  0)
outState( ?SAVE?( q0N0,1,iElem,QQ,QQ,nElems,neigh )) = ff( 0, -1,  0)
outState( ?SAVE?( q010,1,iElem,QQ,QQ,nElems,neigh )) = ff( 0,  1,  0)
outState( ?SAVE?( q00N,1,iElem,QQ,QQ,nElems,neigh )) = ff( 0,  0, -1)
outState( ?SAVE?( q001,1,iElem,QQ,QQ,nElems,neigh )) = ff( 0,  0,  1)
outState( ?SAVE?( q0NN,1,iElem,QQ,QQ,nElems,neigh )) = ff( 0, -1, -1)
outState( ?SAVE?( q01N,1,iElem,QQ,QQ,nElems,neigh )) = ff( 0,  1, -1)
outState( ?SAVE?( q0N1,1,iElem,QQ,QQ,nElems,neigh )) = ff( 0, -1,  1)
outState( ?SAVE?( q011,1,iElem,QQ,QQ,nElems,neigh )) = ff( 0,  1,  1)
outState( ?SAVE?( qN0N,1,iElem,QQ,QQ,nElems,neigh )) = ff(-1,  0, -1)
outState( ?SAVE?( q10N,1,iElem,QQ,QQ,nElems,neigh )) = ff( 1,  0, -1)
outState( ?SAVE?( qN01,1,iElem,QQ,QQ,nElems,neigh )) = ff(-1,  0,  1)
outState( ?SAVE?( q101,1,iElem,QQ,QQ,nElems,neigh )) = ff( 1,  0,  1)
outState( ?SAVE?( qNN0,1,iElem,QQ,QQ,nElems,neigh )) = ff(-1, -1,  0)
outState( ?SAVE?( q1N0,1,iElem,QQ,QQ,nElems,neigh )) = ff( 1, -1,  0)
outState( ?SAVE?( qN10,1,iElem,QQ,QQ,nElems,neigh )) = ff(-1,  1,  0)
outState( ?SAVE?( q110,1,iElem,QQ,QQ,nElems,neigh )) = ff( 1,  1,  0)
outState( ?SAVE?( q1NN,1,iElem,QQ,QQ,nElems,neigh )) = ff( 1, -1, -1)
outState( ?SAVE?( q11N,1,iElem,QQ,QQ,nElems,neigh )) = ff( 1,  1, -1)
outState( ?SAVE?( q1N1,1,iElem,QQ,QQ,nElems,neigh )) = ff( 1, -1,  1)
outState( ?SAVE?( q111,1,iElem,QQ,QQ,nElems,neigh )) = ff( 1,  1,  1)
outState( ?SAVE?( qNNN,1,iElem,QQ,QQ,nElems,neigh )) = ff(-1, -1, -1)
outState( ?SAVE?( qN1N,1,iElem,QQ,QQ,nElems,neigh )) = ff(-1,  1, -1)
outState( ?SAVE?( qNN1,1,iElem,QQ,QQ,nElems,neigh )) = ff(-1, -1,  1)
outState( ?SAVE?( qN11,1,iElem,QQ,QQ,nElems,neigh )) = ff(-1,  1,  1)

    end do nodeloop

  end subroutine cascaded_d3q27
! ****************************************************************************** !

! ****************************************************************************** !
?? copy :: compute_routineHeader( trt_d3q27 )
?? copy :: compute_routineParams
    ! ---------------------------------------------------------------------------
    integer :: iElem
    integer :: nScalars
    real(kind=rk) :: f(-1:1,-1:1,-1:1)
    real(kind=rk) :: u, v, w, u2, v2, w2
    real(kind=rk) :: rho, inv_rho!, rho_omg
    real(kind=rk) :: omega, wP, wN, p_part, n_part
    real(kind=rk) :: X0, X1, XN
    real(kind=rk) :: Y0, Y1, YN
    real(kind=rk) :: Z0, Z1, ZN
    ! ---------------------------------------------------------------------------

    nScalars = varSys%nScalars
    omega = fieldProp(1)%fluid%omLvl( level )
    wP = omega
    wN = fieldProp(1)%fluid%wN( level )

    !CDIR NODEP
?? IF (SOA) THEN
?? copy :: dir_vector
?? ELSE
?? copy :: dir_novec
?? END IF
    nodeloop: do iElem = 1, nSolve

      f(-1, 0, 0) = inState(?FETCH?( qN00, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0,-1, 0) = inState(?FETCH?( q0N0, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 0,-1) = inState(?FETCH?( q00N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 0, 0) = inState(?FETCH?( q100, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 1, 0) = inState(?FETCH?( q010, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 0, 1) = inState(?FETCH?( q001, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0,-1,-1) = inState(?FETCH?( q0NN, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0,-1, 1) = inState(?FETCH?( q0N1, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 1,-1) = inState(?FETCH?( q01N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 1, 1) = inState(?FETCH?( q011, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 0,-1) = inState(?FETCH?( qN0N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 0,-1) = inState(?FETCH?( q10N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 0, 1) = inState(?FETCH?( qN01, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 0, 1) = inState(?FETCH?( q101, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1,-1, 0) = inState(?FETCH?( qNN0, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 1, 0) = inState(?FETCH?( qN10, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1,-1, 0) = inState(?FETCH?( q1N0, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 1, 0) = inState(?FETCH?( q110, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1,-1,-1) = inState(?FETCH?( qNNN, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1,-1, 1) = inState(?FETCH?( qNN1, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 1,-1) = inState(?FETCH?( qN1N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f(-1, 1, 1) = inState(?FETCH?( qN11, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1,-1,-1) = inState(?FETCH?( q1NN, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1,-1, 1) = inState(?FETCH?( q1N1, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 1,-1) = inState(?FETCH?( q11N, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 1, 1, 1) = inState(?FETCH?( q111, 1, iElem, QQ, nScalars, nElems,neigh ))
      f( 0, 0, 0) = inState(?FETCH?( q000, 1, iElem, QQ, nScalars, nElems,neigh ))


      ! calculate rho and velocity ----------------------------------------------
      rho = sum( f )
      inv_rho = 1.0_rk / rho
      u  = ( sum(f(1,:,:)) - sum(f(-1, :, :)) ) * inv_rho
      v  = ( sum(f(:,1,:)) - sum(f( :,-1, :)) ) * inv_rho
      w  = ( sum(f(:,:,1)) - sum(f( :, :,-1)) ) * inv_rho
      ! rho_omg = rho * omega
      ! calculate rho and velocity ----------------------------------------------

      ! u v square
      u2 = u*u
      v2 = v*v
      w2 = w*w

      ! X Y Z components of eq
      ! by equation A.19 - A.21
      X0 = -div2_3 + u2
      X1 = - ( X0 + 1.0d0 + u ) * 0.5d0
      XN = X1 + u

      Y0 = -div2_3 + v2
      Y1 = - ( Y0 + 1.0d0 + v ) * 0.5d0
      YN = Y1 + v

      Z0 = -div2_3 + w2
      Z1 = - ( Z0 + 1.0d0 + w ) * 0.5d0
      ZN = Z1 + w

      ! X0 = -rho_omg * X0
      ! X1 = -rho_omg * X1
      ! XN = -rho_omg * XN

! fEq = -rho * X * Y * Z
! fEq000 = X0 * Y0 * Z0
outState( ?SAVE?( q000,1,iElem,QQ,QQ,nElems,neigh )) = (1.0_rk - omega)*f(0,0,0) - rho*omega*X0*Y0*Z0

p_part = wP * ( ( f(1,0,0) + f(-1,0,0) ) - (-rho*X1*Y0*Z0 - rho*XN*Y0*Z0) ) * div1_2
n_part = wN * ( ( f(1,0,0) - f(-1,0,0) ) - (-rho*X1*Y0*Z0 + rho*XN*Y0*Z0) ) * div1_2
outState( ?SAVE?( q100,1,iElem,QQ,QQ,nElems,neigh )) = f( 1,0,0) - p_part - n_part
outState( ?SAVE?( qN00,1,iElem,QQ,QQ,nElems,neigh )) = f(-1,0,0) - p_part + n_part

p_part = wP * ( ( f(0,1,0) + f(0,-1,0) ) - (-rho*X0*Y1*Z0 - rho*X0*YN*Z0) ) * div1_2
n_part = wN * ( ( f(0,1,0) - f(0,-1,0) ) - (-rho*X0*Y1*Z0 + rho*X0*YN*Z0) ) * div1_2
outState( ?SAVE?( q010,1,iElem,QQ,QQ,nElems,neigh )) = f(0, 1,0) - p_part - n_part
outState( ?SAVE?( q0N0,1,iElem,QQ,QQ,nElems,neigh )) = f(0,-1,0) - p_part + n_part

p_part = wP * ( ( f(0,0,1) + f(0,0,-1) ) - (-rho*X0*Y0*Z1 - rho*X0*Y0*ZN) ) * div1_2
n_part = wN * ( ( f(0,0,1) - f(0,0,-1) ) - (-rho*X0*Y0*Z1 + rho*X0*Y0*ZN) ) * div1_2
outState( ?SAVE?( q001,1,iElem,QQ,QQ,nElems,neigh )) = f(0,0, 1) - p_part - n_part
outState( ?SAVE?( q00N,1,iElem,QQ,QQ,nElems,neigh )) = f(0,0,-1) - p_part + n_part

p_part = wP * ( ( f(0,1,1) + f(0,-1,-1) ) - (-rho*X0*Y1*Z1 - rho*X0*YN*ZN) ) * div1_2
n_part = wN * ( ( f(0,1,1) - f(0,-1,-1) ) - (-rho*X0*Y1*Z1 + rho*X0*YN*ZN) ) * div1_2
outState( ?SAVE?( q011,1,iElem,QQ,QQ,nElems,neigh )) = f(0, 1, 1) - p_part - n_part
outState( ?SAVE?( q0NN,1,iElem,QQ,QQ,nElems,neigh )) = f(0,-1,-1) - p_part + n_part

p_part = wP * ( ( f(0,1,-1) + f(0,-1,1) ) - (-rho*X0*Y1*ZN - rho*X0*YN*Z1) ) * div1_2
n_part = wN * ( ( f(0,1,-1) - f(0,-1,1) ) - (-rho*X0*Y1*ZN + rho*X0*YN*Z1) ) * div1_2
outState( ?SAVE?( q01N,1,iElem,QQ,QQ,nElems,neigh )) = f(0, 1,-1) - p_part - n_part
outState( ?SAVE?( q0N1,1,iElem,QQ,QQ,nElems,neigh )) = f(0,-1, 1) - p_part + n_part

p_part = wP * ( ( f(1,0,1) + f(-1,0,-1) ) - (-rho*X1*Y0*Z1 - rho*XN*Y0*ZN) ) * div1_2
n_part = wN * ( ( f(1,0,1) - f(-1,0,-1) ) - (-rho*X1*Y0*Z1 + rho*XN*Y0*ZN) ) * div1_2
outState( ?SAVE?( q101,1,iElem,QQ,QQ,nElems,neigh )) = f( 1,0, 1) - p_part - n_part
outState( ?SAVE?( qN0N,1,iElem,QQ,QQ,nElems,neigh )) = f(-1,0,-1) - p_part + n_part

p_part = wP * ( ( f(1,0,-1) + f(-1,0,1) ) - (-rho*X1*Y0*ZN - rho*XN*Y0*Z1) ) * div1_2
n_part = wN * ( ( f(1,0,-1) - f(-1,0,1) ) - (-rho*X1*Y0*ZN + rho*XN*Y0*Z1) ) * div1_2
outState( ?SAVE?( q10N,1,iElem,QQ,QQ,nElems,neigh )) = f( 1,0,-1) - p_part - n_part
outState( ?SAVE?( qN01,1,iElem,QQ,QQ,nElems,neigh )) = f(-1,0, 1) - p_part + n_part

p_part = wP * ( ( f(1,1,0) + f(-1,-1,0) ) - (-rho*X1*Y1*Z0 - rho*XN*YN*Z0) ) * div1_2
n_part = wN * ( ( f(1,1,0) - f(-1,-1,0) ) - (-rho*X1*Y1*Z0 + rho*XN*YN*Z0) ) * div1_2
outState( ?SAVE?( q110,1,iElem,QQ,QQ,nElems,neigh )) = f( 1, 1,0) - p_part - n_part
outState( ?SAVE?( qNN0,1,iElem,QQ,QQ,nElems,neigh )) = f(-1,-1,0) - p_part + n_part

p_part = wP * ( ( f(1,-1,0) + f(-1,1,0) ) - (-rho*X1*YN*Z0 - rho*XN*Y1*Z0) ) * div1_2
n_part = wN * ( ( f(1,-1,0) - f(-1,1,0) ) - (-rho*X1*YN*Z0 + rho*XN*Y1*Z0) ) * div1_2
outState( ?SAVE?( q1N0,1,iElem,QQ,QQ,nElems,neigh )) = f( 1,-1,0) - p_part - n_part
outState( ?SAVE?( qN10,1,iElem,QQ,QQ,nElems,neigh )) = f(-1, 1,0) - p_part + n_part

p_part = wP * ( ( f(1,-1,-1) + f(-1,1,1) ) - (-rho*X1*YN*ZN - rho*XN*Y1*Z1) ) * div1_2
n_part = wN * ( ( f(1,-1,-1) - f(-1,1,1) ) - (-rho*X1*YN*ZN + rho*XN*Y1*Z1) ) * div1_2
outState( ?SAVE?( q1NN,1,iElem,QQ,QQ,nElems,neigh )) = f( 1,-1,-1) - p_part - n_part
outState( ?SAVE?( qN11,1,iElem,QQ,QQ,nElems,neigh )) = f(-1, 1, 1) - p_part + n_part

p_part = wP * ( ( f(1,1,-1) + f(-1,-1,1) ) - (-rho*X1*Y1*ZN - rho*XN*YN*Z1) ) * div1_2
n_part = wN * ( ( f(1,1,-1) - f(-1,-1,1) ) - (-rho*X1*Y1*ZN + rho*XN*YN*Z1) ) * div1_2
outState( ?SAVE?( q11N,1,iElem,QQ,QQ,nElems,neigh )) = f( 1, 1,-1) - p_part - n_part
outState( ?SAVE?( qNN1,1,iElem,QQ,QQ,nElems,neigh )) = f(-1,-1, 1) - p_part + n_part

p_part = wP * ( ( f(1,-1,1) + f(-1,1,-1) ) - (-rho*X1*YN*Z1 - rho*XN*Y1*ZN) ) * div1_2
n_part = wN * ( ( f(1,-1,1) - f(-1,1,-1) ) - (-rho*X1*YN*Z1 + rho*XN*Y1*ZN) ) * div1_2
outState( ?SAVE?( q1N1,1,iElem,QQ,QQ,nElems,neigh )) = f( 1,-1, 1) - p_part - n_part
outState( ?SAVE?( qN1N,1,iElem,QQ,QQ,nElems,neigh )) = f(-1, 1,-1) - p_part + n_part

p_part = wP * ( ( f(1,1,1) + f(-1,-1,-1) ) - (-rho*X1*Y1*Z1 - rho*XN*YN*ZN) ) * div1_2
n_part = wN * ( ( f(1,1,1) - f(-1,-1,-1) ) - (-rho*X1*Y1*Z1 + rho*XN*YN*ZN) ) * div1_2
outState( ?SAVE?( q111,1,iElem,QQ,QQ,nElems,neigh )) = f( 1, 1, 1) - p_part - n_part
outState( ?SAVE?( qNNN,1,iElem,QQ,QQ,nElems,neigh )) = f(-1,-1,-1) - p_part + n_part

    end do nodeloop

  end subroutine trt_d3q27
! ****************************************************************************** !

end module mus_d3q27_module
! ****************************************************************************** !
