! See copyright notice in the COPYRIGHT file.
! ****************************************************************************** !
!> author: Kannan Masilamani
!! This module contains the type defintion which stores position of all
!! derive variables in global varSys. 
!!
module mus_derVarPos_type_module
  use env_module, only: rk

  use tem_varSys_module, only: tem_varSys_type

  use mus_scheme_layout_module, only: mus_scheme_layout_type

  implicit none

  private

  public :: mus_derVarPos_type

  !> This type stores the position of each variable in the global sys
  type mus_derVarPos_type
    integer :: pdf               = -1
    integer :: fetch_pdf         = -1
    integer :: omega             = -1
    integer :: density           = -1
    integer :: numDensity        = -1
    integer :: pressure          = -1
    integer :: kinePress         = -1
    integer :: velocity          = -1
    ! all species velocities
    integer :: spc_velocities    = -1
    integer :: velMag            = -1
    integer :: momentum          = -1
    ! all species momentums
    integer :: spc_momenta       = -1
    integer :: shearStress       = -1
    integer :: wss               = -1
    integer :: shearMag          = -1
    integer :: strainRate        = -1
    integer :: shearRate         = -1
    integer :: kineticEnergy     = -1
    integer :: temperature       = -1
    integer :: molefrac          = -1
    integer :: moleflux          = -1
    integer :: equilibrium       = -1
    integer :: nonEquilibrium    = -1
    integer :: equilibriumVel    = -1
    procedure( derive_FromMacro ), pointer, nopass :: equilFromMacro => null()
    ! return velocity of single species
    procedure( derive_FromState ), pointer, nopass :: velFromState   => null()
    procedure( derive_FromState ), pointer, nopass :: equilFromState => null()
    ! return momentum of single species
    procedure( derive_FromState ), pointer, nopass :: momFromState => null()
    ! return velocity of all species
    procedure( derive_FromState ), pointer, nopass :: velocitiesFromState => null()
    ! return momentum of all species
    procedure( derive_FromState ), pointer, nopass :: momentaFromState => null()
  end type mus_derVarPos_type

  !> interface to derive equilibrium from macro.
  !! Mainly used in initial condition and boundary condition routines
  !! to avoid dublication of routines between different scheme kinds.
  !! In this interface, solver definition can be access via 
  !! varSys%method%val(1)%method_data c_ptr
  abstract interface
    subroutine derive_FromMacro( density, velocity, iField, nElems, varSys,    &
      &                          layout, res )
      import :: rk, tem_varSys_type, mus_scheme_layout_type
      !> Array of density.
      !! Single species: dens_1, dens_2 .. dens_n
      !! multi-species: dens_1_sp1, dens_1_sp2, dens_2_sp1, dens_2_sp2 ...
      !!                dens_n_sp1, dens_n_sp2
      !! Access: (iElem-1)*nFields + iField
      real(kind=rk), intent(in) :: density(:)

      !> Array of velocity.
      !! Size: (3, n*nFields)
      !! Access: ( iComp, (iElem-1)*nFields + iField )
      real(kind=rk), intent(in) :: velocity(:, :)

      !> Current field
      integer, intent(in) :: iField

      !> number of elements
      integer, intent(in) :: nElems

      !> variable system which is required to access fieldProp
      !! information via variable method data c_ptr
      type(tem_varSys_type), intent(in) :: varSys

      !> scheme layout contains stencil definition and lattice weights
      type(mus_scheme_layout_type), intent(in) :: layout

      !> Output of this routine
      !! Dimension: n*nComponents of res
      real(kind=rk), intent(out) :: res(:)
    end subroutine derive_FromMacro

    ! Interface that takes state array as input
    ! calculate density, velocity or eq as output
    ! State should be AOS layout
    subroutine derive_FromState( state, iField, nElems, varSys, layout, res )
      import :: rk, tem_varSys_type, mus_scheme_layout_type
      !> Array of state
      !! n * layout%stencil(1)%QQ * nFields
      real(kind=rk), intent(in) :: state(:)

      !> Current field
      integer, intent(in) :: iField

      !> number of elements
      integer, intent(in) :: nElems

      !> variable system which is required to access fieldProp
      !! information via variable method data c_ptr
      type(tem_varSys_type), intent(in) :: varSys

      !> scheme layout contains stencil definition and lattice weights
      type(mus_scheme_layout_type), intent(in) :: layout

      !> Output of this routine
      !! Dimension: n * nComponents of res
      !! Access: (iElem-1)*nComp + iComp
      !! To derive velocities of all species, dimension: n*nFields*nComp
      !! Access: (iElem-1)*nFields*nComp + (iField-1)*nComp + iComp
      real(kind=rk), intent(out) :: res(:)
    end subroutine derive_FromState
  end interface

end module mus_derVarPos_type_module
! ****************************************************************************** !
