! See copyright notice in the COPYRIGHT file.
! **************************************************************************** !
!> author: Kannan Masilamani, Verena Krupp
!! This module contains solver data type with pointers to musubi 
!! [[mus_scheme_type]], [[mus_param_type]], etc which are required for 
!! variable operation method data.
!! Also contains all general routines for the variable system.
!!
?? include 'header/lbm_macros.inc'
?? include 'treelm/source/deriveMacros.inc'
module mus_varSys_module

  use, intrinsic :: iso_c_binding,  only: c_ptr, c_f_pointer, c_loc
  use env_module,                   only: rk, long_k

  use tem_logging_module,           only: logUnit
  use tem_aux_module,               only: tem_abort
  use tem_varSys_module,            only: tem_varSys_type,   &
    &                                     tem_varSys_op_type
  use tem_grow_array_module,        only: grw_realArray_type, &
    &                                     grw_intArray_type, &
    &                                     append, init
  use tem_pointData_module,         only: tem_pointData_list_type, &
    &                                     init,append, truncate
  use treelmesh_module,             only: treelmesh_type
  use tem_geometry_module,          only: tem_CoordOfReal, &
    &                                     tem_PosofId, tem_BaryOfId 
  use tem_time_module,              only: tem_time_type
  use tem_topology_module,          only: tem_IdOfCoord
  use tem_topology_module,          only: tem_levelOf
  use tem_operation_var_module,     only: tem_varSys_op_data_type
  use tem_construction_module,      only: tem_treeIDInTotal
  use tem_element_module,           only: eT_minRelevant, eT_maxRelevant, &
    &                                     eT_fluid

  ! include musubi modules
  use mus_scheme_type_module,  only: mus_scheme_type
  use mus_physics_module,      only: mus_physics_type
  use mus_geom_module,         only: mus_geom_type


  implicit none

  private

  public :: mus_varSys_solverData_type
  public :: mus_varSys_data_type
  public :: mus_get_new_solver_ptr
  public :: mus_init_varSys_solverData
  public :: mus_access_state_ForElement
  public :: mus_stateVar_setupIndices
  public :: mus_stateVar_getValOfIndex
  public :: mus_deriveVar_ForPoint
  public :: mus_deriveVar_getValOfIndex

  !> Method data container for every variable
  type mus_varSys_data_type
    ! type with all solver relevant data for the variable
    type(mus_varSys_solverData_type), pointer :: solverData

    ! the point_datas need to be stored levelwise
    type(tem_pointData_list_type) :: pointData

    !> data array for operation or derived varibales
    !! consists the index arrys for points stored in the
    !! poingtData of input variable
    !! size is number of input variables
    type(tem_varSys_op_data_type) :: opData
  end type mus_varSys_data_type

  !> Contains pointer to musubi data types required for variable operation
  !! method_data
  type mus_varSys_solverData_type
    !> scheme data type
    type(mus_scheme_type), pointer :: scheme => NULL()

    !> contains basic SI units to convert from lattice to physical and
    !! vice versa
    type(mus_physics_type), pointer :: physics => NULL()

    !> Contains geometry information and definitions
    type(mus_geom_type), pointer :: geometry => NULL()

  end type mus_varSys_solverData_type


contains


  ! ************************************************************************* !
  !> Routine to get a pointer to a new instance of mus_varSys_solverData_type 
  !! to be used as method data for a variable in the variable system.
  !!
  !! A new instance is allocated and a c_ptr to this type is returned.
  !! Be aware that local pointer are not automatically deallocated when leaving
  !! the routine
  function mus_get_new_solver_ptr( solver ) result(resPtr)
    ! --------------------------------------------------------------------- !
    !> The prototype is used to initialize the new instance.
    type(mus_varSys_solverData_type), intent(in), optional, target :: solver
    !> Pointer to the newly created instance.
    type(c_ptr) :: resPtr
    ! --------------------------------------------------------------------- !
    !> Local variable to allocate a new instance.
    type(mus_varSys_data_type), pointer :: res
    ! --------------------------------------------------------------------- !

    allocate(res)
    if (present(solver)) res%solverData => solver
   
    !>TODO init the point arrays
    
    resPtr = c_loc(res)

  end function mus_get_new_solver_ptr
  ! ************************************************************************* !


  ! ************************************************************************* !
  ! This routine sets varSys_solverData pointers
  subroutine mus_init_varSys_solverData( me, scheme, physics, geometry )
    ! --------------------------------------------------------------------- !
    type(mus_varSys_solverData_type),  intent(out) :: me
    type(mus_scheme_type),  target, intent(in) :: scheme
    type(mus_physics_type), target, intent(in) :: physics
    type(mus_geom_type),    target, intent(in) :: geometry
    ! --------------------------------------------------------------------- !
    me%scheme => scheme
    me%physics => physics
    me%geometry => geometry
  end subroutine mus_init_varSys_solverData
  ! ************************************************************************* !


  ! ************************************************************************* !
  !> Return the solver state variable for a given set of elements
  !!
  !! The interface has to comply to the abstract interface
  !! [[tem_varSys_module:tem_varSys_proc_element]].
  !!
?? copy :: get_element_headtxt(mus_access_state_ForElement)
    ! --------------------------------------------------------------------- !
    integer :: statePos, iElem, iComp, iLevel
    type(mus_varSys_data_type), pointer :: fPtr
    integer :: nElems_loc
    ! --------------------------------------------------------------------- !
    call C_F_POINTER( fun%method_Data, fPtr )

    ! res is always AOS layout
    res = 0.0_rk
    do iElem = 1, nElems
      ! if state array is defined level wise then use levelPointer(pos)
      ! to access state array
      statePos = fPtr%solverData%geometry%levelPointer( elemPos(iElem) )
      iLevel = tem_levelOf( tree%treeID( elemPos(iElem) ) )
      nElems_loc = fPtr%solverData%scheme%pdf( iLevel )%nSize
      do iComp = 1, fun%nComponents
        res( (iElem-1)*fun%nComponents+iComp ) =                        &
          & fPtr%solverData%scheme%state( iLevel )%val(                 &
          ! position of this state variable in the state array
& ?IDX?( fun%state_varPos(iComp),statePos,varSys%nScalars,nElems_loc ), &
          & fPtr%solverData%scheme%pdf( iLevel )%nNext )
      end do !iComp
    end do !iElem

  end subroutine mus_access_state_ForElement
  ! ************************************************************************* !


  ! ************************************************************************* !
  !> Derive variable for a given set of points using linear interpolation.
  !!
  !! The interface has to comply to the abstract interface
  !! [[tem_varSys_module:tem_varSys_proc_point]].
  !!
?? copy :: get_point_headtxt(mus_deriveVar_ForPoint)
    ! --------------------------------------------------------------------- !
    type(mus_varSys_data_type), pointer :: fPtr
    type(mus_scheme_type), pointer :: scheme
    integer :: statePos, iPnt, iLevel, elemPos
    integer :: coord(4)
    real(kind=rk) :: bary(3)
    integer :: iNeigh, neighPos, neighLvlPos
    integer, allocatable :: srcElemPos(:)
    real(kind=rk), allocatable :: weights(:), srcRes(:), pntVal(:)
    real(kind=rk) :: dist(3), sumWeight
    integer :: iSrc, nSrcElems
    ! --------------------------------------------------------------------- !
    call C_F_POINTER( fun%method_Data, fPtr )
    scheme => fPtr%solverData%scheme
    allocate(srcElemPos(scheme%layout%fStencil%QQ))
    allocate(weights(scheme%layout%fStencil%QQ))
    allocate(srcRes(scheme%layout%fStencil%QQ*fun%nComponents))
    allocate(pntVal(fun%nComponents))
    
    res = 0.0_rk
    do iPnt = 1, nPnts
      weights = 0.0_rk 
      srcRes = 0.0_rk
      coord =  tem_CoordOfReal(tree, point(iPnt,:), tree%global%maxLevel )
      ! returns position of existing element in tree which contains point
      ! 0 if no corresponding node is found,
      ! or the negative of the found ID, if it is a virtual node.
      elemPos = abs(tem_PosofId( tem_IdOfCoord(coord), tree%treeID ))

      ! skip this point if no corresponding node is found
      if (elemPos == 0) cycle

      ! level of existing element 
      iLevel = tem_levelOf( tree%treeID( elemPos ) )

      ! use this element as 1st source element for interpolation
      nSrcElems = 1
      srcElemPos(nSrcElems)  = elemPos

      bary = tem_BaryOfID(tree, tree%treeID(elemPos ))
      dist = abs(bary - point(iPnt,:))
      ! if point is exact bary center of current element then 
      ! no need to do interpolation
      if ( all(dist == 0.0_rk) ) then
        weights(nSrcElems) = 1.0_rk
      else  
        ! position of element in levelDesc total list
        statePos = fPtr%solverData%geometry%levelPointer( elemPos )
        ! get existing neighbor elements in actual tree
        do iNeigh = 1, scheme%layout%fStencil%QQN
          ! neighbor position in levelwise list
          neighLvlPos = scheme%levelDesc( iLevel )%neigh(1) &
            &                 %nghElems( iNeigh, statePos )
          if (neighLvlPos > 0 .and. neighLvlPos <= scheme&
            &                       %levelDesc(iLevel)%offset(2, eT_fluid)) then
            ! position of neighbor in actual tree
            neighPos = scheme%levelDesc(iLevel)%pntTID(neighLvlPos)
            ! neighbor exist in same level
            if (neighPos > 0) then
              nSrcElems = nSrcElems + 1
              srcElemPos(nSrcElems)  = neighPos
            else
              ! if neighbor not in same level then use tem_PosOfId
              neighPos = abs(tem_PosOfId( scheme%levelDesc(iLevel) &
                &                               %total(neighLvlPos),    & 
                &            tree%treeID ))
              if (neighPos > 0) then
                nSrcElems = nSrcElems + 1
                srcElemPos(nSrcElems)  = neighPos 
              end if  
            end if !neighbor in same level
          end if
        end do !iNeigh

        ! calculate weight
        do iSrc = 1, nSrcElems
          bary = tem_BaryOfID(tree, tree%treeID( srcElemPos(iSrc) ))
          dist = abs(bary - point(iPnt,:))
          weights(iSrc) = 1.0_rk/sqrt(dot_product( dist, dist ))
        end do
      end if  

      sumWeight = sum(weights)
      weights = weights/sumWeight

      ! get source element values
      call varSys%method%val(fun%myPos)%get_element( &
         & varSys  = varSys,                         &
         & elemPos = srcElemPos(1:nSrcElems),        &
         & time    = time,                           &
         & tree    = tree,                           &
         & nElems  = nSrcElems,                      &
         & nDofs   = 1,                              &
         & res     = srcRes                          ) 
      

      ! Linear interpolation res = sum(weight_i*phi_i)
      pntVal = 0.0_rk
      do iSrc = 1, nSrcElems
        pntVal(:) = pntVal(:) + weights(iSrc) &
          & * srcRes( (iSrc-1)*fun%nComponents+1 : iSrc*fun%nComponents )
      end do
!      write(*,*) 'Linear intp pntVal ', pntVal
!      pntVal = mus_interpolate_quadratic2d_leastSq( &
!        &          srcMom       = srcRes,           & 
!        &          targetCoord  = point(iPnt,1:2),  &
!        &          nSourceElems = nSrcElems         )
!      write(*,*) 'quadratic intp pntVal ', pntVal
      res( (iPnt-1)*fun%nComponents+1 : iPnt*fun%nComponents ) = pntVal

    end do !iPnt 
  end subroutine mus_deriveVar_ForPoint
  ! ************************************************************************* !


  ! ************************************************************************* !
  !> This routine takes points coordinates, stores them in the method_data and
  !! return indices where points are located in the growing array of points or
  !! values ( sometimes we do not need to store the points )
  !! It is need to setup points for every variable. Points will be provided by 
  !! boundaries or sources depends on what uses the variable. This points do not
  !! change with time . This indices will be stored in corresponding boundary 
  !! or source to evaluate a value on that point later using 
  !! tem_varSys_proc_getValOfIndex.
  subroutine mus_stateVar_setupIndices( fun, varSys, point, offset_bit, &
    &                                   iLevel, tree, nPnts, idx        )
    ! -------------------------------------------------------------------------- !
    !> Description of the method to obtain the variables, for this routine 
    !! we need the location where to store the points.
    class(tem_varSys_op_type), intent(in) :: fun
    !> The variable system to obtain the variable from.
    type(tem_varSys_type), intent(in) :: varSys
    !>  arrays of points for which the indices are returned
    real(kind=rk), intent(in) :: point(:,:)
    !> Offset bit encoded as character for every point.
    !! If not present default is to center i.e offset_bit = achar(1+4+16)
    character, optional, intent(in) :: offset_bit(:)
    !> the point data need to be loaded levelwise, we need the current level
    integer, intent(in) :: iLevel
    !> global treelm mesh info
    type(treelmesh_type), intent(in) :: tree
    !> number of points
    integer, intent(in) :: nPnts
    integer, intent(out) :: idx(:)
    ! -------------------------------------------------------------------------- !
    integer :: iPoint
    integer :: coord(4)
    integer :: elemPos, neighPos, iNeigh
    type(mus_varSys_data_type), pointer :: fPtr
    type(mus_scheme_type), pointer :: scheme
    real(kind=rk) :: bary(3)
    real(kind=rk), allocatable :: weights(:)
    integer, allocatable :: srcElemPos(:)
    real(kind=rk) :: dist(3), sumWeight
    integer :: iSrc, nSrcElems, eType
    integer(kind=long_k) :: treeID
    ! -------------------------------------------------------------------------- !
    
    write(logUnit(4),*) 'setup indices for the points of variable ', &
      &                 trim(varSys%varname%val(fun%myPos))

    call C_F_POINTER( fun%method_Data, fPtr )
    scheme => fPtr%solverData%scheme

    allocate(srcElemPos(scheme%layout%fStencil%QQ))
    allocate(weights(scheme%layout%fStencil%QQ))

    idx = 0
    ! store elemPos of each point in global treeID list
    do iPoint = 1, nPnts
      weights = 0.0_rk

      ! Get the treeID of the element and then get the element position
      ! in the levelwise array
      coord =  tem_CoordOfReal( mesh  = tree,            &
        &                       point = point(iPoint,:), &
        &                       level = iLevel           )
      
      treeID = tem_IdOfCoord(coord)
      ! Find treeID in levelDesc 
      ! KM: Excluded halo elements as they require communication of all links
      do eType = eT_minRelevant, eT_maxRelevant-1
        elemPos = tem_treeIDinTotal( &
          &           tID       = treeID,                        &
          &           levelDesc = scheme%levelDesc(iLevel), &
          &           eType     = eType                          )
        ! found element in total list
        if (elemPos > 0) exit   
      end do  

      ! store point data only if element position is found in total list
      if (elemPos>0) then

        ! append grw array for this point with levelwise element position 
        call append(me  = fPtr%pointData%pntLvl(iLevel)%elemPos, &
          &         val = elemPos                                )
        ! Append points to growing array of points
        call append(me  = fPtr%pointData%pntLvl(iLevel)%grwPnt, &
          &         val = point(iPoint,:)                       )
  
        ! First position of source elements
        call append(me  = fPtr%pointData%pntLvl(iLevel)%srcElem%first,  &
          &         val = fPtr%pointData%pntLvl(iLevel)%srcElem%elemPos &
          &               %nVals + 1                                    )

        ! use this element as 1st source element for interpolation
        nSrcElems = 1
        srcElemPos(nSrcElems)  = elemPos 

        bary = tem_BaryOfID(tree, scheme%levelDesc(iLevel)%total(elemPos))
        dist = abs(bary - point(iPoint,:))
        ! if point is exact bary center of current element then 
        ! no need to do interpolation
        if ( all(dist == 0.0_rk) ) then
          weights(nSrcElems) = 1.0_rk
        else  
          ! get existing neighbor elements in actual tree
          do iNeigh = 1, scheme%layout%fStencil%QQN
            ! neighbor position in levelwise list
            neighPos = scheme%levelDesc( iLevel )%neigh(1) &
              &                   %nghElems( iNeigh, elemPos )
            ! neighbor exist in same level
            if (neighPos > 0) then
              nSrcElems = nSrcElems + 1
              srcElemPos(nSrcElems)  = neighPos
            end if !neighbor in same level
          end do !iNeigh
          ! calculate weight
          do iSrc = 1, nSrcElems
            bary = tem_BaryOfID(tree, scheme%levelDesc(iLevel)    &
              &                              %total( srcElemPos(iSrc) ))
            dist = abs(bary - point(iPoint,:))
            weights(iSrc) = 1.0_rk/sqrt(dot_product( dist, dist ))
          end do
        end if  
  
        ! normalize weights
        sumWeight = sum(weights)
        weights = weights/sumWeight

        call append(me  = fPtr%pointData%pntLvl(iLevel)%srcElem%elemPos, &
          &         val = srcElemPos(1:nSrcElems)                        )

        call append(me  = fPtr%pointData%pntLvl(iLevel)%srcElem%weight, &
          &         val = weights(1:nSrcElems)                          )

        ! last position of source elements  
        call append(me  = fPtr%pointData%pntLvl(iLevel)%srcElem%last,   &
          &         val = fPtr%pointData%pntLvl(iLevel)%srcElem%elemPos &
          &               %nVals                                        )
        ! store the index, 
        idx(iPoint)= fPtr%pointData%pntLvl(iLevel)%elemPos%nVals
      end if
    end do
 
    fPtr%pointData%pntLvl(iLevel)%nPnts = fPtr%pointData%pntLvl(iLevel)%grwPnt &
      &                                       %coordX%nVals
    call truncate(fPtr%pointData%pntLvl(iLevel)%grwPnt)
    call truncate(fPtr%pointData%pntLvl(iLevel)%elemPos)
    call truncate(fPtr%pointData%pntLvl(iLevel)%srcElem%first)
    call truncate(fPtr%pointData%pntLvl(iLevel)%srcElem%last)
    call truncate(fPtr%pointData%pntLvl(iLevel)%srcElem%elemPos)
    call truncate(fPtr%pointData%pntLvl(iLevel)%srcElem%weight)
  
  end subroutine mus_stateVar_setupIndices
  ! ************************************************************************* !


  ! ************************************************************************* !
  !> Routine to get the actual value for a given array of indices. 
  !! The indices belong to the grwarray of points storing levelwise in 
  !! Pointdata%pntLvl(iLevel).
  !! Hence this routines takes the indeices as input, can refer to the pointData
  !! and evaluate the variable and returns the values
  subroutine mus_stateVar_getValOfIndex( fun, varSys, time, iLevel, idx, &
    &                                    idxLen,  nVals,  res            )
    ! -------------------------------------------------------------------------- !
    !> Description of the method to obtain the variables, 
    class(tem_varSys_op_type), intent(in) :: fun

    !> The variable system to obtain the variable from.
    type(tem_varSys_type), intent(in) :: varSys

    !> Point in time at which to evaluate the variable.
    type(tem_time_type), intent(in) :: time
    
    !> Level on which values are requested
    integer, intent(in) :: iLevel

    !> Index of points in the growing array and variable val array to 
    !! return.
    !! Size: n
    integer, intent(in) :: idx(:)

    !> With idx as start index in contiguous memory,
    !! idxLength defines length of each contiguous memory
    !! Size: dependes on number of first index for contiguous array,
    !! but the sum of all idxLen is equal to nVals
    integer, optional, intent(in) :: idxLen(:)

    !> Number of values to obtain for this variable (vectorized access).
    integer, intent(in) :: nVals

    !> Resulting values for the requested variable.
    !!
    !! Dimension: n requested entries x nComponents of this variable
    !! Access: (iElem-1)*fun%nComponents + iComp
    real(kind=rk), intent(out) :: res(:)
    ! -------------------------------------------------------------------------- !
    type(mus_varSys_data_type), pointer :: fPtr
    type(mus_scheme_type), pointer :: scheme
    integer :: iComp, iSrc, iVal, elemPos, first, last, nSize
    integer :: nSrcElems
    real(kind=rk), allocatable :: srcRes(:), pntVal(:)
    real(kind=rk) :: weight
    ! -------------------------------------------------------------------------- !

    write(logUnit(4),*) 'get the values of indices for variable ',  &
      &                  trim(varSys%varname%val(fun%myPos))
      
    call C_F_POINTER( fun%method_Data, fPtr )
    scheme => fPtr%solverData%scheme
    allocate(srcRes(scheme%layout%fStencil%QQ*fun%nComponents))
    allocate(pntVal(fun%nComponents))

    nSize = scheme%pdf( iLevel )%nSize

    res = 0.0_rk

    ! distinguish if we have an array of index or we have contingous memory
    ! access where index are always first entries!
    if (present(idxLen)) then
      call tem_abort('Error: idxLen is not supported in get_valOfIndex for ' &
        &          //'state variable')
    else  
      ! Get the state value at the specific point
      do iVal = 1, nVals
        if (idx(iVal)>0) then
  
          ! elemPos in levelwise list 
          elemPos = fPtr%pointData%pntLvl(iLevel)%elemPos%val(idx(iVal))
  
          first = fPtr%pointData%pntLvl(iLevel)%srcElem%first%val(idx(iVal))
          last = fPtr%pointData%pntLvl(iLevel)%srcElem%last%val(idx(iVal))

          ! get pdf's of source elements
          srcRes = 0.0_rk
          nSrcElems = 0
          do iSrc = first, last
            elemPos = fPtr%pointData%pntLvl(iLevel)%srcElem%elemPos%val(iSrc)
            nSrcElems = nSrcElems + 1
            do iComp = 1, fun%nComponents
              srcRes( (nSrcElems-1)*fun%nComponents + iComp )          &
                & = scheme%state( iLevel )%val(                        &
& ?IDX?( fun%state_varPos(iComp),elemPos,varSys%nScalars,nSize ),      &
                & scheme%pdf( iLevel )%nNext )
            end do !iComp
          end do !iSrc 

          ! Linear interpolation res = sum(weight_i*phi_i)
          pntVal = 0.0_rk
          nSrcElems = 0
          do iSrc = first, last
            weight = fPtr%pointData%pntLvl(iLevel)%srcElem%weight%val(iSrc)
            nSrcElems = nSrcElems + 1
            pntVal(:) = pntVal(:) + weight &
              & * srcRes( (nSrcElems-1)*fun%nComponents+1 &
              &         : nSrcElems*fun%nComponents )
          end do

          ! get the state value for each component of this 
          do iComp = 1, fun%nComponents
            res( (iVal-1)*fun%nComponents+iComp ) = pntVal(iComp)
          end do !iComp

        end if !idx>0
      end do !iVal
    end if ! idx_len  
    
  end subroutine mus_stateVar_getValOfIndex 
  ! ************************************************************************* !

  ! ************************************************************************* !
  !> Routine to get the actual value for a given array of indices for
  !! musubi derive variables
  !! The indices belong to the grwarray of points storing levelwise in 
  !! Pointdata%pntLvl(iLevel).
  !! Hence this routines takes the indeices as input, can refer to the pointData
  !! and evaluate the variable and returns the values
  subroutine mus_deriveVar_getValOfIndex( fun, varSys, time, iLevel, idx,  &
    &                                     idxLen,  nVals,  res       )
    ! -------------------------------------------------------------------------- !
    !> Description of the method to obtain the variables, 
    class(tem_varSys_op_type), intent(in) :: fun

    !> The variable system to obtain the variable from.
    type(tem_varSys_type), intent(in) :: varSys

    !> Point in time at which to evaluate the variable.
    type(tem_time_type), intent(in) :: time
    
    !> Level on which values are requested
    integer, intent(in) :: iLevel

    !> Index of points in the growing array and variable val array to 
    !! return.
    !! Size: n
    integer, intent(in) :: idx(:)

    !> With idx as start index in contiguous memory,
    !! idxLength defines length of each contiguous memory
    !! Size: dependes on number of first index for contiguous array,
    !! but the sum of all idxLen is equal to nVals
    integer, optional, intent(in) :: idxLen(:)

    !> Number of values to obtain for this variable (vectorized access).
    integer, intent(in) :: nVals

    !> Resulting values for the requested variable.
    !!
    !! Dimension: n requested entries x nComponents of this variable
    !! Access: (iElem-1)*fun%nComponents + iComp
    real(kind=rk), intent(out) :: res(:)
    ! -------------------------------------------------------------------------- !
    call tem_abort('Error: in mus_deriveVar_getValOfIndex. Not implemented')
  end subroutine mus_deriveVar_getValOfIndex

end module mus_varSys_module
! **************************************************************************** !
