! See copyright notice in the COPYRIGHT file.
! ****************************************************************************** !
!> author: Manuel Hasert
!! author: Jiaxing Qi
!! Interpolation header to load confiugration and type definition
module mus_interpolate_header_module

  ! include aotus modules
  use aotus_module,     only: flu_State, aot_get_val
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length
  use aot_out_module,   only: aot_out_type, aot_out_val

  ! include treelm modules
  use env_module,            only: rk, labelLen
  use tem_aux_module,        only: tem_abort
  use tem_math_module,       only: matrix_type
  use tem_logging_module,    only: logUnit
  use tem_debug_module,      only: dbgUnit
  use tem_varSys_module,     only: tem_varSys_type
  use tem_construction_module, only: tem_levelDesc_type

  ! include musubi modules
  ! use mus_field_prop_module,     only: mus_field_prop_type
  use mus_fluid_module,          only: fluid_type
  use mus_pdf_module,            only: pdf_data_type
  use mus_scheme_layout_module,  only: mus_scheme_layout_type
  use mus_physics_module,        only: mus_physics_type

  implicit none

  private

  public :: mus_interpolation_type
  public :: mus_load_interpolate
  public :: mus_interpolate_out
  public :: mus_set_nSources

  !> cxDir for compact stencil
  integer, public :: compactNeighDir(4,8)

  !> definition of the used interpolation method
  type mus_interpolation_type

    !> name of this interpolation method
    character(len=labelLen) :: label

    !> Interpolation test by comparing against the initial condition
    logical :: testInterpolation = .false.
    logical :: testEachElement   = .false.
    logical :: testFluids        = .false.
    logical :: noIntpFromFiner   = .false.
    logical :: noIntpFromCoarser = .false.

    logical :: useCompact        = .false.

    !> high order interpolation fine to coarse for ghostFromFiner elements
    procedure( intpRoutine ),nopass,pointer :: fillMineFromFiner => null()
    !> low order interpolation fine to coarse for ghostFromFiner elements
    procedure( intpRoutine ),nopass,pointer :: fillMineFromFiner_low => null()

    !> high order interpolation coarse to fine for ghostFromCoarser elements
    procedure( intpRoutine ),nopass,pointer :: fillFinerFromMe => null()
    !> low order interpolation coarse to fine for ghostFromCoarser elements
    procedure( intpRoutine ),nopass,pointer :: fillFinerFromMe_low => null()

    !> how many source elements are required by HIGH interpolation scheme
    !! After balance, these values are updated inside mus_construct
    integer :: nSourcesFromFiner = 0
    integer :: nSourcesFromCoarser = 0
    ! @todo: this should be removed
    integer :: nSourcesCompactFromCoarser = 0

  end type mus_interpolation_type

! ****************************************************************************** !
  abstract interface
    !> This is the interface for all interpolation methods that
    !  can be called from outside.
    subroutine intpRoutine(  tLevelDesc, omLvl, level, momBuf, &
      &               sState, snSize, tState, tnSize, layout, &
      &               nMaxSources, nTargets, targetList, physics, time, nScalars )
      import :: fluid_type, mus_scheme_layout_type,            &
        &       tem_levelDesc_type, mus_physics_type, rk !, tem_varSys_type

      !> my refinement level
      integer,intent(in) :: level

      !> State vector of SOURCE elements
      real(kind=rk),    intent(in) :: sState(:)
      ! integer,          intent(in) :: sNeigh(:)
      integer,          intent(in) :: snSize

      !> State vector of TARGET GHOST elements
      real(kind=rk), intent(inout) :: tState(:)
      ! integer,          intent(in) :: tNeigh(:)
      integer,          intent(in) :: tnSize

      !> buffer for moments of source elements
      real(kind=rk), intent(inout) :: momBuf(:,:)

      !> level descriptor on target level
      type( tem_levelDesc_type ), intent(in) :: tLevelDesc

      !> Omega on each level
      real(kind=rk), intent(in) :: omLvl(level:level+1)

      !> the layout used
      type( mus_scheme_layout_type ), intent(in) :: layout

      !> Max number of sources for each target ghost
      integer, intent(in) :: nMaxSources

      !> List of target elements ( their position in depSource list )
      integer, intent(in) :: nTargets
      integer, intent(in) :: targetList(nTargets)

      !> physics type to convert lattice to physics SI unit and vice versa
      type(mus_physics_type), intent(in) :: physics

      !> time required by analytical solution
      real(kind=rk), intent(in) :: time

      !> nScalars
      integer, intent(in) :: nScalars

    end subroutine intpRoutine

  end interface
! ****************************************************************************** !

  contains

! ****************************************************************************** !
  !> Read in the type of interpolation scheme
  !!
  !!```lua
  !! interpolation_method = 'linear'               -- simple definition
  !! interpolation_method = {'debug', value = 1.} -- definition in a table
  !!```
  !!
  subroutine mus_load_interpolate( me, conf, parent )
    ! ---------------------------------------------------------------------------
    !> interpolation type to load info to
    type(mus_interpolation_type), intent(out)  :: me
    !> lua state to load from
    type(flu_state)                        :: conf
    !> optional parent table to load from
    integer, optional, intent(in)          :: parent
    ! ---------------------------------------------------------------------------
    character(len=32) :: localKey
    integer :: iError, nEntries, intp_handle
    ! ---------------------------------------------------------------------------

    localKey = 'interpolation_method'

    ! if more than one variable is defined in a table then
    ! ref_value should also be a table with reference value for each variable
    ! defined in the variable table.
    if( present( parent )) then
      call aot_table_open( L=conf, thandle=intp_handle, key=localKey,          &
        &                  parent= parent )
    else
      call aot_table_open( L=conf, thandle=intp_handle, key=localKey)
    endif

    ! not a table, just the name of the interpolation method
    if (intp_handle == 0) then
      call aot_get_val(L=conf, key=localKey,                                   &
        &              val=me%label,                                           &
        &              ErrCode=iError, default='quadratic')
    else
      ! interpolation is defined as a table with optional additional information
      ! Get size of table
      nEntries = aot_table_length(L=conf, thandle=intp_handle)
      call aot_get_val(L=conf, thandle=intp_handle,                            &
        &              val=me%label, ErrCode=iError,                           &
        &              key = 'method', default='quadratic')

      call aot_get_val(L=conf, thandle=intp_handle,                            &
        &              val=me%testInterpolation, ErrCode=iError,               &
        &              key = 'test', default=.false.)
      if( me%testInterpolation ) then
        write(logUnit(1),*)' Activated the Interpolation test.'
        call aot_get_val(L=conf, thandle=intp_handle,                          &
          &              val=me%testEachElement, ErrCode=iError,               &
          &              key = 'testEach', default=.false.)
        if( me%testEachElement   ) then
          write(logUnit(1),*)' Testing each element... '
        end if
        call aot_get_val(L=conf, thandle=intp_handle,                          &
          &              val=me%testFluids, ErrCode=iError,                    &
          &              key = 'testFluids', default=.false.)
        if( me%testFluids ) then
          write(logUnit(1),*)' Testing fluid element... '
        end if
      end if
      call aot_get_val(L=conf, thandle=intp_handle,                            &
        &              val=me%noIntpFromFiner, ErrCode=iError,                 &
        &              key = 'noIntpFromFiner', default=.false.)

      call aot_get_val(L=conf, thandle=intp_handle,                            &
        &              val=me%noIntpFromCoarser, ErrCode=iError,               &
        &              key = 'noIntpFromCoarser', default=.false.)

    endif
    call aot_table_close(L=conf, thandle=intp_handle)

  end subroutine mus_load_interpolate
! ****************************************************************************** !

! ****************************************************************************** !
  subroutine mus_set_nSources( me, nDims, QQ )
    ! ---------------------------------------------------------------------------
    type(mus_interpolation_type) :: me
    integer,     intent(in)  :: nDims, QQ
    ! ---------------------------------------------------------------------------

    ! Determine, how many elements are required
    me%nSourcesFromFiner   = 0
    me%nSourcesFromCoarser = 0
    select case (trim(me%label) )
      case('none', 'TGV_2D' )
      case('debug', 'analytic')
        me%nSourcesFromFiner   = 1
        me%nSourcesFromCoarser = 1
      case('average')
        me%nSourcesFromFiner = 1
        me%nSourcesFromCoarser = 1
      case('eqneq', 'linear')
        me%label = 'linear'
        me%useCompact = .true.
        me%nSourcesFromFiner   = 4
        me%nSourcesFromCoarser = 4
        compactNeighDir = init_cxDirCompact( QQ )
      case('compact', 'compact_moment')
        ! me%label = 'compact'
        me%useCompact = .true.
        me%nSourcesFromFiner   = 4
        me%nSourcesFromCoarser = 4
        compactNeighDir = init_cxDirCompact( QQ )
      case('quadratic' )
        me%nSourcesFromFiner = 4
        if ( QQ == 27 ) then
          me%nSourcesFromCoarser = 19
        else
          me%nSourcesFromCoarser = QQ
        end if
        ! compactNeighDir = init_cxDirCompact( QQ )
      case default
        write(logUnit(1),*)' No valid interpolation method selected'
        call tem_abort()
    end select ! me%label


    if    ( nDims == 1) then
      me%nSourcesCompactFromCoarser = min( 2, me%nSourcesFromCoarser )
    elseif( nDims == 2) then
      me%nSourcesCompactFromCoarser = min( 4, me%nSourcesFromCoarser )
    elseif( nDims == 3) then
      me%nSourcesCompactFromCoarser = min( 7, me%nSourcesFromCoarser )
    endif

    write(logUnit(1),"(A)") 'Setting for interpolation scheme: '//trim(me%label)
    write(logUnit(3),"(A,I0)") '  number of sources from   finer: ', &
      &                        me%nSourcesFromFiner
    write(logUnit(3),"(A,I0)") '  number of sources from coarser: ', &
      &                        me%nSourcesFromCoarser
    write(logUnit(3),"(A,L1)") '  use compact stencil: ', me%useCompact

  end subroutine mus_set_nSources
! ****************************************************************************** !


! ****************************************************************************** !
  !> Dump interpolation method
  subroutine mus_interpolate_out( me, conf )
    ! ---------------------------------------------------------------------------
    !> interpolation type to dump info to
    type(mus_interpolation_type), intent(in)  :: me
    !> aotus type handling the output to the file in lua format
    type(aot_out_type), optional, intent(inout) :: conf
    ! ---------------------------------------------------------------------------
    call aot_out_val( put_conf = conf, vname = 'interpolation_method', &
      &               val = trim(me%label) )

  end subroutine mus_interpolate_out
! ****************************************************************************** !

! ****************************************************************************** !
  function init_cxDirCompact( QQ ) result( me )
    ! ---------------------------------------------------------------------------
    integer, intent(in)  :: QQ
    ! integer, intent(in)  :: cxDir(3,QQ)
    integer              :: me(4,8)
    ! ---------------------------------------------------------------------------
    integer :: iChild
    ! D2Q9 directions
    integer,parameter :: qN0 = 1  !< west             x-
    integer,parameter :: q0N = 2  !< south            y-
    integer,parameter :: q10 = 3  !< east             x+
    integer,parameter :: q01 = 4  !< north            y+
    integer,parameter :: qNN = 5  !<                  y-,x-
    integer,parameter :: qN1 = 6  !<                  y+,x-
    integer,parameter :: q1N = 7  !<                  y-,x+
    integer,parameter :: q11 = 8  !<                  y+,x+
    integer,parameter :: q00 = 9  !< rest
    ! D3Q27 directions
    integer,parameter :: qN00 = 1   !< west             x-
    integer,parameter :: q0N0 = 2   !< south            y-
    integer,parameter :: q00N = 3   !< bottom           z-
    integer,parameter :: q100 = 4   !< east             x+
    integer,parameter :: q010 = 5   !< north            y+
    integer,parameter :: q001 = 6   !< top              z+

    integer,parameter :: q0NN = 7   !<                  z-,y-
    integer,parameter :: q0N1 = 8   !<                  z+,y-
    integer,parameter :: q01N = 9   !<                  z-,y+
    integer,parameter :: q011 = 10  !<                  z+,y+
    integer,parameter :: qN0N = 11  !<                  x-,z-
    integer,parameter :: q10N = 12  !<                  x+,z-
    integer,parameter :: qN01 = 13  !<                  x-,z+
    integer,parameter :: q101 = 14  !<                  x+,z+
    integer,parameter :: qNN0 = 15  !<                  y-,x-
    integer,parameter :: qN10 = 16  !<                  y+,x-
    integer,parameter :: q1N0 = 17  !<                  y-,x+
    integer,parameter :: q110 = 18  !<                  y+,x+

    integer,parameter :: qNNN = 19  !<                  x-,z-
    ! integer,parameter :: qNN1 = 20  !<                  x+,z-
    ! integer,parameter :: qN1N = 21  !<                  x-,z+
    integer,parameter :: qN11 = 22  !<                  x+,z+
    ! integer,parameter :: q1NN = 23  !<                  y-,x-
    integer,parameter :: q1N1 = 24  !<                  y+,x-
    integer,parameter :: q11N = 25  !<                  y-,x+
    ! integer,parameter :: q111 = 26  !<                  y+,x+
    integer,parameter :: q000 = 27  !< rest density is last
    ! ---------------------------------------------------------------------------

    select case (QQ)
    case (9)
      me(:,:) = reshape( [ qNN, q0N, qN0, q00, & ! A
        &                   q0N, q1N, q00, q10, & ! B
        &                   qN0, q00, qN1, q01, & ! C
        &                   q00, q10, q01, q11, & ! D
        &                   qNN, q0N, qN0, q00, & ! A
        &                   q0N, q1N, q00, q10, & ! B
        &                   qN0, q00, qN1, q01, & ! C
        &                   q00, q10, q01, q11  ], [4,8] ) ! D
    case (27)
      me(:,:) = reshape( [ qNNN, q00N, q0N0, qN00, & ! child 1
        &                   q0NN, q10N, q1N0, q000, & ! child 2
        &                   qN0N, q01N, q000, qN10, & ! child 3
        &                   q00N, q11N, q100, q010, & ! child 4
        &                   qNN0, q000, q0N1, qN01, & ! child 5
        &                   q0N0, q100, q1N1, q001, & ! child 6
        &                   qN00, q010, q001, qN11, & ! child 7
        &                   q000, q110, q101, q011 ], [4,8] )
    case default
      !write(logUnit(1),"(A)") 'Compact interpolation requires D2Q9 or D3Q27!'
      call tem_abort()
    end select

    write(dbgUnit(1),"(A)") ''
    write(dbgUnit(1),"(A)") ' compact stencil dir '
    do iChild = 1, 8
      ! me(1:3,iNeigh,iChild) = cxDir(1:3, dir(iNeigh, iChild) )
      write(dbgUnit(1), "(A,I0,A,4I3)")  'childNum: ', iChild, &
        &                                ', dir: ', me(1:4,iChild)
    end do
    write(dbgUnit(1),"(A)") ''
  end function init_cxDirCompact
! ****************************************************************************** !

end module mus_interpolate_header_module
! ****************************************************************************** !
