! See copyright notice in the COPYRIGHT file.
! ************************************************************************** !
!> author: Jiaxing Qi
!! author: Manuel Hasert
!! Linear Interpolation of flow quantities between different grid levels
!!
!! # Interpolation 
!! 
!! The routines defined here, fill up the ghost elements with valid data.
!! Ghost elements are employed at grid level interfaces to provide valid
!! pdf values to the neighboring fluid elements. This way, the solvers can 
!! act on elements of the same size only, treating the levels successively.
!! Target elements are the ghost elements, which have to be filled with 
!! valid values.
!! Source elements are the fluid elements from other levels, from where to 
!! take the input values for the interpolation.
!! The target ghost elements on the target level have corresponding source 
!! fluid elements on the source level.
!!
!! [[tem_topology_module]] For a detailed description of the grid 
!!
!! # Workflow
!!
!! Each interpolation routine acts on a list of ghost elements.
!! This list contains pointers to the position in the total list.
!! For each of these ghost elements, the source elements are identified.
!! Before that, the sourceLevel is identified. However, the code is restricted
!! to work with a level jump of only one level, so the sourceLevel is:\n
!! for fillMyGhostsFromFiner: sourceLevel = targetLevel + 1
!! for fillFinerGhostsFromMe: sourceLevel = targetLevel - 1
!!
!! For an overview over implemented interpolation methods, see
!! [Interpolation methods](../page/features/intp_methods.html)
!!
?? include 'header/lbm_macros.inc'
?? include 'header/lbm_deriveMacros.inc'
?? include 'header/lbm_interfaceMacros.inc'
?? include 'header/lbm_d2q9Macros.inc'
?? include 'treelm/source/logMacros.inc'
module mus_interpolate_linear_module

  ! include treelm modules
  use env_module,            only: rk
  use tem_element_module,    only: eT_GhostFromCoarser, &
    &                              eT_ghostFromFiner
  use tem_param_module,      only: cs2inv, cs2, PI, div1_2, div1_9, div4_9,&
    &                              div1_36, div2_3
  use tem_comm_env_module,   only: tem_comm_env_type
  use tem_debug_module,      only: dbgUnit
  use tem_construction_module, only: tem_levelDesc_type

?? IF (loglvl >= 1) THEN
  use env_module,            only: long_k
  use tem_topology_module,   only: tem_coordOfID
  use tem_param_module,      only: PI
?? ENDIF

  ! include musubi modules
?? IF( .not. SOA) then
  use mus_interpolate_debug_module,  only: TGV_2D
?? ENDIF
  use mus_scheme_layout_module,      only: mus_scheme_layout_type
  use mus_derVarPos_type_module,     only: mus_derVarPos_type
  use mus_derivedQuantities_module2, only: convPrePost, getNEq_diffusive, &
    &                                      getShearRateTensor_diffusive,  &
    &                                      getStrainFacDffs
  use mus_physics_module,            only: mus_physics_type
  ! use mus_moments_module,            only: get_moment

  implicit none

  private

  public :: fillMyGhostsFromFiner_linearDiffusive
  public :: fillFinerGhostsFromMe_linearDiffusive
  public :: fillMyGhostsFromFiner_linear
  public :: fillFinerGhostsFromMe_linear
  public :: fillMyGhostsFromFiner_linearIncomp
  public :: fillFinerGhostsFromMe_linearIncomp
  public :: fillMyGhostsFromFiner_linMoment3d

  contains

! ****************************************************************************** !
  !> [Linear interpolation](../page/features/intp_methods.html) of ghostFromFiner
  !! This routine is used by diffusive linear interpolation scheme.
  !!
?? copy :: interpolateRoutine_header( fillMyGhostsFromFiner_linearDiffusive )
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourcePos      ! source element position in total list
    integer :: targetLevel    ! level of target elements
    integer :: targetPos      ! target element position in total list
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iVal, jVal     ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    integer :: iVar           ! current field variable
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceP(nMaxSources)
    real(kind=rk) :: sourceV(3,nMaxSources)
    real(kind=rk) :: sourceS(3,3, nMaxSources)
    ! macroscopic velocities for all source elements
    real(kind=rk) :: targetR
    real(kind=rk) :: targetP
    real(kind=rk) :: targetV(3)
    real(kind=rk) :: targetS(3,3)
    real(kind=rk) :: fTmp( layout%fStencil%QQ )
    ! equilibrium part of the resulting pdf
    real(kind=rk) :: fEq( layout%fStencil%QQ )
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: fNeq( layout%fStencil%QQ )
    !> Strain tensor of the target element
    real(kind=rk) :: rho0
    real(kind=rk) :: pFac, vFac
    !> debug variables
?? IF (loglvl >= 1) THEN
    integer(kind=long_k) :: targetID
    integer :: coord(4)
    real(kind=rk) :: targetCoord(3), dx
    real(kind=rk) :: mom(layout%fStencil%QQ)
?? ENDIF
    integer :: QQ
    ! ---------------------------------------------------------------------------

    QQ = layout%fStencil%QQ
    sourceLevel = level + 1
    targetLevel = level
    iVar = 1
    rho0 = 1.0_rk

    pFac = physics%pfac( sourceLevel, targetLevel )
    vFac = physics%vfac( sourceLevel, targetLevel )

! ----------------------------------------------------------------------------------
?? IF (loglvl >= 1) THEN
write(dbgUnit(1),*) '----- Enter fillMyGhostsFromFiner_linearDiffusive -----------'
write(dbgUnit(1),*) ''
write(dbgUnit(1),*) 'source level: ', sourceLevel
write(dbgUnit(1),*) '  omega: ', omLvl( sourceLevel )
write(dbgUnit(1),*) '  pressure factor: ',physics%fac( sourceLevel )%press
write(dbgUnit(1),*) '  velocity factor: ',physics%fac( sourceLevel )%vel
write(dbgUnit(1),*) '  strain   factor: ',physics%fac( sourceLevel )%strainRate
write(dbgUnit(1),*) 'target level: ', targetLevel
write(dbgUnit(1),*) '  omega: ', omLvl( targetLevel )
write(dbgUnit(1),*) '  pressure factor: ',physics%fac( targetLevel )%press
write(dbgUnit(1),*) '  velocity factor: ',physics%fac( targetLevel )%vel
write(dbgUnit(1),*) '  strain   factor: ',physics%fac( targetLevel )%strainRate
write(dbgUnit(1),*) 'scale factors:'
write(dbgUnit(1),*) ' pressure: ', pFac
write(dbgUnit(1),*) ' velocity: ', vFac
write(dbgUnit(1),*) 'number of target elements to fill: ', nTargets
write(dbgUnit(1),*) ''
?? ENDIF
! ----------------------------------------------------------------------------------

    ! Treat all coarse target elements
    do indElem = 1, nTargets

      iElem = targetList( indElem )

      ! Get its position in state array
      targetPos = iElem + tLevelDesc%offset( 1, eT_ghostFromFiner)

      ! Find out how many source elements we have for interpolation.
      nSourceElems = tLevelDesc%depFromFiner( iElem )%elem%nVals

! ----------------------------------------------------------------------
?? IF (loglvl >= 1) THEN
! DEBUG: get TGV 2D analytical solution
targetID   = tLevelDesc%total( targetPos )

write(dbgUnit(1),"(A, I0)") 'target treeID: ', targetID
write(dbgUnit(1),"(A, I0)") 'number of source elements: ', nSourceElems
?? ENDIF

?? IF (loglvl >= 2) THEN
! get target bary
coord       = tem_CoordOfId(targetID)
dx          = 2.0_rk * PI / real(2**coord(4), rk)
targetCoord = (real(coord(:3), rk) + 0.5_rk) * dx

! get pressure, velocity, strainRate
mom(1:6) = TGV_2D( targetCoord, time )
write(dbgUnit(1),"(A)") 'Analytical TGV 2D target values:'
write(dbgUnit(1),"(A,  F10.5)") ' p: ', mom(1)
write(dbgUnit(1),"(A, 2F10.5)") ' v: ', mom(2), mom(3)
write(dbgUnit(1),"(A, 3F10.5)") ' s: ', mom(4), mom(5), mom(6)
write(dbgUnit(1),"(A)") ''
?? ENDIF
! ----------------------------------------------------------------------

      sourceP = 0._rk
      sourceV = 0._rk
      sourceS = 0._rk

      ! Now loop over all fine source elements for this target:
      do iSourceElem = 1, nSourceElems

        sourcePos = tLevelDesc%depFromFiner(iElem)%elem%val(iSourceElem)

        ! collect pdf
        do iDir = 1, QQ
           fTmp(iDir) = sstate( ?IDX?(iDir, sourcePos, nScalars, snSize))
        end do

        sourceP(iSourceElem)   = sum( fTmp ) - rho0
        sourceV(1,iSourceElem) = sum( fTmp * layout%fStencil%cxDirRK(1,:) )
        sourceV(2,iSourceElem) = sum( fTmp * layout%fStencil%cxDirRK(2,:) )
        sourceV(3,iSourceElem) = sum( fTmp * layout%fStencil%cxDirRK(3,:) )

        ! Calculate Strain rate
        sourceS(:,:,iSourceElem ) = getShearRateTensor_diffusive( &
          &                               f = fTmp, layout = layout,   &
          &               omega = omLvl( sourceLevel ))

! DEBUG output
?? IF (loglvl >= 1) THEN
write(dbgUnit(1),"(A, F12.8, A, 3F12.8, A, 3F12.8)") &
  &       'Source P: ', sourceP(iSourceElem), &
  &       '   V: ', sourceV(1,iSourceElem), sourceV(2,iSourceElem), sourceV(3,iSourceElem),&
  &       '   S: ', sourceS(1,1,iSourceElem), sourceS(2,2,iSourceElem), sourceS(1,2,iSourceElem)
?? ENDIF
      end do ! iSourceElem

?? IF (loglvl >= 1) THEN
write(dbgUnit(1),"(A)") ''
?? ENDIF

      ! ------------------ LINEAR INTERPOLATION -------------------
      ! This is all done in physical units

      ! pressure: average
      targetP    = sum(sourceP(1:nSourceElems)) / dble(nSourceElems)

      ! velocity: quadratic compact interpolate only works for 2D!
      targetV(1) = sum(sourceV(1,1:nsourceelems)) / dble(nsourceelems)
      targetV(2) = sum(sourceV(2,1:nsourceelems)) / dble(nsourceelems)
      targetV(3) = sum(sourceV(3,1:nsourceelems)) / dble(nsourceelems)

      ! strain rate: average
      do jVal = 1, layout%fStencil%nDims
        do iVal = 1, layout%fStencil%nDims
        targetS(iVal,jVal) =                                           &
          & sum( sourceS(iVal,jVal,1:nSourceElems)) / dble(nSourceElems)
        end do
      end do
      ! ------------------ LINEAR INTERPOLATION -------------------

?? IF (loglvl >= 1) THEN
write(dbgUnit(1),"(A)") 'Averaged target values:'
write(dbgUnit(1),"(A, F12.8, A, 3F12.8, A, 3F12.8)") &
  &       'Target P: ', targetP, &
  &       '   V: ', targetV(1), targetV(2), targetV(3),&
  &       '   S: ', targetS(1,1), targetS(2,2), targetS(1,2)
write(dbgUnit(1),"(A)") ''
?? ENDIF

        ! Scale quantities on source level to target level
        targetR = targetP * pFac + rho0
        targetV = targetV * vFac
        targetS = targetS * pFac

        ! Calculate fEq
?? copy :: Eq_incomp( fEq, targetR, targetV, layout, rho0 )

        ! Calculate fNeq
        fNeq(:) = getNEq_diffusive(                                         &
          &         layout = layout,                                        &
          &         omega  = omLvl( targetLevel ),  &
          &         Sxx    = targetS(:,:))

        ! Now write the resulting pdf in the current direction to the target
        ! Element position resulting pdf = f_eq + f_neq
        do iDir = 1, QQ
          tstate(?IDX?( iDir, targetPos, nScalars, tnSize)) = fEq( iDir ) + fNeq( iDir )
        end do ! iDir

    end do

?? IF (loglvl >= 1) THEN
write(dbgUnit(1), *) '---- Leave fillMyGhostsFromFiner_linearDiffusive -----------'
write(dbgUnit(1), *) ''
?? ENDIF

  end subroutine fillMyGhostsFromFiner_linearDiffusive
! ****************************************************************************** !


! ****************************************************************************** !
  !> [Linear interpolation](../page/features/intp_methods.html) of ghostFromFiner
  !!
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_linearDiffusive )
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourcePos      ! source element position in total list
    integer :: targetLevel    ! level of target elements
    integer :: targetPos      ! target element position in total list
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iVal, jVal     ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceP(nMaxSources)
    real(kind=rk) :: sourceV(3,nMaxSources)
    real(kind=rk) :: sourceS(3,3, nMaxSources)
    ! macroscopic velocities for all source elements
    real(kind=rk) :: targetR
    real(kind=rk) :: targetP
    real(kind=rk) :: targetV(3)
    real(kind=rk) :: targetS(3,3)
    ! temporary array
    real(kind=rk) :: fTmp( layout%fStencil%QQ )
    real(kind=rk) :: fEq( layout%fStencil%QQ )
    real(kind=rk) :: fNeq( layout%fStencil%QQ )
    integer :: iVar ! current field variable
    real(kind=rk) :: weight( nMaxSources )
    real(kind=rk) :: rho0
    real(kind=rk) :: pFac, vFac
    !> debug variables
?? IF (loglvl >= 1) THEN
    integer(kind=long_k) :: treeID
    ! integer :: coord(4)
    ! real(kind=rk) :: targetCoord(3), dx
    ! real(kind=rk) :: mom(layout%fStencil%QQ)
?? ENDIF
    integer :: QQ
    ! --------------------------------------------------------------------------

    QQ = layout%fStencil%QQ

    sourceLevel = level
    targetLevel = level + 1
    iVar = 1
    rho0 = 1.0_rk

    ! macroscopic quantity scale factor from source level to target level
    pFac = physics%pfac( sourceLevel, targetLevel )
    vFac = physics%vfac( sourceLevel, targetLevel )

! DEBUG output
! ----------------------------------------------------------------------------------
?? IF (loglvl >= 1) THEN
write(dbgUnit(1),*) '----- Enter fillFinerGhostsFromMe_linearDiffusive -----------'
write(dbgUnit(1),*) ''
write(dbgUnit(1),*) 'source level: ', sourceLevel
write(dbgUnit(1),*) '  omega: ', omLvl( sourceLevel )
write(dbgUnit(1),*) '  pressure factor: ',physics%fac( sourceLevel )%press
write(dbgUnit(1),*) '  velocity factor: ',physics%fac( sourceLevel )%vel
write(dbgUnit(1),*) '  strain   factor: ',physics%fac( sourceLevel )%strainRate
write(dbgUnit(1),*) 'target level: ', targetLevel
write(dbgUnit(1),*) '  omega: ', omLvl( targetLevel )
write(dbgUnit(1),*) '  pressure factor: ',physics%fac( targetLevel )%press
write(dbgUnit(1),*) '  velocity factor: ',physics%fac( targetLevel )%vel
write(dbgUnit(1),*) '  strain   factor: ',physics%fac( targetLevel )%strainRate
write(dbgUnit(1),*) 'scale factors:'
write(dbgUnit(1),*) ' pressure: ', pFac
write(dbgUnit(1),*) ' velocity: ', vFac
write(dbgUnit(1),*) 'number of target elements to fill: ', nTargets
write(dbgUnit(1),*) ''
?? ENDIF
! ----------------------------------------------------------------------------------

    ! Treat all fine target elements
    do indElem = 1, nTargets

      iElem = targetList( indElem )

      ! Read the target element
      targetPos = iElem + tLevelDesc           &
        &                     %offset( 1, eT_ghostFromCoarser)

      ! Find out how many fine source elements we have for interpolation.
      nSourceElems = tLevelDesc%depFromCoarser( iElem ) &
        &                                         %elem%nVals

      ! Read the pre-calculated weights (like in Average intp.)
      weight(1:nSourceElems) = tLevelDesc        &
        &                          %depFromCoarser( iElem )%weight(:)

! ----------------------------------------------------------------------
?? IF (loglvl >= 1) THEN
! DEBUG: get TGV 2D analytical solution
treeID = tLevelDesc%total( targetPos )
write(dbgUnit(1),"(A, I0)") 'target treeID: ', treeID
write(dbgUnit(1),"(A, I0)") 'number of source elements: ', nSourceElems
write(dbgUnit(1),*) 'weights: ', weight(1:nSourceElems)
write(dbgUnit(1),*) ''
?? ENDIF
! ----------------------------------------------------------------------

      ! Now loop over all fine source elements for this target
      do iSourceElem = 1, nSourceElems

        ! Get the source element position
        sourcePos = tLevelDesc%depFromCoarser( iElem )%elem%val( iSourceElem )

        ! Calculate Rho and Vel
        do iDir = 1, QQ
           fTmp(iDir) = sState(?IDX?(iDir, sourcePos, nScalars, snSize))
        enddo

        sourceP(iSourceElem)   = sum( fTmp ) - rho0
        sourceV(1,iSourceElem) = sum( fTmp * layout%fStencil%cxDirRK(1,:) )
        sourceV(2,iSourceElem) = sum( fTmp * layout%fStencil%cxDirRK(2,:) )
        sourceV(3,iSourceElem) = sum( fTmp * layout%fStencil%cxDirRK(3,:) )

        ! Calculate Strain rate
        sourceS(:,:,iSourceElem) = getShearRateTensor_diffusive( &
          &           f = fTmp, layout = layout,   &
          &           omega = omLvl( sourceLevel ))

! ----------------------------------------------------------------------------
! DEBUG output: check rho, vel, strain of source elements
?? IF (loglvl >= 1) THEN
write(dbgUnit(1),"(A, F12.8, A, 3F12.8, A, 3F12.8)") &
  &       'Source P: ', sourceP(iSourceElem), &
  &       '   V: ', sourceV(1,iSourceElem), sourceV(2,iSourceElem), sourceV(3,iSourceElem),&
  &       '   S: ', sourceS(1,1,iSourceElem), sourceS(2,2,iSourceElem), sourceS(1,2,iSourceElem)
?? ENDIF
! ----------------------------------------------------------------------------
      end do ! iSourceElem

! ----------------------------------------------------------------------------
?? IF (loglvl >= 1) THEN
write(dbgUnit(1),"(A)") ' '
?? ENDIF
! ----------------------------------------------------------------------------

      ! ------------------ LINEAR INTERPOLATION -------------------
      ! linearly interpolate Pressure, Vel, and Strain
      targetP    = sum( weight(1:nSourceElems)*sourceP(1:nSourceElems)   )
      targetV(1) = sum( weight(1:nSourceElems)*sourceV(1,1:nSourceElems) )
      targetV(2) = sum( weight(1:nSourceElems)*sourceV(2,1:nSourceElems) )
      targetV(3) = sum( weight(1:nSourceElems)*sourceV(3,1:nSourceElems) )
      do jVal = 1, layout%fStencil%nDims
        do iVal = 1, layout%fStencil%nDims
        targetS(iVal,jVal) =                                   &
          & sum( weight(1:nSourceElems) * sourceS(iVal,jVal,1:nSourceElems) )
        end do
      end do
      ! ------------------ LINEAR INTERPOLATION -------------------

! ----------------------------------------------------------------------------
! DEBUG output: check rho, vel, strain of target elements
?? IF (loglvl >= 1) THEN
write(dbgUnit(1),"(A, F12.8, A, 3F12.8, A, 3F12.8)") &
  &       'Target P: ', targetP,&
  &       '   V: ', targetV(1), targetV(2), targetV(3),&
  &       '   S: ', targetS(1,1), targetS(2,2), targetS(1,2)
write(dbgUnit(1),"(A)") ''
?? ENDIF
! ----------------------------------------------------------------------------

      ! scale macroscopic quantities (LB unit) from source level to target level
      targetR = targetP * pFac + rho0
      targetV = targetV * vFac
      targetS = targetS * pFac

      ! Calculate fEq and fNeq
?? copy :: Eq_incomp( fEq, targetR, targetV, layout, rho0 )

      fNeq(:) = getnEq_diffusive(                                            &
        &            layout = layout,                                        &
        &            omega  = omLvl( targetLevel ),  &
        &            Sxx    = targetS(:,:))

      ! Now write pdf (fEq+fNeq) to the target element
      do iDir = 1, QQ
        tstate(?IDX?( iDir, targetPos, nscalars, tnSize) ) = fEq( iDir ) + fNeq( iDir )
      end do ! iDir

    end do ! indElem

?? IF (loglvl >= 1) THEN
write(dbgUnit(1),*) '----- Leave fillFinerGhostsFromMe_linearDiffusive -----------'
write(dbgUnit(1),*) ''
?? ENDIF

  end subroutine fillFinerGhostsFromMe_linearDiffusive
! ****************************************************************************** !


! ****************************************************************************** !
  !> [Linear interpolation](../page/features/intp_methods.html) of ghostFromFiner
  !! The interpolation procedure used in this routine is:\n
  !! 1. calculate rho(pressure), velocity, fNeq from source elements
  !! 2. interpolate rho and velocity to target element
  !! 3. calculate fEq of target element from rho and velocity
  !! 4. interpolation fNeq from source elements to target element
  !! 5. calculate pdf of target element = fEq + fac * fNeq
  !!
?? copy :: interpolateRoutine_header( fillMyGhostsFromFiner_linear )
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceVel(3,nMaxSources)
    ! macroscopic density for all source elements
    real(kind=rk) :: sourceRho(nMaxSources)
    real(kind=rk) :: fac, inv_nSource
    ! temporary equilirbium part
    real(kind=rk) :: fEq( layout%fStencil%QQ )
    ! current non-equilibrium part
    real(kind=rk) :: sourceNEq( layout%fStencil%QQ, nMaxSources )
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: fNeq
    real(kind=rk) :: fTmp( layout%fStencil%QQ )
    real(kind=rk) :: rho, vel(3)
    integer :: QQ
    ! ---------------------------------------------------------------------------
    QQ = layout%fStencil%QQ

    sourceLevel = level + 1
    targetLevel = level

    ! Treat all coarse target elements
    do indElem = 1, nTargets

      iElem = targetList( indElem )

      ! Read the target element
      targetElem = iElem + tLevelDesc%offset( 1, eT_ghostFromFiner)

      ! Find out how many fine source elements we have for interpolation.
      nSourceELems = tLevelDesc%depFromFiner( iElem )%elem%nVals

      ! Compute sourceRho and velocity for each field and interpolate pdf
      ! field wise
      fac = ((1._rk - 0.5_rk*omLvl( sourceLevel ))   &
        &       *convPrePost(omLvl( sourceLevel )))  &
        &    /((1._rk - 0.5_rk*omLvl( targetLevel )) &
        &       *convPrePost(omLvl( targetLevel )))

      sourceRho = 0._rk
      sourceVel = 0._rk
      ! Now loop over all fine source elements for this target:
      do iSourceElem = 1, nSourceElems

        ! Get the source element
        sourceElem = tLevelDesc%depFromFiner( iElem )     &
          &                                       %elem%val( iSourceElem )

        ! Calculate rho and vel for equil
        do iDir = 1, QQ
          fTmp(iDir) = sstate(?IDX?(iDir, sourceElem, nscalars, snSize))
        end do
        rho    = sum( fTmp )
        vel(1) = sum( fTmp(:) * layout%fStencil%cxDirRK(1,:) ) / rho
        vel(2) = sum( fTmp(:) * layout%fStencil%cxDirRK(2,:) ) / rho
        vel(3) = sum( fTmp(:) * layout%fStencil%cxDirRK(3,:) ) / rho

        sourceRho( iSourceElem )   = rho - 1.0_rk
        sourceVel( :,iSourceElem ) = vel(:)
?? copy :: Eq_comp( fEq, rho, vel, layout )
        sourceNEq(:,iSourceElem) = fTmp(:) - fEq(:)

      end do  ! iSourceElem

      ! Ok, now I have: rho, vel, fNeq of all my neighbor elements.
      ! Interpolate rho, vel, and fNeq by average
      inv_nSource = 1.0_rk / dble( nSourceElems )
      Rho    = sum(sourceRho(1:nSourceElems))   * inv_nSource + 1.0_rk
      Vel(1) = sum(sourceVel(1,1:nSourceElems)) * inv_nSource
      Vel(2) = sum(sourceVel(2,1:nSourceElems)) * inv_nSource
      Vel(3) = sum(sourceVel(3,1:nSourceElems)) * inv_nSource
      ! calculate eq and store in targetEq
?? copy :: Eq_comp( fEq, rho, vel, layout )

      ! Now write the resulting pdf in the current direction to the target
      ! Element position
      do iDir = 1, QQ
        fNeq = sum( sourceNeq(iDir,1:nSourceElems)) * inv_nSource
        tstate(?IDX?( iDir, targetElem, nscalars, tnSize ) )&
          &   = fEq( iDir ) + fac * fNeq
      enddo ! iDir

    enddo

  end subroutine fillMyGhostsFromFiner_linear
! ****************************************************************************** !


! ****************************************************************************** !
  !> Fill coarse target ghost from fine source fluid by linear interpolation.
  !! 1. For each target, calculate rho, vel and fNeq of its sources.
  !! 2. Average rho and vel, then calculate fEq.
  !! 3. Average fNeq and scale.
  !! 4. set target by f = fEq + fNeq
  !! This routine is used by 2D and 3D incomp acoustic linear interpolation.
  !!
?? copy :: interpolateRoutine_header( fillMyGhostsFromFiner_linearIncomp )
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    integer :: iVar ! current field variable
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceRho
    real(kind=rk) :: sourceVel(3)
    real(kind=rk) :: fNeq( layout%fStencil%QQ )
    ! macroscopic velocities for all source elements
    real(kind=rk) :: rho0, rho, vel(3), inv_nSourceElems
    ! scaling factor for fNeq
    real(kind=rk) :: fac
    ! current non-equilibrium part
    real(kind=rk) :: f( layout%fStencil%QQ ), fEq( layout%fStencil%QQ )
    integer :: QQ
    ! ---------------------------------------------------------------------------
    QQ = layout%fStencil%QQ

    iVar = 1
    sourceLevel = level + 1
    targetLevel = level
    rho0 = 1.0_rk

    fac = getNoneqFac( omLvl( sourceLevel ), &
      &                omLvl( targetLevel )  )

    ! Treat all coarse target elements
    do indElem = 1, nTargets

      iElem = targetList( indElem )

      ! Read the target element treeId
      targetElem = iElem + tLevelDesc%offset( 1, eT_ghostFromFiner)

      ! Find out how many fine source elements we have for interpolation.
      nSourceELems = tLevelDesc%depFromFiner( iElem )%elem%nVals
      inv_nSourceElems = 1.0_rk / dble(nSourceElems)

      sourceRho    = 0._rk
      sourceVel(:) = 0._rk
      fNeq(:)      = 0._rk
      ! Now loop over all fine source elements for this target:
      do iSourceElem = 1, nSourceElems

        ! Get the source element position
        sourceElem = tLevelDesc%depFromFiner( iElem ) &
          &                                       %elem%val( iSourceElem )

        ! Calculate rho and vel for equil
        do iDir = 1, QQ
           f(iDir) = sstate(?IDX?(iDir, sourceElem, nScalars, snSize))
        end do
        rho = sum( f(:) )
        vel(1) = sum( f(:) * layout%fStencil%cxDirRK(1,:) )
        vel(2) = sum( f(:) * layout%fStencil%cxDirRK(2,:) )
        vel(3) = sum( f(:) * layout%fStencil%cxDirRK(3,:) )

        ! sum rho, vel and fNeq of all sources together
        sourceRho    = sourceRho    + rho
        sourceVel(:) = sourceVel(:) + vel(:)
?? copy :: Eq_incomp( fEq, rho, vel, layout, rho0 )
        fNeq(:) = fNeq(:) + f(:) - fEq(:)

      end do ! iSourceElem

      ! interpolate rho, vel, and fNeq by average
      rho     = sourceRho    * inv_nSourceElems
      vel(:)  = sourceVel(:) * inv_nSourceElems
      fNeq(:) = fNeq(:)      * inv_nSourceElems

?? copy :: Eq_incomp( fEq, rho, vel, layout, rho0 )

      do iDir = 1,QQ
        tstate(?IDX?( iDir, targetElem, nScalars, tnSize )) &
          &         = fEq(iDir) + fNeq(iDir) * fac
      end do

    end do

  end subroutine fillMyGhostsFromFiner_linearIncomp
! ****************************************************************************** !


! ****************************************************************************** !
  !> Fill finer ghost from coarser fluid.
  !! This routine is used by 2D and 3D incomp acoustic linear interpolation.
  !!
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_linearIncomp )
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    ! macroscopic velocities for all source elements
    real(kind=rk) :: targetRho, targetVel(3), rho0, rho, vel(3)
    real(kind=rk) :: fNeq( layout%fStencil%QQ )
    real(kind=rk) :: f( layout%fStencil%QQ ), fEq( layout%fStencil%QQ )
    real(kind=rk) :: fac
    real(kind=rk) :: weight( nMaxSources )
    ! current non-equilibrium part
    integer :: iVar ! current field variable
    integer :: QQ
    ! --------------------------------------------------------------------------
    QQ = layout%fStencil%QQ

    iVar = 1
    sourceLevel = level
    targetLevel = level + 1
    rho0 = 1.0_rk

    fac = getNoneqFac( omLvl( sourceLevel ), omLvl( targetLevel )  )

    ! Treat all fine target elements:
    do indElem = 1, nTargets

      iElem = targetList( indElem )

      ! Read the target element treeId
      targetElem = iElem + tLevelDesc%offset( 1, eT_ghostFromCoarser)

      ! Find out how many fine source elements we have for interpolation.
      ! Usually 8, but can differ at corners, obstacles, boundaries...
      nSourceElems = tLevelDesc%depFromCoarser( iElem )%elem%nVals

      ! Read the pre-calculated weights (like in Average intp.)
      weight(1:nSourceElems) = &
        &            tLevelDesc%depFromCoarser( iElem )%weight(1:nSourceElems)

      targetRho    = 0._rk
      targetVel(:) = 0._rk
      fNeq(:)      = 0._rk
      ! Now loop over all fine source elements for this target:
      do iSourceElem = 1, nSourceElems
        ! Get the source element's treeId
        sourceElem = tLevelDesc%depFromCoarser( iElem )   &
          &                                       %elem%val( iSourceElem )

        ! Calculate rho and vel for equil
        do iDir = 1, QQ
           f(iDir) = sstate(?IDX?(iDir, sourceElem, nScalars, snSize ))
        end do
        rho    = sum( f(:) )
        Vel(1) = sum( f(:) * layout%fStencil%cxDirRK(1,:))
        Vel(2) = sum( f(:) * layout%fStencil%cxDirRK(2,:))
        Vel(3) = sum( f(:) * layout%fStencil%cxDirRK(3,:))

?? copy :: Eq_incomp( fEq, rho, vel, layout, rho0 )

        targetRho    =    targetRho + weight(iSourceElem) * rho
        targetVel(:) = targetVel(:) + weight(iSourceElem) * vel(:)
        fNeq(:)      =      fNeq(:) + weight(iSourceElem) * (f(:) - fEq(:))

      end do ! iSourceElem

      ! calculate target Eq
?? copy :: Eq_incomp( fEq, targetRho, targetVel, layout, rho0 )

      do iDir = 1,QQ
        tstate(?IDX?(iDir, targetElem, nScalars, tnSize )) &
          &  = fEq( iDir ) + fac * fNeq( iDir )
      end do

    end do

  end subroutine fillFinerGhostsFromMe_linearIncomp
! ****************************************************************************** !


! ****************************************************************************** !
  !> [Linear interpolation](../page/features/intp_methods.html) of ghostFromFiner
  !!
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_linear )
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iDir           ! current direction (discrete velocity) for loop
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    ! macroscopic velocities for all source elements
    real(kind=rk) :: sourceVel(3,nMaxSources)
    ! macroscopic density for all source elements
    real(kind=rk) :: sourceRho(nMaxSources)
    ! current non-equilibrium part
    real(kind=rk) :: sourceNEq( layout%fStencil%QQ, nMaxSources )
    real(kind=rk) :: fac
    ! temporary equilirbium part
    real(kind=rk) :: fEq( layout%fStencil%QQ )
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: fNeq
    real(kind=rk) :: fTmp( layout%fStencil%QQ )
    real(kind=rk) :: rho, vel(3)
    real(kind=rk) :: weight( nMaxSources )
    integer :: QQ
    ! ---------------------------------------------------------------------------
    QQ = layout%fStencil%QQ

    sourceLevel = level
    targetLevel = level + 1

    ! See explanation for the factor in fineToCoarse
    fac = ((1._rk - 0.5_rk*omLvl(sourceLevel)) &
      &     *convPrePost(omLvl( sourceLevel )))      &
      &   /((1._rk - 0.5_rk*omLvl( targetLevel ))    &
      &     *convPrePost(omLvl( targetLevel )))

    ! Treat all fine target elements:
    do indElem = 1, nTargets

      iElem = targetList( indElem )

      ! Read the target element treeId
      targetElem = iElem + tLevelDesc%offset( 1, eT_ghostFromCoarser)

      ! Find out how many fine source elements we have for interpolation.
      nSourceElems = tLevelDesc%depFromCoarser( iElem )%elem%nVals

      ! See explanation for the factor in fineToCoarse
      sourceRho = 0._rk
      sourceVel = 0._rk

      ! Now loop over all fine source elements for this target:
      do iSourceElem = 1, nSourceElems
        ! Get the source element
        sourceElem = tLevelDesc%depFromCoarser( iElem )%elem%val( iSourceElem )

        ! Read the pre-calculated weights (like in Average intp.)
        weight( iSourceElem ) = tLevelDesc%depFromCoarser( iElem )%weight( iSourceElem )

        ! Calculate rho and vel
        do iDir = 1, QQ
          fTmp(iDir) = sstate( ?IDX?(iDir, sourceElem, nScalars, snSize))
        enddo ! iDir

        rho    = sum( fTmp )
        vel(1) = sum( fTmp * layout%fStencil%cxDirRK(1,:)) / rho
        vel(2) = sum( fTmp * layout%fStencil%cxDirRK(2,:)) / rho
        vel(3) = sum( fTmp * layout%fStencil%cxDirRK(3,:)) / rho

        sourceRho(iSourceElem)   = rho - 1.0_rk
        sourceVel(:,iSourceElem) = vel(:)

?? copy :: Eq_comp( fEq, rho, vel, layout )
        sourceNEq(:,iSourceElem) = fTmp(:) - fEq(:)

! write(dbgUnit(1),"(A, F12.5, A, 3F12.5)") &
!   &       'Source pressPhy: ', sourceRho(iSourceElem), &
!   &       '   vel: ', sourceVel(1,iSourceElem), sourceVel(2,iSourceElem), sourceVel(3,iSourceElem)
      end do ! iSourceElem

      ! Ok, now I have: rho, vel, fNeq of all my neighbor elements.
      ! linearly interpolate rho, vel, and fNeq
      rho    = sum( weight(1:nSourceElems)*sourceRho(  1:nSourceElems)) + 1.0_rk
      vel(1) = sum( weight(1:nSourceElems)*sourceVel(1,1:nSourceElems))
      vel(2) = sum( weight(1:nSourceElems)*sourceVel(2,1:nSourceElems))
      vel(3) = sum( weight(1:nSourceElems)*sourceVel(3,1:nSourceElems))
      ! calculate fEq
?? copy :: Eq_comp( fEq, Rho, Vel, layout )

! write(dbgUnit(1),"(A, F12.5, A, 3F12.5)") &
!   &       'Target pressPhy: ', targetRho,&
!   &       '   vel: ', targetVel(1), targetVel(2), targetVel(3)
!   ! &       '   S: ', targetS(1,1), targetS(2,2), targetS(1,2)
! write(dbgUnit(1),"(A)") ''

      ! Now write the resulting pdf to the target element
      do iDir = 1, QQ
        fNeq = sum( weight(1:nSourceElems)*sourceNeq(iDir,1:nSourceElems) )
        tstate(?IDX?(iDir, targetElem, nScalars, tnSize ) ) = fEq( iDir ) + fac * fNeq
      end do
    enddo ! element loop

  end subroutine fillFinerGhostsFromMe_linear
! ****************************************************************************** !


! ****************************************************************************** !
  !> Doing the linear interpolation just like the quadratic one by using 
  !! the moment representation
?? copy :: interpolateRoutine_header( fillMyGhostsFromFiner_linMoment3d )
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal, iElem, iDir
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: f( layout%fStencil%QQ ) ! pdf to reconstruct
    real(kind=rk) :: fEq( layout%fStencil%QQ )  ! source equilirbium part
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( layout%fStencil%QQ, nMaxSources )
    ! shear stress scaling factor
    real(kind=rk) :: fac, inv_nSource, inv_rho
    real(kind=rk) :: tMom( layout%fStencil%QQ )  ! temp moment calculation
    real(kind=rk) :: rho, vel(3)
    integer :: QQ
    ! real(kind=rk), allocatable :: momBuf(:,:)
    ! ---------------------------------------------------------------------------
    QQ = layout%fStencil%QQ

    sourceLevel = level + 1
    targetLevel = level

    fac = ((1._rk - 0.5_rk*omLvl( sourceLevel ))   &
      &     *convPrePost(omLvl( sourceLevel )))    &
      &   /((1._rk - 0.5_rk*omLvl( targetLevel ))  &
      &     *convPrePost(omLvl( targetLevel )))

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, tLevelDesc%sourceFromFiner%nVals

      ! Get the source element's position in the state vector
      sourceElem = tLevelDesc%sourceFromFiner%val( iSourceElem )

      do iVal = 1, QQ
        f(iVal)  = sstate( ?IDX?(iVal, sourceElem, nScalars, snSize ))
      enddo

      ! \todo: not all moments are needed
      tMom(1:10) = matmul( layout%moment%toMoments%A(1:10,:), f)
      inv_rho = 1.0_rk / tMom(1)
      tMom(2:4) = tMom(2:4) * inv_rho

      ! remove pi(0) parts from the second order moment
      ! that is ((rho u_\alpha * rho u_\beta) / rho)
      ! and additionally rho*cs2 from diagonal entries
      ! Pab = sum( c_ia*c_ib *f_i ) - kronecker_ab*(rho*cs2) - rho*u_a*u_b
      tMom( 5) = tMom( 5) - tMom(2)*tMom(2)*tMom(1) - cs2*tMom(1) !Pxx
      tMom( 6) = tMom( 6) - tMom(3)*tMom(3)*tMom(1) - cs2*tMom(1) !Pyy
      tMom( 7) = tMom( 7) - tMom(4)*tMom(4)*tMom(1) - cs2*tMom(1) !Pzz
      tMom( 8) = tMom( 8) - tMom(2)*tMom(3)*tMom(1)
      tMom( 9) = tMom( 9) - tMom(3)*tMom(4)*tMom(1)
      tMom(10) = tMom(10) - tMom(2)*tMom(4)*tMom(1)

      momBuf(:, iSourceElem ) = tMom(1:10)

    enddo

    ! Treat all fine target elements:
    do indElem = 1, nTargets

      iElem = targetList( indElem )

      ! Read the target element treeId
      targetElem = iElem + tLevelDesc%offset( 1, eT_ghostFromFiner)

      ! Find out how many fine source elements we have for interpolation.
      nSourceElems = tLevelDesc%depFromFiner( iElem )%elem%nVals
      inv_nSource  = 1.0_rk / dble( nSourceElems )

      ! Now loop over all fine source elements for this target:
      do iSourceElem = 1, nSourceElems
        ! Get the source element in the levelwise buffer
        sourceElem = tLevelDesc%depFromFiner( iElem )%elemBuffer%val( iSourceElem )
        ! and read its entries
        sourceMom(:, iSourceElem ) = momBuf( :, sourceElem )
      end do

      do iVal = 1, 10
        tMom(iVal) = sum( sourceMom(iVal, 1:nSourceElems)) &
          &          * inv_nSource
      enddo
      inv_rho = 1.0_rk / tMom(1)

      ! Rescale the shear stress moments
      ! and add back the pi(0) part to the tensor
      tMom( 5) = tMom( 5)*fac + tMom(2)*tMom(2)*inv_rho + cs2*tMom(1) !Pxx
      tMom( 6) = tMom( 6)*fac + tMom(3)*tMom(3)*inv_rho + cs2*tMom(1) !Pyy
      tMom( 7) = tMom( 7)*fac + tMom(4)*tMom(4)*inv_rho + cs2*tMom(1) !Pzz
      tMom( 8) = tMom( 8)*fac + tMom(2)*tMom(3)*inv_rho
      tMom( 9) = tMom( 9)*fac + tMom(3)*tMom(4)*inv_rho
      tMom(10) = tMom(10)*fac + tMom(2)*tMom(4)*inv_rho

      ! And calculate the equilibrium distribution from the macroscopic
      ! quantities in tMom
      rho = tMom(1)
      vel(1:3) = tMom(2:4) * inv_rho
?? copy :: Eq_comp( fEq, rho, vel, layout )

      ! Set the further higher moments to their equilibrium distribution
      ! No need to rescale higher order moments as they were set to
      ! equilibrium
      ! \todo: optimize this
      tMom(11) = sum( fEq * layout%fstencil%cxDirRK(1,:) &
        &                 * layout%fstencil%cxDirRK(1,:) &
        &                 * layout%fstencil%cxDirRK(2,:) )
      tMom(12) = sum( fEq * layout%fstencil%cxDirRK(1,:) &
        &                 * layout%fstencil%cxDirRK(1,:) &
        &                 * layout%fstencil%cxDirRK(3,:) )
      tMom(13) = sum( fEq * layout%fstencil%cxDirRK(2,:) &
        &                 * layout%fstencil%cxDirRK(2,:) &
        &                 * layout%fstencil%cxDirRK(1,:) )
      tMom(14) = sum( fEq * layout%fstencil%cxDirRK(2,:) &
        &                 * layout%fstencil%cxDirRK(2,:) &
        &                 * layout%fstencil%cxDirRK(3,:) )
      tMom(15) = sum( fEq * layout%fstencil%cxDirRK(3,:) &
        &                 * layout%fstencil%cxDirRK(3,:) &
        &                 * layout%fstencil%cxDirRK(1,:) )
      tMom(16) = sum( fEq * layout%fstencil%cxDirRK(3,:) &
        &                 * layout%fstencil%cxDirRK(3,:) &
        &                 * layout%fstencil%cxDirRK(2,:) )
      tMom(17) = sum( fEq * layout%fstencil%cxDirRK(1,:) &
        &                 * layout%fstencil%cxDirRK(1,:) &
        &                 * layout%fstencil%cxDirRK(2,:) &
        &                 * layout%fstencil%cxDirRK(2,:) )
      tMom(18) = sum( fEq * layout%fstencil%cxDirRK(2,:) &
        &                 * layout%fstencil%cxDirRK(2,:) &
        &                 * layout%fstencil%cxDirRK(3,:) &
        &                 * layout%fstencil%cxDirRK(3,:) )
      tMom(19) = sum( fEq * layout%fstencil%cxDirRK(1,:) &
        &                 * layout%fstencil%cxDirRK(1,:) &
        &                 * layout%fstencil%cxDirRK(3,:) &
        &                 * layout%fstencil%cxDirRK(3,:) )

      ! Transform back into pdf space
      ! \todo: optimize this
      f = matmul( layout%moment%toPdf%A, tMom )

      ! Now write the resulting pdf in the current direction to the target
      ! Element position
      do iVal = 1, QQ
        tState( ?IDX?(iVal, targetElem, nScalars, tnSize ))= f(iVal)
      end do
    enddo

    ! deallocate( momBuf )

  end subroutine fillMyGhostsFromFiner_linMoment3d
! ****************************************************************************** !


! ****************************************************************************** !
  !> Calculate the conversion factor for nonEq in difference levels
  pure function getNoneqFac( omegaS, omegaT ) result( fac )
    ! ---------------------------------------------------------------------------
    real(kind=rk), intent(in) :: omegaS  !< omega value on source level
    real(kind=rk), intent(in) :: omegaT  !< omage value on target level
    real(kind=rk) :: fac
    ! ---------------------------------------------------------------------------

??  if( PUSH ) then
      fac = ( 1._rk - 0.5_rk * omegaS ) / ( 1._rk - 0.5_rk * omegaT )
??  else
      fac =   ( 1._rk - 0.5_rk * omegaS ) * ( 1._rk - omegaT ) &
        &   / ( 1._rk - omegaS ) / ( 1._rk - 0.5_rk * omegaT )
??  endif

  end function getNoneqFac
! ****************************************************************************** !

end module mus_interpolate_linear_module
! ****************************************************************************** !
