! See copyright notice in the COPYRIGHT file.
! ****************************************************************************** !
!> author: Manuel Hasert
!! # Interpolation of flow quantities between different grid levels 
!!
!! ## Interpolation
!! 
!! The routines defined here, fill up the ghost elements with valid data.
!! Ghost elements are employed at grid level interfaces to provide valid
!! pdf values to the neighboring fluid elements. This way, the solvers can 
!! act on elements of the same size only, treating the levels successively.
!! Target elements are the ghost elements, which have to be filled with
!! valid values.
!! Source elements are the fluid elements from other levels, from where to
!! take the input values for the interpolation.
!! The target ghost elements on the target level have corresponding source
!! fluid elements on the source level.
!!
!! [[tem_topology_module]] For a detailed description of the grid 
!! 
!! ## Workflow <a id="interpolation-of"></a>
!!
!! Each interpolation routine acts on a list of ghost elements.
!! This list contains pointers to the position in the total list.
!! For each of these ghost elements, the source elements are identified.
!! Before that, the sourceLevel is identified. However, the code is restricted
!! to work with a level jump of only one level, so the sourceLevel is 
!! for sourceLevel = targetLevel+1
!! sourceLevel = targetLevel-1
!!
!! For an overview over implemented interpolation methods, see
!! [Interpolation methods](../page/features/intp_methods.html)
!!
?? include 'header/lbm_macros.inc'
module mus_interpolate_module

  ! include treelm modules
  use env_module,            only: rk
  use tem_aux_module,        only: tem_abort
  use tem_debug_module,      only: main_debug, dbgUnit
  use tem_logging_module,    only: logUnit
  use tem_construction_module, only: tem_levelDesc_type

  ! include musubi modules
  use mus_field_prop_module,            only: mus_field_prop_type
  use mus_interpolate_header_module,    only: mus_interpolation_type
?? if( .not. SOA) then
  use mus_interpolate_debug_module,     only: fillFinerGhostsFromMe_TGV2D, &
    &                                         fillMyGhostsFromFiner_TGV2D
?? endif
  use mus_interpolate_debug_module,     only: do_nothing
  use mus_interpolate_linear_module,    only:                                  &
    &                                   fillFinerGhostsFromMe_linear,          &
    &                                   fillMyGhostsFromFiner_linear,          &
    &                                   fillFinerGhostsFromMe_linearDiffusive, &
    &                                   fillMyGhostsFromFiner_linearDiffusive, &
    &                                   fillMyGhostsFromFiner_linearIncomp,    &
    &                                   fillFinerGhostsFromMe_linearIncomp,    &
    &                                   fillMyGhostsFromFiner_linMoment3d
  use mus_interpolate_quadratic_module, only:                                  &
    ! &                              fillFinerGhostsFromMe_quadratic2d,          &
    &                              fillFinerGhostsFromMe_quadratic3d,          &
    &                              fillFinerGhostsFromMe_quadratic3dIncomp_new,    &
    &                              fillFinerGhostsFromMe_quadratic3dDiffusive
  use mus_interpolate_d2q9_module, only:                                       &
    ! &                                  fillMyGhostsFromFiner_Compact2Dmoment,  &
    ! &                                  fillFinerGhostsFromMe_Compact2Dmoment,  &
    &                                   fillFiner_Linear2D,  &
    &                                   fillFiner_Quadratic2D,  &
    ! &                      fillFiner_Compact2D,   fillCoarser_Compact2D, &
    &                      fillFiner_Compact2D, fillCoarser_Compact2D
  use mus_interpolate_d3q27_module, only: fillCoarser_compactd3q27, &
    &                                     fillFiner_compactD3Q27
  use mus_interpolate_d3q19_module, only: fillCoarser_compactD3Q19
  use mus_scheme_layout_module,     only: mus_scheme_layout_type

  implicit none

  private

  public :: mus_init_interpolate

  contains
! ****************************************************************************** !
  !> This subroutine initialzes the interpolation
  !!
  !! - setting the fillMineFromFiner and fillFinerFromMe function pointers
  !!   to the interpolation function chosen by the user and
  !! - pre-calculating the weights based on the distances between source
  !!   and target nodes
  !! - allocate buffers if required
  !!
  subroutine mus_init_interpolate( intp, levelDesc, kind, layout, &
    &                              nDims, minLevel, maxLevel, scaling )
    ! ---------------------------------------------------------------------------
    !> interpolation type
    type(mus_interpolation_type), intent(inout) :: intp
    integer,                        intent(in) :: nDims, minLevel, maxLevel
    !> level descriptor is actually used
    type( tem_levelDesc_type ), intent(inout) :: levelDesc(minLevel:maxLevel)
    !> the stencil layout representation
    type( mus_scheme_layout_type ), intent(in) :: layout
    !> model kind
    character(len=*),               intent(in) :: kind
    character(len=*),               intent(in) :: scaling
    ! ---------------------------------------------------------------------------
    integer :: iLevel, nVals
    logical :: require_buffer
    ! ---------------------------------------------------------------------------
    ! set the theoretical maximum of source elements
    require_buffer = .false.

    write(logUnit(1),"(A)") 'Initialize interpolation: '//trim(intp%label)
    intp%fillFinerFromMe_low   => do_nothing
    intp%fillMineFromFiner_low => do_nothing

    ! assign functions by label
    select case (trim( intp%label ))
?? if( .not. SOA) then
    case('none')
      write(logUnit(1),*) '   do nothing + do nothing '
      intp%fillFinerFromMe   => do_nothing
      intp%fillMineFromFiner => do_nothing
    case('TGV2D')
      write(logUnit(1),*) ' TGV2D + TGV2D '
      intp%fillMineFromFiner => fillMyGhostsFromFiner_TGV2D
      intp%fillFinerFromMe   => fillFinerGhostsFromMe_TGV2D
?? endif
    case('compact')
      write(logUnit(3),*)' Assign routines for Compact interpolation'

      if ( nDims == 2 ) then
        write(logUnit(3),*) ' Compact 2D + compact 2D'
        intp%fillMineFromFiner => fillCoarser_Compact2D
        ! intp%fillMineFromFiner => fillMyGhostsFromFiner_TGV2D
        ! intp%fillMineFromFiner => fillCoarser_Compact2D

        intp%fillFinerFromMe   => fillFiner_Compact2D
        ! intp%fillFinerFromMe   => fillFinerGhostsFromMe_TGV2D
        ! intp%fillFinerFromMe   => fillFiner_Compact2D
      else ! 3D
        write(logUnit(3),*) ' Compact 3D + compact 3D'
        intp%fillMineFromFiner => fillCoarser_compactd3q27
        ! intp%fillMineFromFiner => fillMyGhostsFromFiner_TGV3D
        ! intp%fillMineFromFiner => do_nothing

        intp%fillFinerFromMe   => fillFiner_compactD3Q27
        ! intp%fillFinerFromMe   => fillFinerGhostsFromMe_TGV3D
        ! intp%fillFinerFromMe   => do_nothing
      end if

      intp%fillFinerFromMe_low   => fillFinerGhostsFromMe_linearDiffusive
      intp%fillMineFromFiner_low => fillMyGhostsFromFiner_linearDiffusive
    ! case('compact_moment')

    !   write(logUnit(3),*) ' Compact moment 2D + compact moment 2D'
    !   if ( nDims == 2 ) then
    !     intp%fillMineFromFiner => fillMyGhostsFromFiner_Compact2Dmoment
    !     ! intp%fillMineFromFiner => fillMyGhostsFromFiner_TGV2D
    !     intp%fillFinerFromMe   => fillFinerGhostsFromMe_Compact2Dmoment
    !     ! intp%fillFinerFromMe   => fillFinerGhostsFromMe_TGV2D
    !   else
    !     write(logUnit(3),*) ' Do not support 3D yet!'
    !     call tem_abort()
    !   end if
    !   intp%fillFinerFromMe_low   => fillFinerGhostsFromMe_linearDiffusive
    !   intp%fillMineFromFiner_low => fillMyGhostsFromFiner_linearDiffusive
    case('linear','quadraticCompact','quadratic')
      if ( layout%fStencil%QQ == 27 ) then
        ! @todo: need to rearrange these assignment!
        intp%fillMineFromFiner => fillCoarser_compactd3q27
        intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadratic3dDiffusive
      else ! D2Q9 or D3Q19
        if ( trim(scaling) == 'diffusive' ) then
          call assign_intp_diffusive( intp, nDims )
        elseif ( trim(scaling) == 'acoustic' ) then
          call assign_intp_acoustic( intp, kind, nDims )
        else
          write(logUnit(1),*) 'Can not assign interpolation for '//trim(scaling)
          call tem_abort()
        end if
      end if
    case default
      write(logUnit(1),*) trim(intp%label)//' is not supported yet!'
      call tem_abort()
    end select

    ! \todo: JQ: intpBufFromCoarser and intpBufFromFiner
    !            can be allocate inside intp routines
    if( require_buffer ) then
      nVals = layout%fStencil%QQ
      do iLevel =  minLevel, maxLevel
        if( allocated( levelDesc( iLevel )%depFromCoarser )) then
          if( size( levelDesc( iLevel )%depFromCoarser ) > 0 ) then
            allocate( levelDesc( iLevel )%intpBufFromCoarser(   &
              &       nVals, levelDesc( iLevel )%sourceFromCoarser%nVals ))
          endif
        endif
        if( allocated( levelDesc( iLevel )%depFromFiner )) then
          if( size( levelDesc( iLevel )%depFromFiner ) > 0 ) then
            allocate( levelDesc( iLevel )%intpBufFromFiner(   &
              &       nVals, levelDesc( iLevel )%sourceFromFiner%nVals ))
          endif
        endif
      enddo
    endif

    if( intp%noIntpFromFiner ) then
      intp%fillMineFromFiner => do_nothing
    end if
    if( intp%noIntpFromCoarser ) then
      intp%fillFinerFromMe   => do_nothing
    end if

    write(logUnit(1),*) 'Done with interpolation method.'

  end subroutine mus_init_interpolate
! ****************************************************************************** !

! ****************************************************************************** !
  !> Set up interpolation routines for diffusive scaling
  subroutine assign_intp_diffusive( intp, nDims )
    ! ---------------------------------------------------------------------------
    !> interpolation type
    type(mus_interpolation_type), intent(inout)    :: intp
    integer,                     intent(in) :: nDims
    ! ---------------------------------------------------------------------------

    write(logUnit(10),*)   ' Assign intp function pointer: diffusive'
    intp%fillFinerFromMe_low   => fillFinerGhostsFromMe_linearDiffusive
    intp%fillMineFromFiner_low => fillMyGhostsFromFiner_linearDiffusive

    select case( trim(intp%label) )
    case( 'linear' )
      if ( nDims == 2 ) then ! 2D
        write(logUnit(5),*)   '  compact 2D + linear 2D'
        intp%fillMineFromFiner => fillCoarser_Compact2D
        intp%fillFinerFromMe   => fillFiner_linear2D
      elseif ( nDims == 3) then ! 3D
        write(logUnit(5),*)   '  linear diffusive + linear diffusive'
        intp%fillMineFromFiner => fillMyGhostsFromFiner_linearDiffusive
        intp%fillFinerFromMe   => fillFinerGhostsFromMe_linearDiffusive
      endif
    case( 'quadratic' )
      if ( nDims == 2 ) then ! 2D
        write(logUnit(1),*) '  compact 2D + quadratic 2D'
        intp%fillMineFromFiner => fillCoarser_Compact2D
        intp%fillFinerFromMe   => fillFiner_quadratic2d
      elseif ( nDims == 3) then ! 3D
        write(logUnit(5),*) '  linear diffusive + quadratic 3D diffusive'
        intp%fillMineFromFiner => fillMyGhostsFromFiner_linearDiffusive
        intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadratic3dDiffusive
      endif
    case default
      write(logUnit(1),*) trim(intp%label)//' is not supported yet!'
      call tem_abort()
    end select

  end subroutine assign_intp_diffusive
! ****************************************************************************** !

! ****************************************************************************** !
  !> Set up interpolation routines for acoustic scaling
  subroutine assign_intp_acoustic( intp, lbm_kind, nDims )
    ! ---------------------------------------------------------------------------
    !> interpolation type
    type(mus_interpolation_type), intent(inout) :: intp
    character(len=*),            intent(in) :: lbm_kind
    integer,                     intent(in) :: nDims
    ! ---------------------------------------------------------------------------
    logical :: incompressible
    ! ---------------------------------------------------------------------------

    write(logUnit(10),*)   ' Assign intp function pointer: acoustic'

    if( trim(lbm_kind) == 'lbm_incomp' .or. trim(lbm_kind) == 'lbm_incomp_nNwtn' ) then
      incompressible = .true.
    else
      incompressible = .false.
    end if

    if( incompressible ) then
      intp%fillFinerFromMe_low   => fillFinerGhostsFromMe_linearIncomp
      intp%fillMineFromFiner_low => fillMyGhostsFromFiner_linearIncomp
    else
      intp%fillFinerFromMe_low   => fillFinerGhostsFromMe_linear
      intp%fillMineFromFiner_low => fillMyGhostsFromFiner_linear
    end if

    select case( trim(intp%label) )
    case( 'linear' ) ! ---------------------------------------------------------
      if( nDims == 2 ) then ! 2D
        if( incompressible ) then
          write(logUnit(5),*)'   linearIncomp + linear 2D Incomp'
          intp%fillMineFromFiner => fillMyGhostsFromFiner_linearIncomp
          intp%fillFinerFromMe   => fillFiner_linear2D
        else
          write(logUnit(5),*)'   linear + linear'
          intp%fillMineFromFiner => fillMyGhostsFromFiner_linear
          intp%fillFinerFromMe   => fillFinerGhostsFromMe_linear
        end if
      else if( nDims == 3) then ! 3D
        if( incompressible ) then
          write(logUnit(5),*)'   linearIncomp + linearIncomp'
          intp%fillMineFromFiner => fillMyGhostsFromFiner_linearIncomp
          intp%fillFinerFromMe   => fillFinerGhostsFromMe_linearIncomp
        else
          write(logUnit(5),*)'   linear + linear'
          intp%fillMineFromFiner => fillMyGhostsFromFiner_linear
          intp%fillFinerFromMe   => fillFinerGhostsFromMe_linear
        end if
      end if
    case( 'quadratic' ) ! ------------------------------------------------------
      if( nDims == 2 ) then ! 2D
        write(logUnit(1),*) 'Compact 2D + Quadratic 2D'
        intp%fillMineFromFiner => fillCoarser_Compact2D
        ! intp%fillMineFromFiner => fillMyGhostsFromFiner_TGV2D
        intp%fillFinerFromMe   => fillFiner_quadratic2D
      elseif( nDims == 3) then ! 3D
        if( incompressible ) then
          write(logUnit(1),*) ' compact D3Q19 + quadratic incomp D3Q19'
          ! intp%fillMineFromFiner => fillMyGhostsFromFiner_linearIncomp
          intp%fillMineFromFiner => fillCoarser_compactD3Q19
          intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadratic3dIncomp_new
        else
          write(logUnit(1),*) ' compact D3Q19 + quadratic D3Q19'
          ! intp%fillMineFromFiner => fillMyGhostsFromFiner_linMoment3d
          intp%fillMineFromFiner => fillCoarser_compactD3Q19
          intp%fillFinerFromMe   => fillFinerGhostsFromMe_quadratic3d
        endif
      end if ! 2D or 3D
    case default
      write(logUnit(1),*) trim(intp%label)//' is not supported yet!'
      call tem_abort()
    end select

  end subroutine assign_intp_acoustic
! ****************************************************************************** !

end module mus_interpolate_module
! ****************************************************************************** !
!> \page intp_methods Interpolation methods
!!
!! \section intp_average Averaging interpolation
!!\verbatim interpolation_method = 'linear'
!!\endverbatim
!! Simple averaging of all quantities with weighted summation
!! \section intp_linear Linear interpolation
!! Linear interpolation of the first two moments (density and momentum) and 
!! separation of equilibrium and non-equilibrium part to account for correct 
!! shear stress calculation
!! This fine->coarse interpolation method is based on Dazhi Yu et. al. (2002):
!! "A multi-block lattice Boltzmann method for viscous fluid flows".
!! The basic idea is to split up the source pdfs into equilibrium- and 
!! non-equilibrum- parts, then interpolate both parts with average interpolation,
!! and apply a corrective scaling factor to the non-equilibrium part before
!! adding both parts up and writing it to the target element.
!! Values are interpolated in a linear fashion by multiplying the quantities
!! of the source elements by weighting factors, which were determined in 
!! \link mus_construction_module::mus_intp_complete_fromcoarserdep  completing the
!! fromCoarser dependencies \endlink
!! in terms of the distance of the source element's barycenters from the
!! target element's barycenters.
!! 
!! \section intp_quadratic Quadratic interpolation
!! Quadratic interpolation based on the compute stencil 
!! \section intp_quadratic_compact Quadratic interpolation with compact stencil
!! Quadratic interpolation of the moment field with derivative information from 
!! the shear stress moments. Only a compact stencil is required.
!!
