! See copyright notice in the COPYRIGHT file.
! ****************************************************************************** !
?? include 'header/lbm_macros.inc'
?? include 'header/lbm_deriveMacros.inc'
?? include 'header/lbm_interfaceMacros.inc'
?? include 'header/lbm_d3q19Macros.inc'
!> summary: Interpolation of flow quantities between different grid levels
!!
!! Ghost elements are employed at grid level interfaces to provide valid
!! pdf values to the neighboring fluid elements. This way, the solvers can 
!! act on elements of the same size only, treating the levels successively.
!! Target elements are the ghost elements, which have to be filled with 
!! valid values.
!! Source elements are the fluid elements from other levels, from where to 
!! take the input values for the interpolation.
!! The target ghost elements on the target level have corresponding source 
!! fluid elements on the source level.
!!
!! [[tem_topology_module]] For a detailed description of the grid
!!
module mus_interpolate_quadratic_module

  ! include treelm modules
  use env_module,              only: rk
  use tem_aux_module,          only: tem_abort
  use tem_param_module,        only: cs2inv, cs2, div1_3, div1_6, div1_9, &
    &                                div2_9, div5_9, div2_3, div1_4, &
    &                                div1_21, div4_21, div5_21, div1_7, &
    &                                div3_7, div1_42, div5_42, div1_2, &
    &                                div1_36, div3_4h
  use tem_element_module,      only: eT_GhostFromCoarser
  use tem_construction_module, only: depSource_type, tem_levelDesc_type
  use tem_debug_module,        only: dbgUnit
  use tem_logging_module,      only: logUnit

  ! include musubi modules
  use mus_pdf_module,                 only: pdf_data_type
  use mus_scheme_layout_module,       only: mus_scheme_layout_type
  use mus_interpolate_header_module,  only: mus_interpolation_type
  use mus_derivedQuantities_module2,  only: getNEq_diffusive,     &
    &                                       getEquilibrium, &
    &                                       getEquilibriumIncomp, &
    &                                       getStrainFacDffs, &
    &                                       getShearRateTensor_diffusive
  use mus_interpolate_compact_module, only: unitTest_interpolate_quadraticCompact
  use mus_interpolate_tools_module,   only: fillMomBuf
  use mus_moments_module,             only: get_moment
  use mus_physics_module,             only: mus_physics_type

  implicit none

  private

  public :: biquadratic_interpolate

  public :: fillFinerGhostsFromMe_quadratic3dDiffusive
  public :: fillFinerGhostsFromMe_quadratic3d
  public :: fillFinerGhostsFromMe_quadratic3dIncomp
  public :: fillFinerGhostsFromMe_quadratic3dIncomp_new

  ! public :: mus_interpolate_quadratic3d_leastSq
  public :: mus_interpolate_quadratic2D_leastSq

  integer, parameter :: QmomentumX = 1
  integer, parameter :: QmomentumY = 2
  integer, parameter :: QmomentumZ = 3
  integer, parameter :: QPxx       = 4
  integer, parameter :: QPyy       = 5
  integer, parameter :: QPzz       = 6
  integer, parameter :: QPxy       = 7
  integer, parameter :: QPyz       = 8
  integer, parameter :: QPzx       = 9
  integer, parameter :: c_x  = 1
  integer, parameter :: c_y  = 2
  integer, parameter :: c_z  = 3
  integer, parameter :: QQ = 19
  real(kind=rk), parameter :: rho0 = 1._rk

  interface mus_interpolate_quadratic3d_leastSq
    module procedure mus_interpolate_quadratic3d_leastSq_scal
    module procedure mus_interpolate_quadratic3d_leastSq_vect
  end interface mus_interpolate_quadratic3d_leastSq

  interface mus_interpolate_quadratic2d_leastSq
    module procedure mus_interpolate_quadratic2d_leastSq_scal
    module procedure mus_interpolate_quadratic2d_leastSq_vect
  end interface mus_interpolate_quadratic2d_leastSq

 contains

! ****************************************************************************** !
  !> This routine is used in diffusive quadratic interpolation scheme.
  !!
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_quadratic3dDiffusive )
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iElem          ! element counter for outer loop
    integer :: indElem        ! element counter for indirection list
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: f( layout%fStencil%QQ ) ! pdf to reconstruct
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom(10, nMaxSources )
    ! interpolated moments
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    integer :: iDir ! current field variable
    real(kind=rk) :: tMom(10)  ! temp moment calculation
    ! non-equilibrium part of the resulting pdf
    real(kind=rk) :: fNeq( layout%fStencil%QQ )
    ! Shear tensor of the source elements
    real(kind=rk) :: strain(3,3)
    integer :: QQ
    ! real(kind=rk) :: momBuf(10,tLevelDesc%sourceFromCoarser%nVals)
    ! ---------------------------------------------------------------------------
    QQ = layout%fStencil%QQ

    sourceLevel = level
    targetLevel = level + 1

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, tLevelDesc%sourceFromCoarser%nVals
      ! Get the source element's position in the state vector
      sourceElem = tLevelDesc%sourceFromCoarser%val( iSourceElem )
      tMom = 0.0d0
      do iDir = 1, QQ
        f(iDir)  = sstate( ?IDX?( iDir, sourceElem, nScalars, snSize ))
        tMom(1)   = tMom(1) + f(iDir)
        tMom(2:4) =   tMom(2:4)           &
          &         + f(iDir) * layout%fStencil%cxDirRK(1:3,iDir)
      enddo

      ! -----------------------------------------
      ! Get the strain rate tensor
      strain(:,:) = getShearRateTensor_diffusive(  f, omLvl( sourceLevel ), layout )
      ! and copy to the moment vector
      tMom( 5) = strain(1,1)
      tMom( 6) = strain(2,2)
      tMom( 7) = strain(3,3)
      tMom( 8) = strain(1,2)
      tMom( 9) = strain(2,3)
      tMom(10) = strain(1,3)

      ! -----------------------------------------
      ! Convert from LB to physical
      ! pressure is actually only the fluctuating part
      ! p = (rho-rho0)*cs2*convFactor with convFactor = rho0*dx^2/dt^2
      tMom(1)    = ( tMom(1) - rho0 ) * cs2 * physics%fac( sourceLevel )%press
      tMom(2:4)  = tMom(2:4)                * physics%fac( sourceLevel )%vel
      ! strain rate conversion
      tMom(5:10) = tMom(5:10)               * physics%fac( sourceLevel )%strainRate
      ! -----------------------------------------

      ! Store all the moments to the current source element's position in the
      ! buffer
      momBuf( :, iSourceElem ) = tMom
    enddo

    ! Treat all fine target elements:
    do indElem = 1, nTargets
      iElem = targetList( indElem )
      ! Read the target element treeId
      targetElem = iElem + tLevelDesc%offset( 1, eT_ghostFromCoarser)
      ! get the target coordinates
      targetCoord = tLevelDesc%depFromCoarser( iElem )%coord
      nSourceElems = tLevelDesc%depFromCoarser( iElem )%elem%nVals

      ! Now loop over all fine source elements for this target:
      do iSourceElem = 1, nSourceElems
        ! Get the source element in the levelwise buffer
        sourceElem = tLevelDesc%depFromCoarser( iElem )%elemBuffer%val( iSourceElem )
        ! and read its entries
        sourceMom(:, iSourceElem ) = momBuf( :, sourceElem )
      end do

      ! interpolate the first ten entries:
      ! pressure(1), velocity(3), strain rate(6)
      tMom(1:10) = mus_interpolate_quadratic3d_leastSq( &
        &               sourceMom(:, 1:19), targetCoord(1:3), 10 )

      ! Copy the shear rate tensor entries to an explicit tensor for the
      ! neq-computation
      strain(1,1) = tMom(5)
      strain(2,2) = tMom(6)
      strain(3,3) = tMom(7)
      strain(1,2) = tMom(8)
      strain(2,1) = tMom(8)
      strain(2,3) = tMom(9)
      strain(3,2) = tMom(9)
      strain(1,3) = tMom(10)
      strain(3,1) = tMom(10)

      ! ---------------------------------------------
      ! Convert from physical to LB
      ! Re-scale the pressure and velocity from physical units to LB units on
      ! the targetlevel
      tMom(1)   = tMom(1)   / physics%fac(targetLevel)%press * cs2inv + rho0
      tMom(2:4) = tMom(2:4) / physics%fac(targetLevel)%vel
      strain(:,:) = strain(:,:) / physics%fac(targetLevel)%strainRate
      ! ---------------------------------------------

      f = getEquilibriumIncomp( tMom(1), tMom(2:4), layout,rho0 )
      fNeq(:) = getnEq_diffusive( layout = layout,                      &
        &                         omega  = omLvl( targetLevel ),  &
        &                         Sxx    = strain(:,:) )

      do iDir = 1, QQ
        tstate(?IDX?(iDir,targetElem,nScalars,tnSize))=f(iDir)+fNeq(iDir)
      end do
    end do

  end subroutine fillFinerGhostsFromMe_quadratic3dDiffusive
! ****************************************************************************** !

! ****************************************************************************** !
  !> Used by Quadratic acoustic incompressible 3D interpolation
  !!
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_quadratic3dIncomp )
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel
    integer :: sourceElem
    integer :: targetLevel
    integer :: targetElem
    integer :: iElem, indElem, iSource, iDir
    real(kind=rk) :: f(QQ), fEq(QQ) ! PDF
    ! moments of the source elements pdf
    real(kind=rk) :: sourceMom( 10, nMaxSources )
    ! interpolated moments
    real(kind=rk) :: tMom(QQ)
    real(kind=rk) :: targetCoord(3) ! coordinates to interpolate to
    real(kind=rk) :: fac, vel(3)
    real(kind=rk) :: cc1, cc2, cc3
    ! ---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    fac = getNoneqFac( omLvl( sourceLevel ), omLvl( targetLevel )  )

    ! First calculate all the required moments for all the source elements
!$omp do schedule(static)
    !DIR$ IVDEP
    !CDIR NODEP
    !CDIR NOLSTVAL
    do iElem = 1, tLevelDesc%sourceFromCoarser%nVals

      sourceElem = tLevelDesc%sourceFromCoarser%val( iElem )
      ! do iDir = 1, QQ
      !   f(iDir) = sState( ?IDX?( iDir, sourceElem, QQ, snSize ))
      ! end do
      f( 1) = sState( ?IDX?(  1, sourceElem, nScalars, snSize ))
      f( 2) = sState( ?IDX?(  2, sourceElem, nScalars, snSize ))
      f( 3) = sState( ?IDX?(  3, sourceElem, nScalars, snSize ))
      f( 4) = sState( ?IDX?(  4, sourceElem, nScalars, snSize ))
      f( 5) = sState( ?IDX?(  5, sourceElem, nScalars, snSize ))
      f( 6) = sState( ?IDX?(  6, sourceElem, nScalars, snSize ))
      f( 7) = sState( ?IDX?(  7, sourceElem, nScalars, snSize ))
      f( 8) = sState( ?IDX?(  8, sourceElem, nScalars, snSize ))
      f( 9) = sState( ?IDX?(  9, sourceElem, nScalars, snSize ))
      f(10) = sState( ?IDX?( 10, sourceElem, nScalars, snSize ))
      f(11) = sState( ?IDX?( 11, sourceElem, nScalars, snSize ))
      f(12) = sState( ?IDX?( 12, sourceElem, nScalars, snSize ))
      f(13) = sState( ?IDX?( 13, sourceElem, nScalars, snSize ))
      f(14) = sState( ?IDX?( 14, sourceElem, nScalars, snSize ))
      f(15) = sState( ?IDX?( 15, sourceElem, nScalars, snSize ))
      f(16) = sState( ?IDX?( 16, sourceElem, nScalars, snSize ))
      f(17) = sState( ?IDX?( 17, sourceElem, nScalars, snSize ))
      f(18) = sState( ?IDX?( 18, sourceElem, nScalars, snSize ))
      f(19) = sState( ?IDX?( 19, sourceElem, nScalars, snSize ))

      ! Calculate the raw moments form the pdf
      ! momBuf(:,iElem) = matmul( layout%moment%toMoments%A(1:10,:), f)

      cc1 = f(7) + f(8) + f(9) + f(10)
      cc2 = f(11) + f(12) + f(13) + f(14)
      cc3 = f(15) + f(16) + f(17) + f(18)

      momBuf(1,iElem) = cc1 + cc2 + cc3 + f(19) + f(1) + f(2) + f(3) + f(4) + f(5) + f(6)
      ! ux uy uz
      momBuf(2,iElem) = -f(1) + f(4) + f(12) + f(14) + f(17) + f(18) - f(11) - f(13) - f(15) - f(16)
      momBuf(3,iElem) = -f(2) + f(5) + f( 9) + f(10) + f(16) + f(18) - f( 7) - f( 8) - f(15) - f(17)
      momBuf(4,iElem) = -f(3) + f(6) + f( 8) + f(10) + f(13) + f(14) - f( 7) - f( 9) - f(11) - f(12)
      ! Sxx Syy Szz
      momBuf(5,iElem) =  f(1) + f(4) + cc2 + cc3 - momBuf(2,iElem)*momBuf(2,iElem) - cs2*momBuf(1,iElem)
      momBuf(6,iElem) =  f(2) + f(5) + cc1 + cc3 - momBuf(3,iElem)*momBuf(3,iElem) - cs2*momBuf(1,iElem)
      momBuf(7,iElem) =  f(3) + f(6) + cc1 + cc2 - momBuf(4,iElem)*momBuf(4,iElem) - cs2*momBuf(1,iElem)
      ! Sxy Syz Sxz
      momBuf(8,iElem) =  f(15) - f(16) - f(17) + f(18) - momBuf(2,iElem)*momBuf(3,iElem)
      momBuf(9,iElem) =  f( 7) - f( 8) - f( 9) + f(10) - momBuf(3,iElem)*momBuf(4,iElem)
     momBuf(10,iElem) =  f(11) - f(12) - f(13) + f(14) - momBuf(2,iElem)*momBuf(4,iElem)

      ! momBuf( 5,iElem) = momBuf( 5,iElem) - momBuf(2,iElem)*momBuf(2,iElem) - cs2*momBuf(1,iElem) !Pxx
      ! momBuf( 6,iElem) = momBuf( 6,iElem) - momBuf(3,iElem)*momBuf(3,iElem) - cs2*momBuf(1,iElem) !Pyy
      ! momBuf( 7,iElem) = momBuf( 7,iElem) - momBuf(4,iElem)*momBuf(4,iElem) - cs2*momBuf(1,iElem) !Pzz
      ! momBuf( 8,iElem) = momBuf( 8,iElem) - momBuf(2,iElem)*momBuf(3,iElem)               !Pxy
      ! momBuf( 9,iElem) = momBuf( 9,iElem) - momBuf(3,iElem)*momBuf(4,iElem)               !Pyz
      ! momBuf(10,iElem) = momBuf(10,iElem) - momBuf(2,iElem)*momBuf(4,iElem)               !Pxz

    end do
!$omp end do
! Here is a implicit barrier, because we can go further until momBuf is complete

    ! Treat all fine target elements:
    ! @todo: indeed all 8 children share the same 19 sources, thus the same polynomial
    !        by this, we can generate the polynomial only once every 8 targets.
!$omp do schedule(static)
    !CDIR NODEP
    do indElem = 1, nTargets

      iElem = targetList( indElem )
      targetElem = iElem + tLevelDesc%offset( 1, eT_ghostFromCoarser)
      targetCoord(1) = tLevelDesc%depFromCoarser( iElem )%coord(1)
      targetCoord(2) = tLevelDesc%depFromCoarser( iElem )%coord(2)
      targetCoord(3) = tLevelDesc%depFromCoarser( iElem )%coord(3)

      ! Now loop over all fine source elements for this target:
      !CDIR unroll=19
      do iSource = 1, 19
        ! Get the source element in the levelwise buffer
        sourceElem = tLevelDesc%depFromCoarser( iElem )%elemBuffer%val( iSource )
        !CDIR unroll=10
        sourceMom(1:10, iSource ) = momBuf(1:10, sourceElem )
      end do

      ! interpolate only the first 10th moments: rho + vel(3) + s(6)
      tMom(1:10) = mus_interpolate_quadratic3d_leastSq( &
        &   srcMom =    sourceMom(:, 1:19),   &
        &   targetCoord =  targetCoord,                 &
        &   nVals  = 10                                 )

      ! Rescale the shear stress moments
      ! and add back the pi(0) part to the tensor
      tMom( 5) = tMom( 5)*fac + tMom(2)*tMom(2) + cs2*tMom(1)
      tMom( 6) = tMom( 6)*fac + tMom(3)*tMom(3) + cs2*tMom(1)
      tMom( 7) = tMom( 7)*fac + tMom(4)*tMom(4) + cs2*tMom(1)
      tMom( 8) = tMom( 8)*fac + tMom(2)*tMom(3)
      tMom( 9) = tMom( 9)*fac + tMom(3)*tMom(4)
      tMom(10) = tMom(10)*fac + tMom(2)*tMom(4)

      vel(1) = tMom(2)
      vel(2) = tMom(3)
      vel(3) = tMom(4)
?? copy :: Eq_incomp( fEq, tMom(1), vel, layout, 1.0_rk )

      ! Calculate higher order moments
      ! tMom(11:QQ) = matmul( layout%moment%toMoments%A(11:QQ,:), fEq )
      tMom(11) = -fEq(15) + fEq(16) - fEq(17) + fEq(18)
      tMom(12) = -fEq(11) - fEq(12) + fEq(13) + fEq(14)
      tMom(13) = -fEq(15) - fEq(16) + fEq(17) + fEq(18)
      tMom(14) = -fEq( 7) + fEq( 8) - fEq( 9) + fEq(10)
      tMom(15) = -fEq(11) + fEq(12) - fEq(13) + fEq(14)
      tMom(16) = -fEq( 7) - fEq( 8) + fEq( 9) + fEq(10)
      tMom(17) =  fEq(15) + fEq(16) + fEq(17) + fEq(18)
      tMom(18) =  fEq( 7) + fEq( 8) + fEq( 9) + fEq(10)
      tMom(19) =  fEq(11) + fEq(12) + fEq(13) + fEq(14)

      ! do iDir = 1, QQ
      !   tState( ?IDX?( iDir, targetElem, QQ, tnSize )) = sum(layout%moment%toPDF%A(iDir,:)*tMom(:))
      ! end do
      tState( ?IDX?( 1, targetElem, nScalars, tnSize )) = ( -tMom(2)+tMom(5)+tMom(13)+tMom(15)-tMom(17)-tMom(19) ) * div1_2
      tState( ?IDX?( 2, targetElem, nScalars, tnSize )) = ( -tMom(3)+tMom(6)+tMom(11)+tMom(16)-tMom(17)-tMom(18) ) * div1_2
      tState( ?IDX?( 3, targetElem, nScalars, tnSize )) = ( -tMom(4)+tMom(7)+tMom(12)+tMom(14)-tMom(18)-tMom(19) ) * div1_2
      tState( ?IDX?( 4, targetElem, nScalars, tnSize )) = (  tMom(2)+tMom(5)-tMom(13)-tMom(15)-tMom(17)-tMom(19) ) * div1_2
      tState( ?IDX?( 5, targetElem, nScalars, tnSize )) = (  tMom(3)+tMom(6)-tMom(11)-tMom(16)-tMom(17)-tMom(18) ) * div1_2
      tState( ?IDX?( 6, targetElem, nScalars, tnSize )) = (  tMom(4)+tMom(7)-tMom(12)-tMom(14)-tMom(18)-tMom(19) ) * div1_2

      tState( ?IDX?( 7, targetElem, nScalars, tnSize )) = (  tMom( 9)-tMom(14)-tMom(16)+tMom(18) ) * div1_4
      tState( ?IDX?( 8, targetElem, nScalars, tnSize )) = ( -tMom( 9)+tMom(14)-tMom(16)+tMom(18) ) * div1_4
      tState( ?IDX?( 9, targetElem, nScalars, tnSize )) = ( -tMom( 9)-tMom(14)+tMom(16)+tMom(18) ) * div1_4
      tState( ?IDX?(10, targetElem, nScalars, tnSize )) = (  tMom( 9)+tMom(14)+tMom(16)+tMom(18) ) * div1_4

      tState( ?IDX?(11, targetElem, nScalars, tnSize )) = (  tMom(10)-tMom(12)-tMom(15)+tMom(19) ) * div1_4
      tState( ?IDX?(12, targetElem, nScalars, tnSize )) = ( -tMom(10)-tMom(12)+tMom(15)+tMom(19) ) * div1_4
      tState( ?IDX?(13, targetElem, nScalars, tnSize )) = ( -tMom(10)+tMom(12)-tMom(15)+tMom(19) ) * div1_4
      tState( ?IDX?(14, targetElem, nScalars, tnSize )) = (  tMom(10)+tMom(12)+tMom(15)+tMom(19) ) * div1_4

      tState( ?IDX?(15, targetElem, nScalars, tnSize )) = (  tMom( 8)-tMom(11)-tMom(13)+tMom(17) ) * div1_4
      tState( ?IDX?(16, targetElem, nScalars, tnSize )) = ( -tMom( 8)+tMom(11)-tMom(13)+tMom(17) ) * div1_4
      tState( ?IDX?(17, targetElem, nScalars, tnSize )) = ( -tMom( 8)-tMom(11)+tMom(13)+tMom(17) ) * div1_4
      tState( ?IDX?(18, targetElem, nScalars, tnSize )) = (  tMom( 8)+tMom(11)+tMom(13)+tMom(17) ) * div1_4

      tState( ?IDX?(19, targetElem, nScalars, tnSize )) = tMom(1)-tMom(5)-tMom(6)-tMom(7)+tMom(17)+tMom(18)+tMom(19)
    end do ! indElem
!$omp end do nowait

  end subroutine fillFinerGhostsFromMe_quadratic3dIncomp
! ****************************************************************************** !


! ****************************************************************************** !
  !> Fill fine ghost from coarse fluid by quadratic interpolation.
  !! 1. Compute moments for all source elements, save in momBuf
  !! 2. For each target, interpolate 1-10 moments (den, vel, tau)
  !! 3. calculate fEq and use it to calculate high order moments(11-19)
  !! 4. convert moments to PDF
  !! This routine is used by 3D acoustic quadratic interpolation.
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_quadratic3d )
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iVal, iElem, indElem, iDir
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    real(kind=rk) :: f( QQ ) ! pdf to reconstruct from
    ! moments of the source elements' pdf
    real(kind=rk) :: sourceMom( QQ, nMaxSources )
    ! shear stress scaling factor
    real(kind=rk) :: fac, inv_rho, rho, vel(3)
    real(kind=rk) :: tMom( QQ )  ! temp moment calculation
    ! ---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    fac = getNoneqFac( omLvl( sourceLevel ), omLvl( targetLevel )  )

    ! First calculate all the required moments for all the source elements
    do iSourceElem = 1, tLevelDesc%sourceFromCoarser%nVals

      ! Get the source element's position in the state vector
      sourceElem = tLevelDesc%sourceFromCoarser%val( iSourceElem )

      do iVal = 1, QQ
        f(iVal) = sState(?IDX?( iVal, sourceElem, nScalars, snSize ))
      enddo

      ! Calculate the raw moments form the pdf
      tMom(:) = matmul( layout%moment%toMoments%A, f)
      inv_rho = 1.0_rk / tMom(1)

      ! Transfer momentum to velocity
      tMom(2:4) = tMom(2:4) * inv_rho
      ! remove pi(0) parts from the second order moment
      ! that is ((rho u_\alpha * rho u_\beta) / rho)
      ! and additionally rho*cs2 from diagonal entries
      ! Pab = sum( c_ia*c_ib *f_i ) - kronecker_ab*(rho*cs2) - rho*u_a*u_b
      tMom( 5) = tMom( 5) - tMom(2)*tMom(2)*tMom(1) - cs2*tMom(1) !Pxx
      tMom( 6) = tMom( 6) - tMom(3)*tMom(3)*tMom(1) - cs2*tMom(1) !Pyy
      tMom( 7) = tMom( 7) - tMom(4)*tMom(4)*tMom(1) - cs2*tMom(1) !Pzz
      tMom( 8) = tMom( 8) - tMom(2)*tMom(3)*tMom(1)               !Pxy
      tMom( 9) = tMom( 9) - tMom(3)*tMom(4)*tMom(1)               !Pyz
      tMom(10) = tMom(10) - tMom(2)*tMom(4)*tMom(1)               !Pxz

      ! Store all the moments to the current source element's position in the
      ! buffer
      momBuf( :, iSourceElem ) = tMom(1:10)
    enddo

    ! Treat all fine target elements:
    do indElem = 1, nTargets

      iElem = targetList( indElem )
      targetElem = iElem + tLevelDesc%offset( 1, eT_ghostFromCoarser)
      nSourceElems = tLevelDesc%depFromCoarser( iElem )%elem%nVals

      do iSourceElem = 1, nSourceElems
        ! Get the source element in the levelwise buffer
        sourceElem = tLevelDesc%depFromCoarser( iElem )%elemBuffer%val( iSourceElem )
        ! and read its entries
        sourceMom(:, iSourceElem) = momBuf(:, sourceElem)
      end do

      tMom(1:10) = mus_interpolate_quadratic3d_leastSq(              &
        &   srcMom      = sourceMom(1:10, 1:19), &
        &   targetCoord = tLevelDesc%depFromCoarser( iElem )%coord, &
        &   nVals       = 10 )

      inv_rho = 1.0_rk / tMom(1)
      ! Transfer velocity to momentum
      tMom(2:4) = tMom(2:4)*tMom(1)

      ! Rescale the shear stress moments
      ! and add back the pi(0) part to the tensor
      tMom( 5) = tMom( 5)*fac + tMom(2)*tMom(2)*inv_rho + cs2*tMom(1)
      tMom( 6) = tMom( 6)*fac + tMom(3)*tMom(3)*inv_rho + cs2*tMom(1)
      tMom( 7) = tMom( 7)*fac + tMom(4)*tMom(4)*inv_rho + cs2*tMom(1)
      tMom( 8) = tMom( 8)*fac + tMom(2)*tMom(3)*inv_rho
      tMom( 9) = tMom( 9)*fac + tMom(3)*tMom(4)*inv_rho
      tMom(10) = tMom(10)*fac + tMom(2)*tMom(4)*inv_rho

      rho = tMom(1)
      vel(1:3) = tMom(2:4) * inv_rho
?? copy :: Eq_comp( f, rho, vel, layout )

      tMom(11:19) = matmul( layout%moment%toMoments%A(11:19,:), f )

      ! transform the moments back to the pdfs
      f = matmul( layout%moment%toPdf%A, tMom )

      ! Now write the resulting pdf in the current direction to the target
      ! Element position
      do iVal = 1, QQ
        tState( ?IDX?( iVal, targetElem, nScalars, tnSize )) = f(iVal)
      enddo

    end do ! indElem

    ! deallocate( momBuf )

  end subroutine fillFinerGhostsFromMe_quadratic3d
! ****************************************************************************** !


! ****************************************************************************** !
  !> Biquadratic interpolation for a single quantity phi
  !!
  ! real(kind=rk), dimension(6,9),parameter  :: intpMaxtrix =  &
  !           reshape((/ &
  !   &  div2_9,  -div1_6,    0.0d0,    div1_6,   -div1_3,   0.0d0, &
  !   &  div2_9,    0.0d0,  -div1_6,   -div1_3,    div1_6,   0.0d0, &
  !   &  div2_9,   div1_6,    0.0d0,    div1_6,   -div1_3,   0.0d0, &
  !   &  div2_9,    0.0d0,   div1_6,   -div1_3,    div1_6,   0.0d0, &
  !   & -div1_9,  -div1_6,  -div1_6,    div1_6,    div1_6,  0.25d0, &
  !   & -div1_9,  -div1_6,   div1_6,    div1_6,    div1_6, -0.25d0, &
  !   & -div1_9,   div1_6,  -div1_6,    div1_6,    div1_6, -0.25d0, &
  !   & -div1_9,   div1_6,   div1_6,    div1_6,    div1_6,  0.25d0, &
  !   &  div5_9,    0.0d0,    0.0d0,   -div1_3,   -div1_3,   0.0d0  &
  !  /),(/6,9/))
  pure function mus_interpolate_quadratic2d_leastSq_vect( srcMom, targetCoord, &
    &                                        nVals, nSourceElems ) result( phi )
    ! ---------------------------------------------------------------------------
    integer, intent(in) :: nVals
    !>
    integer, intent(in) :: nSourceElems
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: srcMom(nVals, nSourceElems)
    !> interpolation location within the square
    real(kind=rk), intent(in) :: targetCoord(2)
    !> interpolated value
    real(kind=rk) :: phi( nVals )
    ! ---------------------------------------------------------------------------
    real(kind=rk) :: lhs(6), sum13, sum24, sum5678
    integer :: iVal
    ! ---------------------------------------------------------------------------

    do iVal = 1, nVals
      ! We extract momentum information completely on the view of the source
      ! coordinate system
      ! Set the right hand side of the equation system
      ! rhs(:)  = srcMom(iVal,:)
      sum5678 = srcMom(iVal,5) + srcMom(iVal,6) + srcMom(iVal,7) + srcMom(iVal,8)
      sum13   = srcMom(iVal,1) + srcMom(iVal,3)
      sum24   = srcMom(iVal,2) + srcMom(iVal,4)
      ! Solve the problem, where b = rhs, x = coefficients
      ! A*x = b
      ! overdetermined, solve the least Square fit problem
      ! (A^T)A*x = (A^T)b
      ! x = ((A^T)A)^-1*(A^T)b
      ! Solve linear system of equation with inverted matrix
! lhs = matmul( matrix, rhs )
lhs(1) = div2_9 * ( sum13 + sum24 ) - div1_9 * sum5678 + div5_9 * srcMom(iVal,9)
lhs(2) = div1_6 * ( - srcMom(iVal,1) + srcMom(iVal,3) &
  &                 - srcMom(iVal,5) - srcMom(iVal,6) &
  &                 + srcMom(iVal,7) + srcMom(iVal,8) )
lhs(3) = div1_6 * ( - srcMom(iVal,2) + srcMom(iVal,4) &
  &                 - srcMom(iVal,5) + srcMom(iVal,6) &
  &                 - srcMom(iVal,7) + srcMom(iVal,8) )
lhs(4) = div1_6 * ( sum13 + sum5678 ) - div1_3 * ( sum24 + srcMom(iVal,9) )
lhs(5) = div1_6 * ( sum24 + sum5678 ) - div1_3 * ( sum13 + srcMom(iVal,9) )
lhs(6) = 0.25d0 * (   srcMom(iVal,5) - srcMom(iVal,6) &
  &                 - srcMom(iVal,7) + srcMom(iVal,8) )

      ! Evaluate the bubble function with the above calculated coefficients
      ! m_pi(x) = a1+a2*xcoord+a3*ycoord+a4*xcoord*xcoord+a5*ycoord*ycoord+a6*xcoord*ycoord;
      phi(iVal) =   lhs(1)                              &
        &   + lhs(2)*targetCoord(c_x)                   &
        &   + lhs(3)*targetCoord(c_y)                   &
        &   + lhs(4)*targetCoord(c_x)*targetCoord(c_x)  &
        &   + lhs(5)*targetCoord(c_y)*targetCoord(c_y)  &
        &   + lhs(6)*targetCoord(c_x)*targetCoord(c_y)
    end do

  end function mus_interpolate_quadratic2d_leastSq_vect
! ****************************************************************************** !

! ****************************************************************************** !
  pure function mus_interpolate_quadratic2d_leastSq_scal( srcMom, targetCoord,      &
    &                                       nSourceElems ) result( phi )
    ! ---------------------------------------------------------------------------
    integer, intent(in) :: nSourceElems
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: srcMom(nSourceElems)
    !> interpolation location within the square
    real(kind=rk), intent(in) :: targetCoord(2)
    !> interpolated value
    real(kind=rk) :: phi
    ! ---------------------------------------------------------------------------
    real(kind=rk) :: lhs(6), sum13, sum24, sum5678 ! Coefficients
    ! ---------------------------------------------------------------------------

    ! We extract momentum information completely on the view of the source
    ! coordinate system
    ! Set the right hand side of the equation system
    ! Solve the problem, where b = rhs, x = coefficients
    ! A*x = b
    ! overdetermined, solve the least Square fit problem
    ! (A^T)A*x = (A^T)b
    ! x = ((A^T)A)^-1*(A^T)b
    ! Solve linear system of equation with inverted matrix
    ! lhs = matmul( matrix, srcMom(:) )
    sum5678 = srcMom(5) + srcMom(6) + srcMom(7) + srcMom(8)
    sum13   = srcMom(1) + srcMom(3)
    sum24   = srcMom(2) + srcMom(4)
lhs(1) = div2_9 * ( sum13 + sum24 ) - div1_9 * sum5678 + div5_9 * srcMom(9)
lhs(2) = div1_6 * ( - srcMom(1) + srcMom(3) &
  &                 - srcMom(5) - srcMom(6) &
  &                 + srcMom(7) + srcMom(8) )
lhs(3) = div1_6 * ( - srcMom(2) + srcMom(4) &
  &                 - srcMom(5) + srcMom(6) &
  &                 - srcMom(7) + srcMom(8) )
lhs(4) = div1_6 * ( sum13 + sum5678 ) - div1_3 * ( sum24 + srcMom(9) )
lhs(5) = div1_6 * ( sum24 + sum5678 ) - div1_3 * ( sum13 + srcMom(9) )
lhs(6) = 0.25d0 * (   srcMom(5) - srcMom(6) &
  &                 - srcMom(7) + srcMom(8) )

    ! Evaluate the bubble function with the above calculated coefficients
    ! m_pi(x) = a1+a2*xcoord+a3*ycoord+a4*xcoord*xcoord+a5*ycoord*ycoord+a6*xcoord*ycoord;
    phi =   lhs(1)                              &
      &   + lhs(2)*targetCoord(c_x)                   &
      &   + lhs(3)*targetCoord(c_y)                   &
      &   + lhs(4)*targetCoord(c_x)*targetCoord(c_x)  &
      &   + lhs(5)*targetCoord(c_y)*targetCoord(c_y)  &
      &   + lhs(6)*targetCoord(c_x)*targetCoord(c_y)

  end function mus_interpolate_quadratic2d_leastSq_scal
! ****************************************************************************** !


! ****************************************************************************** !
  !> Triquadratic interpolation for a single quantity phi
  !!
  pure function mus_interpolate_quadratic3d_leastSq_scal( srcMom, targetCoord ) &
    &                                                               result( phi )
    ! ---------------------------------------------------------------------------
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: srcMom(19)
    !> interpolation location within the square
    real(kind=rk), intent(in) :: targetCoord(3)
    !> interpolated value
    real(kind=rk) :: phi
    ! ---------------------------------------------------------------------------
    real(kind=rk) :: a(10) ! Coefficients
    real(kind=rk) :: sum1_4, sum2_5, sum3_6, sum_7_10, sum11_14, sum15_18
    real(kind=rk) :: m19_div5_21
    ! ---------------------------------------------------------------------------

    ! ---------------------------------------------------------------------------
    ! We extract momentum information completely on the view of the source
    ! coordinate system
    ! Set the right hand side of the equation system
    ! Solve the problem, where b = rhs, x = coefficients
    ! A*x = b
    ! overdetermined, solve the least Square fit problem
    ! (A^T)A*x = (A^T)b
    ! x = ((A^T)A)^-1*(A^T)b
    ! Solve linear system of equation with inverted matrix
    ! a(:)= matmul( matrix, srcMom(:))
    sum1_4 = srcMom(1) + srcMom(4)
    sum2_5 = srcMom(2) + srcMom(5)
    sum3_6 = srcMom(3) + srcMom(6)
    sum_7_10 = srcMom( 7) + srcMom( 8) + srcMom( 9) + srcMom(10)
    sum11_14 = srcMom(11) + srcMom(12) + srcMom(13) + srcMom(14)
    sum15_18 = srcMom(15) + srcMom(16) + srcMom(17) + srcMom(18)
    m19_div5_21 = div5_21 * srcMom(19)

    a(1) =   div4_21 * ( sum1_4 + sum2_5 + sum3_6 )&
      &    - div1_21 * ( sum_7_10 + sum11_14 + sum15_18 ) + div3_7  * srcMom(19)

    a(2) = 0.1d0 * ( - srcMom( 1) + srcMom( 4) &
      &              - srcMom(11) + srcMom(12) - srcMom(13) + srcMom(14) &
      &              - srcMom(15) - srcMom(16) + srcMom(17) + srcMom(18) )
    a(3) = 0.1d0 * ( - srcMom( 2) + srcMom( 5) &
      &              - srcMom( 7) - srcMom( 8) + srcMom( 9) + srcMom(10) &
      &              - srcMom(15) + srcMom(16) - srcMom(17) + srcMom(18) )
    a(4) = 0.1d0 * ( - srcMom( 3) + srcMom( 6) &
      &              - srcMom( 7) + srcMom( 8) - srcMom( 9) + srcMom(10) &
      &              - srcMom(11) - srcMom(12) + srcMom(13) + srcMom(14) )

    a(5) = div1_42 * sum1_4 - div1_7  * ( sum2_5 + sum3_6 ) &
      &  - div1_21 * sum_7_10 + div5_42 * ( sum11_14 + sum15_18 ) - m19_div5_21

    a(6) = div1_42 * sum2_5 - div1_7  * ( sum1_4 + sum3_6 ) &
      &  - div1_21 * sum11_14 + div5_42 * ( sum_7_10 + sum15_18 ) - m19_div5_21

    a(7) = div1_42 * sum3_6 - div1_7  * ( sum1_4 + sum2_5 ) &
      &  - div1_21 * sum15_18 + div5_42 * ( sum_7_10 + sum11_14 ) - m19_div5_21

    a(8) = 0.25d0 * ( srcMom(15) - srcMom(16) - srcMom(17) + srcMom(18) )
    a(9) = 0.25d0 * ( srcMom( 7) - srcMom( 8) - srcMom( 9) + srcMom(10) )
    a(10)= 0.25d0 * ( srcMom(11) - srcMom(12) - srcMom(13) + srcMom(14) )

    ! Evaluate the bubble function with the above calculated coefficients
    ! m_pi(x) = a1+a2*xcoord+a3*ycoord+a4*xcoord*xcoord+a5*ycoord*ycoord+a6*xcoord*ycoord;
    phi =   a( 1)                                   &
      &   + a( 2)*targetCoord(c_x)                  &
      &   + a( 3)*targetCoord(c_y)                  &
      &   + a( 4)*targetCoord(c_z)                  &
      &   + a( 5)*targetCoord(c_x)*targetCoord(c_x) &
      &   + a( 6)*targetCoord(c_y)*targetCoord(c_y) &
      &   + a( 7)*targetCoord(c_z)*targetCoord(c_z) &
      &   + a( 8)*targetCoord(c_x)*targetCoord(c_y) &
      &   + a( 9)*targetCoord(c_y)*targetCoord(c_z) &
      &   + a(10)*targetCoord(c_z)*targetCoord(c_x)

  end function mus_interpolate_quadratic3d_leastSq_scal
! ****************************************************************************** !


! ****************************************************************************** !
  !> Triquadratic interpolation for a single quantity phi
  ! matrixInv%A(1:10,1:19) =
  !      1     2     3     4     5     6     7     8     9    10    11    12    13    14    15    16    17    18    19
  ! [ 4/21  4/21  4/21  4/21  4/21  4/21 -1/21 -1/21 -1/21 -1/21 -1/21 -1/21 -1/21 -1/21 -1/21 -1/21 -1/21 -1/21   3/7]
  ! [-1/10     0     0  1/10     0     0     0     0     0     0 -1/10  1/10 -1/10  1/10 -1/10 -1/10  1/10  1/10     0]
  ! [    0 -1/10     0     0  1/10     0 -1/10 -1/10  1/10  1/10     0     0     0     0 -1/10  1/10 -1/10  1/10     0]
  ! [    0     0 -1/10     0     0  1/10 -1/10  1/10 -1/10  1/10 -1/10 -1/10  1/10  1/10     0     0     0     0     0]
  ! [ 1/42  -1/7  -1/7  1/42  -1/7  -1/7 -1/21 -1/21 -1/21 -1/21  5/42  5/42  5/42  5/42  5/42  5/42  5/42  5/42 -5/21]
  ! [ -1/7  1/42  -1/7  -1/7  1/42  -1/7  5/42  5/42  5/42  5/42 -1/21 -1/21 -1/21 -1/21  5/42  5/42  5/42  5/42 -5/21]
  ! [ -1/7  -1/7  1/42  -1/7  -1/7  1/42  5/42  5/42  5/42  5/42  5/42  5/42  5/42  5/42 -1/21 -1/21 -1/21 -1/21 -5/21]
  ! [    0     0     0     0     0     0     0     0     0     0     0     0     0     0   1/4  -1/4  -1/4   1/4     0]
  ! [    0     0     0     0     0     0   1/4  -1/4  -1/4   1/4     0     0     0     0     0     0     0     0     0]
  ! [    0     0     0     0     0     0     0     0     0     0   1/4  -1/4  -1/4   1/4     0     0     0     0     0]
  !
  pure function mus_interpolate_quadratic3d_leastSq_vect( srcMom, targetCoord, &
    &                                        nVals ) result( phi )
    ! ---------------------------------------------------------------------------
    !> number of quantities to interpolation
    integer, intent(in) :: nVals
    !> source values of the momentum on the square corners
    real(kind=rk), intent(in) :: srcMom(nVals, 19)
    !> interpolation location within the square
    real(kind=rk), intent(in) :: targetCoord(3)
    !> interpolated value
    real(kind=rk) :: phi( nVals )
    ! ---------------------------------------------------------------------------
    real(kind=rk) :: a(10) ! Coefficients
    integer :: iVal
    real(kind=rk) :: sum1_4, sum2_5, sum3_6, sum_7_10, sum11_14, sum15_18
    real(kind=rk) :: m19_div5_21
    ! ---------------------------------------------------------------------------

    ! We extract momentum information completely on the view of the source
    ! coordinate system
    ! Set the right hand side of the equation system
    ! Solve the problem, where b = rhs, x = coefficients
    ! A*x = b
    ! overdetermined, solve the least Square fit problem
    ! (A^T)A*x = (A^T)b
    ! x = ((A^T)A)^-1*(A^T)b
    ! Solve linear system of equation with inverted matrix
    do iVal = 1, nVals
      ! a(:)= matmul( matrix, srcMom(iVal, :))

      sum1_4 = srcMom(iVal,1) + srcMom(iVal,4)
      sum2_5 = srcMom(iVal,2) + srcMom(iVal,5)
      sum3_6 = srcMom(iVal,3) + srcMom(iVal,6)
      sum_7_10 = srcMom(iVal, 7) + srcMom(iVal, 8) + srcMom(iVal, 9) + srcMom(iVal,10)
      sum11_14 = srcMom(iVal,11) + srcMom(iVal,12) + srcMom(iVal,13) + srcMom(iVal,14)
      sum15_18 = srcMom(iVal,15) + srcMom(iVal,16) + srcMom(iVal,17) + srcMom(iVal,18)
      m19_div5_21 = div5_21 * srcMom(iVal,19)

      a(1) =   div4_21 * ( sum1_4 + sum2_5 + sum3_6 )&
        &    - div1_21 * ( sum_7_10 + sum11_14 + sum15_18 ) + div3_7  * srcMom(iVal,19)

      a(2) = 0.1d0 * ( - srcMom(iVal, 1) + srcMom(iVal, 4) &
        &              - srcMom(iVal,11) + srcMom(iVal,12) - srcMom(iVal,13) + srcMom(iVal,14) &
        &              - srcMom(iVal,15) - srcMom(iVal,16) + srcMom(iVal,17) + srcMom(iVal,18) )
      a(3) = 0.1d0 * ( - srcMom(iVal, 2) + srcMom(iVal, 5) &
        &              - srcMom(iVal, 7) - srcMom(iVal, 8) + srcMom(iVal, 9) + srcMom(iVal,10) &
        &              - srcMom(iVal,15) + srcMom(iVal,16) - srcMom(iVal,17) + srcMom(iVal,18) )
      a(4) = 0.1d0 * ( - srcMom(iVal, 3) + srcMom(iVal, 6) &
        &              - srcMom(iVal, 7) + srcMom(iVal, 8) - srcMom(iVal, 9) + srcMom(iVal,10) &
        &              - srcMom(iVal,11) - srcMom(iVal,12) + srcMom(iVal,13) + srcMom(iVal,14) )

      a(5) = div1_42 * sum1_4 - div1_7  * ( sum2_5 + sum3_6 ) &
        &  - div1_21 * sum_7_10 + div5_42 * ( sum11_14 + sum15_18 ) - m19_div5_21

      a(6) = div1_42 * sum2_5 - div1_7  * ( sum1_4 + sum3_6 ) &
        &  - div1_21 * sum11_14 + div5_42 * ( sum_7_10 + sum15_18 ) - m19_div5_21

      a(7) = div1_42 * sum3_6 - div1_7  * ( sum1_4 + sum2_5 ) &
        &  - div1_21 * sum15_18 + div5_42 * ( sum_7_10 + sum11_14 ) - m19_div5_21

      a(8) = div1_4 * ( srcMom(iVal,15) - srcMom(iVal,16) - srcMom(iVal,17) + srcMom(iVal,18) )
      a(9) = div1_4 * ( srcMom(iVal, 7) - srcMom(iVal, 8) - srcMom(iVal, 9) + srcMom(iVal,10) )
      a(10)= div1_4 * ( srcMom(iVal,11) - srcMom(iVal,12) - srcMom(iVal,13) + srcMom(iVal,14) )

      ! Evaluate the bubble function with the above calculated coefficients
      ! m_pi(x) = a1+a2*xcoord+a3*ycoord+a4*xcoord*xcoord+a5*ycoord*ycoord+a6*xcoord*ycoord;
      phi( iVal ) =   a( 1) &
        &           + a( 2)*targetCoord(c_x)                  &
        &           + a( 3)*targetCoord(c_y)                  &
        &           + a( 4)*targetCoord(c_z)                  &
        &           + a( 5)*targetCoord(c_x)*targetCoord(c_x) &
        &           + a( 6)*targetCoord(c_y)*targetCoord(c_y) &
        &           + a( 7)*targetCoord(c_z)*targetCoord(c_z) &
        &           + a( 8)*targetCoord(c_x)*targetCoord(c_y) &
        &           + a( 9)*targetCoord(c_y)*targetCoord(c_z) &
        &           + a(10)*targetCoord(c_z)*targetCoord(c_x)
    enddo

  end function mus_interpolate_quadratic3d_leastSq_vect
! ****************************************************************************** !

! ****************************************************************************** !
  !> Calculate the conversion factor for nonEq in difference levels
  pure function getNoneqFac( omegaS, omegaT ) result( fac )
    ! ---------------------------------------------------------------------------
    real(kind=rk), intent(in) :: omegaS  !< omega value on source level
    real(kind=rk), intent(in) :: omegaT  !< omage value on target level
    real(kind=rk) :: fac
    ! ---------------------------------------------------------------------------

??  if( PUSH ) then
      fac = ( 1._rk - 0.5_rk * omegaS ) / ( 1._rk - 0.5_rk * omegaT )
??  else
      fac =   ( 1._rk - 0.5_rk * omegaS ) * ( 1._rk - omegaT ) &
        &   / ( 1._rk - omegaS ) / ( 1._rk - 0.5_rk * omegaT )
??  endif

  end function getNoneqFac
! ****************************************************************************** !

! ****************************************************************************** !
  !> 2D Biquadratic interpolation for fill fine from coarse.
  !! array f contains function values at 9 coarse elements.
  !! they are arranged as:
  !!   7  8  9
  !!   4  5  6
  !!   1  2  3
  !! position of element 1 is set as original, i.e. (0,0),
  !!             element 9 is set as (1,1)
  !! x,y are the coordinates of target position.
  !! fOut is the interpolated value at target position
  !!
  pure function biquadratic_interpolate( f, x, y ) result( fOut )
    ! ---------------------------------------------------------------------------
    real(kind=rk), intent(in) :: f(9)
    real(kind=rk), intent(in) :: x, y
    ! ---------------------------------------------------------------------------
    real(kind=rk) :: fOut
    real(kind=rk) :: a0, a1, a2, b0, b1, b2, c0, c1, c2, f1, f2, f3, m0, m1, m2
    ! ---------------------------------------------------------------------------

    a0 =          f(1)
    a1 = -3._rk * f(1) + 4._rk * f(2) -         f(3)
    a2 =  2._rk * f(1) - 4._rk * f(2) + 2._rk * f(3)

    b0 =          f(4)
    b1 = -3._rk * f(4) + 4._rk * f(5) -         f(6)
    b2 =  2._rk * f(4) - 4._rk * f(5) + 2._rk * f(6)

    c0 =          f(7)
    c1 = -3._rk * f(7) + 4._rk * f(8) -         f(9)
    c2 =  2._rk * f(7) - 4._rk * f(8) + 2._rk * f(9)

    f1 = a0 + a1 * x + a2 * x * x
    f2 = b0 + b1 * x + b2 * x * x
    f3 = c0 + c1 * x + c2 * x * x

    m0 =          f1
    m1 = -3._rk * f1 + 4._rk * f2 -         f3
    m2 =  2._rk * f1 - 4._rk * f2 + 2._rk * f3

    fOut = m0 + m1 * y + m2 * y * y

  end function biquadratic_interpolate
! ****************************************************************************** !

! ****************************************************************************** !
?? copy :: interpolateRoutine_header( fillFinerGhostsFromMe_quadratic3dIncomp_new )
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel
    integer :: targetLevel
    real(kind=rk) :: fac
    real(kind=rk) :: targetMom(10, nTargets)
    ! ---------------------------------------------------------------------------

    sourceLevel = level
    targetLevel = level + 1
    fac = getNoneqFac( omLvl( sourceLevel ), omLvl( targetLevel )  )

    ! First calculate all the required moments for all the source elements
    call calc_moments( sState, nScalars, snSize, &
      &                tLevelDesc%sourceFromCoarser%val, &
      &                tLevelDesc%sourceFromCoarser%nVals, &
      &                momBuf )

    call intp_target(nTargets, targetList, tLevelDesc%depFromCoarser, momBuf, targetMom)

    call convert_target_pdf( targetMom = targetMom, &
      &                      targetList = targetList, &
      &                      nTargets = nTargets, &
      &                      offset = tLevelDesc%offset( 1, eT_ghostFromCoarser), &
      &                      fac = fac, &
      &                      state = tState, &
      &                      nScalars = nScalars, &
      &                      nSize  = tnSize )

  end subroutine fillFinerGhostsFromMe_quadratic3dIncomp_new
! ****************************************************************************** !

! ****************************************************************************** !
  subroutine calc_moments(state, nScalars, nSize, elementList, nElems, buffer )
    real(kind=rk), intent(in) :: state(:)
    integer, intent(in) :: nScalars
    integer, intent(in) :: nSize
    integer, intent(in) :: nElems
    integer, intent(in) :: elementList(nElems)
    real(kind=rk), intent(out) :: buffer(10,nElems)

    real(kind=rk) :: cc1, cc2, cc3
    real(kind=rk) :: f(QQ)
    integer :: iElem, sourceElem

    !$omp do schedule(static)
    !DIR$ IVDEP
    !CDIR NODEP
    !CDIR NOLSTVAL
    do iElem = 1, nElems

      sourceElem = elementList(iElem)
      f( 1) = state( ?IDX?(  1, sourceElem, nScalars, nSize ))
      f( 2) = state( ?IDX?(  2, sourceElem, nScalars, nSize ))
      f( 3) = state( ?IDX?(  3, sourceElem, nScalars, nSize ))
      f( 4) = state( ?IDX?(  4, sourceElem, nScalars, nSize ))
      f( 5) = state( ?IDX?(  5, sourceElem, nScalars, nSize ))
      f( 6) = state( ?IDX?(  6, sourceElem, nScalars, nSize ))
      f( 7) = state( ?IDX?(  7, sourceElem, nScalars, nSize ))
      f( 8) = state( ?IDX?(  8, sourceElem, nScalars, nSize ))
      f( 9) = state( ?IDX?(  9, sourceElem, nScalars, nSize ))
      f(10) = state( ?IDX?( 10, sourceElem, nScalars, nSize ))
      f(11) = state( ?IDX?( 11, sourceElem, nScalars, nSize ))
      f(12) = state( ?IDX?( 12, sourceElem, nScalars, nSize ))
      f(13) = state( ?IDX?( 13, sourceElem, nScalars, nSize ))
      f(14) = state( ?IDX?( 14, sourceElem, nScalars, nSize ))
      f(15) = state( ?IDX?( 15, sourceElem, nScalars, nSize ))
      f(16) = state( ?IDX?( 16, sourceElem, nScalars, nSize ))
      f(17) = state( ?IDX?( 17, sourceElem, nScalars, nSize ))
      f(18) = state( ?IDX?( 18, sourceElem, nScalars, nSize ))
      f(19) = state( ?IDX?( 19, sourceElem, nScalars, nSize ))

      cc1 = f(7) + f(8) + f(9) + f(10)
      cc2 = f(11) + f(12) + f(13) + f(14)
      cc3 = f(15) + f(16) + f(17) + f(18)

      buffer(1,iElem) = cc1 + cc2 + cc3 + f(19) + f(1) + f(2) + f(3) + f(4) + f(5) + f(6)
      ! ux uy uz
      buffer(2,iElem) = -f(1) + f(4) + f(12) + f(14) + f(17) + f(18) - f(11) - f(13) - f(15) - f(16)
      buffer(3,iElem) = -f(2) + f(5) + f( 9) + f(10) + f(16) + f(18) - f( 7) - f( 8) - f(15) - f(17)
      buffer(4,iElem) = -f(3) + f(6) + f( 8) + f(10) + f(13) + f(14) - f( 7) - f( 9) - f(11) - f(12)
      ! Sxx Syy Szz
      buffer(5,iElem) =  f(1) + f(4) + cc2 + cc3 - buffer(2,iElem)*buffer(2,iElem) - cs2*buffer(1,iElem)
      buffer(6,iElem) =  f(2) + f(5) + cc1 + cc3 - buffer(3,iElem)*buffer(3,iElem) - cs2*buffer(1,iElem)
      buffer(7,iElem) =  f(3) + f(6) + cc1 + cc2 - buffer(4,iElem)*buffer(4,iElem) - cs2*buffer(1,iElem)
      ! Sxy Syz Sxz
      buffer(8,iElem) =  f(15) - f(16) - f(17) + f(18) - buffer(2,iElem)*buffer(3,iElem)
      buffer(9,iElem) =  f( 7) - f( 8) - f( 9) + f(10) - buffer(3,iElem)*buffer(4,iElem)
     buffer(10,iElem) =  f(11) - f(12) - f(13) + f(14) - buffer(2,iElem)*buffer(4,iElem)

    end do
!$omp end do
! Here is a implicit barrier, because we can go further until momBuf is complete

  end subroutine calc_moments
! ****************************************************************************** !

! ****************************************************************************** !
  subroutine convert_target_pdf( targetMom, targetList, nTargets, offset, fac, &
    &                            state, nSize, nScalars )
    integer, intent(in) :: nScalars
    integer, intent(in) :: nSize
    integer, intent(in) :: nTargets
    integer, intent(in) :: targetList(nTargets)
    real(kind=rk), intent(in) :: targetMom(10,nTargets)
    integer, intent(in) :: offset
    real(kind=rk), intent(in) :: fac
    real(kind=rk), intent(inout) :: state(:)

    integer :: iElem, indElem, targetElem
    real(kind=rk) :: tMom(QQ)
    real(kind=rk) :: fEq(QQ)
?? copy :: fEq_d3q19_var

    !DIR$ IVDEP
    !CDIR NODEP
    do indElem = 1, nTargets
      iElem = targetList( indElem )
      targetElem = iElem + offset

      tMom( 1) = targetMom( 1, indElem)
      tMom( 2) = targetMom( 2, indElem)
      tMom( 3) = targetMom( 3, indElem)
      tMom( 4) = targetMom( 4, indElem)
      tMom( 5) = targetMom( 5, indElem)*fac + tMom(2)*tMom(2) + cs2*tMom(1)
      tMom( 6) = targetMom( 6, indElem)*fac + tMom(3)*tMom(3) + cs2*tMom(1)
      tMom( 7) = targetMom( 7, indElem)*fac + tMom(4)*tMom(4) + cs2*tMom(1)
      tMom( 8) = targetMom( 8, indElem)*fac + tMom(2)*tMom(3)
      tMom( 9) = targetMom( 9, indElem)*fac + tMom(3)*tMom(4)
      tMom(10) = targetMom(10, indElem)*fac + tMom(2)*tMom(4)

!!!!!?? copy :: Eq_incomp( fEq, tMom(1), vel, layout, 1.0_rk )
?? copy :: fEq_d3q19_a( fEq, tMom(2), tMom(3), tMom(4), tMom(1), 1.0_rk )

      ! Calculate higher order moments
      tMom(11) = -fEq(15) + fEq(16) - fEq(17) + fEq(18)
      tMom(12) = -fEq(11) - fEq(12) + fEq(13) + fEq(14)
      tMom(13) = -fEq(15) - fEq(16) + fEq(17) + fEq(18)
      tMom(14) = -fEq( 7) + fEq( 8) - fEq( 9) + fEq(10)
      tMom(15) = -fEq(11) + fEq(12) - fEq(13) + fEq(14)
      tMom(16) = -fEq( 7) - fEq( 8) + fEq( 9) + fEq(10)
      tMom(17) =  fEq(15) + fEq(16) + fEq(17) + fEq(18)
      tMom(18) =  fEq( 7) + fEq( 8) + fEq( 9) + fEq(10)
      tMom(19) =  fEq(11) + fEq(12) + fEq(13) + fEq(14)

      state( ?IDX?( 1, targetElem, nScalars, nSize )) = ( -tMom(2)+tMom(5)+tMom(13)+tMom(15)-tMom(17)-tMom(19) ) * div1_2
      state( ?IDX?( 2, targetElem, nScalars, nSize )) = ( -tMom(3)+tMom(6)+tMom(11)+tMom(16)-tMom(17)-tMom(18) ) * div1_2
      state( ?IDX?( 3, targetElem, nScalars, nSize )) = ( -tMom(4)+tMom(7)+tMom(12)+tMom(14)-tMom(18)-tMom(19) ) * div1_2
      state( ?IDX?( 4, targetElem, nScalars, nSize )) = (  tMom(2)+tMom(5)-tMom(13)-tMom(15)-tMom(17)-tMom(19) ) * div1_2
      state( ?IDX?( 5, targetElem, nScalars, nSize )) = (  tMom(3)+tMom(6)-tMom(11)-tMom(16)-tMom(17)-tMom(18) ) * div1_2
      state( ?IDX?( 6, targetElem, nScalars, nSize )) = (  tMom(4)+tMom(7)-tMom(12)-tMom(14)-tMom(18)-tMom(19) ) * div1_2

      state( ?IDX?( 7, targetElem, nScalars, nSize )) = (  tMom( 9)-tMom(14)-tMom(16)+tMom(18) ) * div1_4
      state( ?IDX?( 8, targetElem, nScalars, nSize )) = ( -tMom( 9)+tMom(14)-tMom(16)+tMom(18) ) * div1_4
      state( ?IDX?( 9, targetElem, nScalars, nSize )) = ( -tMom( 9)-tMom(14)+tMom(16)+tMom(18) ) * div1_4
      state( ?IDX?(10, targetElem, nScalars, nSize )) = (  tMom( 9)+tMom(14)+tMom(16)+tMom(18) ) * div1_4

      state( ?IDX?(11, targetElem, nScalars, nSize )) = (  tMom(10)-tMom(12)-tMom(15)+tMom(19) ) * div1_4
      state( ?IDX?(12, targetElem, nScalars, nSize )) = ( -tMom(10)-tMom(12)+tMom(15)+tMom(19) ) * div1_4
      state( ?IDX?(13, targetElem, nScalars, nSize )) = ( -tMom(10)+tMom(12)-tMom(15)+tMom(19) ) * div1_4
      state( ?IDX?(14, targetElem, nScalars, nSize )) = (  tMom(10)+tMom(12)+tMom(15)+tMom(19) ) * div1_4

      state( ?IDX?(15, targetElem, nScalars, nSize )) = (  tMom( 8)-tMom(11)-tMom(13)+tMom(17) ) * div1_4
      state( ?IDX?(16, targetElem, nScalars, nSize )) = ( -tMom( 8)+tMom(11)-tMom(13)+tMom(17) ) * div1_4
      state( ?IDX?(17, targetElem, nScalars, nSize )) = ( -tMom( 8)-tMom(11)+tMom(13)+tMom(17) ) * div1_4
      state( ?IDX?(18, targetElem, nScalars, nSize )) = (  tMom( 8)+tMom(11)+tMom(13)+tMom(17) ) * div1_4

      state( ?IDX?(19, targetElem, nScalars, nSize )) = tMom(1)-tMom(5)-tMom(6)-tMom(7)+tMom(17)+tMom(18)+tMom(19)
    end do

  end subroutine convert_target_pdf
! ****************************************************************************** !

! ****************************************************************************** !
  subroutine intp_target(nTargets, targetList, depFromCoarser, momBuf, targetMom)
    ! ---------------------------------------------------------------------------
    integer, intent(in) :: nTargets
    integer, intent(in) :: targetList(nTargets)
    type( depSource_type ), intent(in) :: depFromCoarser(:)
    real(kind=rk), intent(in) :: momBuf(:,:)
    real(kind=rk), intent(out) :: targetMom(10,nTargets)
    ! ---------------------------------------------------------------------------
    real(kind=rk) :: srcMom(QQ)
    real(kind=rk) :: cx,cy,cz
    real(kind=rk) :: a(10) ! Coefficients
    real(kind=rk) :: sum1_4, sum2_5, sum3_6, sum_7_10, sum11_14, sum15_18
    real(kind=rk) :: m19_div5_21
    integer :: iMom, indElem, iElem, ii !, s(QQ)
    ! ---------------------------------------------------------------------------


    ! The outer loop - target loop (1->nTargets)
    ! and the inner loop - Moment loop (1->10) is merged together
    do ii = 1, 10 * nTargets

        iMom    = mod(ii-1,10)+1
        indElem = ((ii-1)/10)+1

        ! target position in depFromCoarser
        iElem = targetList( indElem )
        cx = depFromCoarser( iElem )%coord(1)
        cy = depFromCoarser( iElem )%coord(2)
        cz = depFromCoarser( iElem )%coord(3)

        srcMom( 1) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val( 1))
        srcMom( 2) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val( 2))
        srcMom( 3) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val( 3))
        srcMom( 4) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val( 4))
        srcMom( 5) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val( 5))
        srcMom( 6) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val( 6))
        srcMom( 7) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val( 7))
        srcMom( 8) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val( 8))
        srcMom( 9) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val( 9))
        srcMom(10) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val(10))
        srcMom(11) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val(11))
        srcMom(12) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val(12))
        srcMom(13) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val(13))
        srcMom(14) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val(14))
        srcMom(15) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val(15))
        srcMom(16) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val(16))
        srcMom(17) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val(17))
        srcMom(18) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val(18))
        srcMom(19) = momBuf(iMom, depFromCoarser( ielem )%elemBuffer%val(19))

        sum1_4   = srcMom( 1) + srcMom( 4)
        sum2_5   = srcMom( 2) + srcMom( 5)
        sum3_6   = srcMom( 3) + srcMom( 6)
        sum_7_10 = srcMom( 7) + srcMom( 8) + srcMom( 9) + srcMom(10)
        sum11_14 = srcMom(11) + srcMom(12) + srcMom(13) + srcMom(14)
        sum15_18 = srcMom(15) + srcMom(16) + srcMom(17) + srcMom(18)
        m19_div5_21 = div5_21 * srcMom(19)

        a(1) =   div4_21 * ( sum1_4 + sum2_5 + sum3_6 )&
          &    - div1_21 * ( sum_7_10 + sum11_14 + sum15_18 ) + div3_7  * srcMom(19)

        a(2) = 0.1d0 * ( - srcMom( 1) + srcMom( 4) &
          &              - srcMom(11) + srcMom(12) - srcMom(13) + srcMom(14) &
          &              - srcMom(15) - srcMom(16) + srcMom(17) + srcMom(18) )
        a(3) = 0.1d0 * ( - srcMom( 2) + srcMom( 5) &
          &              - srcMom( 7) - srcMom( 8) + srcMom( 9) + srcMom(10) &
          &              - srcMom(15) + srcMom(16) - srcMom(17) + srcMom(18) )
        a(4) = 0.1d0 * ( - srcMom( 3) + srcMom( 6) &
          &              - srcMom( 7) + srcMom( 8) - srcMom( 9) + srcMom(10) &
          &              - srcMom(11) - srcMom(12) + srcMom(13) + srcMom(14) )

        a(5) = div1_42 * sum1_4 - div1_7  * ( sum2_5 + sum3_6 ) &
          &  - div1_21 * sum_7_10 + div5_42 * ( sum11_14 + sum15_18 ) - m19_div5_21

        a(6) = div1_42 * sum2_5 - div1_7  * ( sum1_4 + sum3_6 ) &
          &  - div1_21 * sum11_14 + div5_42 * ( sum_7_10 + sum15_18 ) - m19_div5_21

        a(7) = div1_42 * sum3_6 - div1_7  * ( sum1_4 + sum2_5 ) &
          &  - div1_21 * sum15_18 + div5_42 * ( sum_7_10 + sum11_14 ) - m19_div5_21

        a(8) = div1_4 * ( srcMom(15) - srcMom(16) - srcMom(17) + srcMom(18) )
        a(9) = div1_4 * ( srcMom( 7) - srcMom( 8) - srcMom( 9) + srcMom(10) )
        a(10)= div1_4 * ( srcMom(11) - srcMom(12) - srcMom(13) + srcMom(14) )

        ! Evaluate the bubble function with the above calculated coefficients
        ! m_pi(x) = a1+a2*xcoord+a3*ycoord+a4*xcoord*xcoord+a5*ycoord*ycoord+a6*xcoord*ycoord;
        targetMom(iMom,indElem) =   a( 1)       &
          &                       + a( 2)*cx    &
          &                       + a( 3)*cy    &
          &                       + a( 4)*cz    &
          &                       + a( 5)*cx*cx &
          &                       + a( 6)*cy*cy &
          &                       + a( 7)*cz*cz &
          &                       + a( 8)*cx*cy &
          &                       + a( 9)*cy*cz &
          &                       + a(10)*cz*cx

    end do ! ii=10*nTargets
  end subroutine intp_target
! ****************************************************************************** !

end module mus_interpolate_quadratic_module
! ****************************************************************************** !
