! See copyright notice in the COPYRIGHT file.
! ****************************************************************************** !
!> author: Manuel Hasert
!! Interpolation scheme tools
!!
!! For an overview over implemented interpolation methods, see
!! [Interpolation methods](../page/features/intp_methods.html)
!!
?? include 'header/lbm_macros.inc'
module mus_interpolate_tools_module

  ! include treelm modules
  use env_module,              only: rk, long_k, newUnit, pathLen
  use treelmesh_module,        only: treelmesh_type
  use tem_grow_array_module,   only: grw_intArray_type
  use tem_construction_module, only: tem_levelDesc_type
  use tem_element_module,      only: eT_GhostFromCoarser, &
    &                                eT_ghostFromFiner
  use tem_geometry_module,     only: tem_baryOfId, &
    &                                tem_elemSize
  ! use tem_varSys_module,       only: tem_varSys_type

  ! include musubi modules
  use mus_interpolate_header_module, only: mus_interpolation_type
  use mus_pdf_module,                only: pdf_data_type

  implicit none

  private

  public :: dump_intpLists
  public :: debug_dependencies
  public :: fillMomBuf
  public :: get_fNeqFac_f2c
  public :: get_fNeqFac_c2f

  contains
! ****************************************************************************** !
  !> check the dependencies from Finer
  !!
  subroutine debug_dependencies( intp, leveldesc, tree, rank )
    ! ---------------------------------------------------------------------------
    !> state properties
    type( tem_leveldesc_type ), intent(in) :: levelDesc(:)
    !> interpolation method info
    type( mus_interpolation_type ), intent(in) :: intp
    !> global tree information
    type( treelmesh_type ), intent(in) :: tree
    !> musubi mpi communicator environment
    integer, intent(in) :: rank
    ! ---------------------------------------------------------------------------
    integer :: nUnit
    character(len=17)  :: buffer
    ! ---------------------------------------------------------------------------
    nUnit = newunit()
    write(buffer,'(a,i7.7)') 'debugDep. rank: ', rank
    open(file = trim(buffer), unit= nUnit, recl=1024)

    write(nUnit,*)
    write(nUnit,*) 'GhostFromFiner'
    write(nUnit,*)
    write(nUnit,*) '-----------------------------------------------------------'
    call dump_MyGhostsFromFiner( intp, levelDesc, nUnit, tree )
    write(nUnit,*) '-----------------------------------------------------------'
    write(nUnit,*) '-----------------------------------------------------------'
    write(nUnit,*)
    write(nUnit,*) 'GhostFromCoarser'
    write(nUnit,*)
    write(nUnit,*) '-----------------------------------------------------------'
    call dump_FinerGhostsFromMe( intp, levelDesc, nUnit, tree )
    write(nUnit,*) '-----------------------------------------------------------'
    write(nUnit,*) '-----------------------------------------------------------'
    write(nUnit,*)
    write(nUnit,*) 'GhostFromCoarserBuffer'
    write(nUnit,*)
    write(nUnit,*) '-----------------------------------------------------------'
    call dump_FinerGhostsFromMeBuffer( intp, leveldesc, nUnit, tree )
    write(nUnit,*) '-----------------------------------------------------------'

    close( nUnit )

  end subroutine debug_dependencies
! ****************************************************************************** !


! ****************************************************************************** !
  !> check the dependencies from Finer
  !!
  subroutine dump_intpLists( minLevel, maxLevel, levelDesc, rank )
    ! ---------------------------------------------------------------------------
    !> global pdf information
    integer, intent(in) :: minLevel, maxLevel
    !> state properties
    type( tem_leveldesc_type ), intent(in) :: levelDesc(minLevel:maxLevel)
    !> musubi mpi communicator environment
    integer, intent(in) :: rank
    ! ---------------------------------------------------------------------------
    integer :: nUnit, iLevel
    character(len=17)  :: buffer
    ! ---------------------------------------------------------------------------
    nUnit = newunit()
    write(buffer,'(a4,i7.7)') 'dumpIntpLists.', rank
    open(file = trim(buffer), unit= nUnit, recl=1024)

    do iLevel = minLevel, maxLevel
      write(nUnit,*) 
      write(nUnit,*) 'GhostFromFiner      '
      write(nUnit,*) 
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) 'LOW ---fromFiner-----------------------------------------'
      write(nUnit,*) '---------------------------------------------------------'
      call dump_intpList( eType = eT_ghostFromFiner, &
        &    levelDesc = levelDesc(iLevel), &
        &    ind = levelDesc( iLevel)%intpFromFiner_low, nUnit = nUnit )
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) 'HIGH --fromFiner-----------------------------------------'
      write(nUnit,*) '---------------------------------------------------------'
      call dump_intpList( eType = eT_ghostFromFiner, &
        &    levelDesc = levelDesc(iLevel), &
        &    ind = levelDesc( iLevel)%intpFromFiner_high, nUnit = nUnit )
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) 
      write(nUnit,*) 'GhostFromCoarser    '
      write(nUnit,*) 
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) 'LOW ---fromCoarser---------------------------------------'
      write(nUnit,*) '---------------------------------------------------------'
      call dump_intpList( eType = eT_ghostFromCoarser, &
        &    levelDesc = levelDesc(iLevel), &
        &    ind = levelDesc(iLevel)%intpFromCoarser_low, nUnit = nUnit )
      write(nUnit,*) '---------------------------------------------------------'
      write(nUnit,*) 'HIGH --fromCoarser --------------------------------------'
      write(nUnit,*) '---------------------------------------------------------'
      call dump_intpList( eType = eT_ghostFromCoarser, &
        &    levelDesc = levelDesc(iLevel), &
        &    ind = levelDesc( iLevel)%intpFromCoarser_high, nUnit = nUnit )
      write(nUnit,*) '---------------------------------------------------------'
    end do
    close( nUnit )

  end subroutine dump_intpLists
! ****************************************************************************** !


! ****************************************************************************** !
  !> check the dependencies from Finer and write them out so we can compare 
  !!
  subroutine dump_intpList( eType, levelDesc, ind, nUnit )
    ! ---------------------------------------------------------------------------
    !> state properties
    type( tem_levelDesc_type ), intent(in) :: levelDesc
    !> indirectio list
    type( grw_intArray_type ), intent(in)  :: ind
    !>
    integer, intent(in) :: nUnit
    !>
    integer, intent(in) :: eType
    ! ---------------------------------------------------------------------------
    integer :: iElem          ! current target element (for outer loop)
    integer :: indElem        ! element counter for indirection list
    integer :: targetElem     ! element counter for indirection list
    integer(kind=long_k) :: tID
    character(len=pathLen) :: buffer
    ! ---------------------------------------------------------------------------

    buffer = ''
    do indElem = 1, ind%nVals
      iElem = ind%val( indElem )
      targetElem = iElem + levelDesc%offset( 1, eType )
      tID = levelDesc%total( targetElem )
      write(buffer, '(2a, i9)') trim(buffer), ' ', tID
      if( mod( indElem, 32 ) == 0 .or. indElem == ind%nVals ) then
        write(nUnit, *) trim(buffer)
        buffer = ''
      end if
    enddo

  end subroutine dump_intpList
! ****************************************************************************** !


! ****************************************************************************** !
  !> check the dependencies from Finer and write them out so we can compare 
  !!
  subroutine dump_MyGhostsFromFiner( intp, levelDesc, nUnit, tree)
    ! ---------------------------------------------------------------------------
    !> state properties
    type( tem_leveldesc_type ), intent(in) :: leveldesc(:)
    !> interpolation method info
    type( mus_interpolation_type ), intent(in) :: intp
    !> global tree information
    type( treelmesh_type ), intent(in) :: tree
    !> unit to write to
    integer, intent(in) :: nUnit
    ! --------------------------------------------------------------------------
    integer :: ilevel         ! grid refinement level
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iElem          ! current target element (for outer loop)
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    integer(kind=long_k) :: tID(0:intp%nSourcesFromFiner )
    ! --------------------------------------------------------------------------

    do iLevel = tree%global%minLevel, tree%global%maxLevel - 1

      sourceLevel = iLevel + 1
      targetLevel = iLevel

      ! Treat all coarse target elements
      do iElem = 1, levelDesc( targetLevel )%elem%nElems( eT_ghostFromFiner )
        ! Read the target element treeId
        targetElem = iElem + levelDesc( targetlevel )%offset( 1, eT_ghostFromFiner )
        ! Find out how many fine source elements we have for interpolation.
        ! Usually 8, but can differ at corners, obstacles, boundaries...
        nSourceELems = levelDesc( ilevel )%depFromFiner( iElem )%elem%nVals
        tID = 0
        tID( 0 ) = levelDesc( targetLevel )%total( targetElem )

        write(nUnit, '(a,i9,3f10.5)') ' targetGhost  ', tID( 0 ),              &
          &                           tem_baryOfId( tree, tID(0) )

        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems

          ! Get the source element's treeId
          sourceElem = levelDesc( targetLevel )%depFromFiner( iElem )     &
            &                                       %elem%val( iSourceElem )
          tID( iSourceElem ) = levelDesc( sourceLevel )%total( sourceElem )

          call dump_elemDep( targetElem = tID(0),                              &
            &                sourceElem = tID( iSourceElem ),                  &
            &                nUnit = nUnit,                                    &
            &                tree = tree)
        end do  ! iSourceElem
      enddo
    enddo

  end subroutine dump_MyGhostsFromFiner
! ****************************************************************************** !


! ****************************************************************************** !
  !> check the dependencies from Coarser
  !!
  subroutine dump_FinerGhostsFromMe( intp, leveldesc, nUnit, tree)
    ! ---------------------------------------------------------------------------
    !> state properties
    type( tem_levelDesc_type ), intent(in) :: levelDesc(:)
    !> interpolation method info
    type( mus_interpolation_type ), intent(in) :: intp
    !> global tree information
    type( treelmesh_type ), intent(in) :: tree
    !> unit to write to
    integer, intent(in) :: nUnit
    ! ---------------------------------------------------------------------------
    integer :: ilevel         ! grid refinement level
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iElem          ! current target element (for outer loop)
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    integer(kind=long_k) :: tID(0:intp%nSourcesFromCoarser )
    logical :: weights
    ! ---------------------------------------------------------------------------

    do iLevel = tree%global%minLevel, tree%global%maxLevel - 1

      sourceLevel = ilevel
      targetLevel = ilevel + 1

      ! Treat all coarse target elements
      do iElem = 1, levelDesc( targetLevel )%elem%nElems( eT_ghostFromCoarser )
        ! Read the target element treeId
        targetElem = iElem + levelDesc( targetlevel )%offset( 1, eT_ghostFromCoarser)
        ! Find out how many fine source elements we have for interpolation.
        ! Usually 8, but can differ at corners, obstacles, boundaries...
        nSourceELems = levelDesc( targetlevel )%depFromCoarser( iElem )%  &
          &                                                          elem%nVals
        tID = 0
        tID( 0 ) = levelDesc( targetLevel )%total( targetElem )
        write(nUnit, '(a,i9,3f10.5)') ' targetGhost  ', tID( 0 ),              &
          &                                        tem_baryOfId( tree, tID(0) )

        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems

          ! Get the source element's treeId
          sourceElem = levelDesc( targetLevel )%depFromCoarser( iElem )   &
            &                                       %elem%val( iSourceElem )
          tID( iSourceElem ) = levelDesc( sourceLevel )%total( sourceElem )

          if( allocated( levelDesc( targetLevel )                         &
              &       %depFromCoarser( iElem )%weight)) then
            weights = .true.
          else
            weights = .false.
          end if
          if( weights ) then
            call dump_elemDep( targetElem = tID(0),                            &
              &                sourceElem = tID( iSourceElem ),                &
              &                nUnit = nUnit,                                  &
              &                tree = tree,                                    &
              &                weight = levelDesc( targetLevel )          &
              &                   %depFromCoarser( iElem )%weight( iSourceElem))
          else
            call dump_elemDep( targetElem = tID(0),                            &
              &                sourceElem = tID( iSourceElem ),                &
              &                nUnit = nUnit,                                  &
              &                tree = tree )
          end if

        end do ! iSourceElem
      end do ! iElem
    end do ! iLevel

  end subroutine dump_FinerGhostsFromMe
! ****************************************************************************** !


! ****************************************************************************** !
  !> check the dependencies from Coarser
  !!
  subroutine dump_FinerGhostsFromMeBuffer( intp, levelDesc, nUnit, tree )
    ! ---------------------------------------------------------------------------
    !> state properties
    type( tem_leveldesc_type ), intent(in) :: leveldesc(:)
    !> interpolation method info
    type( mus_interpolation_type ), intent(in) :: intp
    !> global tree information
    type( treelmesh_type ), intent(in) :: tree
    !> unit to write to
    integer, intent(in) :: nUnit
    ! ---------------------------------------------------------------------------
    integer :: ilevel         ! grid refinement level
    integer :: sourceLevel    ! level of source elements
    integer :: sourceElem     ! treeId of current source element
    integer :: targetLevel    ! level of target elements
    integer :: targetElem     ! treeId of current source element
    integer :: iElem          ! current target element (for outer loop)
    integer :: iSourceElem    ! current source element (for inner loop)
    integer :: nSourceElems   ! number of source elements for the current target
    integer(kind=long_k) :: tID(0:intp%nSourcesFromCoarser )
    integer :: totalPos ! Position of the element in the total list
    logical :: weights
    ! ---------------------------------------------------------------------------
    do iLevel = tree%global%minLevel, tree%global%maxLevel - 1
      sourceLevel = ilevel
      targetLevel = ilevel + 1

      ! Treat all coarse target elements
      do iElem = 1, levelDesc( targetLevel )%elem%nElems( eT_ghostFromCoarser )
        ! Read the target element treeId
        targetElem = iElem + levelDesc( targetlevel )                     &
          &                      %offset( 1, eT_ghostFromCoarser)
        ! Find out how many fine source elements we have for interpolation.
        ! Usually 8, but can differ at corners, obstacles, boundaries...
        nSourceELems = levelDesc( targetlevel )%depFromCoarser( iElem )%  &
          &                                                          elem%nVals
        tID = 0
        tID( 0 ) = levelDesc( targetLevel )%total( targetElem )
        write(nUnit, '(a,i9,3f10.5)') ' targetGhost  ', tID( 0 ),              &
          &                           tem_baryOfId( tree, tID(0) )

        ! Now loop over all fine source elements for this target:
        do iSourceElem = 1, nSourceElems

          ! Get the source element's treeId
          sourceElem = levelDesc( targetLevel )%depFromCoarser( iElem )   &
            &                                    %elemBuffer%val( iSourceElem )
          totalPos = levelDesc( targetLevel )%sourceFromCoarser%          &
            &                                                 val( sourceElem )
          tID( iSourceElem ) = levelDesc( sourceLevel )%total( totalPos )

          if( allocated( levelDesc( targetLevel )                         &
              &       %depFromCoarser( iElem )%weight)) then
              weights = .true.
          else
            weights = .false.
          end if
          if( weights ) then
            call dump_elemDep( targetElem = tID(0),                            &
              &                sourceElem = tID( iSourceElem ),                &
              &                nUnit = nUnit,                                  &
              &                tree = tree,                                    &
              &                weight = levelDesc( targetLevel )          &
              &                   %depFromCoarser( iElem )%weight( iSourceElem))
          else
            call dump_elemDep( targetElem = tID(0),                            &
              &                sourceElem = tID( iSourceElem ),                &
              &                nUnit = nUnit,                                  &
              &                tree = tree )
          end if

        end do  ! iSourceElem
      enddo
    enddo

  end subroutine dump_FinerGhostsFromMeBuffer
! ****************************************************************************** !


! ****************************************************************************** !
  !> dump dependencies for one element
  !!
  subroutine dump_elemDep( targetElem, sourceElem, nUnit, tree, weight )
    ! ---------------------------------------------------------------------------
    !>
    integer(kind=long_k), intent(in) :: targetElem
    !>
    integer(kind=long_k), intent(in) :: sourceElem
    !>
    integer, intent(in) :: nUnit
    !>
    type( treelmesh_type ), intent(in) :: tree
    !>
    real(kind=rk), optional :: weight
    ! ---------------------------------------------------------------------------
    real(kind=rk) :: xTarget(3), xSource(3)
    real(kind=rk) :: elemSize
    character(len=pathLen) :: buffer
    ! ---------------------------------------------------------------------------
    buffer = ''
    elemSize = tem_elemSize( tree, targetElem )
    xTarget = tem_baryOfId( tree, targetElem )
    xSource = tem_baryOfId( tree, sourceElem )
    write(buffer,'(a,i9, 6f10.5)') '  ', sourceElem,                           &
      &          ((xSource(1:3) - xTarget(1:3))/elemSize), xSource(1:3)
    if( present( weight )) then
      write(buffer, '(a,a,f8.5)') trim( buffer), '    weight: ', weight
    end if
    write(nUnit, '(a)') trim(buffer)

  end subroutine dump_elemDep
! ****************************************************************************** !


! ****************************************************************************** !
  !> Convert PDF of a list of elements to moment, save to buf
  !! @todo: use IDX instead of SAVE macro?
  pure function fillMomBuf( elemList, nElems, state, QQ,  nSize, A) result ( buf )
    ! ---------------------------------------------------------------------------
    integer,       intent(in) :: nElems !< number of source elements
    integer,       intent(in) :: elemList( nElems ) !< elements list
    real(kind=rk), intent(in) :: state(:) !< state array
    integer,       intent(in) :: QQ !< nComponents
    integer,       intent(in) :: nSize
    real(kind=rk), intent(in) :: A(QQ,QQ) !< moments matrix
    real(kind=rk) :: buf(QQ,nElems)
    ! ---------------------------------------------------------------------------
    integer :: iDir, iElem, elemPos
    real(kind=rk) :: PDF(QQ)
    ! ---------------------------------------------------------------------------

    do iElem = 1, nElems
      ! Get the source element's position in the state vector
      elemPos = elemList( iElem )

      ! transfer PDF
      do iDir = 1, QQ
        PDF( iDir ) = state( ?idx?( iDir,elemPos,QQ,nSize))
      end do ! iDir

      ! convert to moment
      buf(:,iElem) = matmul( A, PDF )

    end do ! iElem

  end function fillMomBuf
! ****************************************************************************** !

! ****************************************************************************** !
  !> Calculate fNeq scale factor from fine level to(2) coarse level.
  !! Such scale make sure continuuity of strain rate and shear stress over
  !! levels.
  !! To achive equal strain rate, we want:
  !!   fNeq_c * omega_c / dt_c = fNeq_f * omega_f / dt_f
  !! rearrange this, we get:
  !!   fNeq_c = fNeq_f * ( omega_f / omega_c ) * ( dt_c / dt_f )
  !! dt_c/dt_f is strain rate scale factor (i.e. sFac)
  !!
  !! This function can be used converting fNeq_f to fNeq_c, such as fill coarse
  !! from finer interpolation routines.
  !! How to use this function:
  !!  sFac   = physics%sfac( fineLevel, coarseLevel )
  !!  fNeq_c = fNeq_f * fNeqFac( omega_f, omega_c, sFac )
  !!
  pure function get_fNeqFac_f2c( omegaF, omegaC, sFac ) result ( fac )
    ! ---------------------------------------------------------------------------
    !> omega on fine and coarse level
    real(kind=rk), intent(in) :: omegaF, omegaC
    !> strain rate scaling factor that from fine to coarse
    real(kind=rk), intent(in) :: sFac
    ! ---------------------------------------------------------------------------
    real(kind=rk) :: fac
    ! ---------------------------------------------------------------------------

    fac = omegaF / omegaC * sFac

  end function get_fNeqFac_f2c
! ****************************************************************************** !

! ****************************************************************************** !
  !> Calculate fNeq scale factor from coarse level to(2) fine level.
  !! Such scale make sure continuuity of strain rate and shear stress over
  !! levels.
  !! To achive equal strain rate, we want:
  !!   fNeq_c * omega_c / dt_c = fNeq_f * omega_f / dt_f
  !! rearrange this, we get:
  !!   fNeq_f = fNeq_c * ( omega_c / omega_f ) * ( dt_f / dt_c )
  !! dt_f/dt_c is strain rate from coarse to fine scale factor (i.e. sFac)
  !!
  !! This function can be used converting fNeq_c to fNeq_f,
  !! such as fill fine from coarse interpolation routines.
  !! How to use this function:
  !!  sFac   = physics%sfac( coarseLevel, fineLevel )
  !!  fNeq_f = fNeq_c * fNeqFac( omega_c, omega_f, sFac )
  !!
  pure function get_fNeqFac_c2f( omegaC, omegaF, sFac ) result ( fac )
    ! ---------------------------------------------------------------------------
    !> omega on fine and coarse level
    real(kind=rk), intent(in) :: omegaF, omegaC
    !> strain rate scaling factor that from fine to coarse
    real(kind=rk), intent(in) :: sFac
    ! ---------------------------------------------------------------------------
    real(kind=rk) :: fac
    ! ---------------------------------------------------------------------------

    fac = omegaC / omegaF * sFac

  end function get_fNeqFac_c2f
! ****************************************************************************** !

end module mus_interpolate_tools_module
! ****************************************************************************** !
