! See copyright notice in the COPYRIGHT file.
!! ****************************************************************************** !
!> \ref mus_construct "Creation of the data structures"
!!
?? include 'header/lbm_macros.inc'
?? include 'treelm/source/logMacros.inc'
module mus_construction_module

  ! include treelm modules
  use mpi
  use env_module,              only: rk, long_k, eps, labelLen, my_status_int
  use treelmesh_module,        only: treelmesh_type
  use tem_comm_module,         only: tem_communication_type,                   &
    &                                tem_commpattern_type
  use tem_param_module,        only: childPosition, tem_rotationMatrix, PI
  use tem_topology_module,     only: tem_LevelOf, tem_ChildNumber,             &
    &                                tem_CoordOfId
  use tem_geometry_module,     only: tem_determine_discreteVector,             &
    &                                tem_ElemSizeLevel
  use tem_tools_module,        only: tem_horizontalSpacer
  use tem_grow_array_module,   only: init, append, destroy, empty, truncate
  use tem_dyn_array_module,    only: init, append, expand,  &
    &                                destroy, empty, PositionOfVal, truncate,  &
    &                                dyn_labelArray_type
  use tem_property_module,     only: tem_property_type, prp_solid,             &
    &                                prp_hasQVal, prp_sendHalo
  use tem_bc_prop_module,      only: tem_bc_prop_type, tem_debug_bc_prop
  use tem_construction_module, only: tem_init_elemLevels, tem_find_allElements,&
    &                                tem_build_verticalDependencies,           &
    &                                tem_build_horizontalDependencies,         &
    &                                tem_debug_horizontalDependencies,         &
    &                                tem_levelDesc_type,                       &
    &                                depSource_type,                           &
    &                                depSource_append,                         &
    &                                tem_dumpTreeIDLists
  use tem_element_module,      only: eT_halo, eT_ghostFromCoarser,     &
    &                                eT_minRelevant, eT_maxRelevant,             &
    &                                eT_ghostFromFiner, eT_fluid, destroy
  use tem_timer_module,        only: tem_startTimer, tem_stopTimer,            &
    &                                tem_getTimerVal
  use tem_stencil_module,      only: tem_stencilHeader_type,                   &
    &                                tem_stencil_getLabelForcxDir,             &
    &                                tem_create_d3q81, tem_create_d3q125,      &
    &                                grw_stencilHeaderArray_type, append,      &
    &                                destroy, truncate
  use tem_debug_module,        only: main_debug, dbgUnit
  use tem_aux_module,          only: tem_abort
  use tem_logging_module,      only: logUnit, tem_toStr
  use tem_subTree_module,      only: tem_write_debugMesh
  use tem_comm_env_module,     only: tem_comm_env_type

  ! include musubi modules
  use mus_bc_header_module,          only: boundary_type, glob_boundary_type, &
      &                                      mus_allocate_bc_elems, &
      &                                      mus_set_posInNghElems, &
      &                                      mus_alloc_fieldBC, &
      &                                      debug_glob_boundary_type
  use mus_bc_general_module,         only: set_bouzidi_coeff
  use mus_param_module,              only: mus_param_type
  use mus_geom_module,               only: mus_geom_type
  use mus_scheme_layout_module,      only: mus_scheme_layout_type,             &
    &                                      mus_finalize_layout
  use mus_scheme_type_module,        only: mus_scheme_type
  use mus_pdf_module,                only: pdf_data_type,   &
    &                                      mus_calc_nElems, mus_pdf_allocate,  &
    &                                      allocate_momBuf
  use mus_statistics_module,         only: mus_statistics_type,                &
    &                                      mus_calc_commAmount
?? IF( .not. SOA) then
    use mus_debug_module,              only: debug_normals, &
      &                                      debug_dependenciesFromFiner,        &
      &                                      debug_dependenciesFromCoarser,      &
      &                                      debug_connectivity
?? ENDIF
  use mus_comm_module,               only: mus_init_longBuffers
  use mus_field_module,              only: mus_field_type, remove_solid_in_bc, &
    &                                      mus_check_allWall
  use mus_IBM_module,                only: mus_IBM_globType
  use mus_interpolate_header_module, only: compactNeighDir
  use mus_timer_module,              only: mus_timerHandles

  implicit none

  private

  public :: mus_construct
  public :: communicate_property
  public :: find_fromCoarserDep_compact
  public :: mus_intp_complete_fromCoarserDep
  public :: mus_update_fromFinerDep_2D
  public :: mus_update_fromFinerDep_3D
  public :: calculate_nElems
  public :: construct_connectivity
  public :: set_sendHaloBits

  !> Contains bitmask to reduce halo elements communciation links.
  !! This type is used only in init_levelBuffers
  type halo_commBitmask_type
    !> Bitmask of which direction to send,
    !! only directions of fluid neighbor need to be update
    !! Size: nDofs, nElems
    logical, allocatable :: bitmask(:,:)
  end type halo_commBitmask_type

  type logical_array_type
    logical, allocatable :: val(:)
  end type logical_array_type

contains

! ****************************************************************************** !
  !> Initialize Musubi data strucutres based on data provided by Treelm
  !!
  !! Load the mesh and boundary conditions for this process from disk.
  !! Get the level-wise treeID lists and create the required ghost and halo
  !! elements.
  !!
  !! This is achieved by a two-folded identification of elements.
  !!
  !! - the theoretically required elements are collected based on
  !! [[mus_scheme_layout_module:mus_scheme_layout_type]] "stencil information"
  !! The [[tem_construction_module:tem_findneighbor]] "find neighbor routine"
  !! performs this task for compute fluid elements.
  !! For boundaries which require information from neighbor elements, these
  !! required [[tem_topology_module]] "treeIDs" are collected into the
  !! [[mus_bc_header_module]] "boundary element type"
  !! 
  !! - All required elements are created in the
  !! [[tem_construction_module:tem_createleveldescriptor]]
  !! "Level Descriptor creation routine"
  !!
  !! # Additional Tasks
  !! 
  !! - receive [[tem_construction_module:tem_buildhorizontaldependencies]]
  !! "horizontal"
  !! (within a level for the element updates)
  !! - and [[tem_construction:tem_buildverticaldependencies]] "vertical"
  !! dependencies (between levels for ghost-interpolations).
  !! - The main state vector and the neighbor lists on which the kernel then
  !!    acts is created
  !! - The MPI buffers are created.
  !! - For each [[mus_scheme_module]] "Scheme", the
  !! [[tem_construction_module:level_descriptor_type]] "Level Descriptor"
  !! is created
  !!
  !! # Result
  !!
  !! After this routine, all data structures for starting the main loop of the
  !! solver are allocated and ready.
  !!
  subroutine mus_construct( scheme, geometry, params)
    ! ---------------------------------------------------------------------------
    !> scheme information including fluid, boundary and flow information
    type( mus_scheme_type ), intent(inout) :: scheme
    !> geometric information
    type( mus_geom_type ), intent(inout) :: geometry
    !> run-time Parameters
    type( mus_param_type ), intent(in) :: params

    ! ---------------------------------------------------------------------------
    integer :: iLevel      ! counter for level
    integer :: iStencil    ! lists with different stencils
    integer :: iBC         ! counter for different boundary conditions
    integer :: iField      ! counter for different fields
    logical :: requireAll_loc ! are all links required
    integer :: minLevel, maxLevel
    integer :: ii
    type( mus_statistics_type ) :: stat
    !> mark halo elements which communicate all its links
    type( logical_array_type ), allocatable :: haloRequired(:)
    ! ---------------------------------------------------------------------------
    minLevel = geometry%tree%global%minLevel
    maxLevel = geometry%tree%global%maxLevel

    call tem_horizontalSpacer( fUnit = dbgUnit(1), before = 1 )
    write(dbgUnit(1),*) 'Get into routine: mus_construction'

    call tem_startTimer( timerHandle = mus_timerHandles%initLvlD )


    write(logUnit(1),*) 'Starting to initialize the geometry'
    call tem_horizontalSpacer(fUnit = logUnit(1))

    ! 1. determine the number of elements for each mesh boundary
    ! 2. copy the information to globbc
    ! 3. determine the normal vector and bitmasks for the individual
    !    boundaries
    ! 4. allocate the BC lists in globbc
    ! THIS ROUTINE DOES NOT MODIFY THE BC LISTS IN THE FIELDS!!!
    call build_BClists( globBC    = scheme%globBC,              &
      &                 tree      = geometry%tree,              &
      &                 bc_prop   = geometry%boundary,          &
      &                 minLevel  = minLevel,                   &
      &                 maxLevel  = maxLevel,                   &
      &                 layout    = scheme%layout,              &
      &                 field     = scheme%field(:),            &
      &                 comm      = params%general%proc%comm    )

    ! Treat the boundary lists: in addition to the normal stencil treatment
    ! one has to find all those elements along the element list in the stencil
    ! type, which are boundary elements and give them the special
    ! boundary treatment

    ! 1. define the stencil for the individual boundaries (in the fields)
    !    by setting useAll, requireNeighNeigh, QQ, QQN, nElems, cxDir
    ! 2. fill stencil%elem with the level wise element positions and
    !    stencil%elemLvl with the level wise positions in the primal mesh
    do iField = 1, scheme%nFields
      do iBC = 1, geometry%boundary%nBCtypes
        call mus_alloc_fieldBC( scheme%field( iField )%bc( iBC ),&
          &                     minLevel, maxLevel)
        call mus_build_BCStencils(                                  &
          &      globbc         = scheme%globbc( iBC ),             &
          &      bc             = scheme%field( iField )%bc( iBC ), &
          &      prevailDir     = scheme%layout%prevailDir,         &
          &      prefix         = scheme%field( iField )%label,     &
          &      stencil_labels = scheme%layout%stencil_labels,     &
          &      grwStencil     = scheme%layout%grwStencil,         &
          &      minLevel       = minLevel,                         &
          &      maxLevel       = maxLevel                          )
      enddo

      ! Build and append IBM stencils to stencil array
      call mus_build_IBMStencils(                       &
        &  globIBM    = scheme%field( iField )%globIBM, &
        &  layout     = scheme%layout,                  &
        &  grwStencil = scheme%layout%grwStencil        )
    enddo

    ! finalize scheme layout
    ! copy growing array of stencil to allocatable array of stencil
    ! and destroy growing array
    call mus_finalize_layout( layout       = scheme%layout,        &
      &                       nElemsInTree = geometry%tree%nElems, &
      &                       minLevel     = minLevel,             &
      &                       maxLevel     = maxLevel,             &
      &                       proc         = params%general%proc   )
    ! peakval = my_status_int('VmPeak:')
    ! write(logUnit(10),"(A,I0)") 'After finalize layout, VmPeak: ', peakval

    ! 1. store treeIDs which use this stencil in the
    !    levelDesc( elemLevel )%elem
    ! 2. store the neighbors which can be loaded from the disc or accessed
    !    directly
    ! THIS MEANS THAT FOR ALL ELEMENTS (INCL. BOUNDARY ELEMENTS) NEIGHBORS
    ! ARE STORED BUT THEY DO NOT NEED TO EXIST IN THE MESH AT THIS POINT!!!
    call tem_init_elemLevels( me       = scheme%levelDesc,   &
      &                       boundary = geometry%boundary,       &
      &                       tree     = geometry%tree,           &
      &                       stencils = scheme%layout%stencil(:) )
    ! peakval = my_status_int('VmPeak:')
    ! write(logUnit(10),"(A,I0)") 'After init elemLevels, VmPeak: ', peakval

    ! 1. build all dependencies for fluid, halos and ghosts
    ! 2. build the pointers for each element to its neighbors/stencil elements
    ! All this information is stored in tem_levelDesc_type.
    !
    ! THIS MEANS THAT FOR ALL ELEMENTS FOUND IN em_init_elemLevels THE CORRECT
    ! DEPENDENCIES AND NEIGHBORS ARE SET
    !
    ! This is the heart of the topologic data structure creation and
    ! takes most of the compute time, depending on the mesh you have
    !
    write(logUnit(1),*) 'Creating level descriptor ...'
    write(logUnit(6),*) 'before tem_find_allElements'
    write(dbgUnit(6),*) 'before find_allElements'
    ! The level descriptor is created from the below routine.
    ! the neigh array is created using the LD and communication buffers are filled up. 
    call tem_find_allElements(                                         &
      &                  tree            = geometry%tree,              &
      &                  levelDesc       = scheme%levelDesc,           &
      &                  levelPointer    = geometry%levelPointer,      &
      &                  computeStencil  = scheme%layout%stencil(:),   &
      &                  commPattern     = params%general%commPattern, &
      &                  cleanup         = .true.,                     &
      &                  reqNesting      = params%nNesting,            &
      &                  proc            = params%general%proc         )
    ! peakval = my_status_int('VmPeak:')
    ! write(logUnit(10),"(A,I0)") 'After find allElements, VmPeak: ', peakval

?? IF( .not. SOA) then
      if ( main_debug%dumpTreeIDlists ) then
        call tem_dumpTreeIDlists( minLevel, maxLevel, &
          &                       scheme%levelDesc )
      end if
?? ENDIF

    write(dbgUnit(6),*) 'before horizontal dep'
    write(logUnit(6),*) 'before tem_build_horizontalDependencies'
    do iStencil = 1, scheme%layout%nStencils
      call tem_build_horizontalDependencies(                      &
        &       iStencil       = iStencil,                        &
        &       levelDesc      = scheme%levelDesc,           &
        &       tree           = geometry%tree,                   &
        &       computeStencil = scheme%layout%stencil(iStencil)  )
?? IF( .not. SOA) then
      if( main_debug%checkDependencies ) then
        call tem_debug_HorizontalDependencies( iStencil, &
          &         scheme%levelDesc, geometry%tree, &
          &         scheme%layout%stencil(iStencil) )
      end if
?? ENDIF
    end do
    ! peakval = my_status_int('VmPeak:')
    ! write(logUnit(10),"(A,I0)") 'After build horizontal, VmPeak: ', peakval

    ! Communicate property after setting missing Neighbor in
    ! tem_build_horizontalDependencies routine

    write(logUnit(6),*) 'before communicate_property'
    do iLevel = minLevel, maxLevel
      call communicate_property(                                  &
        &   send     = scheme%levelDesc( iLevel )%sendbuffer,&
        &   recv     = scheme%levelDesc( iLevel )%recvbuffer,&
        &   property = scheme%levelDesc( iLevel )%property,  &
        &   flag     = iLevel,                                    &
        &   proc     = params%general%proc,                       &
        &   pattern  = params%general%commPattern                 )
      call communicate_property(                                  &
        &   send     = scheme%levelDesc( iLevel )%sendbufferFromCoarser,  &
        &   recv     = scheme%levelDesc( iLevel )%recvbufferFromCoarser,  &
        &   property = scheme%levelDesc( iLevel )%property,       &
        &   flag     = iLevel,                                    &
        &   proc     = params%general%proc,                       &
        &   pattern  = params%general%commPattern                 )
      call communicate_property(                                  &
        &   send     = scheme%levelDesc( iLevel )%sendbufferFromFiner,    &
        &   recv     = scheme%levelDesc( iLevel )%recvbufferFromFiner,    &
        &   property = scheme%levelDesc( iLevel )%property,       &
        &   flag     = iLevel,                                    &
        &   proc     = params%general%proc,                       &
        &   pattern  = params%general%commPattern                 )
    end do
    ! peakval = my_status_int('VmPeak:')
    ! write(logUnit(10),"(A,I0)") 'After comm property, VmPeak: ', peakval

    call tem_horizontalSpacer( after = 1, fUnit = logUnit(1) )

    ! Update the positions stored in the bc%elem lists to the positions in the
    ! levelwise total list
    ! 1. allocate and initialize the pre- and post collision buffers as well
    !    as the bcBuffer for the boundary elements
    ! 2. Update the positions stored in the bc%elems lists to the positions
    !    in the levelwise total list as well as the positions of the neighbors
    !    in bc%elems%neigh
    ! IN THIS ROUTINE THE ALREADY EXISTING NEIGHBORS ARE COPIED AND THE ONES
    ! WHICH DO NOT EXIST ARE SET TO REASONABLE ELEMENTS (CURRENT ELEMENT OR
    ! LAST VALID NEIGHBOR)
?? IF( .not. SOA) then
    if( main_debug%dumpBoundaries ) then
      write(dbgUnit(1),*) 'Check BC before update_BCLists, i.e. elemPos ', &
        &                 ' with respect to tree'
      call debug_glob_boundary_type( geometry%boundary%nBCtypes, minLevel, maxLevel, &
        &                            scheme%globBC )
    end if
?? ENDIF
    call update_BClists( scheme       = scheme,                 &
      &                  minLevel     = minLevel,               &
      &                  maxLevel     = maxLevel,               &
      &                  nBCs         = geometry%boundary%nBCtypes, &
      &                  LP           = geometry%levelPointer, &
      &                  remove_solid = params%remove_solid     )

    ! Get vertical dependencies (between levels for interpolation)
    call tem_build_verticalDependencies(          &
      &        levelDesc = scheme%levelDesc, &
      &        minlevel  = minLevel,              &
      &        maxlevel  = maxLevel               )
    ! peakval = my_status_int('VmPeak:')
    ! write(logUnit(10),"(A,I0)") 'After build vertical, VmPeak: ', peakval

    ! Only if multilevel
    if ( minLevel /= maxLevel ) then

      ! initialize the levelwise ghost and source list
      do iLevel = minLevel, maxLevel
        ! fromCoarser
        call init( me = scheme%levelDesc(iLevel)%sourceFromCoarser, &
          & length = scheme%levelDesc(iLevel)%elem%nElems( eT_ghostFromCoarser )   )

        call init( me = scheme%levelDesc(iLevel)%intpFromCoarser_high, &
          & length = scheme%levelDesc(iLevel)%elem%nElems( eT_ghostFromCoarser )      )

        call init( me = scheme%levelDesc(iLevel)%intpFromCoarser_low )

        ! fromFiner
        call init( me = scheme%levelDesc(iLevel)%sourceFromFiner, &
          & length = scheme%levelDesc(iLevel)%elem%nElems( eT_ghostFromFiner ) &
          &          * scheme%intp%nSourcesFromFiner                )

        call init( me  = scheme%levelDesc(iLevel)%intpFromFiner_high, &
          &        length = scheme%levelDesc(iLevel)%elem%nElems(eT_ghostFromFiner))

        call init( me = scheme%levelDesc(iLevel)%intpFromFiner_low )
      end do

      ! Add additional source elements for fromCoarser ghost elements
      if ( scheme%intp%useCompact ) then
        call find_fromCoarserDep_compact( &
          &  levelDesc  = scheme%levelDesc(minLevel:maxLevel),  &
          &  nSourcesFromCoarser = scheme%intp%nSourcesFromCoarser, &
          &  stencil             = scheme%layout%fStencil,     &
          &  minLevel            = minLevel,                   &
          &  maxLevel            = maxLevel                    )
      else
        call mus_intp_complete_fromCoarserDep(                             &
          &  levelDesc  = scheme%levelDesc(minLevel:maxLevel),   &
          &  nSourcesFromCoarser        = scheme%intp%nSourcesFromCoarser, &
          &  nSourcesCompactFromCoarser = scheme%intp                      &
          &                               %nSourcesCompactFromCoarser,     &
          &  stencil = scheme%layout%intpStencil,       &
          &  minLevel = minLevel,                        &
          &  maxLevel = maxLevel                         )
      end if

      ! ------------------------------------------------------------------------
      ! Update the dependencies fromFiner (in 2d, 4 of 8 get removed)
      ! @todo: wrap it as a single routine
      if ( scheme%layout%fStencil%nDims == 2 ) then
        call mus_update_fromFinerDep_2D(                            &
          &    levelDesc = scheme%levelDesc(minLevel:maxLevel), &
          &    nSourcesFromFiner = scheme%intp%nSourcesFromFiner, &
          &    minLevel  = minLevel, &
          &    maxLevel  = maxLevel  )
      else
        call mus_update_fromFinerDep_3D(                            &
          &    levelDesc = scheme%levelDesc(minLevel:maxLevel), &
          &    nSourcesFromFiner = scheme%intp%nSourcesFromFiner, &
          ! &    intpCompact = scheme%intp%useCompact,    &
          &    minLevel  = minLevel, &
          &    maxLevel  = maxLevel  )
      end if

      do iLevel = minLevel, maxLevel
        call truncate( me = scheme%levelDesc(iLevel)%sourceFromCoarser    )
        call truncate( me = scheme%levelDesc(iLevel)%intpFromCoarser_low  )
        call truncate( me = scheme%levelDesc(iLevel)%intpFromCoarser_high )

        call truncate( me = scheme%levelDesc(iLevel)%sourceFromFiner    )
        call truncate( me = scheme%levelDesc(iLevel)%intpFromFiner_low  )
        call truncate( me = scheme%levelDesc(iLevel)%intpFromFiner_high )

        write(logUnit(3),"(A,I0)") '                       level: ', iLevel
        write(logUnit(3),"(A,I0)") ' high order ghostFromCoarser: ', &
          &      scheme%levelDesc( iLevel )%intpFromCoarser_high%nVals
        write(logUnit(3),"(A,I0)") ' low  order ghostFromCoarser: ', &
          &      scheme%levelDesc( iLevel )%intpFromCoarser_low%nVals
        write(logUnit(3),"(A,I0)") '     total sourceFromCoarser: ',      &
          &      scheme%levelDesc( iLevel )%sourceFromCoarser%nVals
        write(logUnit(3),"(A,I0)") ' high order   ghostFromFiner: ', &
          &      scheme%levelDesc( iLevel )%intpFromFiner_high%nVals
        write(logUnit(3),"(A,I0)") ' low  order   ghostFromFiner: ', &
          &      scheme%levelDesc( iLevel )%intpFromFiner_low%nVals
        write(logUnit(3),"(A,I0)") '     total   sourceFromFiner: ',      &
          &      scheme%levelDesc( iLevel )%sourceFromFiner%nVals

      end do ! iLevel = minLevel, maxLevel
      ! ------------------------------------------------------------------------

    end if ! minLevel /= maxLevel
    ! peakval = my_status_int('VmPeak:')
    ! write(logUnit(10),"(A,I0)") 'After update interpolation, VmPeak: ', peakval

    call calculate_nElems( levelDesc = scheme%leveldesc,   &
      &                    proc      = params%general%proc,      &
      &                    minLevel  = minLevel, &
      &                    maxLevel  = maxLevel  )

    ! allocate here since it needs to be deallocated by dynamic load
    ! balancing algorithm
    write(logUnit(6),"(A,I0,A,I0)") 'allocate PDF from Level ', minLevel, &
      &                             ' to ', maxLevel
    allocate( scheme%pdf( minLevel:maxLevel ) )
    allocate( scheme%state( minLevel:maxLevel ) )

    write(logUnit(4),*) 'Allocating PDF state, neighbor and bcBuffer array'
    ! Allocate the PDF state array
    do iLevel = minLevel, maxLevel
      write(dbgUnit(1),*) 'calculate nElems on level: ', iLevel
      call mus_calc_nElems(                                              &
        & me                = scheme%pdf( iLevel ),                      &
        & nFluids           = scheme%levelDesc( iLevel )%elem            &
        &                                %nElems( eT_fluid ),            &
        & nGhostFromCoarser = scheme%levelDesc( iLevel )%elem            &
        &                                %nElems( eT_ghostFromCoarser ), &
        & nGhostFromFiner   = scheme%levelDesc( iLevel )%elem            &
        &                                %nElems( eT_ghostFromFiner ),   &
        & nHalos            = scheme%levelDesc( iLevel )%elem            &
        &                                %nElems( eT_halo )              )

      allocate(scheme%state(iLevel)%val( scheme%pdf(iLevel)%nSize * scheme%varSys%nScalars, 2 ) )
      !For debugging purposes, set complete flow field to invalid
      do ii = 1, scheme%pdf(iLevel)%nSize * scheme%varSys%nScalars
        scheme%state(iLevel)%val(ii, 1) = -1000000.0_rk
        scheme%state(iLevel)%val(ii, 2) = -1000000.0_rk
      end do

      call mus_pdf_allocate(                                           &
        & me              = scheme%pdf( iLevel ),                      &
        & nScalars        = scheme%varSys%nScalars,                    &
        & QQ              = scheme%layout%fStencil%QQ,                 &
        & isPDF           = .true.,                                    &
        & nElems_bcBuffer = scheme%levelDesc( iLevel )                 &
        &                                         %bc_elemBuffer%nVals )

      call allocate_momBuf( scheme%pdf(iLevel),                        &
        &    max( scheme%levelDesc(iLevel)%sourceFromCoarser%nVals,    &
        &         scheme%levelDesc(iLevel)%sourceFromFiner%nVals    )  )

    end do
    ! peakval = my_status_int('VmPeak:')
    ! write(logUnit(10),"(A,I0)") 'After allocate PDF, VmPeak: ', peakval

    write(dbgUnit(6),*) 'clean up elem list in level desctiptor'
    write(logUnit(6),*) 'clean up elem list in level descriptor'
    ! KJ: @todo de-activate for adaptive grid refinement
    do iLevel = minLevel, maxLevel
      call destroy( me = scheme%levelDesc( iLevel )%elem )
    end do

    write(dbgUnit(6),*) 'after clean up'

    ! reconstruct connectivity vector pdf( iLevel )%neigh including
    ! bounce back rules
    call construct_connectivity( scheme   = scheme,                        &
      &                          minLevel = minLevel,                      &
      &                          maxLevel = maxLevel                       )
?? IF( .not. SOA) then
    ! Dump state neighbor array information to debugUnit
    if( main_debug%dumpAuxLists ) then
      call debug_connectivity( scheme   = scheme,   &
        &                      minLevel = minLevel, &
        &                      maxLevel = maxLevel  )
    end if
?? ENDIF

    ! Set halo elements which require all links for communication.
    call set_halo_commLinks( scheme, minLevel, maxLevel, &
      &                      geometry%boundary%nBCtypes, &
      &                      params%comm_reduced,        &
      &                      haloRequired )

    write(dbgUnit(6),*)  'before init levelBuffers'

    ! Communicate all links if any stencil has requireAll true or 
    ! comm_reduced is false
    requireAll_loc = any( scheme%layout%stencil(:)%requireAll ) &
      &             .or. .not. params%comm_reduced
    do iLevel = minLevel, maxLevel
      !    for multi-level, still all the links have to be communicated
      !    so, check for regular Buffer, if the max and min level is the same
      write(dbgUnit(1),*) "Level: ", iLevel
      write(dbgUnit(1),*) "Set send and recv buffer for normal Halo elements"
      call init_levelBuffers(                                 &
        & send       = scheme%levelDesc( iLevel )%sendbuffer, &
        & recv       = scheme%levelDesc( iLevel )%recvbuffer, &
        & pdf        = scheme%pdf(iLevel),                    &
        & requireAll = requireAll_loc,                        &
        & pattern    = params%general%commPattern,            &
        & comm       = params%general%proc%comm,              &
        & scheme     = scheme,                                &
        & stat       = stat,                                  &
        & haloRequired = haloRequired(iLevel)%val,             &
        & offset     = scheme%levelDesc( iLevel )%offset      )


      write(dbgUnit(1),*) "Set send and recv buffer for fromCoarser Halo elements"
      call init_levelBuffers(                                     &
        & send       = scheme%levelDesc( iLevel )%sendbufferFromCoarser, &
        & recv       = scheme%levelDesc( iLevel )%recvbufferFromCoarser, &
        & pdf        = scheme%pdf(iLevel),                        &
        & requireAll = requireAll_loc,                            &
        & pattern    = params%general%commPattern,                &
        & comm       = params%general%proc%comm,                   &
        & scheme     = scheme,                                    &
        & haloRequired = haloRequired(iLevel)%val,             &
        & stat       = stat,                            &
        & offset     = scheme%levelDesc( iLevel )%offset      )

      write(dbgUnit(1),*) "Set MPI buffer for fromFiner Halo elements"
      call init_levelBuffers(                                     &
        & send       = scheme%levelDesc( iLevel )%sendbufferFromFiner, &
        & recv       = scheme%levelDesc( iLevel )%recvbufferFromFiner, &
        & pdf        = scheme%pdf(iLevel),                        &
        & requireAll = requireAll_loc,                            &
        & pattern    = params%general%commPattern,                &
        & comm       = params%general%proc%comm,                  &
        & scheme     = scheme,                                    &
        & stat       = stat,                                      &
        & haloRequired = haloRequired(iLevel)%val,                &
        & offset     = scheme%levelDesc( iLevel )%offset          )

      deallocate( haloRequired(iLevel)%val )
    end do

    deallocate( haloRequired )

    call tem_write_debugMesh( globtree  = geometry%tree,           &
      &                       levelDesc = scheme%levelDesc,   &
      &                       myPart    = params%general%proc%rank )

?? IF( .not. SOA) then
    ! Dump Ghost and Source elements information to debugUnit
    if( main_debug%dumpDependencies) then
      call debug_dependenciesFromFiner( &
        &    scheme%levelDesc(minLevel:maxLevel), &
        &    minLevel, maxLevel )
      call debug_dependenciesFromCoarser( &
        &    scheme%levelDesc(minLevel:maxLevel), &
        &    minLevel, maxLevel )
    end if

    ! Dump Boundary debug information to debugUnit
    if( main_debug%dumpBoundaries ) then
      if ( size( geometry%tree%Property ) > 0 ) then

        write(dbgUnit(1),*) 'Check BC AFTER update_BCLists, i.e. elemPos ', &
          &                 ' with respect to total list'
        call debug_glob_boundary_type( geometry%boundary%nBCtypes, minLevel, maxLevel, &
          &                            scheme%globBC )

        call tem_debug_bc_prop( geometry%boundary )

        call debug_normals( tree   = geometry%tree, &
          &                 scheme = scheme,        &
          &                 nBCs   = geometry%boundary%nBCtypes )

      end if
    end if ! dump Boundaries
?? ENDIF

    do iLevel = minlevel, maxLevel
      ! set the sendHalo bit for all fluid elements which are send
      ! to remote procs
      call set_sendHaloBits( property = scheme%levelDesc( iLevel )%property, &
        &                    sendBuffer = scheme%levelDesc( iLevel )%sendBuffer )
    end do

    call mus_calc_commAmount( stat = stat,     &
      &                       comm = params%general%proc%comm, &
      &                       nprocs = params%general%proc%comm_size )

    call tem_stopTimer( timerHandle = mus_timerHandles%initLvlD )

    call tem_horizontalSpacer( before = 1, fUnit = logUnit(1) )

    write(logUnit(1),"(A,F10.2)") 'Done initializing geometry. Took [s]',  &
      &           tem_getTimerVal( timerHandle =  mus_timerHandles%initLvlD)
    call tem_horizontalSpacer( after  = 1, fUnit = logUnit(1) )

    write(dbgUnit(1),*)  'after the construction'

  end subroutine mus_construct
! ****************************************************************************** !


! ****************************************************************************** !
  !> Initialize the communication buffers for a single level
  subroutine communicate_property( send, recv, property, flag, proc, pattern)
    ! ---------------------------------------------------------------------------
    !> Communication structure to send and receive property
    type(tem_communication_type), intent(inout) :: send, recv
    !> property to communicate
    integer(long_k), intent(inout)              :: property(:)
    !> communication flag
    integer, intent(in)                         :: flag
    !>
    type(tem_commPattern_type), intent(in)      :: pattern
    !> Process description to use.
    type(tem_comm_env_type), intent(in)         :: proc
    ! ---------------------------------------------------------------------------

    call mus_init_longBuffers( comm    = send,                                 &
      &                        pattern = pattern )
    call mus_init_longBuffers( comm    = recv,                                 &
      &                        pattern = pattern )

    call pattern%exchange_long( send         = send,                           &
      &                         recv         = recv,                           &
      &                         state        = property,                       &
      &                         message_flag = flag,                           &
      &                         comm         = proc%comm )

  end subroutine communicate_property
! ****************************************************************************** !


! ****************************************************************************** !
  !> Initialize the communication buffers for a single level
  subroutine init_levelBuffers( send, recv, pdf, pattern, offset, &
    &                           requireAll, scheme, stat, comm, haloRequired )
    ! ---------------------------------------------------------------------------
    !> Communication structure to initialize
    type(tem_communication_type)               :: send, recv
    !> iLevel pdf info with neigh array
    type( pdf_data_type ), intent(in)          :: pdf
    !> communication pattern
    type( tem_commpattern_type ), intent(in)   :: pattern
    !>
    integer, intent(in)         :: offset( 2, eT_minRelevant:eT_maxRelevant )
    !> Statistics
    type( mus_statistics_type ), intent(inout) :: stat
    !> mpi communication enviroment with mpi communicator
    integer, intent(in)                        :: comm
    !> fluid, bnd and flow info
    type( mus_scheme_type ), intent(in)        :: scheme
    !> different place to take values from for interpolation
    logical, intent(in)                        :: requireAll
    !>
    logical, intent(in)                        :: haloRequired(:)
    ! ---------------------------------------------------------------------------
    ! myhalos are to be received from remote and remoteHalos are to be
    ! send to remote
    type( halo_commBitmask_type ), allocatable :: myHalos(:), remoteHalos(:)
    ! ---------------------------------------------------------------------------

    ! initialize bitmask array
    call init_bitmask( Halos    = myHalos,               &
      &                comm     = recv,                  &
      &                nScalars = scheme%varSys%nScalars )
    call init_bitmask( Halos    = remoteHalos,           &
      &                comm     = send,                  &
      &                nScalars = scheme%varSys%nScalars )

    ! only initialize recvBuffers and then communicate the desired
    ! values to the neighbor processes.
    ! @todo: SZ: think about the nDOFs in case more state variables than
    !            the QQ PDFs exist
    call init_recvBuffers(me         = recv,                 &
      &                   myhalos    = myhalos,              &
      &                   neigh      = pdf%neigh,            &
      &                   nSize      = pdf%nSize,            &
      &                 haloRequired = haloRequired,         &
      &                   requireAll = requireAll,           &
      &                   halo_offset= offset(1,eT_halo),    &
      &                   stat       = stat,                 &
      &                   scheme     = scheme,               &
      &                   pattern    = pattern               )

    ! communicate bitmasks
    call communicate_bitmask !myHalos     = myHalos,               &
    !   &                       remoteHalos = remoteHalos,           &
    !   &                       send        = send,                  &
    !   &                       recv        = recv,                  &
    !   &                       comm        = comm,             &
    !   &                       nScalars    = scheme%varSys%nScalars )

    call init_sendBuffers(send        = send,                 &
      &                   remoteHalos = remoteHalos,          &
      &                   neigh       = pdf%neigh,            &
      &                   nSize       = pdf%nSize,            &
      &                   scheme      = scheme,               &
      &                   pattern     = pattern               )

    ! destroy send and recv bitmask since its not needed anymore
    call destroy_bitmask( remoteHalos, send%nProcs )
    call destroy_bitmask( myhalos, recv%nProcs )

    contains

    ! ****************************************************************************** !
      !> Initialize myHalos and remoteHalos bitmasks.
      !! myHalos bitmask are from my halo elements which are to be
      !! received from remote process and remoteHalos bitmask are 
      !! remote process halos which are requested by remote process.
      subroutine init_bitmask( halos, comm, nScalars )
        ! ---------------------------------------------------------------------------
        !> contains bitmask of my halos with array size of recv%nProcs
        type(halo_commBitmask_type), allocatable, intent(out) :: halos(:)
        !> Communication structure to initialize
        type(tem_communication_type), intent(in)              :: comm
        !> number of scalars in variable system to communicate
        integer,                                   intent(in) :: nScalars
        ! ---------------------------------------------------------------------------
        integer :: iProc
        ! ---------------------------------------------------------------------------

        ! number of process my halos depends on
        allocate(halos(comm%nProcs))

        ! Allocate the bitmask for elements to be communicated
        ! from the provider proc
        do iProc = 1, comm%nProcs
          allocate( halos(iProc)%bitmask( nScalars, comm%nElemsProc(iProc) ) )
          halos(iProc)%bitmask = .false.
        end do

      end subroutine init_bitmask
    ! ****************************************************************************** !

    ! ****************************************************************************** !
      !> Destroy allocated send and recv bitmask buffer
      subroutine destroy_bitmask( halos, nProcs )
        ! ---------------------------------------------------------------------------
        !> Communication structure to initialize
        integer, intent(in) :: nProcs
        !> contains bitmask of my halos with array size of recv%nProcs
        type(halo_commBitmask_type), allocatable, intent(inout) :: Halos(:)
        ! ---------------------------------------------------------------------------
        integer :: iProc
        ! ---------------------------------------------------------------------------

        do iProc = 1, nProcs
          deallocate( halos( iProc )%bitmask )
        end do
        deallocate( halos )

      end subroutine destroy_bitmask
    ! ****************************************************************************** !

    ! ****************************************************************************** !
      subroutine communicate_bitmask()
        ! ---------------------------------------------------------------------------
        integer :: iProc
        ! ------ MPI variables
        ! send%nProcs + recv%nProcs )
        integer,allocatable :: rq_handle(:)
        ! mpi_status_size, recv%nProcs + send%nProcs )
        integer,allocatable :: status(:,:)
        integer  :: iErr
        integer  :: message_flag
        integer  :: nScalars
        ! ---------------------------------------------------------------------------

        nScalars = scheme%varSys%nScalars
        ! Choosing random message flag
        message_flag = 1
        allocate( rq_handle( send%nProcs + recv%nProcs))
        allocate( status( mpi_status_size, send%nProcs + recv%nProcs))

        ! First start receive communication of the 'send' bitmasks from the target
        ! Procs. Consumer view: receive from the target Process
        do iProc = 1, send%nProcs
          call mpi_irecv(                              &
            &      remoteHalos( iProc )%bitMask,       & ! me
            &      send%nElemsProc( iProc )*nScalars,  & ! me size
            &      MPI_LOGICAL,                        & ! data type
            &      send%proc( iProc ),                 & ! target me
            &      message_flag,                       & ! flag
            &      comm,                          & ! communicator
            &      rq_handle(iProc),                   & ! handle
            &      iErr  )                               ! error status
        end do

        ! then send the 'recv' bitmask arrays to the neighbor procs
        ! Provider view: send to the source processes
        do iProc = 1, recv%nProcs
          call mpi_isend(                              &
           &      myHalos( iProc )%bitMask,            & ! me
           &      recv%nElemsProc( iProc )*nScalars,   & ! me size
           &      MPI_LOGICAL,                         & ! data type
           &      recv%proc( iProc ),                  & ! target me
           &      message_flag,                        & ! flag
           &      comm,                           & ! communicator
           &      rq_handle( iProc + send%nProcs ),    & ! handle
           &      iErr                    )              ! error status
        end do ! iProc

        ! Wait for above communications to complete
        call mpi_waitall(               &
          & send%nProcs+recv%nProcs,    & ! total number of comm.'s to wait for
          & rq_handle,                  & ! request handles
          & status,                     & ! mpi status
          & iErr         )                ! error status
        ! --------------------------------------------------------------------------
        ! Status: We received the bitmasks from the Provider processes,
        !         from which we can now construct the correct send buffers
        ! --------------------------------------------------------------------------

        deallocate( rq_handle, status )

      end subroutine communicate_bitmask
    ! ****************************************************************************** !

  end subroutine init_levelBuffers
! ****************************************************************************** !


! ****************************************************************************** !
  !> Calculate global and sum of levels numbers of elements
  !! @todo: This routine can be seperated.
  !!
  subroutine calculate_nElems( levelDesc, proc, minLevel, maxLevel )
    ! ---------------------------------------------------------------------------
    !> Level range
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel
    type(tem_levelDesc_type), intent(in) :: levelDesc(minLevel:maxLevel)
    !> mpi communication enviroment with mpi communicator
    type(tem_comm_env_type), intent(in) :: proc
    ! ---------------------------------------------------------------------------
    integer :: iLevel, iErr, nLevels
    integer(kind=long_k) :: ElemCount(minLevel:maxLevel)
    integer(kind=long_k) :: nElems_overall
    integer(kind=long_k) :: nElems_allLevel( minLevel:maxLevel )
    integer(kind=long_k) :: nElems_fluidLevel( minLevel:maxLevel )
    integer :: myGhostFromFiner, myGhostFromCoarser
    integer :: nElems_totalGhostFromFiner
    integer :: nElems_totalGhostFromCoarser
    ! ---------------------------------------------------------------------------

    nLevels = maxLevel - minLevel + 1
    ! Count the amount of different element types
    ! counting levelwise elements ...
    call tem_horizontalSpacer(fUnit = logUnit(6))
    write(logUnit(6),"(A)") 'Elements per level'

    call tem_horizontalSpacer(fUnit = logUnit(6))
    ! done counting levelwise  elements

    ! Calculate the total number of fluid elements in ALL LEVELS and processes

    ! count level-wise global nElems_total
    ElemCount(minLevel:maxLevel) =   &
      &     levelDesc( minLevel:maxLevel )%elem%nElems( eT_fluid ) &
      &   + levelDesc( minLevel:maxLevel )%elem%nElems( eT_ghostFromCoarser ) &
      &   + levelDesc( minLevel:maxLevel )%elem%nElems( eT_ghostFromFiner )  &
      &   + levelDesc( minLevel:maxLevel )%elem%nElems( eT_halo )
    call mpi_allreduce( ElemCount(minLevel:maxLevel), &
      &                 nElems_allLevel(minLevel:maxLevel),             &
      &                 nLevels, mpi_integer8, mpi_sum, proc%comm, iErr )

    ! count level-wise global nElems_fluid
    ElemCount(minLevel:maxLevel) =   &
      &     levelDesc( minLevel:maxLevel )%elem%nElems( eT_fluid )
    call mpi_allreduce( ElemCount(minLevel:maxLevel), &
      &                 nElems_fluidLevel(minLevel:maxLevel), &
      &                 nLevels, mpi_integer8, mpi_sum, proc%comm, iErr )

    ! count my ghost over all level
    myGhostFromFiner   = 0
    myGhostFromCoarser = 0
    do iLevel = minLevel, maxLevel
      myGhostFromFiner   = myGhostFromFiner &
        &                + levelDesc( iLevel )%elem%nElems( eT_ghostFromFiner )
      myGhostFromCoarser = myGhostFromCoarser &
        &                + levelDesc( iLevel )%elem%nElems( eT_ghostFromCoarser )
    end do
    ! count global ghost
    call mpi_allreduce( myGhostFromFiner, nElems_totalGhostFromFiner, &
      &                 1, mpi_integer, mpi_sum, proc%comm, iErr )
    call mpi_allreduce( myGhostFromCoarser, nElems_totalGhostFromCoarser, &
      &                 1, mpi_integer, mpi_sum, proc%comm, iErr )

    nElems_overall = sum( nElems_allLevel( minLevel:maxLevel ))
    do iLevel = minLevel, maxLevel
      write(logUnit(1),"(A,I0)") ' Total number of elements on Level: ', iLevel
      write(logUnit(1),"(A,I0)") '   Fluid:   ', nElems_fluidLevel( iLevel )
    end do

    write(logUnit(3),"(A   )") ' Global Elements information:'
    write(logUnit(3),"(A,I0)") '   nElems overall: ', nElems_overall
    write(logUnit(3),"(A,I0)") '   nElems ghost from finer: ', &
      &                        nElems_totalGhostFromFiner
    write(logUnit(3),"(A,I0)") '   nElems ghost from coarser: ', &
      &                        nElems_totalGhostFromCoarser

    write(logUnit(1),*) ''

  end subroutine calculate_nElems
! ****************************************************************************** !



! ****************************************************************************** !
  !> The required source elements for ghost from coarser elements are
  !! identified in this routine. Moreover, the weights for each sources based
  !! on distance are calculated. \n
  !!
  !! The depdendencies for ghostFromCoarser elements initially only include its
  !! parent element.
  !! Additional dependencies Y for the element x are found which were not
  !! included in the construction routine before hand. However, they are
  !! defined to be inherently part of the stencil and hence exist locally and
  !! can simply be added to the dependencies. Two different ways are
  !! implemented.
  !! 1. Compact interpolation stencil
  !!```
  !!  Y       Y
  !!
  !!        x   o
  !!  Y       O
  !!        o   o
  !!```
  !!
  !! 2. Full interpolation stencil
  !!   the complete compute stencil is included into the dependency list
  !!```
  !!  Y       Y       Y
  !!
  !!        x   o
  !!  Y       O       Y
  !!        o   o
  !!
  !!  Y       Y       Y
  !!```
  !! @todo: to be more clear, this routine can be seperated into two
  !!        one for extend stencil, one for compact stencil
  !! @todo: calculation of weight does not apply to full stencil.
  !!        It needs to be updated.
  !!
  !! If extended stencil is used, all of its sources are identified at first.
  !! If this fails (not all sources can be found), then only identify by
  !! compact stencil
  !!
  !! Note: The view here is according to how the interpolation will be done.
  !!       The coarse level does the interpolation for the finer level.
  !!       Thus, we access the information from the levelDesc with the
  !!       targetLevel and sourceLevel perspective.
  !!
  !! Steps of finding sources:
  !! 1.  loop over all neighbors: iNeigh = 1, QQ
  !! 2.  get neighbor posInTotal: elemPos = neigh(1)%nghElems
  !! 3.  if elemPos > 0,
  !!       add this source into depFromCoarser%elem
  !!       add this source into sourceFromCoarser
  !!       save its position in sourceFromCoarser into depFromCoarser%elemBuffer
  !!       calculate its weight based on relative coordinates in stencil
  !!       weight is inverse proportional to the distance from source to target
  !!       weight is normalized at last, so that sum of weights equals 1
  !! 4.  save its offset on stencil
  !!
  !! After finding sources, append target to low or high interpolation list
  !! based on if all required sources are found.
  !!
  subroutine mus_intp_complete_fromCoarserDep( levelDesc, stencil, &
    &     nSourcesFromCoarser, nSourcesCompactFromCoarser, minLevel, maxLevel )
    ! ---------------------------------------------------------------------------
    integer,                        intent(in) :: minLevel
    integer,                        intent(in) :: maxLevel
    !> Level Descriptor
    type( tem_levelDesc_type )                 :: levelDesc(minLevel:maxLevel)
    integer,                        intent(in) :: nSourcesFromCoarser
    integer,                        intent(in) :: nSourcesCompactFromCoarser
    type( tem_stencilHeader_type ), intent(in) :: stencil
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel  ! level of source elements
    integer :: targetLevel  ! level of target elements
    integer :: targetElem   ! position of target element in total list
    integer :: iElem, iNeigh
    integer :: nGhostElems
    ! maximum number of neighbor elements
    integer :: nNeighElems, nFoundSources
    integer :: posInTotal, parentPosInTotal
    real(kind=rk) :: targetBary(3) ! barycenter of current target element
    integer,allocatable :: mySources(:)
    ! ---------------------------------------------------------------------------

    write(logUnit(1),"(A)") 'Updating dependences of ghost from coarser'

! DEBUG output  ----------------------------------------------------------------
call tem_horizontalSpacer( fUnit = dbgUnit(1), before = 1 )
write(dbgUnit(6),"(A)") "Inside mus_intp_complete_fromCoarserDep routine"
write(dbgUnit(6),"(A)") "Up to now, for each ghostFromCoarser we only have its parent."
write(dbgUnit(6),"(A)") "Here we are trying to find all neighbors of this parent."
write(dbgUnit(6),"(A,I0)") " nSourcesFromCoarser: ", nSourcesFromCoarser
write(dbgUnit(6),"(A,I0)") " nSourcesCompactFromCoarser: ", nSourcesCompactFromCoarser
write(dbgUnit(6),"(A,I0)") ''
! DEBUG output  ----------------------------------------------------------------

    allocate( mySources( nSourcesFromCoarser ))

    do sourceLevel = minLevel, maxLevel - 1

      targetLevel = sourceLevel + 1

write(dbgUnit(6),"(A,I0)") " source level: ", sourceLevel
write(dbgUnit(6),"(A,I0)") " target level: ", targetLevel
write(dbgUnit(6),"(A,I0)") ' nGhostFromCoarser: ', &
  &                        levelDesc( targetLevel )%elem%nElems( eT_ghostFromCoarser )

      ! Treat all the fromCoarser elements
      do iElem = 1, levelDesc( targetLevel )%elem%nElems( eT_ghostFromCoarser )

        parentPosInTotal = &
          & levelDesc( targetLevel )%depFromCoarser( iElem )%elem%val(1)
        targetBary = levelDesc( targetLevel )%depFromCoarser( iElem )%coord(:)

        ! target element treeID
write(dbgUnit(6),"(A,I0)") ''
        ! target element position in total list
        targetElem = iElem + levelDesc( targetLevel )%             &
          &                                offset( 1, eT_ghostFromCoarser)

write(dbgUnit(6),"(A,I0)") ' target treeID: ', leveldesc( targetlevel )%total( targetElem )
write(dbgUnit(6),"(A,I0)") ' child numnber: ', levelDesc( targetLevel )%depFromCoarser( iElem )%childNum
write(dbgUnit(6),"(A,3F8.3)") ' coord relative to parent: ', targetBary(1:3)
write(dbgUnit(6),"(A,I0)") ' number of sources (parent): ', &
  &             levelDesc( targetLevel )%depFromCoarser( iElem )%elem%nVals
write(dbgUnit(6),"(A,I0)") ' parent treeID: ', levelDesc( sourceLevel )%total( parentPosInTotal )
write(dbgUnit(6),"(A   )") ''


        ! --------------- treat parent element (first source) --------------
        ! start with treating the parent element, which is located at the
        ! first position in the dependency list
        ! Later on, proceed to siblings of parent in the target element's
        ! directions from its parent element
        ! add the parent to the list of source elements on this level

        ! Now loop over the maximum number of elements = computeStencil
        nNeighElems = nSourcesFromCoarser
        nFoundSources = 0
        mySources = 0
        do iNeigh = 1, nNeighElems
          if ( iNeigh == Stencil%restPosition ) then
            ! normally parent is the last source due to stencil defination
            posInTotal = parentPosInTotal
          else
            ! this is a neihgbor of the parent element ->
            ! access neighbor information.
            posInTotal = levelDesc( sourceLevel )%neigh(1)%       &
              &                   nghElems( iNeigh, parentPosInTotal )
          end if

          if ( posInTotal > 0 ) then
            nFoundSources = nFoundSources + 1
            mySources( nFoundSources ) = posInTotal
          end if ! posInTotal > 0, i.e. this source is a valid one

        end do ! iNeigh

        if ( nFoundSources == nSourcesFromCoarser ) then
          ! enough source elements
          ! assign to higher order interpolation scheme
          call append( me  = levelDesc( targetLevel )%intpFromCoarser_high, &
            &          val = iElem )
        else
          ! Not enough sources, try compact stencil
          call find_neigh_compact( &
            &    parentPosInTotal, &
            &    nSourcesCompactFromCoarser, &
            &    levelDesc( sourceLevel )%neigh(1)%nghElems, &
            &    compactNeighDir( 1:nSourcesCompactFromCoarser, &
            &                     levelDesc( targetLevel )%depFromCoarser( iElem )%childNum ),&
            &    stencil, &
            &    targetBary,  &
                 ! following are outputs from routine
            &    nFoundSources, mySources, &
            &    levelDesc( targetLevel )%depFromCoarser( iElem )%weight )
          ! assign it to lower order interpolation scheme
          call append( me  = levelDesc( targetLevel )%intpFromCoarser_low, &
            &          val = iElem )

        end if

        call depSource_append( &
          &    me = levelDesc( targetLevel )%depFromCoarser( iElem ),&
          &    sourceList = levelDesc( targetLevel )%sourceFromCoarser,&
          &    mySources = mySources(1:nFoundSources),  &
          &    n = nFoundSources )

      end do ! iElem = 1, nGhostElems

      call truncate( me = levelDesc( targetLevel )%sourceFromCoarser    )
      call truncate( me = levelDesc( targetLevel )%intpFromCoarser_low  )
      call truncate( me = levelDesc( targetLevel )%intpFromCoarser_high )

      write(dbgUnit(6),"(A,I0)") ' level: ', targetLevel
      write(dbgUnit(6),"(A,I0)") ' number of ghost elements: ', nGhostElems
      write(dbgUnit(6),"(A,I0)") ' high order ghost: ', &
        &      levelDesc( targetLevel )%intpFromCoarser_high%nVals
      write(dbgUnit(6),"(A,I0)") ' low  order ghost: ', &
        &      levelDesc( targetLevel )%intpFromCoarser_low%nVals
      write(dbgUnit(6),"(A,I0)") ' total sources: ',      &
        &      levelDesc( targetLevel )%sourceFromCoarser%nVals

    end do ! level loop

    write(dbgUnit(1),"(A)") "Outside of mus_intp_complete_fromCoarserDep routine"
    call tem_horizontalSpacer( fUnit = dbgUnit(1), after = 1 )

  end subroutine mus_intp_complete_fromCoarserDep
! ****************************************************************************** !


! ****************************************************************************** !
  !> All sources (children) have been found in treelm,
  !! updated number of sources needed, based on nDims
  !! collect all source elements into sourceFromFiner
  !! assign ghost to low or high order intp list, based on
  !!
  subroutine mus_update_fromFinerDep_2D( levelDesc, nSourcesFromFiner, &
    &                                    minLevel, maxLevel )
    ! ---------------------------------------------------------------------------
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel
    type( tem_levelDesc_type ) :: levelDesc(minLevel:maxLevel)
    !> When number of sources is lower than this,
    !! mark it as low order interpolation
    integer, intent(in) :: nSourcesFromFiner
    ! ---------------------------------------------------------------------------
    integer :: iLevel, iElem
    integer :: nSources   ! number of source elements actually appended
    integer :: mySources( nSourcesFromFiner )
    ! ---------------------------------------------------------------------------

    write(logUnit(1),"(A)") 'Updating dependencies of ghost from finer.'

    ! all levels except highest level
    do iLevel = minLevel, maxLevel-1

      call init( me  = levelDesc( iLevel )%intpFromFiner_high,                 &
        &        length = levelDesc( iLevel )%elem%nElems( eT_ghostFromFiner ) )

      do iElem = 1, levelDesc( iLevel )%elem%nElems( eT_ghostFromFiner )

        nSources = min( nSourcesFromFiner, &
          &             levelDesc( iLevel )%depFromFiner( iElem )%elem%nVals )
        mySources(1:nSources) = &
          & levelDesc( iLevel )%depFromFiner( iElem )%elem%val(1:nSources)
        levelDesc( iLevel )%depFromFiner( iElem )%coord = [0.5_rk, 0.5_rk, 0.5_rk]

        call depSource_append( me = levelDesc( iLevel )%depFromFiner( iElem ), &
          &                    sourceList = levelDesc( iLevel )%sourceFromFiner, &
          &                    mySources = mySources, &
          &                    n = nSources )

        if ( nSources < nSourcesFromFiner ) then
          ! not enough source elements -> assign to lower order interpolation
          call append( me  = levelDesc( iLevel )%intpFromFiner_low, &
            &          val = iElem )
        else
          ! enough source elements -> assign to higher order interpolation
          call append( me  = levelDesc( iLevel )%intpFromFiner_high, &
            &          val = iElem )
        end if

      end do ! elem%nElems( eT_ghostFromFiner )

    end do ! iLevel = minLevel, maxLevel-1

  end subroutine mus_update_fromFinerDep_2D
! ****************************************************************************** !


! ****************************************************************************** !
  !> Construct the propagation list for each element of 1st field.
  !!
  !! the bounce back rules have to be applied here
  !!
  !! Create connectivity array for one field such that it points to poisition
  !! of 1st Field in state array and use macro NGOFFSET to access other field
  !! states.
  !!
  subroutine construct_connectivity( scheme, minLevel, maxLevel)
    ! ---------------------------------------------------------------------------
    !> Minlevel and maxLevel
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel
    !> Current scheme
    type( mus_scheme_type ) :: scheme
    ! ---------------------------------------------------------------------------
    integer :: iElem, iLevel, iDir, invDir, QQ, nScalars, zeroPos
    ! result: from which direction to take the pdf from (inverse for bounceback)
    integer :: GetFromDir
    ! result: from which element to take the pdf from (local for bounceback)
    integer :: GetFromPos
    integer :: neighPos      ! position of current neighElem in treeID list
    integer, allocatable :: stateVarPos(:)
    integer(kind=long_k) :: neighProp, elemProp
    integer :: nElems, nSize
    ! ---------------------------------------------------------------------------

    write(logUnit(1),*) 'Building state array connectivity ...'

    ! neigh array points to actual position of 1st field in the state 
    ! array with size of nFields*QQ*nElems and
    ! position of current field in the state array can be accessed 
    ! by using offset (iField-1)*QQ

    QQ       = scheme%layout%fStencil%QQ
    nScalars = scheme%varSys%nScalars
    allocate( stateVarPos(QQ) )
    ! state varpos for 1st field since neigh is created only for 1st field
    stateVarPos(:QQ) = scheme%varSys%method%val(               &
      &                      scheme%stateVarMap%varPos%val(1)) &
      &                            %state_varPos(:QQ)

    ! set zero position direction to itself
    ! (e.g. the d3q6 stencil does not have a rest direction at all)
    if ( scheme%layout%fStencil%restPosition > 0 ) then
      zeroPos = scheme%layout%fStencil%restPosition
      do iLevel = minLevel, maxLevel
        nElems = scheme%pdf( iLevel )%nElems_local
        nSize  = scheme%pdf( iLevel )%nSize
        do iElem = 1, nElems
          scheme%pdf( iLevel )%neigh( ?NGPOS?( zeroPos, iElem, nSize )) = &
            &          ?IDX?( zeroPos, iElem, nScalars, nSize )
        end do ! iElem
      end do ! iLevel
    end if


    levelLoop: do iLevel = minLevel, maxLevel
      nElems = scheme%pdf( iLevel )%nElems_local
      nSize  = scheme%pdf( iLevel )%nSize
      elemLoop: do iElem = 1, nElems
        elemProp = scheme%levelDesc( iLevel )%property( iElem )

        ! Loop over every link
        neighLoop: do iDir  = 1, scheme%layout%fStencil%QQN
          ! For push we need to find different neighbors than for pull
          ! -> ngDir
          neighPos = scheme%levelDesc( iLevel )%neigh(1)%            &
            &  nghElems( ?NgDir?( iDir, scheme%layout%fStencil ), iElem )

          ! JQ: get neighbor's property
          if( neighPos > 0 ) then
            neighProp = scheme%levelDesc( iLevel )%property(neighPos)
          else
            neighProp = 0_long_k
          end if

          ! Treat missing neighbor element as solid
          ! neighPos < 0 : neighbor is a boundary
          ! neighPos = 0 : neighbor is nonExisting element
          ! neighProp solid: neighbor was solidified or has missing neigh
          ! elemProp solid: myself was solidified or has missing neigh
          if( (neighPos <= 0) .or. (btest(neighProp, prp_solid))           &
            &  .or. (btest(elemProp, prp_solid))                           &
            & ) then
            ! 1. Fluid element has boundary in current link direction
            ! 2. neighbor element has fluidify property (still solid now)
            ! Get inverse direction of LOCAL element
            invDir = scheme%layout%fStencil%cxDirInv( iDir )
            GetFromDir = stateVarPos(invDir)
            GetFromPos = iElem
            ! I set the source to the target link for halo cells
            ! with no neighbor in that direction
          else
            ! link to get from neighbor
            GetFromDir = stateVarPos(iDir)
            GetFromPos = neighPos
          end if ! has no boundary -> regular treatment

          scheme%pdf( iLevel )%neigh( ?NGPOS?( iDir, iElem, nSize )) = &
            &          ?IDX?( GetFromDir, GetFromPos, nScalars, nSize )
        end do neighLoop
      end do elemLoop
    end do levelLoop
    write(logUnit(1),*) 'Done with state array connectivity.'

  end subroutine construct_connectivity
! ****************************************************************************** !


! ****************************************************************************** !
  !> Select the halo elements which require all links
  !!
  !! Determine which pdfs to communicate
  !!
  subroutine set_halo_commLinks( scheme, minLevel, maxLevel, nBCs, comm_reduced, &
    &                            haloRequired )
    ! ---------------------------------------------------------------------------
    !> scheme information including fluid, boundary and flow information
    type(mus_scheme_type), intent(in) :: scheme
    !> Global information
    integer, intent(in) :: minLevel, maxLevel, nBCs
    !> reduced communication
    logical, intent(in) :: comm_reduced
    !>
    type( logical_array_type ), allocatable, intent(inout) :: haloRequired(:)
    ! ---------------------------------------------------------------------------
    integer :: iElem, iLevel, iSourceElem
    integer :: iBnd, iField, iNeigh, iProc
    integer :: targetElem, sourceElem, targetLevel, sourceLevel
    integer :: nSourceElems
    integer :: elemPos, haloOffset
    ! ---------------------------------------------------------------------------

    write(logUnit(5),"(A)") "Set halo elements comm links"

    allocate( haloRequired(minLevel:maxLevel) )
    do iLevel = minLevel, maxLevel
      allocate( haloRequired(iLevel)%val( scheme%pdf(iLevel)%nElems_halo ) )
      haloRequired(iLevel)%val(:) = .false.
    end do

    do iLevel = minLevel, maxLevel

      ! allocate( haloRequired(iLevel)%val( scheme%pdf(iLevel)%nElems_halo ) )

      if( .not. comm_reduced ) then
        ! Communicate all links of all halos
        ! scheme%pdf(iLevel)%haloRequired(:) = .true.
        haloRequired(iLevel)%val(:) = .true.
      else ! Choose which ones are really required
        ! -----------------------------------------------------------------------
        ! GhostFromFiner
        ! check if there are halo deps
        if( iLevel < maxLevel ) then
          targetLevel = iLevel
          sourceLevel = iLevel + 1
          haloOffset  = scheme%levelDesc( sourceLevel )%offset(1,eT_halo)

          do iElem = 1, scheme%pdf( targetLevel )%nElems_ghostFromFiner
            ! position of coarser ghost element in levelDesc total list
            targetElem = iElem + scheme%levelDesc( targetLevel )          &
              &                      %offset( 1, eT_ghostFromFiner)
            nSourceElems = scheme%levelDesc( targetLevel )%               &
              &                                depFromFiner( iElem )%elem%nVals
            ! number of fine elements(childs)
            do iSourceElem = 1, nSourceElems
              ! position of child element in levelDesc total list
              sourceElem = scheme%levelDesc( targetLevel )%               &
                &                 depFromFiner( iElem )%elem%val( iSourceElem )
              call set_requiredLink(                                           &
                &          haloElem  = sourceElem,                             &
                &          offset    = haloOffset,                             &
                &          linkField = haloRequired( sourceLevel )%val )
            end do ! iSourceElem
          end do ! iElem
        end if !iLevel < maxLevel
        ! GhostFromFiner
        ! -----------------------------------------------------------------------

        ! -----------------------------------------------------------------------
        ! GhostFromCoarser
        ! check if there are halo deps
        if( iLevel < maxLevel ) then
          targetLevel = iLevel + 1
          sourceLevel = iLevel
          haloOffset =scheme%levelDesc( sourceLevel )%offset(1,eT_halo)
          do iElem = 1, scheme%pdf( targetLevel )%nElems_ghostFromCoarser
            ! position of ghost finer element in levelDesc total list
            targetElem = iElem + scheme%levelDesc( targetLevel )          &
              &                      %offset( 1, eT_ghostFromCoarser)
            ! number of coarser elements that finer ghost dependent on.
            ! usually this is just 1.
            nSourceElems = scheme%levelDesc( targetLevel )                &
              &            %depFromCoarser( iElem )%elem%nVals
            do iSourceElem = 1, nSourceElems
              sourceElem = scheme%levelDesc( targetLevel )                &
                &          %depFromCoarser( iElem )%elem%val( iSourceElem )
              ! Check the type of the source element.
              ! If it is a halo -> mark as require all links
              call set_requiredLink(                                           &
                &          haloElem  = sourceElem,                             &
                &          offset    = haloOffset,                             &
                &          linkField = haloRequired( sourceLevel )%val )
            end do ! iSourceElem
          end do ! iElem
        end if !level<maxLevel
        ! GhostFromCoarser
        ! -----------------------------------------------------------------------

        haloOffset = scheme%levelDesc( iLevel )%offset(1, eT_halo)
        ! -----------------------------------------------------------------------
        ! Run over recvbufferFromCoarser
        do iProc = 1, scheme%levelDesc(iLevel)%recvBufferFromCoarser%nProcs
          do iElem = 1, scheme%levelDesc( iLevel )%recvBufferFromCoarser% &
              &                                             nElemsProc( iProc )
            elemPos = scheme%levelDesc( iLevel )%recvBufferFromCoarser%   &
              &                                   elemPos( iProc )%val( iElem )
            call set_requiredLink(                                             &
              &                 haloElem  = elemPos,                           &
              &                 offset    = haloOffset,                        &
              &                 linkField = haloRequired( iLevel )%val )
          end do
        end do
        ! -----------------------------------------------------------------------

        ! -----------------------------------------------------------------------
        ! Run over recvbufferFromFiner
        do iProc = 1, scheme%levelDesc( iLevel )%recvBufferFromFiner%nProcs
          do iElem = 1, scheme%levelDesc( iLevel )%recvBufferFromFiner%   &
            &                                               nElemsProc( iProc )
            elemPos = scheme%levelDesc( iLevel )%recvBufferFromFiner%     &
              &                                   elemPos( iProc )%val( iElem )
            call set_requiredLink(                                             &
              &                 haloElem  = elemPos,                           &
              &                 offset    = haloOffset,                        &
              &                 linkField = haloRequired( iLevel )%val )
          end do
        end do
        ! -----------------------------------------------------------------------

        ! -----------------------------------------------------------------------
        ! Run over all the boundary neighbors
        ! If they require neighbor infos, it is usually integral quantities
        ! ( macrosopic)  and thus require all links
        do iBnd = 1, nBCs
          do iElem = 1, scheme%globBC( iBnd )%nElems( iLevel )
            do iField = 1, scheme%nFields
              do iNeigh = 1, scheme%field( iField )%bc( iBnd )%nNeighs
                ! Get the neighbor source element
                sourceElem = scheme%field( iField )%bc( iBnd )                 &
                  &                 %neigh( iLevel )%posInState( iNeigh, iElem )
                call set_requiredLink(                                         &
                  &             haloElem  = sourceElem,                        &
                  &             offset    = haloOffset,                        &
                  &             linkField = haloRequired( iLevel )%val )
              end do ! iNeigh
            end do  ! iField
          end do ! iElem
        end do ! iBnd

        ! Communicate all links of boundary neighbors which are halo.
        ! There are some boundaries which requires all links of neighbors
        ! like wall_linearInterpolation.
        ! Do this for boundaries other than wall
        do iBnd = 1, nBCs
          if( .not. scheme%globBC(iBnd)%isWall .or. &
            &       scheme%globBC(iBnd)%hasQVal ) then
            do iElem = 1, scheme%globBC( iBnd )%nElems( iLevel )
              elemPos = scheme%globBC(iBnd)%elems(iLevel)%elem(iElem)
              do iNeigh = 1, scheme%layout%fStencil%QQN
                ! Get the neighbor source element
                sourceElem = scheme%levelDesc(iLevel)%neigh(1)%        &
                  & nghElems(?NgDir?(iNeigh, scheme%layout%fStencil),elemPos)
                ! if sourceElem < 0 then this neighbor is boundary
                call set_requiredLink(                                 &
                  &      haloElem  = sourceElem,                       &
                  &      offset    = haloOffset,                       &
                  &      linkField = haloRequired( iLevel )%val )
              end do ! iNeigh
            end do ! iElem
          end if ! not wall
        end do ! iBnd
        ! Boundary neighbors
        ! -----------------------------------------------------------------------
      end if ! comm_reduced?

    end do ! iLevel

  contains

    ! ****************************************************************************** !
    !> set haloRequired link to true
    !!
    subroutine set_requiredLink( haloElem, offset, linkField )
      ! ---------------------------------------------------------------------------
      !> halo element position in the state array 
      integer, intent(in)    :: haloElem
      !> halo element offset in the state array
      !! nr. of elements in state array before halo elements
      integer, intent(in)    :: offset
      !> halo required array used in init_recvBuffers to reduce communication links
      logical, intent(inout) :: linkField(:)
      ! ---------------------------------------------------------------------------
      if( haloElem > offset ) then
        ! haloRequired is only nElems_halo long, so remove the preceeding
        ! elem nums
        linkField( haloElem - offset ) = .true.
      end if

    end subroutine set_requiredLink
    ! ****************************************************************************** !

  end subroutine set_halo_commLinks
! ****************************************************************************** !



! ****************************************************************************** !
  !> Update the neighbor stencil positions for the Boundaries
  !!
  !! Before, positions in the original treeIDs were assigned to
  !! the boundary element lists. Update these now with level-wise positions in
  !! the [[tem_construction_module:level_descriptor_type]] "total list".
  !! When no element is found, the value in nghElems is 0. This has to be
  !! treated in this routine. If an empty neighbor relation is encountered, the
  !! last valid neighbor is used for all other neighbors
  !! Assertion:
  !! Before this routine, we should have:
  !!  scheme%globBC( iBnd )%nElems( iLevel )
  !!  scheme%globBC( iBnd )%elems( iLevel )%elem( iElem ))
  !!
  subroutine update_BClists( scheme, minLevel, maxLevel, nBCs, LP, remove_solid)
    ! ---------------------------------------------------------------------------
    !> scheme information including fluid, boundary and flow information
    type(mus_scheme_type), intent(inout) :: scheme
    !> Level Pointer
    integer, intent(in) :: LP(:)
    !>
    integer, intent(in) :: minLevel, maxLevel, nBCs
    logical, intent(in) :: remove_solid
    ! ---------------------------------------------------------------------------
    integer :: iElem, iLevel, iBnd, iField
    integer :: elemPos, iPos, bcBufferSize
    ! ---------------------------------------------------------------------------

    ! set posInNghElems in scheme%field(iField)%bc
    do iField = 1, scheme%nFields
      do iBnd = 1, nBCs
        call mus_set_posInNghElems( minLevel, maxLevel, &
          &                         scheme%layout%nStencils, &
          &                         scheme%globBC(iBnd), &
          &                         scheme%field(iField)%bc(iBnd) )
      end do
    end do

    if ( remove_solid .and. (nBCs > 0) ) then
      call remove_solid_in_bc( minLevel, maxLevel, nBCs, &
        &                      scheme%nFields, LP, &
        &                      scheme%levelDesc, scheme%globBC, scheme%field )
    end if

    ! 1. Update the element positions from the treeID to levelDesc%total list
    ! 1a. Create dynamic array of bc_elemBuffer with position of
    ! treeID in the levelDescriptor%total list and store the position
    ! in globBC%elems(iLevel)%elembuffer
    do iLevel = minLevel, maxLevel

      ! allocate bc_elemBuffer with initial size of all bc nElems
      bcBufferSize = 0
      do iBnd = 1, nBCs
        bcBufferSize = bcBufferSize + scheme%globBC( iBnd )%nElems( iLevel )
      end do ! iBnd
      call init( me     = scheme%levelDesc( iLevel )%bc_elemBuffer,       &
        &        length = bcBufferSize )

      do iBnd = 1, nBCs

        ! if BC kind of all fields are wall then do nothing for this boundary
        ! i.e do not add this boundary to elemBuffer or bc_elemBuffer or
        ! neigh buffers since wall boundaries are treated explicitly as
        ! bounce back via pdf%neigh array
        if ( .not. scheme%globBC(iBnd)%isWall ) then

          ! allocate neighBuffer pre-collision and postcollision for each
          ! boundary in each field and initialize them
          ! @todo: SZ: think about this allocation in case there are more state
          !            variables than the QQ PDFs
          do iField = 1, scheme%nFields

            if (scheme%field( iField )%bc( iBnd )%requireNeighBufPre) then
              allocate( scheme%field( iField )%bc( iBnd )%                     &
                & neigh( iLevel )%neighBufferPre(                              &
                &         scheme%field( iField )%bc( iBnd )%nNeighs,           &
                &         scheme%globBC( iBnd )%nElems( iLevel )*              &
                &         scheme%layout%fStencil%QQ ) )
              scheme%field(iField)%bc(iBnd)%neigh(iLevel)%neighBufferPre &
                & = -1._rk
            end if
            if (scheme%field( iField )%bc( iBnd )%requireNeighBufPost) then
              allocate( scheme%field( iField )%bc( iBnd )%                     &
                & neigh( iLevel )%neighBufferPost(                             &
                &         scheme%field( iField )%bc( iBnd )%nNeighs,           &
                &         scheme%globBC( iBnd )%nElems( iLevel )*              &
                &         scheme%layout%fStencil%QQ ) )
              scheme%field(iField)%bc(iBnd)%neigh(iLevel)%neighBufferPost &
                & = -1._rk
            end if

            if ( scheme%field(iField)%bc(iBnd)%useComputeNeigh ) then
              allocate( scheme%field( iField )%bc( iBnd )%neigh( iLevel )      &
                &                                        %computeNeighBuf(     &
                &         scheme%globBC( iBnd )%nElems( iLevel )*              &
                &         scheme%layout%fStencil%QQ ) )
              scheme%field( iField )%bc( iBnd )%neigh( iLevel )        &
                &                              %computeNeighBuf = -1._rk
            end if

          end do ! iField

          ! UPDATE globBD%elems%elem and add to bc_elemBuffer
          do iElem = 1, scheme%globBC( iBnd )%nElems( iLevel )
            ! update bc element position in treelm to be pos in total list
            elemPos = LP( scheme%globBC( iBnd )%elems( iLevel )%elem( iElem ) )
            scheme%globBC( iBnd )%elems( iLevel )%elem( iElem ) = elemPos

            ! store elemPos in bc_elemBuffer
            call append( me  = scheme%levelDesc( iLevel )%bc_elemBuffer,  &
              &          val = elemPos,                                        &
              &          pos = iPos )
            ! store the position in the bc_elemBuffer in the elemBuffer
            ! @todo: maybe rename elemeBuffer to posInBCelemBuf
            scheme%globBC( iBnd )%elems( iLevel )%elemBuffer( iElem ) = iPos
          end do ! iElem in globBC%nElems on the current level

          ! elem and elemBuffer for corner nodes
          do iElem = 1, scheme%globBC( iBnd )%cornerBC%nElems( iLevel )
            elemPos = LP( scheme%globBC( iBnd )%           &
              &                  cornerBC%elems( iLevel )%elem( iElem ))

            scheme%globBC( iBnd )%cornerBC%elems( iLevel )%elem( iElem )       &
              &                                                        = elemPos

            iPos = PositionOfVal( me  = scheme%levelDesc( iLevel )%       &
              &                                                bc_elemBuffer,  &
              &                   val = elemPos )
            ! store the position in the bc_elemBuffer in the elemBuffer
            scheme%globBC( iBnd )%cornerBC%elems( iLevel )%                    &
              &                                     elemBuffer( iElem ) = iPos
          end do

        end if ! all Wall
      end do ! iBnd

      call truncate( me = scheme%levelDesc( iLevel )%bc_elemBuffer )
      write(dbgUnit(6),*) 'bc_elemBuffer nVals: ', &
        &                  scheme%levelDesc( iLevel )%bc_elemBuffer%nVals

    end do !iLevel

    ! 2. build up neigh information for higher order boundaries
    ! loop over the field to update array boundary in each field
    do iField = 1, scheme%nFields
      do iBnd = 1, nBCs
        ! Copy the neighboring information to the bc%elem%neigh
        if( scheme%field( iField )%bc( iBnd )%nNeighs > 0 ) then
          ! get position of neighbor element in the levelwise list
          call setFieldBCNeigh(                                       &
            &     fieldBC   = scheme%field( iField )%bc( iBnd ),      &
            &     globBC    = scheme%globBC( iBnd ),                  &
            &     levelDesc = scheme%levelDesc(minLevel:maxLevel),&
            &     minLevel  = minLevel,                     &
            &     maxLevel  = maxLevel                      )
        end if ! nNeigh>0
      end do ! iBnd
    end do ! iField

  end subroutine update_BClists
! ****************************************************************************** !


! ****************************************************************************** !
  !> This routine sets field BC neigh array with position of 
  !! neighbor element in the inward normal direction of boundary 
  !! in the levelwise list.
  !! if valid 1st neighbor does not exist return current element position.
  !! if valid higher order neighbor does not exist return last valid
  !! neighbor
  subroutine setFieldBCNeigh( fieldBC, globBC, levelDesc, minLevel, maxLevel )
    ! ---------------------------------------------------------------------------
    !> Level range
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel
    !> field boundary with boundary neighbor info
    type(boundary_type),      intent(inout) :: fieldBC
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(in)    :: globBC
    !> Level descriptor
    type(tem_levelDesc_type), intent(in)    :: levelDesc(minLevel:maxLevel)
    ! ---------------------------------------------------------------------------
    integer :: iElem, iLevel, iNeigh
    integer :: neighVal
    integer :: posInNghElems, stencilPos
    ! ---------------------------------------------------------------------------

    ! count up the position where to store the neighborhood
    lvlLoop: do iLevel = minLevel, maxLevel
      write(logUnit(6),"(A,I0)") '  Fill field BC neighbor list on level: ', &
        &                        iLevel
      allocate( fieldBC%neigh( iLevel )%posInState( fieldBC%nNeighs,         &
        &                                           globBC%nElems( iLevel ) ))

write(dbgUnit(6),"(A,I0)") ' Fill field BC '//trim(globBC%label)//&
  &                        ' neighbor list on level: ', iLevel
write(dbgUnit(6),"(A,I0,I5)") ' allocate fieldBC%neigh%posInState by:',&
  &                           fieldBC%nNeighs, globBC%nElems( iLevel )

      elemLoop: do iElem = 1, globBC%nElems( iLevel )
!        write(dbgUnit(1),*) 'iElem ', iElem
!        write(dbgUnit(1),*) 'elemID ', &
!          & levelDesc(iLevel)%total(globBC%elems(iLevel)%elem(iElem))
!        write(dbgUnit(1),*) 'normal ', globBC%elems(iLevel)%normal(:, iElem)

        stencilPos = fieldBC%elems(iLevel)%stencilPos(iElem)
        ! current element index in nghElems
        ! counter(iLevel, stencilPos) = counter(iLevel, stencilPos) + 1
        posInNghElems = fieldBC%elems(iLevel)%posInNghElems( iElem )
write(dbgUnit(6),*) 'iElem: ', iElem, ', stencilPos: ', stencilPos
        neighLoop: do iNeigh = 1, fieldBC%nNeighs
          ! neighbor position in the levelwise list
          neighVal = levelDesc( iLevel )%neigh( stencilPos )%                  &
            &                            nghElems( iNeigh, posInNghElems )
write(dbgUnit(6),*) 'iNeigh:', iNeigh, ', posInNghElems:', posInNghElems
write(dbgUnit(6),*) 'neighval:', neighVal

          ! No valid neighbor found. set to boundary element pos if iNeigh = 1
          ! else set last valid neigbor
          if ( neighVal < 1 ) then
            if ( iNeigh == 1 ) then
              !write(dbgUnit(1),*) 'setting current elemPos',&
              !                     globBC%elems( iLevel )%elem( iElem )

              ! Can not even find the first neighbor, i.e. no neighbors at all
              ! so save simply current bnd elem pos (itself as neighbor)
              fieldBC%neigh( iLevel )%posInState( :, iElem ) = &
                &                          globBC%elems( iLevel )%elem( iElem )
            else
              !write(dbgUnit(1),*) 'neighbor fails taking last valid neighbor '
              !write(dbgUnit(1),*) 'setting last valid ', &
              !  &    levelDesc( iLevel )%neigh( stencilPos )%    &
              !  &          nghElems( iNeigh-1, counter(iLevel, stencilPos))

              ! Neighbor fails, so take last neighbor which was valid
              fieldBC%neigh( iLevel )%posInState( iNeigh:, iElem ) = &
                &      fieldBC%neigh( iLevel )%posInState( iNeigh-1, iElem )
            end if ! iNeigh = 1
            ! no need to find further neighbors
            exit neighLoop
          else
            !write(dbgUnit(1),*) 'setting neighVal ', neighVal

            ! regular neighbor: assign
            ! fieldBC%elems( iLevel )%neigh( iNeigh, iElem ) = neighVal
            ! store neighbor position in original treeID list
            fieldBC%neigh( iLevel )%posInState( iNeigh, iElem ) = neighVal
          end if ! neighVal < 1
        end do neighLoop

write(dbgUnit(6),*) 'all neighIDs: ', &
  &    levelDesc(iLevel)%total(fieldBC%neigh(iLevel)%posInState(:,iElem))

      end do  elemLoop
    end do  LvlLoop

  end subroutine setFieldBCNeigh
! ****************************************************************************** !


! ****************************************************************************** !
  !> Assemble the level-wise list of elements which adhere to the boundary
  !! conditions.
  !!
  !! The boundaries will then be treated for each level one by one, running over
  !! the list of elements.
  !! The bitmasks are set for the directions pointing into the domain.
  !! In the LBM, the incoming densities have to be updated.
  !! As boundaries are being set before the kernel, the state arrays have to be
  !! stored to the FETCH position
  !!
  subroutine build_BClists( globBC, tree, bc_prop, minLevel, maxLevel, &
    &                       layout, field, comm )
    ! ---------------------------------------------------------------------------
    !> fluid tree from mesh
    type( treelmesh_type ), intent(in)                  :: tree
    !> boundary information from mesh
    type( tem_bc_prop_type ), intent(in)                :: bc_prop
    !> contains pdf global information
    integer,  intent(in) :: minLevel, maxLevel
    !> scheme layout
    type( mus_scheme_layout_type ), intent(in)          :: layout
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(out), allocatable  :: globBC(:)
    !> field type
    type( mus_field_type ), intent(in)                  :: field(:)
    !> mpi communication enviroment with mpi communicator
    integer, intent(in)                 :: comm
    ! ---------------------------------------------------------------------------
    integer :: iBnd, nBCs
    ! ---------------------------------------------------------------------------

    write(logUnit(1),*) 'Creating boundary lists ...'

    nBCs = bc_prop%nBCtypes
    if( nBCs > 0 ) then

      ! allocate bc states for each boundary with size of nFields
      ! to collect bc states of all fields of this particular boundary
      ! into one global list. Needed when boundary condition is
      ! dependent on all fields.
      ! contains element information about each boundary type
      allocate( globBC( nBCs ) )
      do iBnd = 1, nBCs
        allocate( globBC(iBnd)%elems(             minLevel:maxLevel ) )
        allocate( globBC(iBnd)%nElems(            minLevel:maxLevel ) )
        allocate( globBC(iBnd)%nElems_totalLevel( minLevel:maxLevel ) )
        allocate( globBC(iBnd)%cornerBC%nElems(   minLevel:maxLevel ) )
        allocate( globBC(iBnd)%cornerBC%elems(    minLevel:maxLevel ) )

        globBC(iBnd)%isWall = mus_check_allWall( size(field), field, iBnd )
      end do
      ! copy bc label from mesh boundary to solver globBC
      globBC(:)%label = bc_prop%BC_label(:)

      ! boundary%hasQval is allocated only if prp_hasQval is set in seeder
      ! so initialize them first and use globBC(:)%hasQval in rest of the
      ! routine
      globBC(:)%hasQVal = .false.
      if (allocated(bc_prop%hasQval)) globBC(:)%hasQVal = bc_prop%hasQval(:)

      ! 1. Identify the number of boundary condition elements of each BC type
      call countnBnds( globBC      = globBC,              &
        &              boundaryID  = bc_prop%boundary_ID, &
        &              tree        = tree,                &
        &              stencil     = layout%fStencil,     &
        &              nBCs        = nBCs,                &
        &              comm        = comm                 )

      ! 2. Allocate the BC lists
      call allocateBCList( globBC   = globBC,            &
        &                  nBCs     = nBCs,              &
        &                  minLevel = minLevel,          &
        &                  maxLevel = maxLevel,          &
        &                  QQ       = layout%fStencil%QQ )

      ! 3. Assign the BC lists
      ! Run over all the elements with the property boundary
      ! and check each direction.
      ! Assign all common boundaries to the level-wise representation
      call assignBCList( tree        = tree,            &
        &                stencil     = layout%fStencil, &
        &                bc_prop     = bc_prop,        &
        &                globBC      = globBC           )

      ! 4 Normalize the normal vectors for regular and corner elements
      !   and choose the corresponding prevailing direction from the stencil
      call normalizeBC( globBC   = globBC,   &
        &               field    = field,    &
        &               nBCs     = nBCs,     &
        &               minLevel = minLevel, &
        &               maxLevel = maxLevel, &
        &               layout   = layout    )

    else
      allocate( globBC(0) )
    end if ! nBCs > 0 ?

  end subroutine build_BClists
! ****************************************************************************** !


! ****************************************************************************** !
  !> Identify the number of boundary condition elements of each BC type
  !! and number of elements with multiple BC types.
  !! This is 1st step in Build_BClists
  subroutine countnBnds( globBC, boundaryID, tree, stencil, nBCs, comm )
    ! ---------------------------------------------------------------------------
    !> boundary information from mesh
    integer(kind=long_k),           intent(in) :: boundaryID(:,:)
    !> fluid tree from mesh
    type( treelmesh_type ),         intent(in) :: tree
    !> stencil
    type( tem_stencilHeader_type ), intent(in) :: stencil
    !> contains pdf global information
    integer,                        intent(in) :: nBCs
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(inout)    :: globBC(nBCs)
    !> mpi communication enviroment with mpi communicator
    integer,                        intent(in) :: comm
    ! ---------------------------------------------------------------------------
    integer :: iElem, iDir, iBnd, level, iMeshDir, iErr, iLevel
    integer :: minLevel, maxLevel
    logical :: found( nBCs )
    integer(kind=long_k) :: bID
    !> number of local elements of each boundary type in each level
    integer, allocatable :: nBnds(:,:)
    !> number of local elements with multiple boundary type in each level
    integer, allocatable :: nCornerBnds(:,:)
    integer, allocatable :: nBnds_total(:,:)
    ! ---------------------------------------------------------------------------

    minLevel = tree%global%minLevel
    maxLevel = tree%global%maxLevel

    ! number of local elements of each boundary type in each level
    allocate( nBnds( minLevel:maxLevel, nBCs ))
    ! number of total elements(all processes) of each boundary type in each level
    allocate( nBnds_total( minLevel:maxLevel, nBCs ))
    ! number of elements with multiple boundaries in each level
    allocate( nCornerBnds( minLevel:maxLevel, nBCs ))

    nBnds       = 0
    nCornerBnds = 0

    do iElem = 1, tree%property(1)%nElems
      found = .false.
      level = tem_LevelOf( tree%treeID( tree%property(1)%ElemID(iElem)))
      do iDir = 1, stencil%QQN
        iMeshDir = stencil%map( iDir )
        ! the index in boundary_ID does not correspond to the stencil!
        bID = boundaryID( iMeshDir, iElem )
        if( bID > 0  ) then
          ! checks if number of boundary IDs dumped by seeder is not
          ! greater than number of boundary types in bnd.lua
          if( bID > nBCs ) then
            write(logUnit(1),*) ' found BID:     ', bID
            write(logUnit(1),*) ' where the highest should be ', nBCs
            write(logUnit(1),*) ' at element     ', iElem
            write(logUnit(1),*) ' found higher bID than nr of bcs exist.'
          end if

          if ( .not. found ( bID ) ) then
            ! Increase the counter only, if this bID has not been
            ! identified before at iElem
            nBnds( level, bID ) = nBnds( level, bID )+1
            found( bID ) = .true.
          end if ! found ( bID ) == .false.
        end if ! bid > 0 ?
      end do ! iDir

      ! if element is intersected by more than one boundary
      ! also count it as a corner bc element
      if ( count(found) > 1 ) then
        do iBnd = 1, nBCs
          if (found(iBnd)) then
            nCornerBnds(level, iBnd) = nCornerBnds(level, iBnd) + 1
          endif
        enddo
      endif !count > 1 for corner nodes

    end do ! iElem

    ! Counting number of boundary elements on local partition and global
    ! Also copies field bcstates to globBC bcstates to access all
    ! field bcstates in each boundary routines
    do iBnd = 1, nBCs
      ! number of elements for each iBnd on each level
      globBC( iBnd )%nElems(minLevel:maxLevel) = nBnds(minLevel:maxLevel, iBnd )
      ! Sum up the elements (all levels) of the local partition for each iBnd
      globBC( iBnd )%nElems_local = sum( nBnds(minLevel:maxLevel, iBnd) )

      ! number of elements with multiple boundaries
      globBC( iBnd )%cornerBC%nElems(minLevel:maxLevel) = &
        &                  nCornerBnds(minLevel:maxLevel, iBnd )
    end do ! iBnd

    ! calculate total number of boundary elements in all levels and processes
    call mpi_allreduce( nBnds, nBnds_total, (maxLevel-minLevel+1)*nBCs,  &
      &                 mpi_integer, mpi_sum, comm, iErr      )

    ! Avoid division by 0.
    nBnds_total = max(nBnds_total,1)

    call tem_horizontalSpacer( fUnit = dbgUnit(1), before = 1 )
    write(dbgUnit(1),"(A)") "Number of boundary elements"
    do iBnd = 1, nBCs
      globBC( iBnd )%nElems_totalLevel(:) = nBnds_total( :, iBnd )
      globBC( iBnd )%nElems_total = sum(globBC( iBnd )%nElems_totalLevel(:))
      write(dbgUnit(1),"(A,I0,A)") "iBC: ", iBnd, ", label: "//trim(globBC(iBnd)%label)
      do iLevel = minLevel, maxLevel
        write(dbgUnit(1),"(A,I2,A,I8,A,F5.2)") "level: ", iLevel, &
          &                 ", nElems: ", nBnds(iLevel,iBnd), &
          &                 ", % of nGlobal: ", &
          &     dble(nBnds(iLevel,iBnd))/dble(nBnds_total(iLevel,iBnd)) * 100._rk
      end do
      write(dbgUnit(1),"(A)") ''
    end do
    call tem_horizontalSpacer( fUnit = dbgUnit(1), after = 1 )

   end subroutine countnBnds
! ****************************************************************************** !


! ****************************************************************************** !
  !> Allocate BC lists, 2nd step in Build_BClists
  subroutine allocateBCList( globBC, nBCs, minLevel, maxLevel, QQ )
    ! ---------------------------------------------------------------------------
    integer,                  intent(in)    :: nBCs, minLevel, maxLevel, QQ
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(inout) :: globBC(nBCs)
    ! ---------------------------------------------------------------------------
    integer :: iBC, iLevel
    ! ---------------------------------------------------------------------------

    do iBC = 1, nBCs
      do iLevel = minLevel, maxLevel

        call mus_allocate_bc_elems( globBC(iBC)%elems(iLevel), &
          &                         globBC(iBC)%nElems(iLevel), &
          &                         QQ, globBC(iBC)%hasQVal )

        if( globBC(iBC)%nElems(iLevel) > 0) then
          write(logUnit(6),"(A,A,I2,A,I0)") ' BC: '//trim(globBC(iBC)%label),&
            &  ', level: ', iLevel, ', nElems: ', globBC( iBC )%nElems( iLevel )
          globBC( iBC )%elems( iLevel )%elem(:) = 0
          globBC( iBC )%elems( iLevel )%bitmask = .false.
          globBC( iBC )%elems( iLevel )%normal  = 0
        end if

        ! elements with multiple boundaries
        call mus_allocate_bc_elems( globBC(iBC)%cornerBC%elems(iLevel), &
          &                         globBC(iBC)%cornerBC%nElems(iLevel),&
          &                         QQ, .false. )

        if ( globBC(iBC)%cornerBC%nElems(iLevel) > 0 ) then
          write(logUnit(6),"(A,A,I2,A,I0)") ' BC: '//trim(globBC(iBC)%label),&
            &  ', level: ', iLevel, &
            &  ', corner nElems: ', globBC( iBC )%cornerBC%nElems( iLevel )
          globBC(iBC)%cornerBC%elems( iLevel )%elem(:) = 0
          globBC(iBC)%cornerBC%elems( iLevel )%bitmask = .false.
          globBC(iBC)%cornerBC%elems( iLevel )%normal  = 0
        end if

      end do !iLevel
    end do !iBnd
  end subroutine allocateBCList
! ****************************************************************************** !


! ****************************************************************************** !
  !> This routine assigns the BC lists
  !! Run over all the elements with the property boundary
  !! and check each direction.
  !! Assign all common boundaries to the level-wise representation
  !! 3rd step in build_BCLists
  subroutine assignBCList( tree, bc_prop, stencil, globBC )
    ! ---------------------------------------------------------------------------
    !> fluid tree from mesh
    type( treelmesh_type ), intent(in)         :: tree
    !> boundary information from mesh
    type( tem_bc_prop_type ), intent(in)       :: bc_prop
    !> stencil
    type( tem_stencilHeader_type ), intent(in) :: stencil
    !> boundaries for the elements with bnd property set
    type( glob_boundary_type ), intent(inout)  :: globBC(:)
    ! ---------------------------------------------------------------------------
    integer :: iElem, iDir, iBnd, iMeshDir
    integer(kind=long_k) :: bID
    integer :: normal(3), level, maxLevel, minLevel
    logical :: corner_bitmask( stencil%QQ )
    logical :: found( bc_prop%nBCtypes )
    integer :: iElem_qVal
    integer :: weight(stencil%QQ), length, posInTree
    ! number of elem for each BC on each lvl
    integer, allocatable :: nBnds(:,:), nCornerBnds(:,:)
    ! ---------------------------------------------------------------------------

    minLevel = tree%global%minLevel
    maxLevel = tree%global%maxLevel

    ! number of local elements of each boundary type in each level
    allocate( nBnds( minLevel:maxLevel, bc_prop%nBCtypes ))
    ! number of elements with multiple boundaries in each level
    allocate( nCornerBnds( minLevel:maxLevel, bc_prop%nBCtypes ))
    !Reset the counters
    nBnds       = 0
    nCornerBnds = 0
    iElem_qVal  = 0

    ! calculate weights
    ! Get length of currently treated direction and determine
    ! If it is one of the main coordinate axis
    do iDir = 1, stencil%QQN
      length =   stencil%cxDir( 1, iDir )**2 &
        &      + stencil%cxDir( 2, iDir )**2 &
        &      + stencil%cxDir( 3, iDir )**2

      ! For the normals, we weight the orthogonal directions
      ! of the stencil more heavily
      if( length == 1) then
        weight(iDir) = 4  ! axis orthogonal weighs twice as much as all others
      else if ( length == 2 ) then
        weight(iDir) = 2  ! non-axis orthogonal
      else ! length == 3
        weight(iDir) = 1  ! non-axis orthogonal
      end if
    end do

    do iElem = 1, tree%property(1)%nElems

      posInTree = tree%Property(1)%ElemID(iElem)
      ! Count the elem with qVal property to access the qVal from
      ! bc_prop%qVal
      if( btest( tree%elemPropertyBits( posInTree ), prp_hasQVal) ) &
        & iElem_qVal = iElem_qVal + 1

      found = .false.
      level = tem_LevelOf( tree%treeID( posInTree ) )
      ! run over each direction in the stencil
      do iDir = 1, stencil%QQN
        ! Stencil order of directions might differ from
        ! the treelm numbering. -> look up correct treelm direction number
        iMeshDir = stencil%map( iDir )
        ! Get the boundary ID as stored in the mesh
        bID      = bc_prop%boundary_ID( iMeshDir, iElem )
        ! Exclude no-boundary or periodic boundaries
        if( bID > 0  ) then
          ! If this boundary ID was not found for current element yet,
          ! assign to list
          if ( .not. found ( bID ) ) then
            ! count up and add, only if it has not been added before for iElem
            ! this is before all the helper elements have been identified
            ! therefore only the pointers to the original treeID list are
            ! added. This is done later in the update_BC routine.
            ! Wouldn't it be better to have a dynamic array here instead of
            ! doing it manually
            nBnds( level, bID ) = nBnds( level, bID )+1

            if( .not. allocated( globBC(bID)%elems(level)%elem))then
              write(dbgUnit(1),*) 'Error: bc not allctd lvl ', level,        &
                &                     ' bid ', bID
            end if
            ! Assign the position of the treeID in tree%treeID list
            ! to the boundary elem lists
            globBC( bID )%elems( level )%elem( nBnds( level, bID )) = &
              &                                   posInTree
            found( bID ) = .true.
          end if

          ! Set the bitmask for the incoming directions which have to be
          ! updated, which is opposed to BCid direction.
          globBC( bID )%elems( level )%bitmask(                       &
            & stencil%cxDirInv(iDir), nBnds( level, bID ) ) = .true.

          if( globBC( bID )%hasQVal ) then
            ! Set the qvalue from the boundary array
            ! q-values stored in outgoing direction
            call set_bouzidi_coeff( bc_prop%qVal(iMeshDir, iElem_qVal),     &
              & globBC( bID )%elems( level )%qVal( iDir, nBnds(level,bID)), &
              & globBC( bID )%elems( level )% cIn( iDir, nBnds(level,bID)), &
              & globBC( bID )%elems( level )%cOut( iDir, nBnds(level,bID)), &
              & globBC( bID )%elems( level )%cNgh( iDir, nBnds(level,bID)), &
              & globBC( bID )%elems( level )%cVel( iDir, nBnds(level,bID))  )
          end if ! hasQVal

          ! Add this boundary direction to the sum for the normal vector
          ! use the minus, as we want to go into the flow domain, and the
          ! stencil is pointing outside
          globBC( bID )%elems( level )%normal(:, nBnds( level, bID ))        &
            &  =  globBC( bID )%elems( level )%normal(:, nBnds( level, bID ))&
            &     -  weight(iDir) * stencil%cxDir( :, iDir )
        end if ! bID > 0
      end do  ! iDir

      ! if element is intersected by more than one boundary
      if ( count(found) > 1 ) then
        corner_bitmask = .false.
        normal = 0
        do iBnd = 1, bc_prop%nBCtypes
          if (found(iBnd)) then
            nCornerBnds( level, iBnd ) = nCornerBnds( level, iBnd ) + 1
            ! Assign the position of the treeID to the corner elem list
            globBC( iBnd )%cornerBC%elems( level )              &
              & %elem( nCornerBnds( level, iBnd ) ) = posInTree

            corner_bitmask = corner_bitmask .or. &
              & globBC( iBnd )%elems(level)%bitmask(:, nBnds(level, iBnd))
             normal = normal + &
              & globBC( iBnd )%elems(level)%normal(:, nBnds( level, iBnd ))
          endif
        end do

        do iBnd = 1, bc_prop%nBCtypes
          if (found(iBnd)) then
            globBC( iBnd )%cornerBC%elems( level )       &
              & %bitmask( :, nCornerBnds( level, iBnd ) ) = corner_bitmask
            globBC( iBnd )%cornerBC%elems( level )       &
              & %normal( :, nCornerBnds( level, iBnd ) )  = normal
          endif
        end do
      endif ! count(found) > 1
    end do ! iElem = 1, tree%property(1)%nElems

  end subroutine assignBCList
! ****************************************************************************** !


! ****************************************************************************** !
  !> This routine normalizes the normal vectors of boundary elements including
  !! the corner elements as well as assigns the corresponding prevailing 
  !! direction from the stencil
  subroutine normalizeBC( nBCs, minLevel, maxLevel, globBC, layout, field )
    ! ---------------------------------------------------------------------------
    !> number of boundaries
    integer,                        intent(in) :: nBCs, minLevel, maxLevel
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type),    intent(inout) :: globBC(:)
    !> scheme layout
    type( mus_scheme_layout_type ), intent(in) :: layout
    !> field type
    type( mus_field_type ),         intent(in) :: field(:)
    ! ---------------------------------------------------------------------------
    integer :: iBnd, iLevel, iElem, iDir, iField, counter
    real(kind=rk) :: angle, angleMax, oneDeginRad
    logical :: check_angle(nBCs)
    ! ---------------------------------------------------------------------------
    ! one degree in radian
    oneDegInRad = PI/180.0_rk

    ! check which boundary needs to check for angle
    check_angle = .false.
    do iBnd = 1, nBCs
      do iField = 1, size(field)
        if ( field(iField)%bc(iBnd)%nNeighs > 0 ) then
          check_angle(iBnd) = .true.
        end if
      end do
    end do

    ! normalize all boundary normals
    do iBnd = 1, nBCs
      angleMax = 0.0_rk
      counter = 0
      do iLevel = minLevel, maxLevel
        do iElem = 1, globBC( iBnd )%nElems( iLevel )
          call tem_determine_discreteVector(                         &
            &   globBC( iBnd )%elems( iLevel )%normal( :, iElem ),   &
            &   layout%prevailDir, angle )
          ! find maximum angle between vectors  
          angleMax = max(angleMax, abs(angle))
          ! count number of elements with angle > 1deg
          if (abs(angle) > oneDegInRad) counter = counter + 1
          ! Find the index in the stencil corresponding to the normal
          ! direction
          do iDir = 1, layout%fStencil%QQ
            if( layout%fStencil%cxDir(1, iDir ) ==                 &
              & globBC( iBnd )%elems( iLevel )%normal( 1, iElem )    &
              & .and.                                                &
              & layout%fStencil%cxDir(2, iDir ) ==                 &
              & globBC( iBnd )%elems( iLevel )%normal( 2, iElem )    &
              & .and. layout%fStencil%cxDir(3, iDir ) ==           &
              & globBC( iBnd )%elems( iLevel )%normal( 3, iElem ))   &
              & then
              globBC( iBnd )%elems( iLevel )%normalInd( iElem ) = iDir
            end if
          end do
        end do  ! iElem
      end do  ! iLevel
      ! Normal direction is required only to create stencil for 
      ! boundary elements which requires neighbors so check for
      ! angle only if any field bc has nNeigh > 0
      if ( check_angle(iBnd)  ) then
        ! convert angle in radian to degree
        angleMax = angleMax / oneDegInRad
        ! if angle is > 1 deg then print warning message
        if (angleMax > 1.0_rk) then
          write(logUnit(6),'(a,i0,a)') ' WARNING: Normal direction of ',&
            &  counter, ' elements of "'// trim(globBC(iBnd)%label)//'" boundary' 
          write(logUnit(6),*) 'deviates from stencil discrete vectors.'
          write(logUnit(6),'(a,f5.2)') ' Max. angle between bnd normal and '//&
            & 'discrete vector is:', angleMax
          write(logUnit(6),*) 'This might cause some problems in '// &
            & 'outlet BC.'
          write(logUnit(6),*) 'Solution: use outlet_pab'
        end if
      end if
    end do ! iBnd


    ! normalize normal of elements with multiple boundaries
    do iBnd = 1, nBCs
      do iLevel = minLevel, maxLevel
        do iElem = 1, globBC(iBnd)%cornerBC%nElems( iLevel )
          call tem_determine_discreteVector(                                   &
            &   globBC( iBnd)%cornerBC%elems( iLevel )%normal( :, iElem ),     &
            &   layout%prevailDir)
          ! Find the index in the stencil corresponding to the normal
          ! direction
          do iDir = 1, layout%fStencil%QQ
            if( layout%fStencil%cxDir(1, iDir) ==                            &
              & globBC(iBnd)%cornerBC%elems(iLevel)%normal(1, iElem)           &
              & .and.                                                          &
              & layout%fStencil%cxDir(2, iDir) ==                            &
              & globBC(iBnd)%cornerBC%elems(iLevel)%normal(2, iElem)           &
              & .and.                                                          &
              & layout%fStencil%cxDir(3, iDir) ==                            &
              & globBC(iBnd)%cornerBC%elems(iLevel)%normal(3, iElem))          &
              & then
            globBC(iBnd)%cornerBC%elems(iLevel)%normalInd(iElem) = iDir
            end if
          end do
        end do ! iElem
      end do ! iLevel
    end do ! iBnd

  end subroutine normalizeBC
! ****************************************************************************** !







! ****************************************************************************** !
  !> Create the communication buffers
  !!
  !! Assign the positions in state vectors, where to fetch from or save to
  !! Solver specific part to identify the actual data locations
  !! to send use within the communication buffers.
  !! @todo: this should be pattern specific, with typed exchange we
  !!        do not need to store the pos, as it is done inside MPI
  !!        with the defined datatype
  !!
  !! Access to state array: SAVE
  !! ---------------------------
  !! Why? Because we need to work on the pdfs on which the
  !! compute kernel was working. The kernel stored to SAVE
  !! All actions taking place on the state array after the kernel
  !! need to be performed with SAVE access
  !!
  subroutine init_recvBuffers( me, myHalos, pattern, requireAll, haloRequired, &
    &                          halo_offset, scheme, stat, neigh, nSize )
    ! ---------------------------------------------------------------------------
    !> receive communication buffer
    type(tem_communication_type), intent(inout) :: me
    !> contains bitmask of my halos with array size of recv%nProcs
    type(halo_commBitmask_type), intent(inout)  :: myHalos(:)
    !> communication pattern
    type(tem_commPattern_type), intent(in)      :: pattern
    !> requires complete element information or only required links?
    logical, intent(in) :: requireAll
    !> Level descriptor
    integer, intent(in) :: halo_offset
    !> Scheme information on fluid, boundary and flow properties
    type( mus_scheme_type ), intent(in)         :: scheme
    !> Statistics
    type(mus_statistics_type), intent(inout)    :: stat
    !> PDF neighbor array
    integer, intent(in) :: neigh(:)
    !> number of Elements in neigh array
    integer, intent(in) :: nSize
    logical, intent(in) :: haloRequired(:)
    ! ---------------------------------------------------------------------------
    ! positions, in state vectors, of links with fluid neighbors
    integer, allocatable :: pos(:)
    integer :: iElem, iDir, iProc, elemPos
    integer :: counter ! number of total links that acturally needs to update
    integer :: iField, nScalars, varPos, neighDir, QQ
    ! Include the current direction for the communciation
    logical :: includeInComm
    integer :: nghElem ! neighbor element position in total list
    integer :: neighPos ! position of neighbor in the neigh array list
    ! ---------------------------------------------------------------------------

    nScalars = scheme%varSys%nScalars
    QQ = scheme%layout%fStencil%QQ

    allocate(pos(maxval(me%nElemsProc*nScalars)))

    do iProc = 1, me%nProcs
      counter = 0
      do iElem = 1, me%nElemsProc( iProc )
        ! halo element position in the levelDesc total list
        elemPos = me%elemPos( iProc )%val( iElem )
        ! Here the check for the bitmasks has to be performed
        ! Check only pdf states. No need to communicate omega
        do iDir = 1, QQ
          includeInComm = .false.
          ! Now check all the neighbors of the current halo element
          ! - If we encounter fluid or ghost element, then the direction is
          !   required to exchange for      -> bitmask( iDir, iElem ) = .true.
          ! - If a halo element is ecountered, there is no need for
          !   communication                -> bitmask( iDir, iElem ) = .false.
          ! Get the element position of the neighbor element
          ! By accessing the neighbor array, get the position of the neighbor
          ! element's direction and reconstruct the neighbor element position
          ! in the total List
          neighDir = ?NgDir?( iDir, scheme%layout%fStencil)
          neighPos = ?NGPOS?( neighDir, elemPos, nSize )
          nghElem  = ?ElemIDX?( neigh( neighPos ), nScalars, nSize )
          ! Now check, what kind of element it is. Must be anything but halo.
          ! requireAll is true for IBM stencil or comm_reduced is false
          if( nghElem <= halo_offset .or. requireAll ) then
            includeInComm = .true.
          end if

          ! If it's a ghostHalo or boundary halo, then communicate all
          if( haloRequired( elemPos - halo_offset )) then
            includeInComm = .true.
          end if

          ! If current neighbor is a ghosthalo or boundary halo, 
          ! then communicate all
          if( nghElem > halo_offset ) then
            if( haloRequired( nghElem  - halo_offset )) then
              includeInComm = .true.
            end if
          end if

          ! Field loop is inside direction loop since neighbor element
          ! identification is independent of fields
          if( includeInComm ) then
            do iField = 1, scheme%nFields
              counter = counter + 1
              ! ? SAVE ? ... for regular halos
              ! ? IDX ? ... for ghost-halos
              ! - > These two are the same for PULL!! so, no difference.
              ! Why SAVE? Because we need to work on the pdfs on which the
              ! compute kernel was working. The kernel stored to SAVE
              ! All actions taking place on the state array after the kernel
              ! need to be performed with SAVE access
              pos(counter) = ?SAVE?(iDir,iField,elemPos,QQ,nScalars,nSize,neigh)

              varPos = scheme%varSys%method%val(                      &
                &             scheme%stateVarMap%varPos%val(iField) ) &
                &                   %state_varPos(iDir)
              myHalos( iProc )%bitMask( varPos, iElem ) = .true.
            end do ! field loop
          end if
        end do ! iDir
      end do ! iElem = 1, me%nElemsProc( iProc )

      stat%nLinks_total = stat%nLinks_total + scheme%nFields * QQ * me%nElemsProc(iProc)
      stat%nLinks_comm = stat%nLinks_comm + counter
      ! copy position array to me%pos, allocate me%val array
      write(dbgUnit(1),*) "Set recv buffer, iProc: ", iProc, ", counter: ", counter
      call pattern%initBuf_real( me%buf_real( iProc ), pos, counter )

    end do ! iProc
    deallocate(pos)

  end subroutine init_recvBuffers
! ****************************************************************************** !


! ****************************************************************************** !
  !> Create the communication buffers
  !!
  !! Receive buffers were before created. Now receive the buffers as
  !! send buffers from the remote processes.
  !! Assign the positions, where to fetch from and where to save to
  !! Solver specific part to identify the actual data locations
  !! to send use within the communication buffers.
  !! @todo: this should be pattern specific, with typed exchange we
  !!        do not need to store the pos, as it is done inside MPI
  !!        with the defined datatype
  !! @todo: does PULL and PUSH give the same results? as SAVE is used here
  !!
  !! Access to state array: SAVE
  !! ---------------------------
  !! Why? Because we need to work on the pdfs on which the
  !! compute kernel was working. The kernel stored to SAVE
  !! All actions taking place on the state array after the kernel
  !! need to be performed with SAVE access
  !!
  subroutine init_sendBuffers( send, remoteHalos, neigh, pattern, scheme, nSize )
    ! ---------------------------------------------------------------------------
    !> send communication buffer
    type(tem_communication_type), intent(inout) :: send
    !> contains bitmask of remote halos with array size send%nProcs
    type(halo_commBitmask_type), intent(in)     :: remoteHalos(:)
    !> neighbor array for state array
    integer, intent(in)                         :: neigh(:)
    !> communication pattern
    type(tem_commPattern_type), intent(in)      :: pattern
    !> fluid, boundary and flow information
    type( mus_scheme_type ), intent(in)         :: scheme
    integer, intent(in)                         :: nSize
    ! ---------------------------------------------------------------------------
    ! Element positions
    integer, allocatable :: pos(:)
    integer :: iElem ! count variable for elements
    integer :: iDir, iProc, counter, elemPos
    integer :: iField, nScalars, QQ, varPos
    ! ---------------------------------------------------------------------------

    nScalars = scheme%varSys%nScalars
    QQ = scheme%layout%fStencil%QQ
    allocate( pos(maxval(send%nElemsProc*nScalars)))
    ! allocate( send%buf_real( send%nProcs ))

    ! Then assign the send buffer positions
    do iProc = 1, send%nProcs
      counter = 0
      do iElem = 1, send%nElemsProc( iProc )
        ! Now assign the exact positions in the sendBuf
        elemPos = send%elemPos( iProc )%val( iElem )
        ! Check only pdf states. No need to communicate omega
        do iDir = 1, QQ
          ! Field loop is inside direction loop to maintain the order of pos
          ! as in init_recvBuffers
          do iField = 1, scheme%nFields
            varPos = scheme%varSys%method%val(                     &
              &            scheme%stateVarMap%varPos%val(iField) ) &
              &                  %state_varPos(iDir)
            if( remoteHalos( iProc )%bitMask( varPos, iElem )) then
              counter = counter + 1
              ! ? SAVE ? ... for regular halos
              ! ? IDX ? ... for ghost-halos
              ! -> These two are the same for PULL!! so, no difference.
              pos( counter ) = ?SAVE?(iDir,iField,elemPos,QQ,nScalars,nSize,neigh)
            endif
          end do ! field loop
        end do ! Dir loop
      end do ! iElem

      write(dbgUnit(1),*) "Set send buffer, iProc: ", iProc, ", counter: ", counter
      call pattern%initBuf_real( send%buf_real( iProc ), pos, counter )
    end do ! iProc
    write(dbgUnit(1),"(A)" ) ''

    deallocate(pos)

  end subroutine init_sendBuffers
! ****************************************************************************** !


! ****************************************************************************** !
  !> subroutine to find neighbours of element with individual (for each element)
  !! stencil definitions.
  !! Unique stencil label for boundary stencils are created with boundary label
  !! and stencil%cxDir therefore each stencil is limited to one boundary type
  subroutine mus_build_BCStencils( globBC, bc, prevailDir, prefix, minLevel,    &
    &                              maxLevel, stencil_labels, grwStencil )
    ! ---------------------------------------------------------------------------
    !> boundaries for the elements with bnd property set
    type(glob_boundary_type), intent(in)             :: globBC
    !> field boundary with boundary neighbor info
    type(boundary_type), intent(inout)               :: bc
    !> scheme layout
    real(kind=rk),                  intent(in)       :: prevailDir(:,:)
    !> field label
    character(len=*), intent(in)                     :: prefix
    !> min and max level
    integer,                        intent(in)       :: minLevel, maxLevel
    !> dynamic array of stencil labels
    type(dyn_labelArray_type), intent(inout)         :: stencil_labels
    !> growing array of stencils
    type(grw_stencilHeaderArray_type), intent(inout) :: grwStencil
    ! ---------------------------------------------------------------------------
    integer :: iElem, elemPos, iLevel, iNeigh, stencilPos
    integer :: normal(3)
    type(tem_stencilHeader_type) :: stencil
    logical :: wasAdded
    character(len=labelLen) :: stnLabel
    ! ---------------------------------------------------------------------------

    ! only need stencil when BC require neighbors
    if ( bc%nNeighs > 0 ) then

      stencil%label = trim(prefix)//trim(bc%label)
      stencil%useAll = .false.
      ! only true for nNeighs == 2?
      stencil%requireNeighNeigh = .true.
      stencil%QQ     = bc%nNeighs
      stencil%QQN    = bc%nNeighs
      allocate( stencil%cxDir( 3, stencil%QQN ))
      stencil%cxDir = 0

      write(logUnit(5),"(A)")    ' Set BC stencil: '//trim(stencil%label)
      write(logUnit(6),"(A,I0)") '             QQ: ', stencil%QQ

      allocate( stencil%elemLvl( minLevel:maxLevel ) )

      do iLevel = minLevel, maxLevel

        ! allocate field bc stencilPos to size of globBC nElems of that level
        allocate( bc%elems(iLevel)%stencilPos(    globBC%nElems( iLevel ) ) )
        allocate( bc%elems(iLevel)%posInNghElems( globBC%nElems( iLevel ) ) )

        do iElem = 1, globBC%nElems( iLevel )
          ! Get the normal direction of this boundary
          ! only for higher order boundaries
          normal(:) = globBC%elems( iLevel )%normal( :, iElem )
          ! And store the normal direction into the stencil
          ! QQN stands for the number of neighbors
          ! Here it means, how many neighbors in the normal direction
          do iNeigh = 1, stencil%QQN
            ! ?: what is the purpose of multiplying iStencilElem with normal
            !    direction?
            ! !: We are going along the normal direction for more than
            ! one neighbor in one direction (required for higher order boundary)
            stencil%cxDir(:, iNeigh) = normal * iNeigh
          end do !iNeigh

          ! Get unique stencil label from cxDir
          stnLabel = tem_stencil_getLabelForcxDir(  me         = stencil,   &
            &                                       prevailDir = prevailDir )

          ! extend stencil label with bc label such that
          ! a stencil is restricted strictly to one boundary type
          stnLabel = trim(stencil%label)//'_'//trim(stnLabel)

!wri  te(dbgUnit(1),*) 'stencil label ', trim(stnLabel)
          ! append stencil label to dynamic array of stencil labels,
          ! if label is new then wasAdded is true else it is false
          call append( me       = stencil_labels, &
            &          val      = stnlabel,       &
            &          pos      = stencilPos,     &
            &          wasAdded = wasAdded        )
!wri  te(dbgUnit(1),*) 'wasAdded ', wasAdded, ' stencilPos ', stencilPos
          if ( wasAdded ) then
            write(logUnit(10),"(A)") "stencil: "//trim(stnlabel)//" was added"
            write(logUnit(10),"(A,I0)") "stencilPos: ", stencilPos
          endif

          ! new stencil, append this stencil to stencil array
          ! and initialize elemLvl and elem in stencil type
          if ( wasAdded ) then
            write(logUnit(10),"(A)") "append stencil to grwStencil"
            call append( me  = grwStencil, &
              &          val = stencil     )

            ! initialize elemLvl and elem list for new stencil
            write(logUnit(10),"(A)") "init grwStencil%elemLvl"
            if ( .not. allocated(grwStencil%val(stencilPos)%elemLvl) ) then
              write(logUnit(10),"(A)") "grwStencil%elemlvl not allocated"
              ! @todo: not allocated for dynamic load balance, needs check!
              allocate( grwStencil%val(stencilPos)%elemLvl( minLevel:maxLevel ) )
            end if
            call init( me     = grwStencil%val(stencilPos)%elemLvl( iLevel ), &
              &        length = 8                                              )
            write(logUnit(10),"(A)") "init grwStencil%elem"
            call init( me     = grwStencil%val(stencilPos)%elem, &
              &        length = 8                                 )

          end if

          ! Get the position of the treeID in the primal mesh
          elemPos = globBC%elems( iLevel )%elem( iElem )

          ! append element position in treeID list
          ! to level wise list and level independent list
          call append( me  = grwStencil%val(stencilPos)%elemLvl( iLevel ), &
            &          val = elemPos                                        )
          call append( me  = grwStencil%val(stencilPos)%elem, &
            &          val = elemPos                           )

          ! store stencil position current boundary element
          bc%elems(iLevel)%stencilPos(iElem) = stencilPos
        end do !iElem
      end do !iLevel

      deallocate( stencil%elemLvl )

    end if ! bc%nNeighs > 0


  end subroutine mus_build_BCStencils
! ****************************************************************************** !


! ****************************************************************************** !
  !> This routine build and append IBM stencils to scheme stencil array
  subroutine mus_build_IBMStencils( globIBM, layout, grwStencil )
    ! ---------------------------------------------------------------------------
    !> datatype to store the surface information
    type( mus_IBM_globType ), intent(inout) :: globIBM
    !> scheme stencil layout
    type( mus_scheme_layout_type ), intent(inout) :: layout
    !> contains array of stencils
    type( grw_stencilHeaderArray_type ), intent(inout) :: grwStencil
    ! ---------------------------------------------------------------------------
    type( tem_stencilHeader_type ) :: IBM_stencil
    character(len=labelLen) :: label
    integer :: iIBM, stencilPos
    logical :: wasAdded
    ! ---------------------------------------------------------------------------
    if (globIBM%nIBMs > 0) &
      write(logUnit(10),*) 'Build and append IBM stencil'

    do iIBM = 1, globIBM%nIBMs
      if( globIBM%IBM(iIBM)%movPredef )then
        call tem_create_d3q81( IBM_stencil )
      else
        call tem_create_d3q125( IBM_stencil )
      end if

      ! Get unique stencil label from cxDir
      !label = tem_stencil_getLabelForcxDir( me         = IBM_stencil,      &
      !  &                                   prevailDir = layout%prevailDir )
      label = IBM_stencil%label

      ! append stencil label to dynamic array of stencil labels,
      ! if label is new then wasAdded is true else it is false
      call append( me       = layout%stencil_labels, &
        &          val      = label,                 &
        &          pos      = stencilPos,            &
        &          wasAdded = wasAdded               )

      ! if stencil is already there then just store the position
      ! else append stencil to stencil array and store its position
      if ( wasAdded ) then
        call append( me  = grwStencil, &
          &          val = IBM_stencil )
      end if
      ! store position of IBM stencil in stencil array
      globIBM%IBM(iIBM)%stencilPos = stencilPos

      write(logUnit(10),*) 'Stored stencil for IBM ', iIBM, ' at position ',   &
        &                 globIBM%IBM(iIbm)%stencilPos
    end do
  end subroutine mus_build_IBMStencils
! ****************************************************************************** !


! ****************************************************************************** !
  !> set the sendHalo bit for all fluid elements which are send
  !! to remote procs
  subroutine set_sendHaloBits( property, sendBuffer )
    ! ---------------------------------------------------------------------------
    type( tem_communication_type ), intent(in) :: sendBuffer
    integer(kind=long_k), intent(inout) :: property(:)
    ! ---------------------------------------------------------------------------
    integer :: iProc, iElem, pos
    ! ---------------------------------------------------------------------------

    ! loop over the integer send buffers ...
    do iProc = 1, sendBuffer%nProcs
      ! ... and over the corresponding elements ...
      do iElem = 1, sendBuffer%elemPos(iProc)%nVals
        pos = sendBuffer%elemPos(iProc)%val(iElem)
        ! ... and set the property bits
        property(pos) = ibset(property(pos), prp_sendHalo )
      end do
    end do

  end subroutine set_sendHaloBits
! ****************************************************************************** !

! ****************************************************************************** !
  !> Find source elements for ghostFromCoarser element using compact stencil
  !!
  subroutine find_fromCoarserDep_compact( levelDesc, stencil, &
      &                                   nSourcesFromCoarser, &
      &                                   minLevel, maxLevel  )
    ! ---------------------------------------------------------------------------
    !> Min and max level
    integer,                        intent(in) :: minLevel, maxLevel
    !> Level Descriptor
    type( tem_levelDesc_type )                 :: levelDesc(minLevel:maxLevel)
    !> Number of source elements required by high order interpolation.
    !! If not enough source elements can be found, mark the ghost as low order.
    integer,                        intent(in) :: nSourcesFromCoarser
    type( tem_stencilHeader_type ), intent(in) :: stencil
    ! ---------------------------------------------------------------------------
    integer :: sourceLevel  ! level of source elements
    integer :: targetLevel  ! level of target elements
    integer :: iElem, ii
    integer :: nFound, nGhostElems
    integer :: parentPosInTotal, childNum, ghostPosInTotal
    real(kind=rk) :: targetBary(3) ! barycenter of current target element
    integer, allocatable :: mySources(:)
    real(kind=rk) :: childCoord(3,8) = reshape( [ 0.75_rk, 0.75_rk, 0.75_rk, &
      &                                        0.25_rk, 0.75_rk, 0.75_rk, &
      &                                        0.75_rk, 0.25_rk, 0.75_rk, &
      &                                        0.25_rk, 0.25_rk, 0.75_rk, &
      &                                        0.75_rk, 0.75_rk, 0.25_rk, &
      &                                        0.25_rk, 0.75_rk, 0.25_rk, &
      &                                        0.75_rk, 0.25_rk, 0.25_rk, &
      &                                        0.25_rk, 0.25_rk, 0.25_rk], [3,8] )
    ! ---------------------------------------------------------------------------

    write(logUnit(1),"(A)") 'Finding source for ghostFromCoarser using compact stencil'

    allocate( mySources( nSourcesFromCoarser ))

    do sourceLevel = minLevel, maxLevel - 1

      targetLevel = sourceLevel + 1
      nGhostElems = levelDesc( targetLevel )%elem%nElems( eT_ghostFromCoarser )

      ! Treat all ghostFromCoarser elements
      do iElem = 1, nGhostElems

        parentPosInTotal = &
          &  levelDesc( targetLevel )%depFromCoarser( iElem )%elem%val(1)
        ghostPosInTotal = iElem + levelDesc( targetLevel )%offset( 1, eT_ghostFromCoarser)

        targetBary = levelDesc( targetLevel )%depFromCoarser( iElem )%coord(:)
        childNum = levelDesc( targetLevel )%depFromCoarser( iElem )%childNum
        levelDesc( targetLevel )%depFromCoarser( iElem )%coord(:) = &
          & childCoord(1:3, childNum )

        ! --------------- treat parent element (first source) --------------
        call find_neigh_compact( &
          &    parent = parentPosInTotal, &
          &    nNeighs = nSourcesFromCoarser, &
          &    nghElems = levelDesc( sourceLevel )%neigh(1)%nghElems, &
          &    neighDir = compactNeighDir(1:nSourcesFromCoarser, childNum ), &
          &    stencil = stencil, &
          &    targetBary = targetBary, &
          &    nFound = nFound, mySources = mySources, &
          &    weight = levelDesc( targetLevel )%depFromCoarser( iElem )%weight )

        levelDesc( targetLevel )%depFromCoarser( iElem )%parentPosInSource = -1
        do ii = 1, nFound
          if( mySources(ii) == parentPosInTotal ) then
            levelDesc( targetLevel )%depFromCoarser( iElem )%parentPosInSource = &
              &  ii
            exit
          end if
        end do

! ------------------------------------------------------------------------------
write(dbgUnit(5),"(A,I0)") 'ghostID: ', &
  &                        levelDesc( targetLevel )%total( ghostPosInTotal )
write(dbgUnit(5),"(A,I0)") 'parentID: ', &
  &                        levelDesc( sourceLevel )%total( parentPosInTotal )
write(dbgUnit(5),"(A,I0)") 'childNum: ', childNum
write(dbgUnit(5),"(A,3F6.3)") 'coord: ', childCoord(1:3, childNum )
! write(dbgUnit(5),"(2A)") 'sourceIDs: ', &
!   &  trim(tem_toStr( levelDesc(sourceLevel)%total(mySources(1:nFound)),','))
write(dbgUnit(5),"(A,4I10)") 'sourceIDs: ', &
  &  levelDesc(sourceLevel)%total(mySources(1:nFound))
! ------------------------------------------------------------------------------


        ! Assign to high or low list
        if ( nFound == nSourcesFromCoarser ) then
          call append( me  = levelDesc( targetLevel )%intpFromCoarser_high, &
            &          val = iElem )
        else
          call append( me  = levelDesc( targetLevel )%intpFromCoarser_low, &
            &          val = iElem )
        end if

        call depSource_append(                                               &
          &    me         = levelDesc( targetLevel )%depFromCoarser( iElem ),&
          &    sourceList = levelDesc( targetLevel )%sourceFromCoarser,      &
          &    mySources  = mySources(1:nFound),                      &
          &    n          = nFound )

      end do ! iElem = 1, nGhostElems

    end do ! level loop

  end subroutine find_fromCoarserDep_compact
! ****************************************************************************** !

! ****************************************************************************** !
  subroutine find_neigh_compact( parent, nNeighs, nghElems, neighDir, stencil, &
    &                            targetBary, nFound, mySources, weight )
    ! ---------------------------------------------------------------------------
    !> Parent position in total list
    integer, intent(in) :: parent
    !> Neighbor list for all elements
    integer, intent(in) :: nghElems(:,:)
    !> number of neighbors
    integer, intent(in) :: nNeighs
    integer :: neighDir(nNeighs)
    !> child bary relative to parent
    real(kind=rk), intent(in) :: targetBary(3)
    !> stencil
    type( tem_stencilHeader_type ), intent(in) :: stencil
    !> number of source elements actually found
    integer, intent(out) :: nFound
    !> position of found source elements in total list
    integer, intent(out) :: mySources(:)
    !> weights of source elements
    real(kind=rk), allocatable, intent(out) :: weight(:)
    ! ---------------------------------------------------------------------------
    integer :: iNeigh, posInTotal
    ! resulting distance of bary centers from target to uncle
    real(kind=rk) :: distance(3)
    ! intermediate storage, used for recovery when 1st round search fails
    real(kind=rk) :: temp_weights(nNeighs), sumWeight
    ! ---------------------------------------------------------------------------

call tem_horizontalSpacer( fUnit = dbgUnit(7), before = 1 )
write(dbgUnit(7),"(A)") 'Get into find_neigh_compact'
write(dbgUnit(7),"(A)") 'to Find all coarse sources for a given fine ghost'

    nFound       = 0
    mySources    = 0
    temp_weights = 0._rk

    do iNeigh = 1, nNeighs

write(dbgUnit(7),"(A, I0)") '      iNeigh: ', iNeigh
write(dbgUnit(7),"(A, I0)") '    neighDir: ', neighDir(iNeigh)
write(dbgUnit(7),"(A,3I3)") ' NeighOffset: ', stencil%cxDir(:, neighDir(iNeigh))

      ! if ( foundDir(iNeigh, childNum) == Stencil%restPosition ) then
      if ( neighDir(iNeigh) == Stencil%restPosition ) then
        ! normally parent is the last source due to stencil defination
        posInTotal = parent
      else
        ! this is a neihgbor of the parent element ->
        ! access neighbor information.
        posInTotal = nghElems( neighDir(iNeigh), parent )
      end if

write(dbgUnit(7),"(A, I0)") '  posInTotal: ', posInTotal

      if ( posInTotal > 0 ) then ! this neighbor exist
        nFound = nFound + 1
        mySources(nFound) = posInTotal

        ! distance from this source to target
        ! coordinates of source is indeed its offset to parent of target
        ! For 2D: use only X and Y
        distance(1:3) = abs( dble(stencil%cxDir(1:3, neighDir(iNeigh))) - targetBary(1:3) )

        ! weight is inverser proportional to distance
        temp_weights( nFound ) =   ( 1._rk - distance(1) ) &
          &                      * ( 1._rk - distance(2) ) &
          &                      * ( 1._rk - distance(3) )

write(dbgUnit(7),"(A, I0)")   '      nFound: ', nFound
write(dbgUnit(7),"(A,3F7.3)") '    distance: ', distance(1:3)
write(dbgUnit(7),"(A,3F7.3)") '      weight: ', temp_weights(nFound)

      end if ! posInTotal > 0, i.e. this source is a valid one
write(dbgUnit(7),"(A)") ''
    end do ! iNeigh

    if ( nFound /= 0 ) then
      ! normalize weights
      allocate( weight( nFound ))
      sumWeight = sum( temp_weights( 1:nFound ))
      weight(:) = temp_weights( 1:nFound ) / sumWeight
    else ! nFound == 0
      write(dbgUnit(1),*) 'Can NOT find any source for this ghostFromCoarser.'
      write(dbgUnit(1),*) 'Thus can NOT do interpolation.'
      write(logUnit(1),*) 'Can NOT find any source for this ghostFromCoarser.'
      write(logUnit(1),*) 'Thus can NOT do interpolation. Check debug file'
      call tem_abort()
    end if

    call tem_horizontalSpacer( fUnit = dbgUnit(7), after = 1 )

  end subroutine find_neigh_compact
! ****************************************************************************** !

! ****************************************************************************** !
  !> All sources (children) have been found in treelm,
  !! updated number of sources needed, based on nDims
  !! collect all source elements into sourceFromFiner
  !! assign ghost to low or high order intp list, based on
  !!
  subroutine mus_update_fromFinerDep_3D( levelDesc, nSourcesFromFiner, &
    &                                    minLevel, maxLevel )
    ! ---------------------------------------------------------------------------
    integer, intent(in) :: minLevel
    integer, intent(in) :: maxLevel
    type( tem_levelDesc_type ) :: levelDesc(minLevel:maxLevel)
    !> When number of sources is lower than this,
    !! mark it as low order interpolation
    integer, intent(in) :: nSourcesFromFiner
    ! logical, intent(in) :: intpCompact
    ! ---------------------------------------------------------------------------
    integer :: iLevel, iElem
    integer :: nSources   ! number of source elements actually appended
    integer :: mySources(8)
    ! ---------------------------------------------------------------------------

    write(logUnit(1),"(A)") 'Updating dependencies of ghost from finer 3D.'

    ! all levels except highest level
    do iLevel = minLevel, maxLevel-1

      call init( me     = levelDesc( iLevel )%intpFromFiner_high,                 &
        &        length = levelDesc( iLevel )%elem%nElems( eT_ghostFromFiner ) )

      do iElem = 1, levelDesc( iLevel )%elem%nElems( eT_ghostFromFiner )

        if ( levelDesc( iLevel )%depFromFiner( iElem )%elem%nVals == 8 ) then
        !   &  .and. intpCompact ) then
          nSources = nSourcesFromFiner
          mySources(1:nSources) = &
            & [ levelDesc( iLevel )%depFromFiner( iElem )%elem%val(1), &
            &   levelDesc( iLevel )%depFromFiner( iElem )%elem%val(4), &
            &   levelDesc( iLevel )%depFromFiner( iElem )%elem%val(6), &
            &   levelDesc( iLevel )%depFromFiner( iElem )%elem%val(7)  ]
          call append( me  = levelDesc( iLevel )%intpFromFiner_high, &
            &          val = iElem )
        ! else if ( levelDesc( iLevel )%depFromFiner( iElem )%elem%nVals == 8 ) then
        !   nSources = levelDesc( iLevel )%depFromFiner( iElem )%elem%nVals
        !   mySources(1:nSources) = &
        !     & levelDesc( iLevel )%depFromFiner( iElem )%elem%val(1:nSources)
        !   call append( me  = levelDesc( iLevel )%intpFromFiner_high, &
        !     &          val = iElem )
        else
          nSources = levelDesc( iLevel )%depFromFiner( iElem )%elem%nVals
          mySources(1:nSources) = &
            & levelDesc( iLevel )%depFromFiner( iElem )%elem%val(1:nSources)
          call append( me  = levelDesc( iLevel )%intpFromFiner_low, &
            &          val = iElem )
        end if

        levelDesc( iLevel )%depFromFiner( iElem )%coord = [0.5_rk, 0.5_rk, 0.5_rk]
        call depSource_append( &
          &    me = levelDesc( iLevel )%depFromFiner( iElem ), &
          &    sourceList = levelDesc( iLevel )%sourceFromFiner, &
          &    mySources = mySources, &
          &    n = nSources )

      end do ! elem%nElems( eT_ghostFromFiner )

    end do ! iLevel = minLevel, maxLevel-1

  end subroutine mus_update_fromFinerDep_3D
! ****************************************************************************** !


end module mus_construction_module
! ****************************************************************************** !
