! See copyright notice in the COPYRIGHT file.
! ****************************************************************************** !
!> This module deals with the calculation of moments from pdfs
!!
module mus_moments_module

  ! include treelm modules
  use env_module,         only: rk, labelLen
  use tem_aux_module,     only: tem_abort
  use tem_math_module,    only: invert_matrix
  use tem_logging_module, only: logUnit

  ! include musubi modules
  use mus_moments_type_module,  only: mus_moment_type

  implicit none

  private

  public :: get_moment
  public :: get_momentVector
  public :: mus_init_moments
  public :: set_momentIndices

 contains

! ****************************************************************************** !
  !> Initialize the moment space
  !!
  subroutine mus_init_moments( me, QQ, cxDir, label )
    ! ---------------------------------------------------------------------------
    type( mus_moment_type ) :: me
    !>
    integer, intent(in) :: QQ
    integer, intent(in) :: cxDir(3, QQ)
    character(len=labelLen) :: label
    ! ---------------------------------------------------------------------------

    write(logUnit(1),"(A)") 'Initializing the moment matrix by stencil: '&
      &                     //trim(label)

    ! Set both entries for the matrix to the stencil size (quadratic matrix)
    write(logUnit(10),"(A,I0)") 'Set nEntries: ', QQ
    me%toMoments%nEntries = QQ
    me%toPdf%nEntries = QQ

    write(logUnit(10),"(A,2I4)") 'allocate toMoment: ', &
      &                         me%toPdf%nEntries(1), me%toPdf%nEntries(2)
    if ( allocated( me%toMoments%A ) ) deallocate( me%toMoments%A )
    if ( allocated( me%toPdf%A     ) ) deallocate( me%toPdf%A     )
    allocate( me%toMoments%A( me%toMoments%nEntries(1), &
      &                       me%toMoments%nEntries(2) ))
    allocate( me%toPdf%A( me%toPdf%nEntries(1), me%toPdf%nEntries(2) ))

    ! Allocate and reset the moment labels
    write(logUnit(10),"(A,I0)") 'allocate momLabel: ', QQ
    if(allocated( me%momLabel )) deallocate( me%momLabel )
    allocate( me%momLabel(QQ))

    me%momLabel = ''
    call init_transformation_matrix( QQ, cxDir, label, me, me%toMoments%A, me%toPdf%A )

  end subroutine mus_init_moments
! ****************************************************************************** !


! ****************************************************************************** !
  !> Intialize the moment transformation matrix
  !!
  subroutine init_transformation_matrix( QQ, cxDir, label, me, toMoment, toPdf )
    ! ---------------------------------------------------------------------------
    !>
    integer, intent(in) :: QQ
    integer, intent(in) :: cxDir(3, QQ)
    character(len=labelLen) :: label
    type( mus_moment_type ), intent(inout) :: me
    real(kind=rk), intent(inout) :: toMoment( me%toMoments%nEntries(1),&
      &                                       me%toMoments%nEntries(2) )
    real(kind=rk), intent(inout) :: toPdf( me%toPDF%nEntries(1), &
      &                                    me%toPDF%nEntries(2)  )
    ! ---------------------------------------------------------------------------
    logical :: requireMom
    ! ---------------------------------------------------------------------------

    write(logUnit(10),"(A)") 'init transformation matrix'
    ! require moments for stencil
    requireMom = .true.

    ! Set the moments one after another on the matrix. 
    ! The moment description might come from some place else.
    ! Lua? -> check mus_scheme_type_module the type(mus_momentSpace_type)
    select case( trim(label) )
    case('d2q9')
      me%momLabel(1) = 'density'
      toMoment( 1,:) = get_momentVector( QQ,cxDir, (/0,0,0/)) ! m0
      me%momLabel(2) = 'momX'
      toMoment( 2,:) = get_momentVector( QQ,cxDir, (/1,0,0/)) ! mx
      me%momLabel(3) = 'momY'
      toMoment( 3,:) = get_momentVector( QQ,cxDir, (/0,1,0/)) ! my
      me%momLabel(4) = 'momXX'
      toMoment( 4,:) = get_momentVector( QQ,cxDir, (/2,0,0/)) ! mxx
      me%momLabel(5) = 'momYY'
      toMoment( 5,:) = get_momentVector( QQ,cxDir, (/0,2,0/)) ! myy
      me%momLabel(6) = 'momXY'
      toMoment( 6,:) = get_momentVector( QQ,cxDir, (/1,1,0/)) ! mxy
      me%momLabel(7) = 'momXXY'
      toMoment( 7,:) = get_momentVector( QQ,cxDir, (/2,1,0/)) ! mxxy
      me%momLabel(8) = 'momXYY'
      toMoment( 8,:) = get_momentVector( QQ,cxDir, (/1,2,0/)) ! mxyy
      me%momLabel(9) = 'momXXYY'
      toMoment( 9,:) = get_momentVector( QQ,cxDir, (/2,2,0/)) ! mxxyy
      ! set moments positions
      !position of each moments pos in D3Q19 QQ,cxDir
      allocate(me%first_moments(2))
      me%first_moments = (/ 1, 2 /)
      allocate(me%second_moments(3))
      me%second_moments = (/ 1, 2, 4 /)
      allocate(me%third_moments(2))
      me%third_moments = (/ 1, 3 /)
      allocate(me%fourth_moments(1))
      me%fourth_moments = (/ 1 /)

    case('d2q9_reference')

      toMoment(1,:) = (/ 1._rk,  1._rk,  1._rk,  1._rk,  1._rk,  1._rk,  1._rk,  1._rk,  1._rk /)
      toMoment(2,:) = (/-1._rk,  0._rk,  1._rk,  0._rk, -1._rk, -1._rk,  1._rk,  1._rk,  0._rk /)
      toMoment(3,:) = (/ 0._rk, -1._rk,  0._rk,  1._rk, -1._rk,  1._rk, -1._rk,  1._rk,  0._rk /)
      toMoment(4,:) = (/ 1._rk,  0._rk,  1._rk,  0._rk,  1._rk,  1._rk,  1._rk,  1._rk,  0._rk /)
      toMoment(5,:) = (/ 0._rk,  1._rk,  0._rk,  1._rk,  1._rk,  1._rk,  1._rk,  1._rk,  0._rk /)
      toMoment(6,:) = (/ 0._rk,  0._rk,  0._rk,  0._rk,  1._rk, -1._rk, -1._rk,  1._rk,  0._rk /)
      toMoment(7,:) = (/ 0._rk,  0._rk,  0._rk,  0._rk, -1._rk,  1._rk, -1._rk,  1._rk,  0._rk /)
      toMoment(8,:) = (/ 0._rk,  0._rk,  0._rk,  0._rk, -1._rk, -1._rk,  1._rk,  1._rk,  0._rk /)
      toMoment(9,:) = (/ 0._rk,  0._rk,  0._rk,  0._rk,  1._rk,  1._rk,  1._rk,  1._rk,  0._rk /)

      ! set moments positions
      !position of each moments pos in D3Q19 QQ,cxDir
      allocate(me%first_moments( 2))
      allocate(me%second_moments(3))
      allocate(me%third_moments( 2))
      allocate(me%fourth_moments(1))
      me%first_moments  = (/ 1, 2 /)
      me%second_moments = (/ 1, 2, 4 /)
      me%third_moments  = (/ 1, 3 /)
      me%fourth_moments = (/ 1 /)

    case('d3q19')

      me%momLabel(1) = 'density'
      toMoment( 1,:) = get_momentVector( QQ,cxDir, (/0,0,0/)) ! m0
      me%momLabel(2) = 'momX   '
      toMoment( 2,:) = get_momentVector( QQ,cxDir, (/1,0,0/)) ! mx
      me%momLabel(3) = 'momY   '
      toMoment( 3,:) = get_momentVector( QQ,cxDir, (/0,1,0/)) ! my
      me%momLabel(4) = 'momZ   '
      toMoment( 4,:) = get_momentVector( QQ,cxDir, (/0,0,1/)) ! mz
      me%momLabel(5) = 'momXX  '
      toMoment( 5,:) = get_momentVector( QQ,cxDir, (/2,0,0/)) ! mxx
      me%momLabel(6) = 'momYY  '
      toMoment( 6,:) = get_momentVector( QQ,cxDir, (/0,2,0/)) ! myy
      me%momLabel(7) = 'momZZ  '
      toMoment( 7,:) = get_momentVector( QQ,cxDir, (/0,0,2/)) ! mzz
      me%momLabel(8) = 'momXY  '
      toMoment( 8,:) = get_momentVector( QQ,cxDir, (/1,1,0/)) ! mxy
      me%momLabel(9) = 'momYZ  '
      toMoment( 9,:) = get_momentVector( QQ,cxDir, (/0,1,1/)) ! myz
      me%momLabel(10) = 'momXZ'
      toMoment(10,:) = get_momentVector( QQ,cxDir, (/1,0,1/)) ! mzx
      me%momLabel(11) = 'momXXY '
      toMoment(11,:) = get_momentVector( QQ,cxDir, (/2,1,0/)) ! mxxy
      me%momLabel(12) = 'momXXZ '
      toMoment(12,:) = get_momentVector( QQ,cxDir, (/2,0,1/)) ! mxxz
      me%momLabel(13) = 'momYYX '
      toMoment(13,:) = get_momentVector( QQ,cxDir, (/1,2,0/)) ! myyx
      me%momLabel(14) = 'momYYZ '
      toMoment(14,:) = get_momentVector( QQ,cxDir, (/0,2,1/)) ! myyz
      me%momLabel(15) = 'momZZX '
      toMoment(15,:) = get_momentVector( QQ,cxDir, (/1,0,2/)) ! mzzx
      me%momLabel(16) = 'momZZY '
      toMoment(16,:) = get_momentVector( QQ,cxDir, (/0,1,2/)) ! mzzy
      me%momLabel(17) = 'momXXYY'
      toMoment(17,:) = get_momentVector( QQ,cxDir, (/2,2,0/)) ! mxxyy
      me%momLabel(18) = 'momYYZZ'
      toMoment(18,:) = get_momentVector( QQ,cxDir, (/0,2,2/)) ! myyzz
      me%momLabel(19) = 'momZZXX'
      toMoment(19,:) = get_momentVector( QQ,cxDir, (/2,0,2/)) ! mzzxx
      ! set moments positions
      !position of each moments pos in D3Q19 layout
      allocate(me%first_moments(3))
      me%first_moments = (/ 1, 2, 3 /)
      allocate(me%second_moments(6))
      me%second_moments = (/ 1, 2, 3, 4, 5, 6/)
      allocate(me%third_moments(6))
      me%third_moments = (/ 1, 2, 3, 4, 5, 6/)
      allocate(me%fourth_moments(3))
      me%fourth_moments = (/ 1, 2, 3 /)

    case('d3q27')

      me%momLabel(1) = 'density'
      toMoment( 1,:) = get_momentVector( QQ,cxDir, [0,0,0]) ! m0
      me%momLabel(2) = 'momX   '
      toMoment( 2,:) = get_momentVector( QQ,cxDir, [1,0,0]) ! mx
      me%momLabel(3) = 'momY   '
      toMoment( 3,:) = get_momentVector( QQ,cxDir, [0,1,0]) ! my
      me%momLabel(4) = 'momZ   '
      toMoment( 4,:) = get_momentVector( QQ,cxDir, [0,0,1]) ! mz
      me%momLabel(5) = 'momXX  '
      toMoment( 5,:) = get_momentVector( QQ,cxDir, [2,0,0]) ! mxx
      me%momLabel(6) = 'momYY  '
      toMoment( 6,:) = get_momentVector( QQ,cxDir, [0,2,0]) ! myy
      me%momLabel(7) = 'momZZ  '
      toMoment( 7,:) = get_momentVector( QQ,cxDir, [0,0,2]) ! mzz
      me%momLabel(8) = 'momXY  '
      toMoment( 8,:) = get_momentVector( QQ,cxDir, [1,1,0]) ! mxy
      me%momLabel(9) = 'momYZ  '
      toMoment( 9,:) = get_momentVector( QQ,cxDir, [0,1,1]) ! myz
      me%momLabel(10) = 'momXZ'
      toMoment(10,:) = get_momentVector( QQ,cxDir, [1,0,1]) ! mzx

      toMoment(11,:) = get_momentVector( QQ,cxDir, [2,1,0]) ! mxxy
      toMoment(12,:) = get_momentVector( QQ,cxDir, [2,0,1]) ! mxxy
      toMoment(13,:) = get_momentVector( QQ,cxDir, [1,2,0]) ! mxxy
      toMoment(14,:) = get_momentVector( QQ,cxDir, [0,2,1]) ! mxxy
      toMoment(15,:) = get_momentVector( QQ,cxDir, [1,0,2]) ! mxxy
      toMoment(16,:) = get_momentVector( QQ,cxDir, [0,1,2]) ! mxxy
      toMoment(17,:) = get_momentVector( QQ,cxDir, [1,1,1]) ! mxxy

      toMoment(18,:) = get_momentVector( QQ,cxDir, [2,2,0]) ! mxxy
      toMoment(19,:) = get_momentVector( QQ,cxDir, [2,0,2]) ! mxxy
      toMoment(20,:) = get_momentVector( QQ,cxDir, [0,2,2]) ! mxxy
      toMoment(21,:) = get_momentVector( QQ,cxDir, [2,1,1]) ! mxxy
      toMoment(22,:) = get_momentVector( QQ,cxDir, [1,2,1]) ! mxxy
      toMoment(23,:) = get_momentVector( QQ,cxDir, [1,1,2]) ! mxxy

      toMoment(24,:) = get_momentVector( QQ,cxDir, [2,2,1]) ! mxxy
      toMoment(25,:) = get_momentVector( QQ,cxDir, [2,1,2]) ! mxxy
      toMoment(26,:) = get_momentVector( QQ,cxDir, [1,2,2]) ! mxxy

      toMoment(27,:) = get_momentVector( QQ,cxDir, [2,2,2]) ! mxxy
    case default
      requireMom = .false.
    end select

    if ( requireMom ) then
      toPdf = invert_matrix( toMoment )
    end if

  end subroutine init_transformation_matrix
! ****************************************************************************** !


! ****************************************************************************** !
  !> The integer moment vector for a given cxDir and order.
  !!
  !! Assuming 0**0 = 1 here.
  !!
  pure function mus_iMomVector( cxDir, expX, QQ ) result(iMom)
    ! ---------------------------------------------------------------------------
    !> number of velocity channels (include rest)
    integer, intent(in) :: QQ
    !> discrete velocity
    integer, intent(in) :: cxDir(:,:)
    !> order in each direction
    integer, intent(in) :: expX(3)
    !>
    integer :: iMom(QQ)
    ! ---------------------------------------------------------------------------
    integer :: nIndices
    integer :: iVal
    integer :: iX
    integer :: X_idx(3)
    integer :: xx(3)
    ! ---------------------------------------------------------------------------

    nIndices = 0
    do iX=1,3
      if (expX(iX) > 0) then
        nIndices = nIndices + 1
        X_idx(nIndices) = iX
        xx(nIndices) = expX(iX)
      end if
    end do

    select case(nIndices)
    case(0)
      imom = 1
    case(1)
      do iVal=1,QQ
        imom(iVal) = cxDir(X_idx(1),iVal)**xx(1)
      end do
    case(2)
      do iVal=1,QQ
        imom(iVal) = cxDir(X_idx(1),iVal)**xx(1) &
          &        * cxDir(X_idx(2),iVal)**xx(2)
      end do
    case(3)
      do iVal=1,QQ
        imom(iVal) = cxDir(1,iVal)**xx(1) &
          &        * cxDir(2,iVal)**xx(2) &
          &        * cxDir(3,iVal)**xx(3)
      end do
    end select

  end function mus_iMomVector
! ****************************************************************************** !


! ****************************************************************************** !
  !> Calculate the moment of a centain order
  !! The moment of a distribution \( f_i \) is defined as:\n
  !! \[
  !!    m_{x^{p}y^{q}z^{r}} = \sum_{i}^{Q} c^{p}_{xi} c^{q}_{yi} c^{r}_{zi} f_i
  !! \]
  !! The fucntion argument `expX` is array of size 3,
  !! which contains the values of \f$p, q, r\f$
  !!
  pure function get_moment( QQ, cxDir, expX, pdf ) result( mom )
    ! ---------------------------------------------------------------------------
    integer, intent(in) :: QQ
    !> distribution value
    real(kind=rk), intent(in) :: pdf(QQ)
    !>
    integer, intent(in) :: cxDir(3, QQ)
    integer, intent(in) :: expX(3) !< exponents of the moments
    real(kind=rk) :: mom     !< moment
    ! ---------------------------------------------------------------------------

    !HK: assuming 0**0 should be 1 here
    mom = sum( pdf * get_momentVector( QQ, cxDir, expX ) )

  end function get_moment
! ****************************************************************************** !


! ****************************************************************************** !
  !> get the moment vector to calculate the moment from the pdf
  !!
  pure function get_momentVector( QQ, cxDir, expX ) result( mom )
    ! ---------------------------------------------------------------------------
    !>
    integer, intent(in) :: QQ
    integer, intent(in) :: cxDir(3, QQ)
    !> exponents of the moments
    integer, intent(in) :: expX(3)
    !> moment vector
    real(kind=rk) :: mom(QQ)
    ! ---------------------------------------------------------------------------

    !HK: assuming 0**0 should be 1 here
    Mom = real( mus_iMomVector(cxDir = cxDir,  &
      &                        expX  = expX,   &
      &                        QQ    = QQ),    kind = rk )

  end function get_momentVector
! ****************************************************************************** !


! ****************************************************************************** !
  !> set indices for accessing the pressure, velocity and the shear from a 1d
  !! vector
  !!
  subroutine set_momentIndices( nDims, iPress, iVelMin, iVelMax, iSMin, iSMax )
    ! ---------------------------------------------------------------------------
    !> number of dimensions
    integer, intent(in) :: nDims
    !> index for the pressure / density
    integer, intent(out) :: iPress
    !> starting index for velocity
    integer, intent(out) :: iVelMin
    !> ending index for velocity
    integer, intent(out) :: iVelMax
    !> starting index for shear
    integer, intent(out) :: iSMin
    !> ending index for shear
    integer, intent(out) :: iSMax
    ! ---------------------------------------------------------------------------

    ! Pressure / density is always first entry
    iPress = 1
    ! Indices from where / to where to read / store the velocity quantities
    iVelMin = 2
    iVelMax = nDims + 1
    ! Indices from where / to where to read / store the stress   quantities
    iSMin = iVelMax + 1
    iSMax = iSMin   + max(3*(nDims-1)-1, 0)

  end subroutine set_momentIndices
! ****************************************************************************** !


end module mus_moments_module
! ****************************************************************************** !
