! See copyright notice in the COPYRIGHT file.
! ***************************************************************************** !
!> author: Kannan Masilamani
!! Module containing subroutines for initialize Musubi source 
!! variables and update source terms
!!
module mus_source_module
  ! include treelm modules
  use env_module,               only: rk
  use tem_aux_module,           only: tem_abort
  use tem_varSys_module,        only: tem_varSys_type
  use tem_time_module,          only: tem_time_type
  use treelmesh_module,         only: treelmesh_type
  use tem_logging_module,       only: logUnit
  use tem_construction_module,  only: tem_levelDesc_type
  use tem_geometry_module,      only: tem_baryOfID
  use tem_timer_module,         only: tem_startTimer, tem_stopTimer
  use tem_tools_module,         only: tem_horizontalSpacer

  ! include musubi modules
  use mus_pdf_module,        only: pdf_data_type
  use mus_field_module,      only: mus_field_type
  use mus_source_var_module, only: mus_source_type, mus_source_op_type
  use mus_timer_module,      only: mus_timerHandles

  implicit none
  private

  public :: mus_init_sourceTerms
  public :: mus_apply_sourceTerms

contains


  ! ************************************************************************** !
  !> This routine does set_params and setupIndices for all sources terms
  !! by gathering points to apply souce term before.
  subroutine mus_init_sourceTerms( field, nFields, globSrc, varSys, &
    &                              tree, nElems_solve, levelDesc )
    ! -------------------------------------------------------------------------- 
    !> Number of fields
    integer, intent(in)                    :: nFields

    !> contains sources of all fields
    type(mus_field_type), intent(inout)   :: field(:)

    !> global source
    type(mus_source_type), intent(inout)   :: globSrc

    !> global variable system
    type(tem_varSys_type), intent(in)      :: varSys

    !> global treelm mesh
    type( treelmesh_type ), intent(in)     :: tree
    
    !> Number of elements to solve in all levels
    !! nFluids + nGhosts
    integer, intent(in) :: nElems_solve(tree%global%minLevel:)
    
    !> Level descriptors
    type( tem_levelDesc_type ), intent(in) :: levelDesc(tree%global%minLevel:)
    ! -------------------------------------------------------------------------- 
    integer :: iLevel, iField, iSrc
    integer :: nSolve, minLevel, maxLevel
    real(kind=rk), allocatable :: bary(:,:)
    ! -------------------------------------------------------------------------- 
    call tem_horizontalSpacer(fUnit = logUnit(1))
    write(logUnit(1),*)'Initializing Source terms ...'
   
    ! immediately exit this routine if there are no source terms to apply
    if ( all(field(:)%source%varDict%nVals == 0) &
      & .and. globSrc%varDict%nVals == 0 ) then
      write(logUnit(1),*) 'No active source terms'
      return
    end if  

    minLevel = tree%global%minLevel
    maxLevel = tree%global%maxLevel

    ! allocate source element array
    do iField = 1, nFields
      do iSrc = 1, field(iField)%source%varDict%nVals
        allocate( field(iField)%source%method(iSrc)%elemLvl(minLevel:maxLevel) )
      end do  
    end do

    do iSrc = 1, globSrc%varDict%nVals
      allocate( globSrc%method(iSrc)%elemLvl(minLevel:maxlevel) )
    end do  

    ! get source parameter to store in data variable for coupling
    write(logUnit(10),*) ' Setup indices for sources'
    do iLevel = minLevel, maxLevel
      write(logUnit(10),*) 'iLevel: ', iLevel

      nSolve = nElems_solve(iLevel)
      ! Gets barycenter of all elements to solve i.e fluid+ghost to apply
      ! source terms
      call mus_getPoints_forSources( bary      = bary,             &
        &                            nSolve    = nSolve,           &
        &                            tree      = tree,             &
        &                            levelDesc = levelDesc(iLevel) )

      do iField = 1, nFields
        ! Store idx for active field source variables
        call mus_setupIndices_forSrc( source     = field(iField)%source, &
          &                           varSys     = varSys,               &
          &                           tree       = tree,                 &
          &                           nSolve     = nSolve,               &
          &                           bary       = bary,                 &
          &                           iLevel     = iLevel                )
      end do

      ! Store idx for active glob source variables
      call mus_setupIndices_forSrc( source     = globSrc, &
        &                           varSys     = varSys,  &
        &                           tree       = tree,    &
        &                           nSolve     = nSolve,  &
        &                           bary       = bary,    &
        &                           iLevel     = iLevel   )

    end do ! iLevel 
    call tem_horizontalSpacer(fUnit = logUnit(1))

  end subroutine mus_init_sourceTerms
  ! *************************************************************************** !


  ! *************************************************************************** !
  !> This routines does setup indices for given source within a field or 
  !! global. Index are stored for points which source term is active
  subroutine mus_setupIndices_forSrc(source, varSys, nSolve, bary, iLevel, &
    &                                tree)
    ! -------------------------------------------------------------------------- 
    !> Source term to fill in
    type(mus_source_type), intent(inout) :: source

    !> global variable system
    type(tem_varSys_type), intent(in)      :: varSys

    !> global treelm mesh
    type( treelmesh_type ), intent(in)     :: tree
    
    !> Number of elements to apply source term  on this level
    integer, intent(in) :: nSolve

    !> Space coordinates to apply source terms
    real(kind=rk), intent(in) :: bary(nSolve,3)

    !> Current level
    integer, intent(in) :: iLevel
    ! -------------------------------------------------------------------------- 
    integer :: iVar, iElem, counter, src_nElems
    integer :: data_varPos
    integer, allocatable :: idx(:)
    ! -------------------------------------------------------------------------- 
    allocate(idx(nSolve))

    do iVar = 1, source%varDict%nVals
      idx = 0
      data_varPos = source%method(iVar)%data_varPos
      ! set params
      call varSys%method%val(data_varPos)%set_params( &
        & varSys   = varSys,                          &
        & instring = 'isSurface = false'              )
  
      call varSys%method%val(data_varPos)%setup_indices( &
        & varSys     = varSys,                           &
        & point      = bary,                             &
        & iLevel     = iLevel,                           &
        & tree       = tree,                             &
        & nPnts      = nSolve,                           &
        & idx        = idx                               )
  
      ! Store only valid idx to apply source term only on shapes defined
      ! for a data variable
      src_nElems = count(idx>0)
      source%method(iVar)%elemLvl(iLevel)%nElems = src_nElems 
      allocate(source%method(iVar)%elemLvl(iLevel)%posInTotal(src_nElems))
      allocate(source%method(iVar)%elemLvl(iLevel)%idx(src_nElems))

      counter = 0
      do iElem = 1, nSolve
        if (idx(iElem) > 0) then
          counter = counter + 1
          source%method(iVar)%elemLvl(iLevel)%posInTotal( counter ) = iElem
          source%method(iVar)%elemLvl(iLevel)%idx( counter ) = idx(iElem)
        end if
      end do
    end do !iVar

  end subroutine mus_setupIndices_forSrc
  ! *************************************************************************** !

  ! *************************************************************************** !
  !> This routine extract barycenter of elements from total list 
  !! for source terms
  subroutine mus_getPoints_forSources( bary, nSolve, tree, levelDesc )
    ! -------------------------------------------------------------------------- 
    !> Space coordinates to apply source terms
    real(kind=rk), intent(out), allocatable :: bary(:,:)

    !> Number of elems/points to apply source terms on this level
    integer, intent(in) :: nSolve

    !> global treelm mesh
    type( treelmesh_type ), intent(in)     :: tree
    
    !> Level descriptors
    type( tem_levelDesc_type ), intent(in) :: levelDesc
    ! -------------------------------------------------------------------------- 
    integer :: iElem
    ! -------------------------------------------------------------------------- 

    allocate(bary(nSolve, 3))
    do iElem = 1, nSolve
      bary(iElem, :) = tem_BaryOfId(tree, levelDesc%total(iElem))
    end do

  end subroutine mus_getPoints_forSources
  ! *************************************************************************** !


  ! *************************************************************************** !
  !> Apply all source terms i.e field specific source and global source on
  !! all fields.
  !! NEXT timestep t+1.
  !!
  subroutine mus_apply_sourceTerms( field, nFields, globSrc, pdf, varSys, &
    &                               iLevel, time, tree, state )
    ! --------------------------------------------------------------------------
    !> Number of fields
    integer, intent(in)                :: nFields

    !> contains sources of all fields
    type(mus_field_type), intent(in)  :: field(nFields)

    !> global source
    type(mus_source_type), intent(in)  :: globSrc

    !> global tree information
    type(treelmesh_type), intent(in)   :: tree

    !> pdf datatype 
    type(pdf_data_type), intent(inout) :: pdf

    !> global variable system
    type(tem_varSys_type), intent(in)  :: varSys

    !> current level
    integer, intent(in)                :: iLevel

    !> current timing information
    type(tem_time_type), intent(in)    :: time

    !> state type containing the state vector to update
    real(kind=rk), intent(inout) :: state(:,:)
    ! -------------------------------------------------------------------------- 
    ! counter variables
    integer :: iVar, iField
    integer :: now, next
    ! -------------------------------------------------------------------------- 
    call tem_startTimer( timerHandle = mus_timerHandles%source )
    ! \todo create buffer for source to copy state values before applying 
    ! source term to avoid using state which overwritten by source variable

    now = pdf%nNow
    next = pdf%nNext

    ! loop over all active sourve variable to apply source
    do iField = 1, nFields
      do iVar = 1, field(iField)%source%varDict%nVals
        call field(iField)%source%method(iVar)%applySrc( &
          & inState    = state(:, now),              &
          & outState   = state(:, next),             &
          & neigh      = pdf%neigh(:),                   &
          & nPdfSize   = pdf%nSize,                      &  
          & iLevel     = iLevel,                         &
          & varSys     = varSys,                         &
          & time       = time,                           &
          & tree       = tree                            )
      end do
    end do   

    ! apply global source
    do iVar = 1, globSrc%varDict%nVals
      call globSrc%method(iVar)%applySrc(  &
        & inState    = state(:, now),  &
        & outState   = state(:, next), &
        & neigh      = pdf%neigh(:),       &
        & nPdfSize   = pdf%nSize,          &  
        & iLevel     = iLevel,             &
        & varSys     = varSys,             &
        & time       = time,               &
        & tree       = tree                )
    end do

    call tem_stopTimer( timerHandle = mus_timerHandles%source )
  end subroutine mus_apply_sourceTerms
! ***************************************************************************** !

end module mus_source_module
! ***************************************************************************** !
