-- This file contains important parameters
-- get mesh refine level
if os.getenv('LEVEL') == nil or os.getenv('LEVEL') == "" then
  level = 4
else
  level = tonumber( os.getenv('LEVEL'))
end

-- get Reynolds number
if os.getenv('Re') == nil or os.getenv('Re') == "" then
  Re = 25
else
  Re = tonumber( os.getenv('Re'))
end

-- get Reynolds number
if os.getenv('KIND') == nil or os.getenv('KIND') == "" then
  kind = 'lbm_incomp'
else
  kind = os.getenv('KIND')
end

-- get Reynolds number
if os.getenv('GLOBAL') == "TRUE" then
  track_global = true
else
  track_global = false
end

refinementLevel = 1
minLevel = level
minlevel = minLevel
maxLevel = level+refinementLevel

-- some constant parameters
sqrt3 = math.sqrt(3.0)
twoPi = 2.0 * math.pi

-- flow domain [0, 2*pi]
length = twoPi
dx     = length / (2^minLevel)
dxMin  = length / (2^maxLevel)
dxDash = 0.5 * dxMin

folder = 'mesh/'
mesh = folder
simulation_name = 'TGV'
scheme_label = 'fluid'
-- ebug = {debugMode = true, debugFiles = true, debugMesh = 'debug/'}

origin = {0., 0., 0.}      -- bounding box origin
periodic_z1 = origin[3] + 2.*dx + dxDash
periodic_z2 = origin[3] + dx - dxDash
refine_origin = { length / 4 + dxDash, length / 4 + dxDash, dx + dxDash }
refine_length = length / 2 - dxMin
refine_vec1 = { refine_length,           0.0,  0.0 }
refine_vec2 = {           0.0, refine_length,  0.0 }
refine_vec3 = {           0.0,           0.0,  dxMin }

uc = 0.00
theta = math.pi / 3.0
ucx = uc * math.cos(theta)
ucy = uc * math.sin(theta)

---------------------------------------------------------------------
-- Simulation setup

intp       = 'linear'
scaling    = 'diffusive'
--scaling    = 'acoustic'
relaxation = 'bgk'
layout     = 'd2q9'
--layout     = 'd3q19'
interpolation_method = { method = intp }

-- Scaling and initialization settings
u0      = 1.0
rho0LB  = 1.0
rhoPhy  = 1.0
nuPhy   = 1.0 / Re
viscPhy = nuPhy * rhoPhy

-- if scaling == 'acoustic' then
--   -- keep velocity constant, compute omega from dt
--   uLB = 0.05
--   dt  = uLB / u0 * dx
--   nuLB = nuPhy * dt/dx/dx
--   omega = 1.0 / ( 3.0 * nuLB + 0.5)
-- else
  -- keep omage constant, compute dt from omage
  omega = 1.8
  nuLB  = ( 2.0/omega - 1.0 ) / 6.0
  dt  = nuLB / nuPhy * dx * dx
  uLB = u0 * dt / dx
-- end

-- origin is the center of first element
x0 = { 0.5*dx, 0.5*dx, 1.5*dx }

xProbe = { x0[1] + math.pi/3, x0[2] + math.pi/3, x0[3] }

cs2 = 1./3.
p0 = rho0LB / rhoPhy * dx^2 / dt^2 * cs2

kx = twoPi / length -- wave number in x direction
ky = twoPi / length -- wave number in y direction
tD = 1.0 / nuPhy / (kx^2+ky^2) -- rate of decay coefficient

tEnd     = math.ceil(tD/dx) * dx
interval = tEnd / 10.0

-- Analytical solution with respect to x0
function pressure(x,y,z,t)
  -- when kx == ky
  p = (math.cos(2.0*kx*(x-x0[1]))+math.cos(2.0*ky*(y-x0[2]))) * math.exp(-2.0*t/tD)
  return p0 - u0^2 * p * rhoPhy / 4.0
end
function ic_pressure(x,y,z)
  return pressure(x,y,z,0)
end

-- velocity X
function velocityX(x,y,z,t)
  velX = math.cos(kx*(x-x0[1])) * math.sin(ky*(y-x0[2])) * math.exp(-t/tD)
  return u0 * (-math.sqrt(ky/kx)) * velX + ucx
end
function ic_velocityX(x,y,z)
  return velocityX(x,y,z,0)
end

-- velocity Y
function velocityY(x,y,z,t)
  velY = math.sin(kx*(x-x0[1])) * math.cos(ky*(y-x0[2])) * math.exp(-t/tD)
  return u0 * (math.sqrt(kx/ky)) * velY + ucy
end
function ic_velocityY(x,y,z)
  return velocityY(x,y,z,0)
end

-- velocity Z
function velocityZ(x,y,z,t)
  return 0
end
function ic_velocityZ(x,y,z)
  return velocityZ(x,y,z,0)
end

--velocity mag
function Velocity(x, y, z, t)
  velX = velocityX( x,y,z,t )
  velY = velocityY( x,y,z,t )
  return math.sqrt( velX^2 + velY^2 )
end

-- Sxx
function Sxx(x,y,z,t)
  sxx = math.sin(kx*(x-x0[1])) * math.sin(ky*(y-x0[2])) * math.exp(-t/tD)
  return u0 * sxx
end
function ic_Sxx(x,y,z)
  return Sxx(x,y,z,0)
end
-- Syy
function Syy(x,y,z,t)
  return -Sxx(x,y,z,t)
end
function ic_Syy(x,y,z)
  return Syy(x,y,z,0)
end
-- Sxy
function Sxy(x,y,z,t)
  return 0.0  -- when kx == ky
end
function ic_Sxy(x,y,z)
  return Sxy(x,y,z,0)
end
