
require 'seeder'
-- Musubi configuration file for the permeability.
-- Please adapt only the parameters: 
-- 

scaling = 'diffusive'
setInPhysicalUnits = true
-- phyiscal parameters
viscPhys = 0.00001   -- fluid viscosity m^2/s
u0Phys = 0.001      -- velocity m/s
rho0Phys = 1.0     -- reference density kg/m^3

-- characteristic time 
--tChar = length/u0Phys
--tEnd = tChar*100.   -- run up to which simulation time?
deltapPhys = 0.01

-- LB parameters
-- reference omega, velocity and density
omega = 1.8
u0LB  = 0.0003
rho0LB  = 1.       
cs2LB = 1./3.

model = 'lbm_incomp'
relaxation = 'bgk'
interpolation_method = 'quadratic'
if scaling == 'acoustic' then
  uLB=u0LB
  dt = uLB/u0Phys*dx
  viscLB = viscPhys*dt/dx/dx
  omega = 1./(3.*viscLB + 0.5)
else
  viscLB  = 1./6.*(2./omega - 1.)
  dt = viscLB/viscPhys*dx*dx  
  uLB=u0LB*2^(refLevel-level)
end
if model == 'lbm' then
  p0LB = rho0LB*cs2LB
else
  p0LB = 0.
end
p0Phys = p0LB*dx*dx/dt/dt*rho0Phys
-- Number of iterations
if setInPhysicalUnits then
  print( ' setting physical pressure ')
  p0 = p0Phys
  pressIn =  p0Phys + deltapPhys
  pressOut = p0Phys
else
  p0 = p0LB
  deltapLB = deltapPhys*dt*dt/dx/dx/rho0Phys
  pressIn = (p0LB+deltapLB)/cs2LB
  pressOut = (p0LB)/cs2LB
end
-- characteristic time in LB iterations
iChar = nElemsMax*5 --tChar/dt
nIters = iChar*10

u0Phys = uLB*dx/dt
Re = u0Phys*height/viscPhys
-------------------------------------------------------------------------------
-- Do not change the script from here on unless you know what you do.
-------------------------------------------------------------------------------
minInterval = 5
if Re > 1.4 then
  print('WARNING: Reynolds Number > 1 !   Re = '..Re )
end
if refinement == 0 then
  label = ''
else
  label = 'RM_'..interpolation_method
end
-- Simulation name
simulation_name = simname    
mesh = 'mesh/'  -- MESH
-- Time step settings
tmax  =  nIters   -- total iteration number
tRamp = 100 --math.max( 100, iChar/10)
interval = tmax/100
sim_control = {
  time_control = {
    max = {iter= tmax},
      interval = interval
      }
  }  

physics = {dt = dt, rho0 = rho0Phys, setPhysicalQuantities = setInPhysicalUnits }

ebug = { debugMode = true, debugFiles = true, dumpTreeIDs = true, verbosity = 100,
          ebugMesh = './debug/mesh_'} 

-- Initial condition 
initial_condition = { pressure = p0, 
                      velocityX = 0.0,
                      velocityY = 0.0,
                      velocityZ = 0.0 }
-- wall boundary condition type. 
-- should be wall, but can be set to wall_higerhOrder.
-- make sure to have the qVals ready in your mesh then
bcWallKind = 'wall' 

-- Boundary conditions
boundary_condition = {  
  { label = obj_label,  kind = bcWallKind},
  { label = 'channel', kind = bcWallKind} 
, { label = 'inlet', 
--  kind = 'inlet_ubb', 
    kind = 'outlet_expol', 
    pressure = 
    { predefined = 'combined',
      temporal = { predefined='smooth', min_factor = pressOut,
                  max_factor=pressIn, from_time=0, to_time= math.min(500*dt, tRamp) }, 
  	 },
      velocityY = 0.0, velocityZ = 0.0 }
,{ label = 'outlet', 
  kind = 'outlet_expol', pressure = pressOut}
}

fluid = { omega = omega, rho0 = rho0LB, bulk_viscosity = 1./36. }
identify = {
  kind = model, 
  relaxation = relaxation, 
  label = idLabel, 
  layout = 'd3q19' 
  }

-- Output settings
--utput = { active = true, -- VTK output activated?
--  folder = './output/',     -- Output location
--  vtk = true,             -- VTK output activated?
--  ascii = true,           -- VTK output activated?
--  dumpGhosts = true, dumpHalos = true,
--  time = {interval = interval, min = 0, max = -1}
--}

-- Tracking              
tracking = {
  -- Probe the pressure at two points
  -- 1. in the middle of the starting free flow section
  -- 2. in the middle of the ending free flow section
{ label = 'probe', 
  variable = {'pressure_phy', 'velocity_phy'}, 
  shape = {kind = 'canoND', 
    object = {{origin =
         {startX+0.5*startLength,0.,0.} 
         },{origin =
         {endX[1]-0.5*endLength,0.,0.} 
         }}},
  time_control = {interval = {iter= minInterval}, min = {iter= min }, max ={iter= tmax}},
  format='ascii',
 folder='tracking/'
  }
,{ label = 'convergence',
  variable = {'pressure_phy'}, 
  shape = {kind = 'canoND', 
    object = {{origin =
         {startX+0.5*startLength,0.,0.} 
         }}},
  time_control = {interval= {iter= math.max( 1, iChar/10)}, min={iter= iChar/2.}, max ={iter= tmax}},
  format='convergence',
  convergence = {norm='average', nvals = 500,
    condition = { threshold = 1.e-10, operator = '<=' }}
  }
  -- Pressure drop along a line
,{ label = 'dpdx', 
  variable = {'pressure_phy', 'pressure'}, 
  shape = {kind = 'canoND', object = {origin ={startX,0.,0.}, 
                                      vec = {length, 0., 0.},
                                      segments = nElemsMax } },
  time_control = {interval = {iter= minInterval}, min= {iter = tmax}, max = {iter= tmax}},
  format='asciiSpatial'
  , folder='tracking/'
  }
, 
  -- Inflow: Measure the average of 
  -- - volumetric flux q = \dot{V}/A = A*v/A = v
{ label = 'flowRate', 
  variable = {'velocity_phy'}, 
  reduction = {'average'}, 
  shape = {kind = 'canoND', object = 
    {  origin = {
       startX + 0.5*startLength,
      -height*0.5,
      -height*0.5}, 
      vec  = {
        {0., 0., height}, 
        {0., height, 0.}}, 
      segments={nElemsMax, nElemsMax}, 
      distribution='equal'} } , 
  format='ascii', folder='tracking/', 
  time_control = {interval = {iter= interval}, min = {iter = 0}, max = {iter= tmax }},
}

  -- Inflow: Measure the average of 
  -- - pressure
,{ label = 'pressureIn', 
  variable = {'pressure_phy'}, 
  reduction = {'average'}, 
  shape = {kind = 'canoND', object = 
    {  origin = {
       startX + 0.5*startLength,
      -height*0.5,
      -height*0.5}, 
      vec  = {
        {0., 0., height}, 
        {0., height, 0.}}, 
      segments={nElemsMax, nElemsMax}, 
      distribution='equal'} } , 
  format='ascii', folder='tracking/', 
  time_control = { interval= {iter= interval}, min ={iter= 0}, max ={iter = tmax }},
}
-- Outflow: Measure the average of 
  -- - pressure
  -- - volumetric flux q = \dot{V}/A = A*v/A = v
,{ label = 'pressureOut', 
  variable = {'pressure_phy','velocity_phy'}, 
  reduction = {'average', 'average'}, 
  shape = {kind = 'canoND', object = 
    {  origin = {
       endX[1] - 0.5*endLength,
      -height*0.5,
      -height*0.5}, 
      vec  = {
        {0., 0., height}, 
        {0., height, 0.}}, 
      segments={nElemsMax, nElemsMax}, 
      distribution='equal'} } , 
  format='ascii', folder='tracking/', 
  time_control = { interval = {iter= interval}, min = {iter=0}, max= {iter= tmax }}
},
  -- Measure pressure and velocity at cut planes 
{ label ='hvs', 
  variable = {
    'pressure_phy',
    'velocity_phy',
    'pressure',
    'velocity'},
  shape = {{kind = 'canoND', object = {origin = {startX, -height*0.5,0.}, 
                                      vec  = {{length,0,0.},{0.0, height, 0.0}}, 
                                      segments={nElemsMax, nElemsMax}, 
                                      distribution='equal'} }
          ,{kind = 'canoND', object = {origin = {startX, 0.0, -length*0.5}, 
                                      vec  = {{length,0,0.},{0.0, 0.0, length}}, 
                                      segments={nElemsMax, nElemsMax}, 
                                      distribution='equal'} 
          }
          }
  , format='harvester', folder='tracking/', 
  time_control = { interval = {iter=interval*10}, min= {iter= 0}, max= {iter= tmax }},
},
  -- measure the porosity
{ label ='porosity',
  variable = {'volFrac'} ,
  reduction = {'sum'},
  shape = {{kind = 'canoND', object = {
    origin = {originPorous[1]+dx, -wallPos[2]+dx,-wallPos[3]+dx}, 
    vec  = {
      {sizePorous[1]-2*dx,0,0.},
      {0.0, 2*(wallPos[2]-dx), 0.0},
      {0.0, 0.0, 2*(wallPos[3]-dx)}
    }, 
    segments={nElemsMax*1.2, nElemsMax*1.2, nElemsMax*1.2}, 
    distribution='equal'} }
          }
  , format='harvester', folder='tracking/', 
  time_control = {interval= {iter= interval}, min= {iter= 0}, max= {iter= 0 }},
},
} -- tracking table

-- Restart settings
estart = {
  -- If this table is available it will write a restart file
  -- with the specified options.
  ead  = {
    folder    = 'restart/', -- the folder the restart files are written to
    file_name = 'lastHeader.lua' -- the folder the restart files are written to
  },
    write = 'restart/', -- the folder the restart files are written to
    time_control = {interval= {iter= tmax}, min= {iter= tmax}, max= {iter= tmax}},
 }
if verbose then
  --print(' tChar  '..tChar)
  --print(' pressIn '..pressIn)
  --print(' pressOut'..pressOut)
  print('LB parameters:')
  print('  u =  '..uLB)
  print('  om = '..omega)
  print('Physical      ')
  print('  viscPhys', viscPhys )

  print('Reynolds number')
  print(' Re LB ', uLB*(height/dx)/viscLB)
  print(' RePhys ', Re )
end 


