! See copyright notice in the COPYRIGHT file.
!> This module provides a spatial function to describe a 2D polygon.
!!
!! The spatial function will provide different values for points within the
!! polygon and those outside.
!!
!! A polygon is defined in the configuration by the following definition:
!!
!!```lua
!!   -- Vector to return inside the polygon
!!   inval = {1.0, 0.0, 2.0}
!!
!!   -- It also can be a scalar, if none is provided a scalar 1 is assumed.
!!   inval = 1.0 -- This is the value to return for points inside the polygon
!!               -- defaults to 1.
!!
!!   -- Vector to return outside the polygon
!!   outval = {0.0, 0.0, 0.0}
!!   -- Needs to conform to the definition of inval, that is, if inval is scalar
!!   -- or not given, outval needs to be a scalar.
!!   -- Defaults to all zero components for a vector of the length of inval.
!!   -- If inval is a scalar, outval has to be given as a scalar aswell:
!!   outval = 0.0 -- The value to return for points outside the polygon,
!!                -- defaults to 0.
!!
!!   -- List of 2D Points to be used as vertices for the polygon.
!!   -- The polygon will be closed by going from the last point back to the
!!   -- first one.
!!   vertex = { { 1.0,  0.0},
!!              { 0.0, -1.0},
!!              {-1.0,  0.0},
!!              { 0.0,  1.0},
!!            }
!!```
!!
module tem_polygon_material_module
  use env_module, only: rk
  use aotus_module, only: flu_state, aot_get_val, aoterr_Fatal
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length

  use tem_aux_module, only: tem_abort
  use tem_logging_module, only: logunit

  implicit none

  private

  !> Description of a 2D closed polygon.
  type tem_polygon_material_type
    !> Number of vertices in the polygon.
    integer :: nVertices

    integer :: nComponents

    !> Value of Material inside the polygon.
    real(kind=rk), allocatable :: inval(:)

    !> Value of Material outside the polygon.
    real(kind=rk), allocatable :: outval(:)

    !> 2D Coordinates of the vertices.
    !!
    !! First index runs over number of vertices, second from 1 to 2.
    real(kind=rk), allocatable :: vertex(:,:)
  end type tem_polygon_material_type

  public :: tem_polygon_material_type
  public :: tem_polygon_material_load
  public :: tem_polygon_material_value
  public :: tem_eval_polygon_material
  public :: tem_eval_polygon_material_scal

  public :: tem_polygon_material_test_angle
  public :: tem_polygon_material_test_value

  ! Some private constants

  !> Definition of Pi
  real(kind=rk), parameter :: PI = 2*asin(1.0_rk)

  !> Overestimating tolerance factor for comparisons of reals
  real(kind=rk), parameter :: tolp = (1.0_rk + epsilon(1.0_rk))

  !> Underestimating tolerance factor for comparisons of reals
  real(kind=rk), parameter :: tolm = (1.0_rk - epsilon(1.0_rk))


contains


  !> Load a polygon definition from the user configuration (a Lua script).
  subroutine tem_polygon_material_load(me, conf, thandle)
    !> Polygon data structure to fill with information provided
    !! by the user in config.
    type(tem_polygon_material_type), intent(out) :: me

    !> Handle to the Lua script containing the polygon definition
    type(flu_state) :: conf

    !> Handle for the table containing the polygon definition.
    integer, intent(in), optional :: thandle

    ! Local Variables
    real(kind=rk), allocatable :: defout(:)
    integer :: vertex_table, valtable
    integer :: iVertex
    integer :: iError
    integer, allocatable :: vError(:)
    integer :: iError_v(2)

    write(logUnit(1),*) 'Loading predefined function polygonal material:'

    call aot_table_open( L       = conf,    &
      &                  parent  = thandle, &
      &                  key     = 'inval', &
      &                  thandle = valtable )

    if (valtable  == 0) then
      ! inval not provided as a table, try to read it as a scalar.
      allocate(me%inval(1))
      call aot_get_val( L         = conf,        &
        &               thandle   = thandle,     &
        &               key       = 'inval',     &
        &               val       = me%inval(1), &
        &               default   = 1.0_rk,      &
        &               ErrCode   = iError       )

      if (btest(iError, aoterr_Fatal)) then
        write(logunit(1),*) 'ERROR in tem_polygon_material_load: Not able to'
        write(logunit(1),*) '      get value for inval!'
        call tem_abort()
      end if

      me%nComponents = 1
      allocate(me%outval(me%nComponents))

      ! Outval needs to be consistent with the inval definition, if inval was
      ! defined as a scalar, outval also has to be a scalar!
      ! We do not check for tables with single entries in this case.
      call aot_get_val( L       = conf,         &
        &               thandle = thandle,      &
        &               key     = 'outval',     &
        &               val     = me%outval(0), &
        &               default = 0.0_rk,       &
        &               ErrCode = iError        )

      if (btest(iError, aoterr_fatal)) then
        write(logunit(1),*) 'ERROR in tem_polygon_material_load: Not able to'
        write(logunit(1),*) '      get a value for outval!'
        write(logunit(1),*) '      Note, that outval needs be a scalar, as'
        write(logunit(1),*) '      inval is provided as a scalar.'
        write(logunit(1),*) '      This also applies if no inval is provided.'
        call tem_abort()
      end if

    else
      
      ! Intable is a table, close it an read it into an array.
      call aot_table_close(L = conf, thandle = valtable)

      ! Value to use inside the polygon
      call aot_get_val( L         = conf,     &
        &               thandle   = thandle,  &
        &               key       = 'inval',  &
        &               val       = me%inval, &
        &               maxlength = 20,       &
        &               default   = [1.0_rk], &
        &               ErrCode   = vError    )

      if (any(btest(vError, aoterr_Fatal))) then
        write(logunit(1),*) 'ERROR in tem_polygon_material_load: Not able to'
        write(logunit(1),*) '      get values for inval!'
        call tem_abort()
      end if

      me%nComponents = size(me%inval)
      deallocate(vError)

      ! Definition of outval needs to be consistent with inval, it has to have
      ! the same number of components, and also needs to be a vector.
      ! However, we define a default of all zeroes, so if outval is 0 for all
      ! components, this definition can be omitted in the user definition.
      allocate(me%outval(me%nComponents))
      allocate(vError(me%nComponents))
      allocate(defout(me%nComponents))

      defout = 0.0_rk

      ! Value to use outside the polygon
      call aot_get_val( L       = conf,      &
        &               thandle = thandle,   &
        &               key     = 'outval',  &
        &               val     = me%outval, &
        &               default = defout,    &
        &               ErrCode = vError     )

      if (any(btest(vError,aoterr_Fatal))) then
        write(logunit(1),*) 'ERROR in tem_polygon_material_load: Not able to'
        write(logunit(1),*) '      get a value for outval!'
        write(logunit(1),*) '      Note, that outval needs to have the same'
        write(logunit(1),*) '      length as inval.'
        call tem_abort()
      end if

      deallocate(vError)
      deallocate(defout)

    end if


    call aot_table_open( L       = conf,        &
      &                  parent  = thandle,     &
      &                  key     = 'vertex',    &
      &                  thandle = vertex_table )

    if (vertex_table == 0) then
      write(logunit(1),*) 'ERROR in tem_polygon_material_load: No vertices'
      write(logunit(1),*) '      defined, unable to set up a polynomial!'
      write(logunit(1),*) '      Please define a vertex table with a list of'
      write(logunit(1),*) '      2D points.'
      call tem_abort()
    end if

    me%nVertices = aot_table_length( L       = conf,        &
      &                              thandle = vertex_table )

    allocate(me%vertex(me%nVertices, 2))

    do iVertex=1,me%nVertices
      call aot_get_val( val     = me%vertex(iVertex, :), &
        &               ErrCode = iError_v,              &
        &               L       = conf,                  &
        &               thandle = vertex_table,          &
        &               pos     = iVertex                )
      if (any(iError_v /= 0)) then
        write(logunit(1),*) 'ERROR in tem_polygon_material_load: Not able to'
        write(logunit(1),*) '      obtain vertex ', iVertex, '!'
        write(logunit(1),*) '      Vertices have to be vectors of length 2,'
        write(logunit(1),*) '      with real numbers as entries.'
        call tem_abort()
      end if
    end do

    call aot_table_close( L       = conf,        &
      &                   thandle = vertex_table )

  end subroutine tem_polygon_material_load



  !> Evaluate a list of points, and return inval for each that is within me
  !! and outval for all other points.
  function tem_eval_polygon_material(me, coord, n) result(res)
    !> Description of the polygon to evaluate
    type(tem_polygon_material_type), intent(in) :: me

    !> Number of points to get a value for.
    integer, intent(in) :: n

    !> Coordinates for which the function should be evaluated.
    real(kind=rk), intent(in) :: coord(n,3)

    !> Resulting value at each point.
    real(kind=rk) :: res(n, me%nComponents)

    integer :: iPoint

    do iPoint=1,n
      res(iPoint,:) = tem_polygon_material_value( polygon = me,              &
        &                                         point   = coord(iPoint,:2) )
    end do

  end function tem_eval_polygon_material



  !> Evaluate a list of points, and return first component of inval for
  !! each that is within me and first component of outval for all
  !! other points.
  function tem_eval_polygon_material_scal(me, coord, n) result(res)
    !> Description of the polygon to evaluate
    type(tem_polygon_material_type), intent(in) :: me

    !> Number of points to get a value for.
    integer, intent(in) :: n

    !> Coordinates for which the function should be evaluated.
    real(kind=rk), intent(in) :: coord(n,3)

    !> Resulting value at each point.
    real(kind=rk) :: res(n)

    real(kind=rk) :: loc(me%nComponents)

    integer :: iPoint

    do iPoint=1,n
      loc = tem_polygon_material_value( polygon = me,              &
        &                               point   = coord(iPoint,:2) )
      res(iPoint) = loc(1)
    end do

  end function tem_eval_polygon_material_scal



  !> Return the material value for point, based on the position in relation to
  !! the polygon.
  !!
  !! The point containment is decided with the help of the Gauss-Bonnet
  !! theorem, which is highly robust, but requires the polygon vertices to
  !! follow an ordering along the polygon lines.
  !!
  !! The returned value will be polygon%inval, if the point is inside and
  !! polygon%outside otherwise.
  function tem_polygon_material_value(polygon, point) result(res)
    !> Polygon to describe the material shape.
    type(tem_polygon_material_type), intent(in) :: polygon

    !> Point to check against the polygon.
    real(kind=rk), intent(in) :: point(:)

    !> Material value at point, as defined by the polygon.
    real(kind=rk) :: res(polygon%nComponents)

    real(kind=rk) :: diffa(polygon%nVertices,2)
    real(kind=rk) :: diffb(polygon%nVertices,2)
    real(kind=rk) :: anglesum

    ! Compute difference between all vertices of the polygon and the given
    ! point.
    diffa(:,1) = polygon%vertex(:,1) - point(1)
    diffa(:,2) = polygon%vertex(:,2) - point(2)

    ! Copy the differences to a shifted array, to allow the simultaneous
    ! computation of all the angles for polygon segments.
    ! After the last point, we jump back to the first index to close the
    ! polygon again.
    diffb(:polygon%nVertices-1,:) = diffa(2:,:)
    diffb(polygon%nVertices,:) = diffa(1,:)

    ! With the Gauss-Bonnet theorem, we can decide the containment of the
    ! point within the polygon, based on the sum of all angles under which
    ! the polygon segments are seen from the point.
    anglesum = sum( angle_between(va_x = diffa(:,1), va_y = diffa(:,2), &
      &                           vb_x = diffb(:,1), vb_y = diffb(:,2)) )

    if (abs(anglesum) > 3.0_rk) then
      ! The point is inside, if the sum of angles is 2 Pi actually, but to
      ! allow for numerical inaccuracies and points on the polygon segments
      ! we just check for greater 3 here.
      res = polygon%inval
    else
      ! Points with a smaller value are deemed to be outside the polygon.
      ! They might actually be located on a vertex, but the angle at the
      ! vertex would be greater for the outside part than the inside, so it
      ! should definitely be ok to assume this point as outside.
      res = polygon%outval
    end if

  end function tem_polygon_material_value


  !> A subroutine to test the tem_polygon_material_value function in
  !! tem_polygon_material_test.
  !!
  !! The only argument will be true, if all calculations result in the
  !! expected values.
  subroutine tem_polygon_material_test_value(success)
    !> Indicator if all tests were computed correctly.
    logical, intent(out) :: success

    type(tem_polygon_material_type) :: polygon

    real(kind=rk) :: matval(1)

    ! Use simple polygon with 5 vertices to check the containment function.
    ! It has roughly this shape:
    !
    !  x-------x
    !   \      |
    !    \     |
    !     x    |
    !    /     |
    !   /      |
    !  x-------x
    !
    ! With the 'x' indicating the 5 vertices.
    polygon%nVertices = 5
    allocate(polygon%Vertex(polygon%nVertices,2))

    polygon%Vertex(1,:) = [ 0.0_rk,  0.0_rk]
    polygon%Vertex(2,:) = [-1.0_rk, -1.0_rk]
    polygon%Vertex(3,:) = [ 1.0_rk, -1.0_rk]
    polygon%Vertex(4,:) = [ 1.0_rk,  1.0_rk]
    polygon%Vertex(5,:) = [-1.0_rk,  1.0_rk]

    ! Defining returned values for inside and outside the polygon:
    polygon%nComponents = 1
    allocate(polygon%inval(polygon%nComponents))
    allocate(polygon%outval(polygon%nComponents))

    polygon%inval = 1
    polygon%outval = 0

    success = .true.
    ! (If any subsequent test fails, we set this to false.)

    write(*,*) 'Checking point containment for this 5 point polygon:'
    write(*,*)
    write(*,*) '(-1,1)---(1,1)'
    write(*,*) '    \      |'
    write(*,*) '     \     |'
    write(*,*) '    (0,0)  |'
    write(*,*) '     /     |'
    write(*,*) '    /      |'
    write(*,*) '(-1,-1)--(1,-1)'
    write(*,*)

    ! Check a point inside the polygon:
    matval = tem_polygon_material_value( polygon = polygon,         &
      &                                  point   = [0.5_rk, 0.5_rk] )

    if (matval(1) < 1.0_rk) success = .false.
    write(*,*) 'Point (0.5, 0.5) has value:', matval

    ! Check a point outside the polygon:
    matval = tem_polygon_material_value( polygon = polygon,         &
      &                                  point   = [-1.0_rk, 0.0_rk] )

    if (matval(1) > epsilon(1.0_rk)) success = .false.
    write(*,*) 'Point (-1.0, 0.0) has value:', matval

    ! Check a point on a side of the polygon (should be considered inside):
    matval = tem_polygon_material_value( polygon = polygon,         &
      &                                  point   = [-0.5_rk, -0.5_rk] )

    if (matval(1) < 1.0_rk) success = .false.
    write(*,*) 'Point (-0.5, -0.5) has value:', matval

    ! Check an outward pointing corner (should be outside):
    matval = tem_polygon_material_value( polygon = polygon,         &
      &                                  point   = [1.0_rk, 1.0_rk] )

    if (matval(1) > epsilon(1.0_rk)) success = .false.
    write(*,*) 'Point (1.0, 1.0) has value:', matval

    ! Check an inward pointing corner (should be inside):
    matval = tem_polygon_material_value( polygon = polygon,         &
      &                                  point   = [0.0_rk, 0.0_rk] )

    if (matval(1) < 1.0_rk) success = .false.
    write(*,*) 'Point (0.0, 0.0) has value:', matval

  end subroutine tem_polygon_material_test_value



  !> Compute the angle between to vectors (they should not both be the 0
  !! vector).
  !!
  !! This function uses the crossproduct and arcsin to find the angle between
  !! two 2D vectors. It takes the four components of the vectors as scalar
  !! arguments to allow vectorized evaluation of multiple vector pairs at once.
  !! None of the vectors is to be a zero vector!
  elemental function angle_between(va_x, va_y, vb_x, vb_y) result(angle)
    !> The first vector va
    real(kind=rk), intent(in) :: va_x, va_y

    !> The second vector vb
    real(kind=rk), intent(in) :: vb_x, vb_y

    !> The angle betweend va and vb
    real(kind=rk) :: angle

    ! The angle is arctan( (va X vb)/(va*vb) ):
    ! Using atan2 here for numerical stability and correct signs.
    angle = atan2( (va_x*vb_y - va_y*vb_x), &
      &            (va_x*vb_x + va_y*vb_y)  )

    ! If the angle is close enough to Pi, consider this as a colinear
    ! 0 angle (avoid multiplicities of Pi in the summed angle).
    if (abs(angle) > tolm*Pi) angle = 0.0_rk

  end function angle_between




  !> A subroutine to test the angle_between function in
  !! tem_polygon_material_test.
  !!
  !! The only argument will be true, if all calculations result in the
  !! expected values.
  subroutine tem_polygon_material_test_angle(success)
    !> Indicator if all tests were computed correctly.
    logical, intent(out) :: success

    real(kind=rk) :: res

    success = .true.

    write(*,*) 'Checking angle computation between two 2D Vectors:'
    ! 90 degree
    res = angle_between( va_x  = 1.0_rk, va_y = 0.0_rk, &
      &                  vb_x  = 0.0_rk, vb_y = 1.0_rk  )

    if (res < 0.5_rk*tolm*PI .or. res > 0.5_rk*tolp*PI) success = .false.
    write(*,*) '90 deg:', res


    ! 45 degree
    res = angle_between( va_x  = 10.0_rk, va_y = 0.0_rk, &
      &                  vb_x  =  2.0_rk, vb_y = 2.0_rk  )

    if (res < 0.25_rk*tolm*PI .or. res > 0.25_rk*tolp*PI) success = .false.
    write(*,*) '45 deg:', res


    ! -90 degree
    res = angle_between( va_x  = 10.0_rk, va_y =  0.0_rk, &
      &                  vb_x  =  0.0_rk, vb_y = -3.0_rk  )

    if (res < -0.5_rk*tolp*PI .or. res > -0.5_rk*tolm*PI) success = .false.
    write(*,*) '-90 deg:', res


    ! -135 degree
    res = angle_between( va_x  = 10.0_rk, va_y =  0.0_rk, &
      &                  vb_x  = -3.0_rk, vb_y = -3.0_rk  )

    if (res < -0.75_rk*tolp*PI .or. res > -0.75_rk*tolm*PI) success = .false.
    write(*,*) '-135 deg:', res


    ! 120 degree
    res = angle_between( va_x  =  0.0_rk,       va_y =  4.5_rk, &
      &                  vb_x  = -sqrt(3.0_rk), vb_y = -1.0_rk  )

    if ( res < 2.0_rk/3.0_rk*tolm*PI &
      &  .or. res > 2.0_rk/3.0_rk*tolp*PI) success = .false.
    write(*,*) '120 deg:', res

    ! Zero vector first
    res = angle_between( va_x  = 0.0_rk, va_y =  0.0_rk, &
      &                  vb_x  = 3.0_rk, vb_y = -1.0_rk  )

    if (res > epsilon(1.0_rk)) success = .false.
    write(*,*) 'Zero vector(first):', res

    ! Zero vector last
    res = angle_between( va_x  = -1.0_rk, va_y = -1.0_rk, &
      &                  vb_x  =  0.0_rk, vb_y =  0.0_rk  )

    if (res > epsilon(1.0_rk)) success = .false.
    write(*,*) 'Zero vector(last):', res

  end subroutine tem_polygon_material_test_angle

end module tem_polygon_material_module
