! See copyright notice in the COPYRIGHT file.
! ****************************************************************************** !
!> declaration of the element property header data, which is globally defined
!! for all processes.
!!
module tem_prophead_module

  ! include treelm modules
  use mpi
  use env_module, only: long_k

  ! include aotus modules
  use aot_out_module,   only: aot_out_type, aot_out_val,                       &
    &                         aot_out_open_table, aot_out_close_table
  use aot_table_module, only: aot_table_open, aot_table_close, aot_get_val
  use aotus_module,     only: flu_State

  implicit none

  private

  public :: tem_prophead_type, init_tem_prophead
  public :: load_tem_prophead, dump_tem_prophead

  type tem_prophead_type
    !> A label to describe this property
    character(len=64) :: label

    !> The position of the bit, which flags the presence
    !! of this property for an element.
    integer :: bitpos

    !> Total number of elements with this property
    integer(kind=long_k) :: nElems
  end type tem_prophead_type

  contains

! ****************************************************************************** !
  !> @todo MC: this subroutine maybe dropped because nelems is not known
  !! beforehand while label and bitpos are known priori, or called from
  !! [[tem_init_vrtx_prop]]
  !! left as it is, boundary property module uses it.
  subroutine init_tem_prophead( me, label, bitpos, part_nElems, comm )
    ! ---------------------------------------------------------------------------
    !> Property to init
    type(tem_prophead_type), intent(out) :: me
    !> Descriptive label
    character(len=*), intent(in) :: label
    !> Position to flag for this prop
    integer, intent(in) :: bitpos
    !> Elements with this property on the local partition
    integer, intent(in) :: part_nElems
    !> MPI communicator to use
    integer, intent(in) :: comm
    ! ---------------------------------------------------------------------------
    integer(kind=long_k) :: myElems
    integer :: iError
    ! ---------------------------------------------------------------------------

    myElems = part_nElems

    me%label = label
    me%bitpos = bitpos
    call MPI_Allreduce( myElems, me%nElems, 1, MPI_INTEGER8, MPI_SUM,          &
      &                 comm, iError )

  end subroutine init_tem_prophead
! ****************************************************************************** !


! ****************************************************************************** !
  !> Load the property header.
  !!
  subroutine load_tem_prophead(me, myPart, comm, conf, root)
    ! ---------------------------------------------------------------------------
    !> Property to load
    type(tem_prophead_type), intent(out) :: me(:)
    !> The process, calling the routine
    integer, intent(in) :: myPart
    !> Relevant communicator
    integer, intent(in) :: comm
    !> lua flu state to read lua file
    type( flu_State ) :: conf
    !> The root process doing the IO
    integer, optional, intent(in) :: root
    ! ---------------------------------------------------------------------------
    integer :: locroot, iprop, sub_handle
    integer :: iError
    integer :: thandle ! handle of property table
    ! ---------------------------------------------------------------------------

    if (present(root)) then
      locroot = root
    else
      locroot = 0
    end if

    if (myPart == locroot) then
      !open properties table
      call aot_table_open( L       = conf,                                     &
        &                  thandle = thandle,                                  &
        &                  key     = 'property' )
      do iprop = 1, size(me)
        call aot_table_open( L       = conf,                                   &
          &                  parent  = thandle,                                &
          &                  thandle = sub_handle,                             &
          &                  pos     = iprop )
        call aot_get_val( L       = conf,                                      &
          &               thandle = sub_handle,                                &
          &               key     = 'label',                                   &
          &               val     = me(iprop)%label,                           &
          &               ErrCode = iError )
        call aot_get_val( L       = conf,                                      &
          &               thandle = sub_handle,                                &
          &               key     = 'bitpos',                                  &
          &               val     = me(iprop)%bitpos,                          &
          &               ErrCode = iError )
        call aot_get_val( L       = conf,                                      &
          &               thandle = sub_handle,                                &
          &               key     = 'nElems',                                  &
          &               val     = me(iprop)%nElems,                          &
          &               ErrCode = iError )
        call aot_table_close( L = conf, thandle = sub_handle )
      end do
      call aot_table_close( L = conf, thandle = thandle )
    end if

    do iprop = 1, size(me)
      call MPI_Bcast( me(iprop)%nElems, 1, MPI_INTEGER8, locroot, comm, iError)
      call MPI_Bcast( me(iprop)%bitpos, 1, MPI_INTEGER, locroot, comm, iError)
      call MPI_Bcast( me(iprop)%label, 64, MPI_CHARACTER, locroot, comm, iError)
    end do

  end subroutine load_tem_prophead
! ****************************************************************************** !


! ****************************************************************************** !
  !> Subroutine to dump header informations on a given property
  !!
  subroutine dump_tem_prophead(me, conf)
    ! ---------------------------------------------------------------------------
    !> Property to write
    type(tem_prophead_type), intent(in) :: me(:)
    !> aotus lua state to write output
    type(aot_out_type) :: conf
    ! ---------------------------------------------------------------------------
    integer :: iprop
    ! ---------------------------------------------------------------------------
    ! open property table
    call aot_out_open_table( conf, 'property' )
    do iprop=1, size(me)
      call aot_out_open_table( conf )
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'label',                                    &
        &               val      = trim(me(iprop)%label) )
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'bitpos',                                   &
        &               val      = me(iprop)%bitpos )
      call aot_out_val( put_conf = conf,                                       &
        &               vname    = 'nElems',                                   &
        &               val      = me(iprop)%nElems )
      call aot_out_close_table(conf)
    end do
    call aot_out_close_table(conf)

  end subroutine dump_tem_prophead
! ****************************************************************************** !


end module tem_prophead_module
! ****************************************************************************** !
