! See copyright notice in the COPYRIGHT file.
! ****************************************************************************** !
!> author: Melven Zoellner
!! author: Jiaxing Qi
!! summary: This module reduces values in time
!!
!! Module takes data from all timesteps and different cells
!! (one after another, but may store data of previous timesteps)
!! and reduces the the data to one value per cell for each variable.
!! Currently the only implemented operations are average, min, max
!! and sum.
!! Desired operation should be specified as
!!```lua
!! transient_reduction = {
!!   'min', 'max', 'sum' or 'average', 'none'
!! }
!!```
!!
module tem_transient_reduction_module

  ! include treelm modules
  use env_module,         only: rk, labelLen
  use tem_aux_module,     only: tem_abort
  use tem_varSys_module,  only: tem_varSys_type
  use tem_logging_module, only: logUnit
  use tem_time_module,    only: tem_time_type
  use tem_float_module,   only: operator(.fne.)

  ! include aotus modules
  use aotus_module,     only: aot_get_val, flu_State
  use aot_table_module, only: aot_table_open, aot_table_close, aot_table_length

  implicit none

  private

  public :: tem_transient_reduction_type
  public :: tem_load_transient_reduction
  public :: tem_transient_reduction_init
  public :: tem_transient_reduction_open
  public :: tem_transient_reduction_apply
  public :: tem_transient_reduction_close
  public :: tem_transient_reduction_reset

  !> all data needed for a transient reduction,
  !! operation to perform and necessary data from previous timesteps
  type tem_transient_reduction_type

    !> Which operation to perform
    character(len=labelLen) :: operation = ''

    !> number of elements
    integer :: nElems = 0

    !> number of components
    integer :: nComponents = 0

    !> Number of degrees of freedom
    integer :: nDofs

    !> number of "recorded" previous timesteps
    integer :: nTimes = 0

    !> array to store data from previous timesteps
    !! size (max nTimes, nComponents*nElems)
    real(kind=rk), allocatable :: val(:,:)

    !> array to store time at previous (and current) timestep
    !! size: (0:max nTimes+1)
    real(kind=rk), allocatable :: time(:)

  end type tem_transient_reduction_type

contains

! ****************************************************************************** !
  !> read configuration file
  !!
  subroutine tem_load_transient_reduction( conf, me, parent, key, iError )
    ! ---------------------------------------------------------------------------
    !> handle for lua file
    type(flu_State)                      :: conf
    !>
    type(tem_transient_reduction_type), intent(out), allocatable :: me(:)
    !> handle for reduce table
    integer, optional,intent(in)         :: parent
    !> which key to open
    character(len=*),optional,intent(in) :: key
    !> handle for reduce table
    integer, optional,intent(out)        :: iError
    ! ---------------------------------------------------------------------------
    integer :: nEntries, handle, iErr, iPos, subhandle
    character(len=labelLen) :: localKey
    ! ---------------------------------------------------------------------------
    if( present( key )) then
      localKey = key
    else
      localKey = 'transient_reduction'
    endif
    nEntries = 1
    allocate( me( nEntries))
    iPos = 1

    ! First check, if it is NOT a table
    if ( present(parent) ) then
      ! ...but still inside a table
      call tem_load_transient_reduction_single( conf    = conf,                &
        &                                       handle  = parent,              &
        &                                       me      = me(iPos),            &
        &                                       key     = localKey,            &
        &                                       iError  = iErr )
    else
      ! ... or at the root level
      call aot_get_val( L       = conf,                                        &
        &               key     = localKey,                                    &
        &               val     = me(iPos)%operation,                          &
        &               ErrCode = iErr,                                        &
        &               default = '' )
    end if

    if( (iErr /= 0) .or. (trim(me(iPos)%operation)=='') ) then
      ! ok, that failed. Let's try opening as a table.
      call aot_table_open( L       = conf,                                     &
        &                  thandle = handle,                                   &
        &                  parent  = parent,                                   &
        &                  key     = localKey )

      ! count number of entries
      nEntries = aot_table_length(L=conf, thandle=handle)
      deallocate( me )
      allocate( me( nEntries ))

      ! try to open subtables
      call aot_table_open( L       = conf,                                     &
        &                  parent  = handle,                                   &
        &                  thandle = subhandle,                                &
        &                  pos     = 1 )

      if ( subhandle .eq. 0 ) then

        ! only a single table, but multiple entries, read by positions
        call aot_table_close( L = conf, thandle = subhandle )
        do iPos = 1, nEntries
          call tem_load_transient_reduction_single( conf    = conf,            &
            &                                       handle  = handle,          &
            &                                       me      = me(iPos),        &
            &                                       pos     = iPos,            &
            &                                       iError  = iErr )
        end do
        if( iErr .ne. 0 ) then
          ! Error, allocate with 0 size
          nEntries = 0
          deallocate( me )
          allocate( me(nEntries))
        end if

      else ! it has subtables, get into subtable and read by key

        call aot_table_close( L = conf, thandle = subhandle )
        do iPos = 1, nEntries
          call aot_table_open( L       = conf,                                 &
            &                  parent  = handle,                               &
            &                  thandle = subhandle,                            &
            &                  pos     = iPos )
          call tem_load_transient_reduction_single( conf    = conf,            &
            &                                       handle  = subhandle,       &
            &                                       me      = me(iPos),        &
            &                                       key     = localKey,        &
            &                                       iError  = iErr )
          call aot_table_close( L = conf, thandle = subhandle )
        end do

      end if ! open subtable

      call aot_table_close(conf, handle)

    end if ! open transient_reduction table

    do iPos = 1, nEntries
      write(logUnit(1),*) ' Load transient reduction: '//                      &
        &                   trim(me(iPos)%operation)//'. nTimes: ',            &
        &                   me(iPos)%nTimes
    enddo

    if( present( iError )) iError = iErr
  end subroutine tem_load_transient_reduction
! ****************************************************************************** !


! ****************************************************************************** !
  !> Read a single entry of transient reductions from the lua file
  !!
  subroutine tem_load_transient_reduction_single( conf, me, handle, key, pos,  &
    &                                             iError )
    ! ---------------------------------------------------------------------------
    !> handle for lua file
    type(flu_State),intent(inout)                    :: conf
    !>
    type(tem_transient_reduction_type),intent(inout) :: me
    !> handle for reduce table
    integer, optional,intent(in)                     :: handle
    !> which key to open
    character(len=*),optional,intent(in)             :: key
    !> which key to open
    integer,optional,intent(in)                      :: pos
    !> handle for reduce table
    integer, optional,intent(out)                    :: iError
    ! ---------------------------------------------------------------------------
    integer :: iErr
    ! ---------------------------------------------------------------------------
    if ( present( key )) then

      call aot_get_val( L       = conf,      &
        &               thandle = handle,    &
        &               val     = me%nTimes, &
        &               ErrCode = iErr,      &
        &               key     = "nTimes",  &
        &               default = 0 )

      call aot_get_val( L       = conf,         &
        &               thandle = handle,       &
        &               val     = me%operation, &
        &               ErrCode = iErr,         &
        &               key     = "operation" )

    else if( present( pos ))then

      call aot_get_val( L       = conf,         &
        &               thandle = handle,       &
        &               val     = me%operation, &
        &               ErrCode = iErr,         &
        &               pos     = pos )

    else

      iErr = -1

    end if

    if( present( iError )) then
      iError = iErr
    end if

  end subroutine tem_load_transient_reduction_single
! ****************************************************************************** !


! ****************************************************************************** !
  !> Initialize transient reduction objects according to the number of values we
  !! get each timestep
  !!
  subroutine tem_transient_reduction_init( me, nElems, varSys, varPos, time,   &
    &                                      nDofs )
    ! ---------------------------------------------------------------------------
    !> transient reduction type to be initialized
    type(tem_transient_reduction_type ), intent(inout) :: me(:)
    !> number of elements
    integer, intent(in) :: nElems
    !> corresponding variable system
    type(tem_varSys_type), intent(in) :: varSys
    !> current time information
    type(tem_time_type), intent(in) :: time
    !> position of variable to reduce in the global varSys
    integer, intent(in) :: varPos(:)
    !> Number of degrees of freedom per variable
    integer, optional, intent(in) :: nDofs
    ! ---------------------------------------------------------------------------
    integer :: iRed, loc_nDofs
    ! ---------------------------------------------------------------------------
    write(logUnit(1),*)'Initializing transient reductions ...'

    if (present(nDofs)) then
      loc_nDofs = nDofs
    else
      loc_nDofs = 1
    end if

    do iRed = 1, size(me)
      ! number of values
      me(iRed)%nElems = nElems
      me(iRed)%nComponents = varSys%method%val( varPos(iRed) )%nComponents
      me(iRed)%nDofs = loc_nDofs

      select case(me(iRed)%operation)

        case('none')

        case('min')
          allocate(me(iRed)%val(1,nElems*me(iRed)%nComponents*loc_nDofs))
          me(iRed)%val(1,:) = huge(1.0_rk)

        case('max')
          allocate(me(iRed)%val(1,nElems*me(iRed)%nComponents*loc_nDofs))
          me(iRed)%val(1,:) = tiny(1.0_rk)

        case('sum')
          allocate(me(iRed)%val(1,nElems*me(iRed)%nComponents*loc_nDofs))
          me(iRed)%val(1,:) = 0.0_rk

        case('average')
          allocate(me(iRed)%val(1,nElems*me(iRed)%nComponents*loc_nDofs))
          me(iRed)%val(1,:) = 0.0_rk
          allocate(me(iRed)%time(0:2)) ! store t0 and current time
          ! t0 is reference time point
          ! t1 is the last reduction time point
          ! t2 is current time
          me(iRed)%time(0) = time%sim
          me(iRed)%time(1:2) = 0.0_rk

        case('osi') ! oscillatory shear index
          allocate(me(iRed)%val( me(iRed)%nTimes,                              &
            &                    nElems*me(iRed)%nComponents*loc_nDofs))
          me(iRed)%val(:,:) = 0.0_rk

        case default
          write(logUnit(1),*)'unknown transient reduction operation: '//       &
            &            me(iRed)%operation 
          call tem_abort()

      end select
    end do

  end subroutine tem_transient_reduction_init
! ****************************************************************************** !


! ****************************************************************************** !
  !> resets the "recorded" data of the reduction
  !!
  subroutine tem_transient_reduction_reset( me )
    ! ---------------------------------------------------------------------------
    !> transient reduction type
    type(tem_transient_reduction_type ),intent(inout) :: me(:)
    ! ---------------------------------------------------------------------------
    integer :: iRed
    ! ---------------------------------------------------------------------------

    do iRed = 1, size(me)
      select case(me(iRed)%operation)
        case('min')
          me(iRed)%val(1,:) = huge(1.0_rk)

        case('max')
          me(iRed)%val(1,:) = tiny(1.0_rk)

        case('sum')
          me(iRed)%val(1,:) = 0.0_rk

        case('average')
          ! reset reference time point
          ! t0 is reference time point
          ! t1 is the last reduction time point
          ! t2 is current time
          me(iRed)%val(1,:) = 0.0_rk
          me(iRed)%time(0) = me(iRed)%time(0) + me(iRed)%time(1)
          me(iRed)%time(1) = 0.0_rk

      end select
    end do

  end subroutine tem_transient_reduction_reset
! ****************************************************************************** !


! ****************************************************************************** !
  !> "open" the reduction to prepare it
  !!
  subroutine tem_transient_reduction_open( me, time )
    ! ---------------------------------------------------------------------------
    !> transient reduction type
    type(tem_transient_reduction_type ),intent(inout) :: me(:)
    !>
    real(kind=rk), intent(in)                         :: time
    ! ---------------------------------------------------------------------------
    integer :: iRed
    ! ---------------------------------------------------------------------------
    do iRed = 1, size(me)
      select case(me(iRed)%operation)
        case('average')
          ! t0 is reference time point
          ! t1 is the last reduction time point
          ! t2 is current time
          ! me(iRed)%time(2) = time - me(iRed)%time(0)
          me(iRed)%time(1) = me(iRed)%time(2)
          me(iRed)%time(2) = time
      end select
    end do

  end subroutine tem_transient_reduction_open
! ****************************************************************************** !


! ****************************************************************************** !
  !> apply the transient reduction
  !!
  subroutine tem_transient_reduction_apply( me, chunk, offset, nChunkElems,    &
    &                                       varSys, varPos )
    ! ---------------------------------------------------------------------------
    !> The transient reduction type to work on
    type( tem_transient_reduction_type ), intent(inout) :: me(:)
    !> apply the reduction to the data from the current timestep
    real(kind=rk), intent(inout)                        :: chunk(:)
    !> offset of chunk in total number of elements
    integer, intent(in)                                 :: offset
    !> number of elements the chunk has
    integer, intent(in)                                 :: nChunkElems
    !> global variable system defined in solver
    type( tem_varSys_type ), intent(in)                 :: varSys
    !> position of variable to reduce in the global varSys
    integer, intent(in)                                 :: varPos(:)
    ! ---------------------------------------------------------------------------
    integer :: iRed, iComp, iElem, iDof
    integer :: varPos_red, chunkVarPos
    integer :: compOff, elemsize, e_pos, d_pos
    integer :: nScalars
    ! ---------------------------------------------------------------------------
    nScalars = sum(varSys%method%val(varPos(:))%nComponents)


    ! calculate result and update me%val
    compOff = 0
    do iRed = 1, size( me )
      ! Size of a single element. 
      ! nDofs is same for all variables
      elemsize = nScalars*me(iRed)%nDofs

      varPos_red = offset * me( iRed )%nComponents
      do iElem = 1, nChunkElems
        e_pos = (iElem-1)*elemsize
        do iDof = 1, me( iRed )%nDofs
          d_pos = (iDof-1)*nScalars + compOff
          do iComp = 1, me( iRed )%nComponents
            chunkVarPos = e_pos + d_pos + iComp
            varPos_red = varPos_red + 1
            !varPos_red = (offset + iElem - 1)*varSys%variable%val( iRed )%nComponents + iComp
            select case( me( iRed )%operation )

              case('none')

              case('min')
                me(iRed)%val(1,varPos_red) = min( me(iRed)%val(1,varPos_red),  &
                  &                           chunk(chunkVarPos))
                chunk(chunkVarPos) = me(iRed)%val(1,varPos_red)

              case('max')
                me(iRed)%val(1,varPos_red) = max( me(iRed)%val(1,varPos_red),  &
                  &                           chunk(chunkVarPos))
                chunk(chunkVarPos) = me(iRed)%val(1,varPos_red)

              case('sum')
                me(iRed)%val(1,varPos_red) = me(iRed)%val(1,varPos_red)        &
                  &                    + chunk(chunkVarPos)
                chunk(chunkVarPos) = me(iRed)%val(1,varPos_red)

              case('average')
                if(me(iRed)%time(2) .fne. 0._rk) then
                  ! the average value up to now
                  ! t1 is the last reduction time point
                  ! t2 is current time
                  ! average = ( t1*val + (t2-t1)*chunk ) / t2
                  me(iRed)%val(1,varPos_red) = 1./me(iRed)%time(2)             &
                    & * ( me(iRed)%time(1) * me(iRed)%val(1,varPos_red)        &
                    & + (me(iRed)%time(2) - me(iRed)%time(1))                  &
                    &                                     * chunk(chunkVarPos) )

                  chunk(chunkVarPos) = me(iRed)%val(1,varPos_red)
                end if

            end select
          end do !iComp
        end do !iDof
      end do !iElem
      compOff = compOff + me( iRed )%nComponents
    end do !iRed

  end subroutine tem_transient_reduction_apply
! ****************************************************************************** !


! ****************************************************************************** !
  !> close the reduction after it has been applied to all chunks
  !!
  subroutine tem_transient_reduction_close( me )
    ! ---------------------------------------------------------------------------
    !> transient reduction type to be closed
    type(tem_transient_reduction_type ),intent(inout) :: me(:)
    ! ---------------------------------------------------------------------------
    integer :: iRed
    ! ---------------------------------------------------------------------------

    do iRed = 1, size(me)
      select case(me(iRed)%operation)
        case('average')
          ! record current time for future usage
          me(iRed)%time(1) = me(iRed)%time(2)
      end select
    end do

  end subroutine tem_transient_reduction_close
! ****************************************************************************** !


end module tem_transient_reduction_module
! ****************************************************************************** !
